
/*
 * bs_string.h:
 *      header file for bs_string.cpp. See remarks there.
 *
 *      Warning: if you include this header, you'll get a
 *      typedef for "string" to BSString.
 *
 *@@include #include "helpers\xstring.h"
 *@@include #include "base\bs_string.h"
 */

/*
 *      This file Copyright (C) 1999-2000 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#ifndef WARPIN_STRING_HEADER_INCLUDED
    #define WARPIN_STRING_HEADER_INCLUDED

    // STL support
    #include <map>

    typedef unsigned long size_type;

    /*
     *@@ BSString:
     *      simple string class, modeled after the C++
     *      standard library "string" class, which is
     *      not implemented with IBM VAC++ 3.0.
     *
     *      See bs_string.cpp for an introduction.
     */

    class BSString
    {
        #ifdef __IBMCPP__
            friend class BSString;
        #endif
        // 2000-01-15: This line is extraordinary useless - class instances
        //             of the same type are always friends (czw)
        // V0.9.1 (2000-02-01) [umoeller]: Hi Jens. Not with VAC++, apparently,
        //             because without this line, this wouldn't compile.

        protected:
            XSTRING         _str;

            size_t          *_pShiftTable;
            int             _fRepeat;

            void Init();
            void Cleanup();
            void ClearShiftTable();

            void CopyFrom(const char *psz);
            void CopyFrom(const BSString &s,
                          size_type ulPos = 0,
                          size_type n = npos);

        public:
            static const size_type npos;

            BSString();
            ~BSString();

            // default copy constructor;
            // required for exception handling!
            BSString(const BSString &s);

            BSString(const BSString &s,
                     size_type ulPos,
                     size_type n = npos);
            BSString(const char *psz);
            BSString(const char *p1, const char *p2);

            BSString& assign(const BSString &s);
            BSString& assign(const BSString &s,
                             size_type ulPos,
                             size_type n = npos);

            BSString& operator=(const BSString &s);
            BSString& operator=(const char *psz);
            BSString& operator=(char c);

            /*
             *@@ c_str:
             *      returns the string member as a
             *      classic const C string. Note that
             *      this _never_ returns NULL. If the
             *      member string is empty, this
             *      returns a pointer to a null byte.
             */

            const char* c_str()
                        const
            {
                if (_str.psz)
                    return (_str.psz);
                else
                    return "";
            }

            /*
             *@@ size:
             *      returns the no. of characters in
             *      the string (excluding our internal
             *      null terminator) or 0 if the string
             *      is empty.
             */

            inline size_type size()
                             const
            {
                return (_str.ulLength);
            }

            /*
             *@@ length:
             *      STL string compatibility function,
             *      same as BSString::size().
             */

            inline size_type length()
                             const
            {
                return (_str.ulLength);
            }

            /*
             *@@ capacity:
             *      returns the amount of memory presently
             *      allocated for the string. This is at
             *      least what size() or length() return.
             *
             *@@added V0.9.6 (2000-10-31) [umoeller]
             */

            inline size_type capacity()
                             const
            {
                return (_str.cbAllocated);
            }

            void reserve(size_type stExtra = 0);

            /*
             *@@ operator():
             *      returns 1 (TRUE) if the string
             *      contains something, 0 (FALSE)
             *      otherwise.
             *
             *      Note: as far as I know, this operator
             *      is NOT defined with the C++ string class.
             */

            inline int operator()()
                       const
            {
                return (size() != 0);
            }

            /*
             *@@ empty:
             *      returns 1 (TRUE) if the string
             *      contains nothing, 0 (FALSE) otherwise.
             */

            inline int empty()
                       const
            {
                return (size() == 0);
            }

            char operator[](const size_type ul) const;

            int compare(const BSString &s) const;
            int compare(size_type ulPos, size_type n, const BSString &s) const;

            BSString& append(const BSString &s);
            BSString& append(const char *psz);
            BSString& append(char c);

            /*
             *@@ operator+=:
             *      shortcut to append(BSString &s).
             *
             *@@changed V0.9.2 (2000-03-29) [umoeller]: added return *this
             */

            BSString& operator+=(const BSString &s)     // in: string to append
            {
                append(s);
                return *this;
            }

            /*
             *@@ operator+=:
             *      shortcut to append(const char *psz).
             *
             *@@changed V0.9.2 (2000-03-29) [umoeller]: added return *this
             */

            BSString& operator+=(const char *psz)     // in: string to append
            {
                append(psz);
                return *this;
            }

            /*
             *@@ operator+=:
             *      shortcut to append(char c).
             *
             *@@changed V0.9.2 (2000-03-29) [umoeller]: added return *this
             */

            BSString& operator+=(char c)     // in: character to append
            {
                append(c);
                return *this;
            }

            BSString& erase(size_type ulPos = 0, size_type n = npos);

            size_type find(const BSString &strFind, size_type ulPos = 0) const;
            size_type find(const char *pszFind, size_type ulPos = 0) const;
            size_type find(char c, size_type ulPos = 0) const;

            size_type rfind(char c, size_type ulPos = npos) const;

            size_type find_first_of(const char *achChars, size_type ulPos = 0) const;

            size_type find_first_not_of(char c, size_type ulPos = 0) const;
            size_type find_first_not_of(const char *achChars, size_type ulPos = 0) const;

            BSString& replace(size_type ulPosThis,
                              size_type nThis,
                              const BSString &strReplace);

            /* BSString& replace(size_type ulPosThis,
                              size_type nThis,
                              const BSString &strReplace,
                              size_type ulPosReplace = 0,
                              size_type nReplace = npos);

            BSString& replace(size_type ulPosThis,
                              size_type nThis,
                              const char *pszReplace,
                              size_type nReplace);

            BSString& replace(size_type ulPosThis,
                              size_type nThis,
                              const char *pszReplace); */

            BSString substr(size_type ulPos = 0,
                            size_type n = npos)
                     const;

            /*
             *@@ swap:
             *      swaps the contents of this instance
             *      with that of another BSString.
             *
             *      This is very fast because only the
             *      pointers are exchanged.
             *
             *@@added V0.9.6 (2000-10-31) [umoeller]
             */

            void swap(BSString &str)
            {
                XSTRING strTemp;
                memcpy(&strTemp, &str._str, sizeof(XSTRING));
                memcpy(&str._str, &_str, sizeof(XSTRING));
                memcpy(&_str, &strTemp, sizeof(XSTRING));
                // we've changed, so invalidate our own shift table
                ClearShiftTable();
                str.ClearShiftTable();
            }

            size_type _printf(const char *pszFormatString, ...);

            size_type _find_word(BSString &strFind,
                                 size_type ulPos = 0,
                                 const char *pcszBeginChars = "\x0d\x0a ()/\\-,.;*",
                                 const char *pcszEndChars   = "\x0d\x0a ()/\\-,.:;*&'")
                      const;

            size_type _find_word(const char *pszFind,
                                 size_type ulPos = 0,
                                 const char *pcszBeginChars = "\x0d\x0a ()/\\-,.;*",
                                 const char *pcszEndChars   = "\x0d\x0a ()/\\-,.:;*&'")
                      const;

            size_type _find_replace(BSString &strFind,
                                    const BSString &strReplaceWith,
                                    size_type *pulPos);
            size_type _find_replace(const char *pszFind,
                                    const BSString &strReplaceWith,
                                    size_type *pulPos);
            size_type _find_replace(char cFind,
                                    char cReplace,
                                    size_type *pulPos);

            int _extract_from(const char *pStart,
                              const char *pEnd);

            int _extract_word(unsigned long ulIndex,
                              BSString &strTarget,
                              size_type ulPos = 0,
                              const char *pcszBeginChars = "\x0d\x0a ()/\\-,.;*",
                              const char *pcszEndChars   = "\x0d\x0a ()/\\-,.:;*&'")
                const;

            void _format();
    };

    int operator==(const BSString &s1, const BSString &s2);

    int operator==(const char *psz1, const BSString &s2);

    int operator==(const BSString &s1, const char *psz2);

    int operator!=(const BSString &s1, const BSString &s2);

    int operator!=(const char *psz1, const BSString &s2);

    int operator!=(const BSString &s1, const char *psz2);

    int operator<(const BSString &s1, const BSString &s2);

    BSString operator+(const BSString &s1, const BSString &s2);

    BSString operator+(const char *psz1, const BSString &s2);

    BSString operator+(const BSString &s1, const char *psz2);

    /*
     *@@ string:
     *      typedef for using BSStrings like the
     *      standard string class.
     */

    typedef BSString string;

    /*
     *@@ BSStringMap:
     *      typedef for a BSString map.
     *
     *@@added V0.9.2 (2000-03-11) [umoeller]
     */

    typedef map< BSString,      // key class
                 BSString,      // data class
                 less<BSString>
               > BSStringMap;

#endif

