
/*
 * fe_job.h:
 *      header file for fe_job.cpp.
 *
 *@@include #define INCL_WINSHELLDATA
 *@@include #define INCL_WINWORKPLACE
 *@@include #include <os2.h>
 *@@include #include <string.h>
 *@@include #include "base\bs_errors.h"
 *@@include #include "base\bs_config.h"
 *@@include #include "frontend\fe_job.h"
 */

/*
 *      Copyright (C) 1998-2000 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#ifndef WARPIN_FEWORKER_HEADER_INCLUDED
    #define WARPIN_FEWORKER_HEADER_INCLUDED

    /* ******************************************************************
     *
     *  Forward declarations
     *
     ********************************************************************/

    class FEDatabase;
    class FEPackageID;
    class FEArcPackagePck;
    class FEArcPackageGroup;
    class FEDBPackage;

    class FEInstallJob;
    class FEGroupJob;

    /* ******************************************************************
     *
     *  FEJob class
     *
     ********************************************************************/

    // job class
    #define JOBCLASS_ERROR                   0
    #define JOBCLASS_INSTALLJOB     0x00010000
    #define JOBCLASS_GROUPJOB       0x00020000

    // flags for FEInstallJob.ulJob
    #define JOB_UNDEFINED                    0

    #define JOB_INSTALL_IGNORE      0x00010000
    #define JOB_INSTALL             0x00010001
    #define JOB_DEINSTALL           0x00010002

    #define JOB_GROUP_MIXED         0x00020004

    /*
     *@@ FEJobBase:
     *      job base class, of which the other job
     *      classes are derived. No objects of this
     *      class exist, but only of the subclasses
     *      (FEInstallJob or FEGroupJob).
     *
     *      This class is abstract anyway, so don't
     *      try to create instances of this.
     */

    class FEJobBase
    {
        protected:
            ULONG       _ulJobClass;

            ULONG       _ulJobID;
                    // If this job contains an FEArcPackagePck,
                    // one of the following:
                    // -- JOB_INSTALL_IGNORE: well, do nothing
                    // -- JOB_INSTALL: install package
                    // -- JOB_DEINSTALL: remove package
                    // If this job contains an FEArcPackageGroup,
                    // just as above, but in addition:
                    // -- JOB_GROUP_MIXED: sub-packages have
                    //    varying flags.

            FEJobBase(FEGroupJob *pGroupJob_);

            virtual ~FEJobBase() {};

        public:
            FEGroupJob  *_pGroupJob;
                    // the group job we belong to or NULL

            PVOID       _pvGuiData;
                    // for GUI usage: reverse linkage to PACKAGERECORD in cnr

            virtual FEInstallJob* IsInstallJob();
            virtual FEGroupJob* IsGroupJob();

            ULONG QueryJobID();

            virtual ULONG ToggleSelection() = 0;

            virtual BOOL Select(ULONG ulJobNew) = 0;
    };

    /* ******************************************************************
     *
     *  FEInstallJob class
     *
     ********************************************************************/

    // flags for FEInstallJob.ulInstallStatus
    #define INSTALLED_UNKNOWN       0
    #define INSTALLED_NO            1
    #define INSTALLED_IS_OLDER      2
    #define INSTALLED_IS_NEWER      3
    #define INSTALLED_IS_SAME       4

    // flags for FEPackage.ulConfigureFlags
    #define JOB_CONFIGSYS    0x01    // CONFIG.SYS can be updated
    #define JOB_WPSCLASSES   0x02    // WPS classes can be installed
    #define JOB_WPSOBJECTS   0x04    // WPS objects can be created

    /*
     *@@ FEInstallJob:
     *      "install job" class, which represents
     *      a package to be installed. pPackage
     *      points to the package from the archive
     *      which this job represents.
     *
     *      Note that for groups, an instance of
     *      FEGroupJob is created instead.
     */

    class FEInstallJob : virtual public FEJobBase
    {
        protected:
            BSString    _strTargetPath;
                    // target path;
                    // initially copied from FEArcPackagePck,
                    // can be changed later here

        public:
            FEArcPackagePck *_pArcPackage;
                    // package to be installed or removed
                    // (points to a list item in an FEArchive);
                    // this will always be a package, never a
                    // group, since we have an install job

            FEPackageID *_pPackageID;
                    // package ID of the member package;
                    // constructed by CheckInstall
            FEPackageID *_pPackageIDInstalled;
                    // package ID of installed package;
                    // constructed by CheckInstall, if package is installed already;
                    // otherwise it's NULL

            ULONG       _ulInstallStatus;
                    // install status, valid after CheckInstall call.
                    // -- INSTALLED_UNKNOWN: CheckInstall not yet called. (99-11-01) [umoeller]
                    // -- INSTALLED_NO: package not installed.
                    // -- INSTALLED_IS_OLDER: package is installed and older than in archive;
                    //                     it should be updated then.
                    // -- INSTALLED_IS_NEWER: package is installed and newer than in archive
                    //                     it should _not_ be updated then.
                    // -- INSTALLED_IS_SAME: package is installed and same as in archive.

            FEInstallJob(FEArcPackagePck *pPackage,
                         FEGroupJob *pGroupJob);

            virtual ~FEInstallJob();

            virtual FEInstallJob* IsInstallJob();

            BOOL CheckInstall(FEDatabase *pDatabase);

            const char* QueryTargetPath() const;

            const char* QueryPckPackageID() const;

            ULONG QueryPckSizeRounded() const;

            const char* QueryPckTitle() const;

            ULONG QueryPckFilesCount() const;

            ULONG QueryPckOrigSize() const;

            BOOL SetTargetPath(const char *pszNewTargetPath,
                               list<FEJobBase*> &JobsList);

            BOOL DefaultSelect();

            BOOL RepairSelect();

            BOOL UpdateSelect();

            virtual BOOL Select(ULONG ulJobNew);

            virtual ULONG ToggleSelection();

            BOOL Is2BeInstalled();

            BOOL Is2BeRemoved();

            VOID Unpack();

            VOID AddFileHeader(WIFileHeader* pwifh);

            BOOL Store(FEDatabase *pDatabase);
    };

    /*
     *@@ FEGroupJob:
     *      "group job" class which represents a
     *      group on the "container" pages.
     */

    class FEGroupJob : virtual public FEJobBase
    {
        public:
            // list of packages which are part of this
            // group; the items on this list can either
            // be FEInstallJob's or FEGroupJob's
            list<FEJobBase*> _GroupMembersList;

            const FEArcPackageGroup *_pGroupPackage;
                    // group which is represented by this job

            FEGroupJob(FEArcPackageGroup *pGroup_,
                       FEGroupJob *pGroupJob_);

            virtual FEGroupJob* IsGroupJob();

            virtual ULONG ToggleSelection();

            virtual ULONG Select(ULONG ulJobNew);

            VOID Update();
    };

    /*
     *@@ FEConfigJob:
     *
     *@@added V0.9.1 (2000-01-08) [umoeller]
     */

    class FEConfigJob : virtual public FEJobBase
    {
        FEConfigJob() : FEJobBase(NULL) { };
    };

#endif
