
/*
 * fe_package_arc.h:
 *      header file for fe_package_arc.cpp.
 *
 *@@include #define INCL_WINSHELLDATA
 *@@include #define INCL_WINWORKPLACE
 *@@include #include <os2.h>
 *@@include #include <string.h>
 *@@include #include "base\bs_errors.h"
 *@@include #include "base\bs_logger.h"
 *@@include #include "base\bs_config.h"
 *@@include #include "frontend\fe_package_arc.h"
 *@@include #include "frontend\warpin.h"
 */

/*
 *      Copyright (C) 1998-2000 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#ifndef WARPIN_FEPACKAGEARC_HEADER_INCLUDED
    #define WARPIN_FEPACKAGEARC_HEADER_INCLUDED

    class BSPckDeclBase;
    class BSGroupDeclBase;

    /*
     *@@ FEArcPackageBase:
     *      parent class for the packages used in "full install"
     *      and "add/remove" modes, that is, every time an archive
     *      is involved.
     *
     *      This is for stuff which is shared by both "real" packages
     *      (FEArcPackagePck) and groups (FEArcPackageGroup), which
     *      both inherit from this class.
     *
     *      The constructor is protected because instances of
     *      this class shall never be created directly.
     *
     *      However, this class is used by the lists in the
     *      FEInstallJob and FEArchive classes to store packages.
     *      If you need to find out whether the package is a
     *      group or not, call FEArcPackageBase::IsPackage().
     *      Depending on the return value, you can manually cast
     *      to FEArcPackagePck or FEArcPackageGroup.
     *
     *      Note that FEInstallJob::Is2BeInstalled() returns TRUE
     *      for instances of FEArcPackagePck also, so that one
     *      might be quicker.
     *
     *@@added V0.9.0 (99-11-06) [umoeller]
     *@@changed V0.9.1 (2000-01-07) [umoeller]: renamed from FEArcPackage
     */

    class FEArcPackageBase : public FEPackageBase
    {
        public:
            FEArchive           *_pArchive;
                    // the archive this package belongs to;
                    // this always points to a valid archive!

            FEArcPackageGroup   *_pGroup;
                    // the group this package/group belongs to or NULL

            ULONG               _ulSizeRounded;
                    // total uncompressed size in KB (set by FEArchive)

        protected:
            FEArcPackageBase(FEArchive *pArchive,
                             FEArcPackageGroup *pGroup,
                             ULONG ulType);

        public:
            /*
             *@@ IsPackage:
             *      this virtual method returns NULL, unless
             *      this instance is of the FEArcPackagePck
             *      subclass, where "this" is returned already
             *      typecast to FEArcPackagePck*.
             *
             *@@added V0.9.1 (2000-01-08) [umoeller]
             */

            virtual FEArcPackagePck* IsPackage() const
            {
                return NULL;
            }

            ULONG QueryCodepage() const;
    };

    /*
     *@@ FEArcPackagePck:
     *      FEArcPackageBase subclass for packages in an archive.
     *
     *      These things are created by FEArchive::ParsePackages
     *      according to the install script while WarpIN initializes
     *      itself in "full install" or "add/remove" mode.
     *
     *@@added V0.9.0 (99-11-06) [umoeller]
     *@@changed V0.9.1 (2000-02-07) [umoeller]: added lotsa new config stuff
     *@@changed V0.9.1 (2000-02-08) [umoeller]: reduced list<> code bloat
     */

    class FEArcPackagePck : public FEArcPackageBase
    {
        public:
            BSPckDeclBase       *_pDecl;
                    // package declaration from the script
                    // moved this here from FEPackageBase
                    // V0.9.12 (2001-05-17) [umoeller]

            FEArcPackagePck(FEArchive *pArchive,
                            FEArcPackageGroup *pGroup,
                            BSPckDeclBase *pDecl,
                            PFNWICALLBACK pWICallback);

            /*
             *@@ IsPackage:
             *      as opposed to the parent method
             *      (FEArcPackageBase::IsPackage),
             *      this returns "this".
             */

            virtual FEArcPackagePck* IsPackage() const
            {
                return (FEArcPackagePck*)this;
            }

            VOID Unpack(VOID);

            VOID AddFileHeader(WIFileHeader* pwifh);
    };

    /*
     *@@ FEArcPackageGroup:
     *      FEArcPackageBase subclass for groups.
     *
     *      These things are created by FEArchive::ParsePackages
     *      according to the install script while WarpIN initializes
     *      itself in "full install" or "add/remove" mode.
     *
     *      Groups are only used on the "Container" page in
     *      "full install" and "add/remove" modes. Groups
     *      will never be installed, because they're only
     *      for grouping FEArcPackage's logically.
     *
     *      That's why we have the public GroupMembersList
     *      member here, which is really the only thing
     *      groups are good for.
     *
     *      Note that even though this class inherits
     *      from FEPackageBase, many of the member fields
     *      in there are not used by groups. To avoid this,
     *      we'd have to use multiple inheritance, which
     *      I prefer to avoid. ;-)
     *
     *@@added V0.9.0 (99-11-06) [umoeller]
     */

    class FEArcPackageGroup : public FEArcPackageBase
    {
        public:
            BSGroupDeclBase         *_pGroupDecl;

            list<FEArcPackageBase*> _GroupMembersList;
                    // list of packages which are part of this
                    // group; the items on this list can either
                    // be FEArcPackage's or FEArcPackageGroup's.
                    // This list is added to by FEArcPackageBase::FEArcPackageBase.

            FEArcPackageGroup(FEArchive *pArchive,
                              FEArcPackageGroup *pGroup,
                              BSGroupDeclBase *pDecl);
    };

#endif
