
/*
 *@@sourcefile bs_config.cpp:
 *      this has the code for the classes which manipulate CONFIG.SYS,
 *      the WPS class list, and create WPS objects.
 *
 *      These classes are designed to work independently of WarpIN
 *      and could be used in any program.
 *      I have attempted to design these classes for easy, yet flexible use.
 *
 *      The general idea of these classes is that you have one class
 *      for each manipulation to be done and another class which undoes
 *      the change again. (The exception is the BSConfigSys class; see
 *      remarks below).
 *
 *      For each group of classes, there is one AddToUndoList static
 *      (class) method, which creates a list of Undo objects from the
 *      logger which was used to do the manipulation before.
 *      See BSCfgSysManip::BSCfgSysManip for an example usage.
 *
 *      The loggers are all direct, unmodified descendants of the
 *      BSMemLoggerBase class (bs_logger.cpp) and only defined to use C++
 *      type checking.
 *
 *      Some config methods throw instances of the BSConfigExcpt class def'd in
 *      bs_config.h, which you should handle. If so, a "throw (BSConfigExcpt)"
 *      has been added to the respective function headers.
 *
 *      The manipulation method of each of the manipulation classes
 *      take some kind of BSMemLoggerBase instance as input (see the description
 *      below). This instance logs all the changes made and can then
 *      be passed to the "Undo" class/method to undo the change again.
 *
 *      These are the classes implemented here:
 *
 *      --  BSCfgSysManip:
 *          a "manipulator" object of this class holds information
 *          for a CONFIG.SYS change, which is passed to BSConfigSys::Manipulate.
 *          See BSCfgSysManip::BSCfgSysManip for a description.
 *
 *      --  BSConfigSys:
 *          the actual class which holds the CONFIG.SYS text file
 *          and allows changes to be made.
 *          This class does not have a corresponding "Undo" class.
 *          Instead, use AddToCfgSysManipUndoList.
 *
 *      --  BSRegisterClass:
 *          this allows registering classes with the WPS.
 *          See BSRegisterClass::Register for an example usage.
 *
 *      --  BSDeregisterClass:
 *          the "Undo" class to the previous.
 *
 *      --  BSReplaceClass:
 *          this allows WPS classes to be replaced.
 *          See BSReplaceClass::Replace for an example usage.
 *
 *      --  BSUnreplaceClass:
 *          the "Undo" class to the previous.
 *
 *      --  BSCreateWPSObject:
 *          for creating a WPS object.
 *
 *      --  BSDeleteWPSObject:
 *          the "Undo" class for the previous, to delete a WPS object again.
 *
 *      See bs_config.h for the declarations of these classes.
 *
 *@@header "base\bs_config.h"
 *@@header "base\bs_config_impl.h"
 */

/*
 *      This file Copyright (C) 1999-2000 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define OS2EMX_PLAIN_CHAR
    // this is needed for "os2emx.h"; if this is defined,
    // emx will define PSZ as _signed_ char, otherwise
    // as unsigned char

// all these are required for winh.h
#define INCL_DOSERRORS
#define INCL_WINWINDOWMGR
#define INCL_WINMESSAGEMGR
#define INCL_WINDIALOGS
#define INCL_WINSTDCNR
#define INCL_WININPUT
#define INCL_WINSYS
#define INCL_WINSHELLDATA
#define INCL_WINWORKPLACE
#include <os2.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "setup.h"

#include <list>

// include's from helpers
#include "helpers\dosh.h"
#include "helpers\winh.h"
#include "helpers\prfh.h"
#include "helpers\stringh.h"
#include "helpers\xstring.h"

#include "helpers\configsys.h"

// front-end includes
#include "base\bs_string.h"
#include "base\bs_errors.h"

#include "base\bs_logger.h"
#include "base\bs_config.h"
#include "base\bs_config_impl.h"

#pragma hdrstop

/* ******************************************************************
 *
 *  BSCfgSysManip class
 *
 ********************************************************************/

/*
 *@@ BSCfgSysManip(char* pszConfigSys):
 *      this constructor translates a CONFIGSYS attribute (as used
 *      in the PCK tag and stored in the database) into a CONFIGSYSITEM
 *      structure.
 *
 *      The BSCfgSysManip class is designed for use with the
 *      BSConfigSys class and describes manipulations to be done upon
 *      the CONFIG.SYS file (represented by the BSConfigSys class).
 *
 *      Manipulating CONFIG.SYS works just as follows:
 *
 *      1)  Create an instance of BSConfigSys:
 *
 +              BSConfigSys *pConfigSys = new BSConfigSys;
 *
 *          This will load your CONFIG.SYS file into the instance's memory.
 *
 *      2)  Create an instance of BSCfgSysManip and specify the manipulation
 *          in the constructor:
 *
 +              BSCfgSysManip *pManip = new BSCfgSysManip("SET TEST=YES | UNIQUE");
 *
 *          The BSCfgSysManip class has a single constructor which takes a PSZ
 *          (char*) as input. This input string has exactly the format like
 *          with the CONFIGSYS attribute to the PCK tag in WarpIN installation
 *          scripts, like this (see the WarpIN Programmer's Reference for details):
 *
 +              statement [| modifiers]
 *
 *          "modifiers" can be one of the following:
 +              [UNIQUE[(xxx)]] [vertical]
 +              ADDRIGHT [vertical]
 +              ADDLEFT [vertical]
 +              REMOVELINE
 +              REMOVEPART
 *          with "xxx" being a search string.
 *
 *          "vertical" can be one of the following:
 +              ADDTOP
 +              ADDAFTER(xxx)
 +              ADDBEFORE(xxx)
 *          with "xxx" being a search string.
 *
 *          Examples:
 +              "BASEDEV=NEWDRIVR.SYS /WHATEVER | UNIQUE ADDAFTER(IBM1S506.ADD)"
 +              "SET PATH=C:\BLAH | ADDRIGHT"
 *
 *          After this constructor has successfully converted pszConfigSys,
 *          all the instance data is valid (see cfgsys.h).
 *
 *          However, this does _not_ handle macro resultion like in WarpIn
 *          scripts (which is done in warpin.cpp before calling the Manipulate
 *          method), because in this file scope we know nothing about the
 *          PackageInfo instances.
 *
 *      3)  Create a CfgSysDone logger instance:
 +              BSCfgSysDoneLogger logger;
 *
 *      4)  Invoke the BSConfigSys::Manipulate method with the BSCfgSysManip
 *          instance (this will add data to the logger):
 +              pConfigSys->Manipulate(logger, pManip);
 *
 *      5)  Write CONFIG.SYS back to disk and clean up.
 +              pConfigSys->Flush(TRUE);
 +              delete pBSCfgSysManip;
 +              delete pConfigSys;
 *
 *      Now, if you want to undo the changes later, call the static
 *      AddToCfgSysManipUndoList method with the logger passed to
 *      to the various constructor calls (above), which will create
 *      a list of BSCfgSysManip instance which will be able to undo
 *      the changes again:
 +          list<BSCfgSysManip*> UndoList;
 +          AddToCfgSysManipUndoList(UndoList, logger);
 *      and iterate over the list and call BSConfigSys::Manipulate with
 *      the objects on that list to undo the changes.
 *
 *      Throws:
 *      -- BSConfigExcpt.
 *
 *@@changed V0.9.1 (2000-01-06) [umoeller]: added UNIQUE(xxx) support
 *@@changed V0.9.5 (2000-08-26) [umoeller]: fixed UNIQUE(xxx) and REMOVELINE, which deleted wrong lines
 */

BSCfgSysManip::BSCfgSysManip(const char *pcszConfigSys)
               : BSConfigBase(CFGT_CFGSYS)
{
    if (!pcszConfigSys)
        throw(BSConfigExcpt(CFGEXCPT_SYNTAX, 0));

    // initialize all fields to zero
    memset(&_ConfigManip, 0, sizeof(_ConfigManip));
            // V0.9.12 (2001-05-22) [umoeller]

    // now check if we have modifiers
    PSZ     pSep = strchr(pcszConfigSys, '|');
    if (!pSep)
        // no modifiers: just copy
        _ConfigManip.pszNewLine = strdup(pcszConfigSys);
    else
    {
        // we do have modifiers:
        BOOL    fVerticalsAllowed = TRUE;

        // get rid of spaces before '|'
        PSZ     pSep2 = pSep-1;
        while (*pSep2 == ' ')
            pSep2--;

        // get the "statement" part
        _ConfigManip.pszNewLine = strhSubstr(pcszConfigSys, pSep2 + 1);

        // remember the modifiers position
        PSZ     pModifiers = pSep + 1;
        BOOL    fReplaceModeFound = FALSE;
        // now check for the replace mode;
        // the default is 0 (CFGRPL_ADD)
        _ConfigManip.iReplaceMode = CFGRPL_ADD;
        fVerticalsAllowed = TRUE;

        PSZ pszUnique;
        if ((pszUnique = strhistr(pModifiers, "UNIQUE")))
        {
            _ConfigManip.iReplaceMode = CFGRPL_UNIQUE;
            fVerticalsAllowed = TRUE;
            fReplaceModeFound = TRUE;   // block future replacement modifiers

            // check if we have a "UNIQUE(xxx)" syntax
            _ConfigManip.pszUniqueSearchString2 = strhExtract(pszUnique, '(', ')', NULL);
                // if found, this extracts the stuff between ( and );
                // if not, this returns NULL
        }
        if (strhistr(pModifiers, "ADDRIGHT"))
        {
            if (!fReplaceModeFound)
            {
                _ConfigManip.iReplaceMode = CFGRPL_ADDRIGHT;
                fVerticalsAllowed = TRUE;
                fReplaceModeFound = TRUE; // block future replacement modifiers
            }
            else
                // double replace mode found:
                throw(BSConfigExcpt(CFGEXCPT_DOUBLEREPLACEMODE,
                                    strdup(pModifiers)));
        }
        if (strhistr(pModifiers, "ADDLEFT"))
        {
            if (!fReplaceModeFound)
            {
                _ConfigManip.iReplaceMode = CFGRPL_ADDLEFT;
                fVerticalsAllowed = TRUE;
                fReplaceModeFound = TRUE; // block future replacement modifiers
            }
            else
                // double replace mode found:
                throw(BSConfigExcpt(CFGEXCPT_DOUBLEREPLACEMODE,
                                    strdup(pModifiers)));
        }
        if (strhistr(pModifiers, "REMOVELINE"))
        {
            if (!fReplaceModeFound)
            {
                _ConfigManip.iReplaceMode = CFGRPL_REMOVELINE;
                fVerticalsAllowed = FALSE;
                fReplaceModeFound = TRUE; // block future replacement modifiers

                // check also for stuff after "="
                PSZ p = strchr(_ConfigManip.pszNewLine, '=');
                if (p)
                    _ConfigManip.pszUniqueSearchString2 = strdup(p + 1);
            }
            else
                // double replace mode found:
                throw(BSConfigExcpt(CFGEXCPT_DOUBLEREPLACEMODE,
                                  strdup(pModifiers)));
        }
        if (strhistr(pModifiers, "REMOVEPART"))
        {
            if (!fReplaceModeFound)
            {
                _ConfigManip.iReplaceMode = CFGRPL_REMOVEPART;
                fVerticalsAllowed = FALSE;
                fReplaceModeFound = TRUE;
            }
            else
                // double replace mode found:
                throw(BSConfigExcpt(CFGEXCPT_DOUBLEREPLACEMODE,
                                  strdup(pModifiers)));
        }

        // now parse vertical modifiers
        _ConfigManip.iVertical = CFGVRT_BOTTOM;      // default
        BOOL fVerticalFound = FALSE;
        if (strhistr(pModifiers, "ADDTOP"))
        {
            _ConfigManip.iVertical = CFGVRT_TOP;
            fVerticalFound = TRUE;
        }
        PSZ p2;
        if ((p2 = strhistr(pModifiers, "ADDAFTER(")))
        {
            if (!fVerticalFound)
            {
                PSZ pEndOfSearch = strchr(p2, ')');
                if (!pEndOfSearch)
                    throw(BSConfigExcpt(CFGEXCPT_INVALIDSEARCH, strdup(p2)));
                else
                    _ConfigManip.pszVerticalSearchString
                            = strhSubstr(p2 + 9,        // strlen("ADDAFTER(")
                                         pEndOfSearch); // excluding that char

                _ConfigManip.iVertical = CFGVRT_AFTER;
                fVerticalFound = TRUE;
            }
            else
                throw(BSConfigExcpt(CFGEXCPT_INVALIDVERTICAL, strdup(p2)));
        }
        if ((p2 = strhistr(pModifiers, "ADDBEFORE(")))
        {
            if (!fVerticalFound)
            {
                PSZ pEndOfSearch = strchr(p2, ')');
                if (!pEndOfSearch)
                    throw(BSConfigExcpt(CFGEXCPT_INVALIDSEARCH, strdup(p2)));
                else
                    _ConfigManip.pszVerticalSearchString = strhSubstr(p2 + 10,       // strlen("ADDBEFORE(")
                                                          pEndOfSearch); // excluding that char

                _ConfigManip.iVertical = CFGVRT_BEFORE;
                fVerticalFound = TRUE;
            }
            else
                throw(BSConfigExcpt(CFGEXCPT_INVALIDVERTICAL, strdup(p2)));
        }

        // finally check if vertical modifier is allowed at all
        if ( (fVerticalFound) && (!fVerticalsAllowed) )
            throw(BSConfigExcpt(CFGEXCPT_INVALIDVERTICAL, strdup(pModifiers)));

    } // end elseif (!pSep)
}

/*
 *@@ ~BSCfgSysManip:
 *      destructor.
 *
 *@@changed V0.9.1 (2000-01-06) [umoeller]: added UNIQUE(xxx) support
 */

BSCfgSysManip::~BSCfgSysManip()
{
    if (_ConfigManip.pszNewLine)
        free(_ConfigManip.pszNewLine);
    if (_ConfigManip.pszVerticalSearchString)
        free(_ConfigManip.pszVerticalSearchString);
    if (_ConfigManip.pszUniqueSearchString2)
        free(_ConfigManip.pszUniqueSearchString2);
}

/*
 *@@ DescribeType:
 *      describes the current manipulator type to the GUI.
 *
 *@@added V0.9.2 (2000-02-19) [umoeller]
 */

const char* BSCfgSysManip::DescribeType()
{
    return ("CONFIG.SYS manipulation");
}

/*
 *@@ DescribeData:
 *      describes the current manipulator data to the GUI.
 *
 *@@added V0.9.2 (2000-02-19) [umoeller]
 *@@changed V0.9.5 (2000-08-26) [umoeller]: UNIQUE wasn't reported right
 */

BSString BSCfgSysManip::DescribeData()
{
    BSString    str;
    BOOL        fAddSpace = TRUE;

    switch (_ConfigManip.iReplaceMode)
    {
        case CFGRPL_UNIQUE:
            str += "UNIQUE";
            if (_ConfigManip.pszUniqueSearchString2)
            {
                str += "(";
                str += _ConfigManip.pszUniqueSearchString2;
                str += ")";
            }
        break;

        case CFGRPL_ADDLEFT: str += "ADDLEFT"; break;
        case CFGRPL_ADDRIGHT: str += "ADDRIGHT"; break;
        case CFGRPL_REMOVELINE: str += "REMOVELINE"; break;
        case CFGRPL_REMOVEPART: str += "REMOVEPART"; break;

        default:
            fAddSpace = FALSE;
        break;
    }

    if (fAddSpace)
        str += " ";

    switch (_ConfigManip.iVertical)
    {
        case CFGVRT_TOP: str += "ADDTOP "; break;

        case CFGVRT_BEFORE:
            str += "BEFORE(";
            str += _ConfigManip.pszVerticalSearchString;
            str += ") ";
        break;

        case CFGVRT_AFTER:
            str += "AFTER ";
            str += _ConfigManip.pszVerticalSearchString;
            str += ") ";
        break;
    }

    str += _ConfigManip.pszNewLine;

    return (str);
}

/* ******************************************************************
 *
 *  BSConfigSys class
 *
 ********************************************************************/

/*
 *@@ BSConfigSys:
 *      the constructor, which loads the current CONFIG.SYS
 *      file from OS/2 boot drive into instance memory.
 *
 *      Throws:
 *      -- BSConfigExcpt.
 *
 *@@changed V0.9.7 (2001-01-15) [umoeller]: now using csysLoadConfigSys
 */

BSConfigSys::BSConfigSys()
{
    _pszContent = NULL;

    /* sprintf(_szFilename, "%c:\\config.sys", doshQueryBootDrive());
    // now read CONFIG.SYS file to initialize the dlg items
    APIRET arc = doshLoadTextFile(_szFilename, &_pszContent); */

    APIRET arc = csysLoadConfigSys(NULL,        // default CONFIG.SYS
                                   &_pszContent);

    if (arc != NO_ERROR)
        throw(BSConfigExcpt(CFGEXCPT_OPEN, arc));

    _fDirty = FALSE;
}

/*
 *@@ ~ConfigSys:
 *      the destructor.
 *
 *      This does _not_ write the file. Use BSConfigSys::Flush()
 *      before deleting an instance of this.
 */

BSConfigSys::~BSConfigSys()
{
    if (_pszContent)
        free(_pszContent);
}

/*
 *@@ Manipulate:
 *      this monster method changes the data in our memory copy of
 *      CONFIG.SYS according to the BSCfgSysManip object, which you must
 *      have created before. See BSCfgSysManip::BSCfgSysManip for a
 *      description of this usage.
 *
 *      This also takes a BSCfgSysDoneLogger as input where all the
 *      changes made to the CONFIG.SYS memory copy are logged. This
 *      logger can then be passed to BSConfigSys::Undo to have the
 *      changes undone again.
 *
 *      Call BSConfigSys::Flush to have CONFIG.SYS written back to disk.
 *
 *      Returns NO_ERROR if everything was OK or an error code.
 */

int BSConfigSys::Manipulate(BSCfgSysManip &Manip,       // in: manipulator object
                            BSCfgSysDoneLogger *pLogger, // in: logger object to append items to (can be NULL)
                            BSFileLogger *pLogFile)     // in: file logger (log file), can be NULL
{
    int         irc = 0;

    XSTRING     strLogger;
    xstrInit(&strLogger, 0);

    irc = csysManipulate(&_pszContent,
                         &Manip._ConfigManip,
                         &_fDirty,
                         &strLogger);

    // now parse the logger...
    if (strLogger.ulLength)
    {
        // something to be logged:
        PSZ p = strLogger.psz;

        while (p)
        {
            PSZ pEOL = strhFindEOL(p, NULL);
            if (pEOL && *pEOL)
            {
                BSString strTemp(p, pEOL);

                if (strTemp())
                {
                    if (pLogFile)
                        // write to log file
                        pLogFile->Append("Updated CONFIG.SYS: %s\n",
                                         strTemp.c_str());

                    if (pLogger)
                        // write to "done" logger:
                        pLogger->Append(strTemp);
                }
                else
                    break;

                p = pEOL + 1;
            }
            else
                break;
        }
    }

    xstrClear(&strLogger);

    return (irc);
}

/*
 *@@ Flush:
 *      this rewrites the CONFIG.SYS file on disk with the
 *      data we have in memory.
 *
 *      This makes a backup copy in "CONFIG.003" style if
 *      (pszBackup != NULL). See doshCreateBackupFileName
 *      for details.
 *
 *      Returns:
 *
 *      --  0: contents were dirty, written back to disk.
 *
 *      --  1: no error, but contents were clean, so no write-back
 *             was needed.
 *
 *      Throws:
 *      -- BSConfigExcpt(CFGEXCPT_WRITE, APIRET).
 *
 *@@changed V0.9.3 (2000-05-03) [umoeller]: added more error checking
 *@@changed V0.9.3 (2000-05-12) [umoeller]: added pszBackup
 *@@changed V0.9.6 (2000-10-27) [umoeller]: added check if contents are dirty
 */

int BSConfigSys::Flush(PSZ pszBackup,           // in/out: create backup?
                       BSFileLogger *pLogFile)  // in: file logger (log file), can be NULL
                 const
{
    int irc = 0;

    if (_fDirty)
    {
        APIRET arc = csysWriteConfigSys(NULL,       // default _szFilename,
                                        _pszContent,
                                        pszBackup);        // create backup

        if (arc)
        {
            if (pLogFile)
                pLogFile->Append("Error %d occured writing CONFIG.SYS back to disk.\n",
                                 arc);

            throw(BSConfigExcpt(CFGEXCPT_WRITE, arc));
        }

        if (pLogFile)
            pLogFile->Append("CONFIG.SYS file written back to disk, backup is %s.\n",
                             pszBackup);
    }
    else
        irc = 1;

    return (irc);
}

/* ******************************************************************
 *
 *  BSRegisterClass class
 *
 ********************************************************************/

/*
 *@@ BSRegisterClass:
 *      this constructor translates a REGISTERCLASS attribute (as used
 *      in the PCK tag and stored in the database) into the BSRegisterClass
 *      instance data.
 *
 *      Throws:
 *      -- BSConfigExcpt.
 */

BSRegisterClass::BSRegisterClass(const char *pcszRegisterClass)
                 : BSConfigBase(CFGT_REGISTERCLASS)
{
    _pszClassName = 0;
    _pszModule = 0;

    // find separator
    PSZ pBegin = (PSZ)pcszRegisterClass;
    PSZ pEnd = strchr(pBegin, '|');
    if (pEnd)
    {
        // strip trailing spaces
        PSZ pEnd2 = pEnd;
        while ((*pEnd2 == ' ') && (pEnd2 > pcszRegisterClass))
            pEnd2--;

        _pszClassName = strhSubstr(pBegin, pEnd2);
        if (!_pszClassName)
            throw(BSConfigExcpt(REGEXCPT_SYNTAX, strdup(pcszRegisterClass)));
        else
        {
            pBegin = pEnd + 1;
            // strip leading spaces
            while ((*pBegin) && (*pBegin == ' '))
                pBegin++;
            _pszModule = strdup(pBegin);
            if (!_pszModule)
                throw(BSConfigExcpt(REGEXCPT_SYNTAX, strdup(pcszRegisterClass)));
        }
    }
    else
        throw(BSConfigExcpt(REGEXCPT_SYNTAX, strdup(pcszRegisterClass)));
}

/*
 *@@ ~BSRegisterClass:
 *      the destructor.
 */

BSRegisterClass::~BSRegisterClass()
{
    if (_pszClassName)
        free(_pszClassName);
    if (_pszModule)
        free(_pszModule);
}

/*
 *@@ DescribeType:
 *      describes the current manipulator type to the GUI.
 *
 *@@added V0.9.9 (2001-03-27) [umoeller]
 */

const char* BSRegisterClass::DescribeType()
{
    return ("WPS class registration");
}

/*
 *@@ DescribeData:
 *      describes the current manipulator data to the GUI.
 *
 *@@added V0.9.9 (2001-03-27) [umoeller]
 */

BSString BSRegisterClass::DescribeData()
{
    BSString    str = _pszClassName;
    str += " in ";
    str += _pszModule;

    return (str);
}

/*
 *@@ IsRegistered:
 *      this returns TRUE if a class of the same WPS class name
 *      as this instance is already registered with the WPS.
 *      In that case, the module of the registered class is
 *      copied into pszBuffer, which should be sufficient in size.
 *
 *      This throws a BSConfigExcpt with REGEXCPT_QUERYCLASSLIST
 *      if the class list could not be queried.
 */

bool BSRegisterClass::IsRegistered(char* pszBuffer)
                      const
{
    BOOL    fInstalled = FALSE;
    PBYTE   pClassList = winhQueryWPSClassList(),
            pClassThis = 0;
    if (pClassList)
    {
        if ((pClassThis = winhQueryWPSClass(pClassList, _pszClassName)))
        {
            fInstalled = TRUE;
            if (pszBuffer)
                strcpy(pszBuffer, ((POBJCLASS)pClassThis)->pszModName);
        }
        free(pClassList);
    }
    else
        throw(BSConfigExcpt(REGEXCPT_QUERYCLASSLIST, 0));

    return (fInstalled);
}

/*
 *@@ Register:
 *      this attempts to register the class.
 *
 *      If (fReplace == TRUE), we do not call
 *      BSRegisterClass::IsRegistered before registering
 *      this, i.e. we will always register the class,
 *      even if it's already registered.
 *
 *      If (fReplace == FALSE) and BSRegisterClass::IsRegistered
 *      returned TRUE, we throw a BSConfigExcpt with
 *      REGEXCPT_ALREADYREGISTERED and pszSubstr set to the
 *      file name of the registered DLL. Note that IsRegistered might
 *      in turn throw BSConfigExcpt with REGEXCPT_QUERYCLASSLIST.
 *
 *      If registering the class failed, this throws a
 *      BSConfigExcpt with REGEXCPT_REGISTER and iData set
 *      to the APIRET of winhRegisterClass.
 *
 *      This method can take a BSDoneLoggerBase object as input, which
 *      can later be used with the AddToDeregisterClassUndoList
 *      static class method to easily create a list of BSDeregisterClass
 *      objects to undo the changes made.
 *
 *      Example usage (exception handling omitted):
 +          BSRegisterDoneLogger logger;
 +                      // create logger instance
 +          BSRegisterClass RegClass("XFolder;C:\XFOLDER\XFLDR.DLL");
 +                      // create BSRegisterClass instance:
 +          RegClass.Register(TRUE, logger);
 +          ... // register more classes with the same logger
 +
 +          // now create undo list
 +          list<DeregisterClass*> List;
 +          AddToDeregisterClassUndoList(List, logger);
 +          list<DeregisterClass*>::iterator DeregStart = List.begin(),
 +                                           DeregEnd = List.end();
 +          for (; DeregStart != DeregEnd; DeregStart++)
 +              (**DeregStart).Deregister;
 */

int BSRegisterClass::Register(bool fReplace,
                              BSRegisterDoneLogger *pLogger, // in: logger object to append items to (can be NULL)
                              BSFileLogger *pLogFile)     // in: file logger (log file), can be NULL
                     const
{
    int irc = 0;

    if (fReplace == FALSE)
    {
        char szBuffer[300];
        if (IsRegistered(szBuffer))
            throw(BSConfigExcpt(REGEXCPT_ALREADYREGISTERED, strdup(szBuffer)));
    }

    CHAR   szBuf[1000];
    APIRET arc = winhRegisterClass(_pszClassName,
                                   _pszModule,
                                   szBuf,
                                   sizeof(szBuf));
    // always record in logger, even if failed,
    // because the class is in the class list anyway
    if (pLogger)
        pLogger->Append(_pszClassName);

    if (arc != NO_ERROR)
        throw(BSConfigExcpt(REGEXCPT_REGISTER, arc));

    return (irc);
}

/* ******************************************************************
 *
 *  BSDeregisterClass class
 *
 ********************************************************************/

/*
 *@@ BSDeregisterClass:
 *      the constructor, which in this case takes a simple
 *      PSZ class name as input.
 */

BSDeregisterClass::BSDeregisterClass(const char *pcszDeregisterClass)
                   : BSConfigBase(CFGT_DEREGISTERCLASS)
{
    _pszClassName = strdup(pcszDeregisterClass);
}

/*
 *@@ ~BSDeregisterClass:
 *      the destructor.
 */

BSDeregisterClass::~BSDeregisterClass()
{
    if (_pszClassName)
        free(_pszClassName);
}

/*
 *@@ DescribeType:
 *      describes the current manipulator type to the GUI.
 *
 *@@added V0.9.2 (2000-02-19) [umoeller]
 */

const char* BSDeregisterClass::DescribeType()
{
    return ("WPS class deregistration");
}

/*
 *@@ DescribeData:
 *      describes the current manipulator data to the GUI.
 *
 *@@added V0.9.2 (2000-02-19) [umoeller]
 */

BSString BSDeregisterClass::DescribeData()
{
    BSString    str = _pszClassName;

    return (str);
}

/*
 *@@ Deregister:
 *      this attempts to deregister the class.
 *
 *      Throws:
 *      -- BSConfigExcpt.
 */

int BSDeregisterClass::Deregister(BSFileLogger *pLogFile)       // in: file logger (log file), can be NULL
                       const
{
    int irc = 0;

    if (!WinDeregisterObjectClass(_pszClassName))
    {
        if (pLogFile)
            pLogFile->Append("An error occured deregistering WPS class %s.\n",
                             _pszClassName);

        throw(BSConfigExcpt(REGEXCPT_DEREGISTER, 0));
    }

    if (pLogFile)
        pLogFile->Append("Deregistered WPS class %s.\n",
                         _pszClassName);

    return (irc);
}

/* ******************************************************************
 *
 *  BSReplaceClass class
 *
 ********************************************************************/

/*
 *@@ BSReplaceClass:
 *      this constructor translates a REPLACECLASS attribute (as used
 *      in the PCK tag and stored in the database) into the BSReplaceClass
 *      instance data.
 */

BSReplaceClass::BSReplaceClass(const char *pcszReplaceClass)
                : BSConfigBase(CFGT_REPLACECLASS)
{
    _pszOldClassName = 0;
    _pszNewClassName = 0;

    // find separator
    PSZ pBegin = (PSZ)pcszReplaceClass;
    PSZ pEnd = strchr(pBegin, '|');
    if (pEnd)
    {
        // strip trailing spaces
        PSZ pEnd2 = pEnd;
        while ((*pEnd2 == ' ') && (pEnd2 > pcszReplaceClass))
            pEnd2--;

        _pszOldClassName = strhSubstr(pBegin, pEnd2);
        if (!_pszOldClassName)
            throw(BSConfigExcpt(REGEXCPT_SYNTAX, strdup(pcszReplaceClass)));
        else
        {
            pBegin = pEnd + 1;
            // strip leading spaces
            while ((*pBegin) && (*pBegin == ' '))
                pBegin++;
            _pszNewClassName = strdup(pBegin);
            if (!_pszNewClassName)
                throw(BSConfigExcpt(REGEXCPT_SYNTAX, strdup(pcszReplaceClass)));
        }
    }
    else
        throw(BSConfigExcpt(REGEXCPT_SYNTAX, strdup(pcszReplaceClass)));
}

/*
 *@@ ~BSReplaceClass:
 *      the destructor.
 */

BSReplaceClass::~BSReplaceClass()
{
    if (_pszOldClassName)
        free(_pszOldClassName);
    if (_pszNewClassName)
        free(_pszNewClassName);
}

/*
 *@@ DescribeType:
 *      describes the current manipulator type to the GUI.
 *
 *@@added V0.9.2 (2000-02-19) [umoeller]
 */

const char* BSReplaceClass::DescribeType()
{
    return ("WPS class replacement");
}

/*
 *@@ DescribeData:
 *      describes the current manipulator data to the GUI.
 *
 *@@added V0.9.2 (2000-02-19) [umoeller]
 *@@changed V0.9.5 (2000-08-26) [umoeller]: UNIQUE wasn't reported right
 */

BSString BSReplaceClass::DescribeData()
{
    BSString    str = _pszOldClassName;
    str += " with ";
    str += _pszNewClassName;

    return (str);
}

/*
 *@@ Replace:
 *      this attempts to replace the class or undo
 *      an existing replacement (if fReplace == FALSE).
 *
 *      If replacing the class failed, this throws a
 *      BSConfigExcpt with REGEXCPT_REPLACE.
 *
 *      Example usage (exception handling omitted):
 +          BSReplaceDoneLogger logger;
 +                      // create logger instance
 +          BSReplaceClass ReplClass("WPFolder;XFolder");
 +                      // create BSReplaceClass instance:
 +          ReplClass.Replace(TRUE, logger);
 +          ... // replace more classes with the same logger
 +
 +          // now create undo list
 +          list<UnreplaceClass*> List;
 +          AddToUnreplaceClassUndoList(List, logger);
 +          list<UnreplaceClass*>::iterator UnrplStart = List.begin(),
 +                                          UnrplEnd = List.end();
 +          for (; UnrplStart != UnrplEnd; UnrplStart++)
 +              (**UnrplStart).Unreplace;
 *
 *@@changed V0.9.1 (2000-01-05) [umoeller]: finally added logging, which was missing
 *@@changed V0.9.9 (2001-03-30) [umoeller]: fixed wrong logging if replace failed
 */

int BSReplaceClass::Replace(BSReplaceDoneLogger *pLogger,   // in: logger object to append items to (can be NULL)
                            BSFileLogger *pLogFile)     // in: file logger (log file), can be NULL
                    const
{
    int irc = 0;

    BOOL brc = WinReplaceObjectClass(_pszOldClassName, _pszNewClassName,
                                     TRUE);       // replace
    // always record in logger, even if failed,
    // because the class is in the class list anyway
    // V0.9.9 (2001-03-30) [umoeller]
    if (pLogger)
    {
        BSString str2Append(_pszOldClassName);
        str2Append += "|";
        str2Append += _pszNewClassName;
        pLogger->Append(str2Append.c_str());
    }


    if (!brc)
    {
        if (pLogFile)
            pLogFile->Append("An error occured replacing WPS class %s with %s.\n",
                             _pszOldClassName,
                             _pszNewClassName);

        throw(BSConfigExcpt(REGEXCPT_REPLACE, 0));
    }

    // success:
    if (pLogFile)
        pLogFile->Append("Replaced WPS class %s with %s.\n",
                         _pszOldClassName,
                         _pszNewClassName);

    return (irc);
}

/* ******************************************************************
 *
 *  BSUnreplaceClass class
 *
 ********************************************************************/

/*
 *@@ DescribeType:
 *      describes the current manipulator type to the GUI.
 *
 *@@added V0.9.2 (2000-02-19) [umoeller]
 */

const char* BSUnreplaceClass::DescribeType()
{
    return ("WPS class un-replacement");
}

/*
 *@@ DescribeData:
 *      describes the current manipulator data to the GUI.
 *
 *@@added V0.9.2 (2000-02-19) [umoeller]
 */

BSString BSUnreplaceClass::DescribeData()
{
    BSString    str = _pszOldClassName;
    str += " with ";
    str += _pszNewClassName;

    return (str);
}

/*
 *@@ Unreplace:
 *      this attempts to unreplace the class.
 *      Throws a BSConfigExcpt upon failure.
 */

int BSUnreplaceClass::Unreplace(BSFileLogger *pLogFile) // in: file logger (log file), can be NULL
                      const
{
    int irc = 0;

    if (!WinReplaceObjectClass(_pszOldClassName, _pszNewClassName, FALSE))
    {
        if (pLogFile)
            pLogFile->Append("Error undoing WPS class replacement of %s with %s.\n",
                             _pszOldClassName,
                             _pszNewClassName);

        throw(BSConfigExcpt(REGEXCPT_UNREPLACE, 0));
    }

    // success:
    if (pLogFile)
        pLogFile->Append("Undid WPS class replacement of %s with %s.\n",
                         _pszOldClassName,
                         _pszNewClassName);

    return (irc);
}

/* ******************************************************************
 *
 *  BSCreateWPSObject class
 *
 ********************************************************************/

/*
 *@@ BSCreateWPSObject:
 *      this constructor translates a CREATEOBJECT attribute (as used
 *      in the PCK tag and stored in the database) into the BSCreateWPSObject
 *      instance data.
 *
 *      Throws:
 *      -- BSConfigExcpt.
 *
 *@@changed V0.9.3 (2000-04-08) [umoeller]: "REPLACE" wasn't evaluated; fixed. Thanks, Cornelis Bockemhl.
 */

BSCreateWPSObject::BSCreateWPSObject(const char *pcszCreateObject)
                   : BSConfigBase(CFGT_CREATEOBJECT)
{
    // extract class name
    PSZ pBegin = (PSZ)pcszCreateObject;

    if (strcmp(pBegin, "REPLACE ") == 0)
    {
        // added V0.9.3 (2000-04-08) [umoeller]
        _fReplace = TRUE;
        pBegin += strlen("REPLACE");
        while (   (*pBegin)
               && (*pBegin == ' ')
              )
            pBegin++;
    }
    else
        _fReplace = FALSE;

    PSZ pEnd = strchr(pBegin, '|');
    if (pEnd)
    {
        _pszClassName = strhSubstr(pBegin, pEnd);

        // extract title
        pBegin = pEnd+1;
        pEnd = strchr(pBegin, '|');
        if (pEnd)
        {
            _pszTitle = strhSubstr(pBegin, pEnd);

            // extract location
            pBegin = pEnd+1;
            pEnd = strchr(pBegin, '|');
            if (pEnd)
            {
                // yet another separator found: we then have a config parameter
                _pszLocation = strhSubstr(pBegin, pEnd);

                // extract config (rest of pszCreateObject)
                pBegin = pEnd+1;
                _pszSetupString = strdup(pBegin);
            }
            else
                // no separator after "location" found:
                // duplicate whole string
                      _pszLocation = strdup(pBegin);

            if (!_pszLocation)
                throw(BSConfigExcpt(WPOEXCPT_NOLOCATION, strdup(pcszCreateObject)));
        }
        else
            throw(BSConfigExcpt(WPOEXCPT_NOTITLE, strdup(pcszCreateObject)));
    }
    else
        throw(BSConfigExcpt(WPOEXCPT_NOCLASS, strdup(pcszCreateObject)));
}

/*
 *@@ ~BSCreateWPSObject:
 *      the destructor.
 */

BSCreateWPSObject::~BSCreateWPSObject()
{
    if (_pszClassName)
        free(_pszClassName);
    if (_pszTitle)
        free(_pszTitle);
    if (_pszSetupString)
        free(_pszSetupString);
    if (_pszLocation)
        free(_pszLocation);
}

/*
 *@@ DescribeType:
 *      describes the current manipulator type to the GUI.
 *
 *@@added V0.9.9 (2001-03-27) [umoeller]
 */

const char* BSCreateWPSObject::DescribeType()
{
    return ("WPS object");
}

/*
 *@@ DescribeData:
 *      describes the current manipulator data to the GUI.
 *
 *@@added V0.9.9 (2001-03-27) [umoeller]
 */

BSString BSCreateWPSObject::DescribeData()
{
    BSString    str = _pszClassName;
    str += " ";
    str += _pszTitle;
    str += " ";
    str += _pszLocation;
    str += " ";
    str += _pszSetupString;

    return (str);
}

/*
 *@@ CreateObject:
 *      this attempts to create the WPS object
 *      with the instance data.
 *
 *      Throws a BSConfigExcpt with WPOEXCPT_CREATE
 *      if this fails.
 */

void BSCreateWPSObject::CreateObject(BSWPSObjectsDoneLogger *pLogger, // in: logger object to append items to (can be NULL)
                                     BSFileLogger *pLogFile)    // in: file logger (log file), can be NULL
{
    _hobj = WinCreateObject(_pszClassName,
                            _pszTitle,
                            _pszSetupString,
                            _pszLocation,
                            (_fReplace)
                               ? CO_REPLACEIFEXISTS
                               : CO_UPDATEIFEXISTS);
    if (_hobj == NULLHANDLE)
    {
        if (pLogFile)
            pLogFile->Append("Error creating WPS object \"%s\", class \"%s\", location \"%s\", setup \"%s\".\n",
                             _pszTitle,
                             _pszClassName,
                             _pszLocation,
                             _pszSetupString);

        throw(BSConfigExcpt(WPOEXCPT_CREATE, strdup(_pszTitle)));
    }

    if (pLogFile)
    {
        pLogFile->Append("Created WPS object \"%s\", class \"%s\", location \"%s\", setup \"%s\".\n",
                         _pszTitle,
                         _pszClassName,
                         _pszLocation,
                         _pszSetupString);
        pLogFile->Append("HOBJECT is 0x%lX.\n", _hobj);
    }

    // store the object ID in the logger
    PSZ pObjectID = strhistr(_pszSetupString, "OBJECTID=<");
    if (pObjectID)
    {
        PSZ pBegin = pObjectID + 9;
        PSZ pEnd = strchr(pBegin, '>');
        if (pEnd)
        {
            PSZ pszObjectID = strhSubstr(pBegin, pEnd+1);
            if (pLogger)
                pLogger->Append(pszObjectID);
            free(pszObjectID);
        }
    }
}

/* ******************************************************************
 *
 *  BSDeleteWPSObject class
 *
 ********************************************************************/

/*
 *@@ BSDeleteWPSObject:
 *      the constructor, which in this case takes a simple
 *      PSZ object ID as input.
 */

BSDeleteWPSObject::BSDeleteWPSObject(const char *pcszID2Delete)
                   : BSConfigBase(CFGT_DELETEOBJECT)
{
    _pszObjectID = strdup(pcszID2Delete);
}

/*
 *@@ ~BSDeleteWPSObject:
 *      the destructor.
 */

BSDeleteWPSObject::~BSDeleteWPSObject()
{
    if (_pszObjectID)
        free(_pszObjectID);
}

/*
 *@@ DescribeType:
 *      describes the current manipulator type to the GUI.
 *
 *@@added V0.9.2 (2000-02-19) [umoeller]
 */

const char* BSDeleteWPSObject::DescribeType()
{
    return ("WPS object deletion");
}

/*
 *@@ DescribeData:
 *      describes the current manipulator data to the GUI.
 *
 *@@added V0.9.2 (2000-02-19) [umoeller]
 */

BSString BSDeleteWPSObject::DescribeData()
{
    BSString    str = _pszObjectID;

    return (str);
}

/*
 *@@ Delete:
 *      this attempts to delete the object.
 *      Throws a BSConfigExcpt upon failure.
 */

int BSDeleteWPSObject::Delete(BSFileLogger *pLogFile)   // in: file logger (log file), can be NULL
                       const
{
    int irc = 0;

    HOBJECT     hobj = WinQueryObject(_pszObjectID);
    if (!hobj)
    {
        if (pLogFile)
            pLogFile->Append("Error deleting WPS object \"%s\".\n",
                             _pszObjectID);

        throw(BSConfigExcpt(WPOEXCPT_DELETEOBJECT, 0));
    }

    if (!WinDestroyObject(hobj))
    {
        if (pLogFile)
            pLogFile->Append("Error deleting WPS object, HOBJECT was 0x%lX.\n",
                             hobj);

        throw(BSConfigExcpt(WPOEXCPT_DELETEOBJECT, 0));
    }

    return (irc);
}

/* ******************************************************************
 *
 *  BSClearProfile class
 *
 ********************************************************************/

/*
 *@@ BSClearProfile:
 *
 *@@added V0.9.1 (2000-02-07) [umoeller]
 */

BSClearProfile::BSClearProfile(const char *pcszClearProfile)
                : BSConfigBase(CFGT_CLEARPROFILE)
{
    _pszProfile = NULL;
    _pszApplication = NULL;
    _pszKey = NULL;

    // get profile
    PSZ pFirstBackslash = strchr(pcszClearProfile, '\\');
    if (pFirstBackslash)
    {
        // extract profile
        _pszProfile = strhSubstr(pcszClearProfile, pFirstBackslash);

        // get application
        PSZ pSecondBackslash = strchr(pFirstBackslash + 2, '\\');
        if (pSecondBackslash)
        {
            // key specified:
            // extract application up to key
            _pszApplication = strhSubstr(pFirstBackslash + 1, pSecondBackslash);

            // extract key (all the rest)
            _pszKey = strhdup(pSecondBackslash + 1);
        }
        else
            // key not specified:
            _pszApplication = strhdup(pFirstBackslash + 1);
    }

    if (    (_pszProfile == NULL)
         || (_pszApplication == NULL)
       )
        // any error:
        throw(BSConfigExcpt(PRFEXCPT_SYNTAX, strdup(pcszClearProfile)));
}

/*
 *@@ ~BSClearProfile:
 *      destructor.
 *
 *@@added V0.9.1 (2000-02-07) [umoeller]
 */

BSClearProfile::~BSClearProfile()
{
    if (_pszProfile)
        free(_pszProfile);
    if (_pszApplication)
        free(_pszApplication);
    if (_pszKey)
        free(_pszKey);
}

/*
 *@@ DescribeType:
 *      describes the current manipulator type to the GUI.
 *
 *@@added V0.9.2 (2000-02-19) [umoeller]
 */

const char* BSClearProfile::DescribeType()
{
    return ("Profile data");
}

/*
 *@@ DescribeData:
 *      describes the current manipulator data to the GUI.
 *
 *@@added V0.9.2 (2000-02-19) [umoeller]
 */

BSString BSClearProfile::DescribeData()
{
    return (DescribePrfKey());
}

/*
 *@@ DescribePrfKey:
 *      returns a descriptive string describing the
 *      member fields, a la "USER\app\key".
 *
 *@@added V0.9.1 (2000-02-12) [umoeller]
 */

BSString BSClearProfile::DescribePrfKey() const
{
    BSString str(_pszProfile);
    if (_pszApplication)
    {
        str += "\\";
        str += _pszApplication;

        if (_pszKey)
        {
            str += "\\";
            str += _pszKey;
        }
    }

    return (str);
}

/*
 *@@ Clear:
 *      deletes the corresponding profile entry.
 *
 *@@added V0.9.5 (2000-08-26) [umoeller]
 */

int BSClearProfile::Clear(HAB hab,
                          BSFileLogger *pLogFile)   // in: file logger (log file), can be NULL
    const
{
    BOOL    fOK = FALSE;
    HINI    hini = NULLHANDLE,
            hini2Close = NULLHANDLE;

    if (strcmp(_pszProfile, "USER") == 0)
        hini = HINI_USER;
    else if (strcmp(_pszProfile, "SYSTEM") == 0)
        hini = HINI_SYSTEM;
    else
    {
        hini = PrfOpenProfile(hab,
                              _pszProfile);
        hini2Close = hini;
    }

    if (hini != NULLHANDLE)
    {
        ULONG cb = 0;
        // does key exist?
        if (PrfQueryProfileSize(hini,
                                _pszApplication,
                                _pszKey,                // can be NULL
                                &cb))
        {
            // key exists:
            if (PrfWriteProfileString(hini,
                                      _pszApplication,
                                      _pszKey,          // can be NULL
                                      NULL))
                fOK = TRUE;
        }
        else
            // key does not exist:
            fOK = TRUE;
    }

    if (!fOK)
    {
        BSString strClear = DescribePrfKey();
        if (pLogFile)
            pLogFile->Append("Error deleting profile key %s.\n",
                             strClear.c_str());

        // error:
        throw(BSConfigExcpt(PRFEXCPT_PRFERROR, strdup(strClear.c_str())));
    }

    if (pLogFile)
    {
        BSString strClear = DescribePrfKey();
        pLogFile->Append("Deleted profile key %s.\n",
                         strClear.c_str());
    }

    return (1);
}

/* ******************************************************************
 *
 *  BSWriteProfile class
 *
 ********************************************************************/

/*
 *@@ BSWriteProfile:
 *
 *@@added V0.9.1 (2000-02-07) [umoeller]
 */

BSWriteProfile::BSWriteProfile(const char *pcszWriteProfile)
                : BSClearProfile(CFGT_WRITEPROFILE)
{
    _pszKey = NULL;

    PSZ     pSlash = strchr(pcszWriteProfile, '|');
    if (pSlash)
    {
        // copy write data (after '|')
        _pszWriteString = strdup(pSlash + 1);

        // get profile
        PSZ pFirstBackslash = strchr(pcszWriteProfile, '\\');
        if (pFirstBackslash)
        {
            // extract profile
            _pszProfile = strhSubstr(pcszWriteProfile, pFirstBackslash);

            // get application
            PSZ pSecondBackslash = strchr(pFirstBackslash + 2, '\\');
            if (pSecondBackslash)
            {
                // extract application
                _pszApplication = strhSubstr(pFirstBackslash + 1, pSecondBackslash);

                // extract key (up to '|')
                _pszKey = strhSubstr(pSecondBackslash + 1, pSlash);
            }
        }
    }

    if (_pszKey == NULL)
        // any error:
        throw(BSConfigExcpt(PRFEXCPT_SYNTAX, strdup(pcszWriteProfile)));
}

/*
 *@@ ~BSWriteProfile:
 *      destructor.
 *
 *@@added V0.9.1 (2000-02-07) [umoeller]
 */

BSWriteProfile::~BSWriteProfile()
{
    if (_pszWriteString)
        free(_pszWriteString);
}

/*
 *@@ Write:
 *      this actually writes data to a profile, all
 *      as specified by the member variables.
 *
 *      Note that this stores the "done" stuff in
 *      a BSClearPrfAttrsLogger for undoing later.
 *
 *      Throws:
 *      -- BSConfigExcpt.
 *
 *@@added V0.9.1 (2000-02-07) [umoeller]
 */

int BSWriteProfile::Write(HAB hab,      // in: anchor block (for PrfOpenProfile)
                          BSClearPrfAttrsLogger *pLogger,
                          BSFileLogger *pLogFile)   // in: file logger (log file), can be NULL
                    const
{
    BOOL    fOK = FALSE;
    HINI    hini = NULLHANDLE,
            hini2Close = NULLHANDLE;

    if (strcmp(_pszProfile, "USER") == 0)
        hini = HINI_USER;
    else if (strcmp(_pszProfile, "SYSTEM") == 0)
        hini = HINI_SYSTEM;
    else
    {
        hini = PrfOpenProfile(hab,
                              _pszProfile);
        hini2Close = hini;
    }

    if (hini != NULLHANDLE)
    {
        if (PrfWriteProfileString(hini,
                                  _pszApplication,
                                  _pszKey,
                                  _pszWriteString))
        {
            fOK = TRUE;
        }
    }

    // recompose string for BSClearProfile
    BSString strClear = _pszProfile;
    strClear.append("\\");
    strClear.append(_pszApplication);
    strClear.append("\\");
    strClear.append(_pszKey);

    if (fOK)
    {
        if (pLogFile)
            pLogFile->Append("Wrote profile key %s.\n",
                             strClear.c_str());

        if (pLogger)
            pLogger->Append(strClear.c_str());
    }
    else
    {
        // error:
        if (pLogFile)
            pLogFile->Append("Error writing profile key %s.\n",
                             strClear.c_str());

        throw(BSConfigExcpt(PRFEXCPT_PRFERROR, strdup(strClear.c_str())));
    }

    return (1);
}

/* ******************************************************************
 *
 *  BSExecute class
 *
 ********************************************************************/

/*
 *@@ BSExecute:
 *      this constructor takes an EXECUTE attribute (of the PCK tag)
 *      as input and sets up the instance data accordingly.
 *
 *      Throws:
 *      -- BSConfigExcpt.
 *
 *@@added V0.9.1 (2000-02-07) [umoeller]
 *@@changed V0.9.3 (2000-04-08) [umoeller]: "|" was a requirement, contradicting the docs; changed. Thanks, Cornelis Bockemhl.
 *@@changed V0.9.4 (2000-07-10) [umoeller]: _ulExecType wasn't always initialized; thanks Yuri Dario
 *@@changed V0.9.4 (2000-07-26) [umoeller]: now ORing the flags
 *@@changed V0.9.9 (2001-03-27) [umoeller]: fixed EXECUTE parameters
 */

BSExecute::BSExecute(const char *pcszExecute)
           : BSConfigBase(CFGT_EXECUTE)
{
    BOOL    fContextFound = FALSE,
            fContinue = TRUE;
    PSZ     pSearch = (PSZ)pcszExecute;

    _ulExecType = 0;
    _arc = 0;       // NO_ERROR

    while (fContinue)
    {
        // skip leading spaces
        while (    (*pSearch)
                && (*pSearch == ' ')
              )
            pSearch++;

        if (strncmp(pSearch, "CONFIGSYS", 9) == 0)
        {
            _ulExecType |= CFGT_CFGSYS;
            fContextFound = TRUE;
            pSearch += 9; // strlen("CONFIGSYS");
        }
        else if (strncmp(pSearch, "REGISTERCLASS", 13) == 0)
        {
            _ulExecType |= CFGT_REGISTERCLASS;
            fContextFound = TRUE;
            pSearch += 13; // strlen("REGISTERCLASS");
        }
        else if (strncmp(pSearch, "CREATEOBJECT", 12) == 0)
        {
            _ulExecType |= CFGT_CREATEOBJECT;
            fContextFound = TRUE;
            pSearch += 12; // strlen("CREATEOBJECT");
        }
        else
            // other: stop looping
            fContinue = FALSE;
    } // end while (fContinue)

    if (!(*pSearch))
        throw(BSConfigExcpt(EXEEXCPT_SYNTAX, strdup(pcszExecute)));

    // pSearch now points to the first non-space
    // character which does not introduces a keyword;
    // check if this is a "|" separator

    if (*pSearch == '|')
        // skip '|'
        pSearch++;

    // skip following spaces
    while (    (*pSearch)
            && (*pSearch == ' ')
          )
        pSearch++;

    // NOW we finally point to the executable

    _pszParams = NULL;

    if (*pSearch)
    {
        // separate executable and parameters:
        // This can be separated by another '|',
        // or a space.

        // NOTE: The '|' separator is used by the "resolved"
        // logger, from which BSDeExecute instances are
        // created on de-install. But the documentation states
        // spaces should be used... as long as a user doesn't
        // use spaces in the (unresolved) filename, we're fine.
        // For example we get:
        // 1)  on install: $(1)\install.cmd PARAMETER
        //      works nicely, and $(1) may contain spaces.
        // 2)  Logger stores C:\PATH1 WITH SPACES\install.cmd|PARAMETER
        //     in database.
        // 3)  on deinstall, BSDeExecutes get created, and
        //     we now have nice separation and get PARAMETER
        //     properly separated.
        const char *p = strchr(pSearch, '|');
        if (!p)
            p = strchr(pSearch, ' ');
        if (p)
        {
            const char *pSep = p;
            // we have a space --> parameters:
            while (    (*p)
                    && (*p == ' ')
                  )
                p++;

            if (*p)
            {
                _pszParams = strdup(p);
                _pszExecutable = strhSubstr(pSearch, pSep);
            }
            else
                _pszExecutable = strdup(pSearch);
        }
        else
            _pszExecutable = strdup(pSearch);
    }
    else
        throw(BSConfigExcpt(EXEEXCPT_SYNTAX, strdup(pcszExecute)));
}

/*
 *@@ ~BSExecute:
 *      destructor.
 *
 *@@added V0.9.1 (2000-02-07) [umoeller]
 */

BSExecute::~BSExecute()
{
    if (_pszExecutable)
        free(_pszExecutable);
    if (_pszParams)
        free(_pszParams);
}

/*
 *@@ DescribeType:
 *      describes the current manipulator type to the GUI.
 *
 *@@added V0.9.9 (2001-03-27) [umoeller]
 */

const char* BSExecute::DescribeType()
{
    return ("Execute program");
}

/*
 *@@ DescribeData:
 *      describes the current manipulator data to the GUI.
 *
 *@@added V0.9.9 (2001-03-27) [umoeller]
 */

BSString BSExecute::DescribeData()
{
    BSString    str = "\"";
    str += _pszExecutable;
    str += "\"";

    if (_pszParams)
    {
        str += " \"";
        str += _pszParams;
        str += "\"";
    }

    if (_ulExecType == 0)
        str += " (no flags)";
    else
    {
        if (_ulExecType & CFGT_CFGSYS)
            str += " CONFIGSYS";
        if (_ulExecType & CFGT_REGISTERCLASS)
            str += " REGISTERCLASS";
        if (_ulExecType & CFGT_CREATEOBJECT)
            str += " CREATEOBJECT";
    }

    return (str);
}

/*
 *@@ Execute:
 *      executes the member program. This does not return
 *      until the program terminates. As a result, do not
 *      call this on a PM thread, because this gets blocked.
 *
 *      Throws:
 *      -- BSConfigExcpt.
 *
 *@@added V0.9.1 (2000-02-07) [umoeller]
 */

int BSExecute::Execute(BSFileLogger *pLogFile)  // in: file logger (log file), can be NULL
{
    BSString    strParams = "/c ";
    strParams.append(_pszExecutable);
    if (_pszParams)
    {
        strParams += " ";
        strParams += _pszParams;
    }

    ULONG       sid = 0;
    PID         pid = 0;

    _arc = doshQuickStartSession("cmd.exe",
                                 (PSZ)strParams.c_str(),
                                 FALSE,   // foreground
                                 0,       // auto-close, visible
                                 TRUE,    // wait
                                 &sid,
                                 &pid);
    if (_arc != NO_ERROR)
    {
        if (pLogFile)
            pLogFile->Append("Error %d executing \"CMD.EXE %s\".\n",
                             _arc,
                             strParams.c_str());

        throw(BSConfigExcpt(EXEEXCPT_DOSERROR, _arc));
    }

    if (pLogFile)
        pLogFile->Append("Executed \"CMD.EXE %s\".\n",
                         strParams.c_str());

    return (0);
}

/* ******************************************************************
 *
 *  BSDeExecute class
 *
 ********************************************************************/

/*
 *@@ DescribeType:
 *      describes the current manipulator type to the GUI.
 *
 *@@added V0.9.2 (2000-02-19) [umoeller]
 */

const char* BSDeExecute::DescribeType()
{
    return ("External program for deinstall");
}

/*
 *@@ DescribeData:
 *      describes the current manipulator data to the GUI.
 *
 *@@added V0.9.2 (2000-02-19) [umoeller]
 *@@changed V0.9.5 (2000-08-26) [umoeller]: UNIQUE wasn't reported right
 */

BSString BSDeExecute::DescribeData()
{
    BSString    str = "\"";
    str += _pszExecutable;
    str += "\"";

    if (_pszParams)
    {
        str += " \"";
        str += _pszParams;
        str += "\"";
    }

    return (str);
}

/* ******************************************************************
 *
 *  BSKillProcess class
 *
 ********************************************************************/

/*
 *@@ BSKillProcess:
 *
 *      Throws:
 *      -- BSConfigExcpt.
 *
 *@@added V0.9.1 (2000-02-12) [umoeller]
 */

BSKillProcess::BSKillProcess(const char *pcszKillProcess)
               : BSConfigBase(CFGT_KILLPROCESS)
{
    if (!pcszKillProcess)
        throw(BSConfigExcpt(KILLEXCPT_SYNTAX, 0));

    _pszKillProcess = strdup(pcszKillProcess);
}

/*
 *@@ ~BSKillProcess:
 *      destructor.
 *
 *@@added V0.9.1 (2000-02-12) [umoeller]
 */

BSKillProcess::~BSKillProcess()
{
    if (_pszKillProcess)
        free(_pszKillProcess);
}

/*
 *@@ DescribeType:
 *      describes the current manipulator type to the GUI.
 *
 *@@added V0.9.4 (2000-07-01) [umoeller]
 */

const char* BSKillProcess::DescribeType()
{
    return ("Kill process");
}

/*
 *@@ DescribeData:
 *      describes the current manipulator data to the GUI.
 *
 *@@added V0.9.4 (2000-07-01) [umoeller]
 */

BSString BSKillProcess::DescribeData()
{
    return (BSString)_pszKillProcess;
}


