
/*
 *@@sourcefile bs_logger.cpp:
 *      this implements the various logger classes.
 *
 *      See BSLoggerRoot and subclasses for details.
 *
 *      The base logger class was renamed from INILogger and moved
 *      to this separate file (99-11-01) [umoeller].
 *      More logger classes were added with V0.9.9 (2001-03-30) [umoeller].
 *
 *@@added V0.9.0 (99-11-01) [umoeller]
 */

/*
 *      This file Copyright (C) 1999-2000 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define OS2EMX_PLAIN_CHAR
    // this is needed for "os2emx.h"; if this is defined,
    // emx will define PSZ as _signed_ char, otherwise
    // as unsigned char

#define INCL_DOSSEMAPHORES
#define INCL_WINSHELLDATA
#include <os2.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>

#include "setup.h"

#include <list>

// include's from helpers
#include "helpers\prfh.h"
#include "helpers\xstring.h"

// base includes
#include "base\bs_string.h"
#include "base\bs_errors.h"

#include "base\bs_logger.h"
#include "base\bs_config.h"

#pragma hdrstop

/* ******************************************************************
 *
 *  BSLoggerRoot implementation
 *
 ********************************************************************/

/* ******************************************************************
 *
 *  BSMemLoggerBase implementation
 *
 ********************************************************************/

/*
 *@@ BSMemLoggerBase:
 *      default constructor for creating an empty log string.
 *
 *      See BSMemLoggerBase::Append for example usage.
 *
 *      BSMemLoggerBase::Store and BSMemLoggerBase::Load can be used for
 *      storing a logger into or retrieving one from an INI file.
 */

BSMemLoggerBase::BSMemLoggerBase()
{
    _pabLogString = 0;
    _cbLogString = 0;
}

/*
 *@@ ~BSMemLoggerBase:
 *      destructor
 */

BSMemLoggerBase::~BSMemLoggerBase()
{
    if (_pabLogString)
        free(_pabLogString);
}

/*
 *@@ StoreData:
 *      implementation for the pure virtual method
 *      in BSLoggerRoot to store a chunk of data.
 *
 *      The implementation in BSMemLoggerBase stores
 *      this in the logger's memory block.
 *
 *@@added V0.9.9 (2001-03-30) [umoeller]
 */

void BSMemLoggerBase::StoreData(const char *pabData, unsigned long cbData)
{
    if (pabData)
    {
        if (_pabLogString == NULL)
        {
            // first run:
            _pabLogString = (char*)malloc(cbData);
            memcpy(_pabLogString, pabData, cbData);
            _cbLogString = cbData;
        }
        else
        {
            // subsequent runs: append after the existing data
            // V0.9.9 (2001-03-30) [umoeller]: now using realloc

            _pabLogString = (char*)realloc(_pabLogString, _cbLogString + cbData);
            // copy new attribs behind existing data
            memcpy(_pabLogString + _cbLogString,
                   pabData,
                   cbData);
            _cbLogString += cbData;

            /* char*   pszTemp = (char*)malloc(_cbLogString + cbData);
            // copy old buffer
            memcpy(pszTemp, _pabLogString, _cbLogString);
            // copy new attribs behind last null byte
            memcpy(pszTemp + _cbLogString,
                   pszData,
                   cbData);
            // set new buffer
            free(_pabLogString);
            _pabLogString = pszTemp;
            _cbLogString += cbData; */
        }
    }
}

/*
 *@@ Append:
 *      this appends binary data to the log string.
 *      This can be any data, but you should be able
 *      to decode it again so if each data is variable
 *      in length, you should add some markers.
 *
 *      This makes a copy of the binary data and appends
 *      it to the logger, if data already exists.
 *
 *      Example 1: you could add plain text strings.
 *                 Don't forget to store the null terminator then too.
 *
 +          const char *pcsz = "Hello";
 +          Logger.Append(pcsz, strlen(pcsz) + 1);
 *
 *      Example 2: you could add binary data, with
 *                 each item having a predefined length.
 *
 +          char    *pabData = (char*)malloc(10);
 +          memset(pabData, 0, 10);
 +          Logger.Append(pabData, 10);
 +          free(pabData);
 *
 *@@changed V0.9.1 (2000-01-05) [umoeller]: fixed memory allocation problems
 *@@changed V0.9.9 (2001-03-30) [umoeller]: now using realloc
 */

void BSMemLoggerBase::Append(const char *pabData,     // in: data block to append (must be terminated with null)
                             unsigned long cbData)    // in: sizeof(*pszData), including the null byte
{
    StoreData(pabData, cbData);
}

/*
 *Append:
 *      overloaded Append, this does a strdup on the string
 *      and finds the length automatically  (including the null terminator).
 *
 *      Example:
 +          Logger.Append("Hello")
 *      is equal to
 +          Logger.Append("Hello", strlen("Hello") + 1);
 *
 *changed V0.9.9 (2001-02-28) [umoeller]: removed excessive memory allocations
 *changed V0.9.9 (2001-03-30) [umoeller]: removed
 */

void BSMemLoggerBase::Append(const char *pcsz)
{
    if (pcsz)
        StoreData(pcsz,
                  strlen(pcsz) + 1);        // include null terminator
}

/*
 *@@ Append:
 *      overloaded Append; like the "const char*" version, but
 *      this takes a BSString as input. This is faster than
 *      the const char* method because BSString maintains the
 *      length of the string automatically.
 *
 *      Example:
 +          BSString str("Hello");
 +          Logger.Append(str);
 *      is equal to
 +          Logger.Append("Hello", strlen("Hello") + 1);
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 *@@changed V0.9.12 (2001-05-22) [umoeller]: fixed missing null terminator (CONFIG.SYS garbage in database)
 */

void BSMemLoggerBase::Append(BSString &str)
{
    StoreData(str.c_str(),
              str.size() + 1);      // null terminator was missing,
                                    // fixed V0.9.12 (2001-05-22) [umoeller]
}

/*
 *@@ Store:
 *      this stores the whole BSMemLoggerBase in the given profile key.
 *      Returns TRUE if successfully written.
 *
 *      Note: if the logger is currently empty, this will delete
 *      the given INI key.
 */

BOOL BSMemLoggerBase::Store(HINI hini,            // in: INI handle
                            const char *pszApp,   // in: INI application
                            const char *pszKey)   // in: INI key
                   const
{
    return (PrfWriteProfileData(hini, pszApp, pszKey, _pabLogString, _cbLogString));
}

/*
 *@@ Load:
 *      reverse to BSMemLoggerBase::Store, this loads the data from INI.
 *      This will overwrite the current contents of the logger.
 *
 *      Returns TRUE if data was found.
 */

BOOL BSMemLoggerBase::Load(HINI hini,             // in: INI handle
                        const char *pszApp,    // in: INI application
                        const char *pszKey)    // in: INI key
{
    if (_pabLogString)
        free(_pabLogString);
    _pabLogString = prfhQueryProfileData(hini,
                                         (PSZ)pszApp,
                                         (PSZ)pszKey,
                                         &_cbLogString);
    return (_pabLogString != 0);
}

/* ******************************************************************
 *
 *  BSFileLogger implementation
 *
 ********************************************************************/

HMTX G_hmtxFileLoggers = NULLHANDLE;

/*
 *@@ LockFileLoggers:
 *
 *@@added V0.9.12 (2001-05-31) [umoeller]
 */

BOOL LockFileLoggers(VOID)
{
    if (!G_hmtxFileLoggers)
        return (!DosCreateMutexSem(NULL,
                                   &G_hmtxFileLoggers,
                                   0,
                                   TRUE));

    return (!DosRequestMutexSem(G_hmtxFileLoggers, SEM_INDEFINITE_WAIT));
}

/*
 *@@ UnlockFileLoggers:
 *
 *@@added V0.9.12 (2001-05-31) [umoeller]
 */

VOID UnlockFileLoggers(VOID)
{
    DosReleaseMutexSem(G_hmtxFileLoggers);
}

/*
 *@@ BSFileLogger:
 *      constructor to open a file logger with the
 *      specified file name.
 *
 *@@added V0.9.9 (2001-03-30) [umoeller]
 *@@changed V0.9.12 (2001-05-31) [umoeller]: added mutex
 */

BSFileLogger::BSFileLogger(const char *pcszFilename)
{
    LockFileLoggers();

    _Pmpf(("BSFileLogger: constructor"));

    _File = NULL;
    _strFileName = pcszFilename;

    _indent = 0;

    if (pcszFilename)
    {
        _File = fopen(pcszFilename,
                      "a");
        if (!_File)
        {
            UnlockFileLoggers();
            CHAR szError[400];
            sprintf(szError,
                    "Cannot open log file \"%s\" for writing.",
                    pcszFilename);
            throw(BSLoggerExcpt(szError));
        }
    }
    else
    {
        UnlockFileLoggers();
        throw(BSLoggerExcpt("pcszFilename is NULL."));
    }

    DATETIME dt;
    DosGetDateTime(&dt);

    fprintf(_File,
            "\n\nWarpIN install log opened\n"
            "Date: %04d-%02d-%02d, Time: %02d:%02d:%02d\n"
            "--------------------------------\n",
            dt.year, dt.month, dt.day,
            dt.hours, dt.minutes, dt.seconds);

    UnlockFileLoggers();
}

/*
 *@@ ~BSFileLogger:
 *      destructor. Closes the log file.
 *
 *@@added V0.9.9 (2001-03-30) [umoeller]
 *@@changed V0.9.12 (2001-05-31) [umoeller]: added mutex
 */

BSFileLogger::~BSFileLogger()
{
    LockFileLoggers();

    _Pmpf(("BSFileLogger: destructor"));

    if (_File)
    {
        fprintf(_File, "\nInstall log closed.\n");
        fclose(_File);
        _File = NULL;
    }

    UnlockFileLoggers();
}

/*
 *@@ IncIndent:
 *      modifies indentation for the file output
 *      (spaces between date/time and the actual
 *      logger output).
 *
 *      A positive "i" increases the spacing, a
 *      negative decreases it.
 *
 *@@added V0.9.9 (2001-03-30) [umoeller]
 *@@changed V0.9.12 (2001-05-31) [umoeller]: added mutex
 */

void BSFileLogger::IncIndent(int i)
{
    LockFileLoggers();

    _indent += i;

    if (i < 0)
        i = 0;

    UnlockFileLoggers();
}

/*
 *@@ Append:
 *      yet another overloaded Append, which allows you
 *      to use printf formatting.
 *
 *      You must provide a sufficiently large buffer
 *      in pszBuf yourself. This is only used for
 *      temporary storage.
 *
 *      Returns the return value from vsprintf.
 *
 *      To differentiate this from the const char*
 *      version, a dummy pointer has been added as
 *      the first parameter, which should be NULL.
 *
 *      NOTE: This does not append \n for each string.
 *
 *@@added V0.9.9 (2001-03-30) [umoeller]
 *@@changed V0.9.12 (2001-05-31) [umoeller]: added mutex
 */

size_type BSFileLogger::Append(const char *pcszFormat,  // in: format string
                               ...)                     // in: additional printf args
{
    LockFileLoggers();

    _Pmpf(("BSFileLogger::Append: got \"%s\"", pcszFormat));

    DATETIME dt;
    DosGetDateTime(&dt);

    fprintf(_File,
            "%02d:%02d:%02d ",
            dt.hours, dt.minutes, dt.seconds);

    if (_indent)
    {
        for (int i = 0;
             i < _indent;
             i++)
            fprintf(_File, " ");
    }

    va_list arg_ptr;
    va_start(arg_ptr, pcszFormat);
    int iWritten = vfprintf(_File, pcszFormat, arg_ptr);
    va_end(arg_ptr);

    UnlockFileLoggers();
}


