
/*
 *@@sourcefile bs_rexx.cpp:
 *      REXX support for WarpIN scripts.
 *
 *      This was added with V0.9.2 by Cornelis Bockemhl and moved
 *      to the base code (src\base) with V0.9.9.
 *
 *      WarpIN REXX support now comes in two layers:
 *
 *      --  The code here in src\base\bs_rexx.cpp gets called during
 *          script parsing to store REXX code in a map according
 *          to the respective function names and then execute it
 *          to replace parts in the scripts
 *
 *          This REXX support is encapsulated in the FERexx class, of
 *          which one single instance is created by main() at WarpIN
 *          startup.
 *
 *      --  The code in src\frontend\fe_rexx.cpp has the exported
 *          REXX functions because that code needs access to the GUI.
 *
 *@@header "base\bs_rexx.h"
 *@@added V0.9.2 (2000-03-10) [cbo]
 */

/*
 *      Copyright (C) 2000 Cornelis Bockemhl.
 *      Copyright (C) 2000-2001 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define OS2EMX_PLAIN_CHAR
    // this is needed for "os2emx.h"; if this is defined,
    // emx will define PSZ as _signed_ char, otherwise
    // as unsigned char

// REXX support
#define INCL_REXXSAA
#include <os2.h>
#ifdef __IBMCPP__
    #include <rexxsaa.h>    // EMX has this in os2.h already
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// include all these headers because we need calbacks.h
#include "setup.h"

// include's from helpers
#include "helpers\stringh.h"
#include "helpers\xstring.h"

// back-end includes
// #include "wiarchive\wiarchive.h"

// base includes
#include "base\bs_string.h"
#include "base\bs_rexx.h"

/* ******************************************************************
 *
 *  FERexx global variables
 *
 ********************************************************************/

static BSStringMap G_RexxCodeStore;

// WarpIN REXX support; this is implemented thru a
// global object. We use new() to create the FERexx
// instance, because if we did this statically and
// errors occured, WarpIN would crash even before
// main() is reached, which makes debugging impossible.
// So instead, this is created in main() now.
FERexx          *G_pFERexx = 0;

/* ******************************************************************
 *
 *  FERexx implementation
 *
 ********************************************************************/

/*
 *@@ FERexx:
 *      constructor.
 *
 *      Registers the functions in the
 *      rexx functions array passed to it.
 *
 *      This implementation has been changed to allow moving
 *      REXX support to the WPI runtime (src\base). The REXX
 *      functions array is now passed to this constructor
 *      from main() when REXX is initialized.
 *
 *@@added V0.9.2 (2000-03-10) [cbo]
 *@@changed V0.9.9 (2001-02-28) [umoeller]: added array parameter
 */

FERexx::FERexx(WIREXXFUNCTION *paFunctions, // in: array of REXX functions
               ULONG cFunctions)            // in: array item count (NOT array size)
{
    ULONG   rc,
            ul;

    // register REXX functions
    for (ul = 0;
         ul < cFunctions;
         ul++)
    {
        rc = RexxRegisterFunctionExe(paFunctions[ul].pcszFunctionName,
#ifdef __IBMCPP__
                                     (PFN)paFunctions[ul].pfn);
#else
                                     paFunctions[ul].pfn);
#endif
    }

    // store these in instance so we can deregister them later
    _paFunctions = paFunctions;
    _cFunctions = cFunctions;
}

/*
 *@@ ~FERexx:
 *
 *@@added V0.9.2 (2000-03-10) [cbo]
 */

FERexx::~FERexx()
{
    ULONG rc,
          ul;

    // deregister REXX functions
    for (ul = 0;
         ul < _cFunctions;
         ul++)
    {
        rc = RexxDeregisterFunction(_paFunctions[ul].pcszFunctionName);
    }
}

/*
 *@@ StoreRexxCode:
 *      store a REXX code string under the given name. The code is not
 *      verified for validity in any way, but some reformatting occurs
 *      in order to allow for "usual" REXX code (line ends separating
 *      commands, comma as line continuation mark): In this context,
 *      the REXX interpreter only allows semicolons ";" as command
 *      separators and the whole code must not contain any line breaks.
 *
 *@@added V0.9.2 (2000-03-10) [cbo]
 *@@changed V0.9.2 (2000-03-13) [cbo]: allows line breaks and line continuation
 */

void FERexx::StoreRexxCode(const char *name,
                           const char *code)
{
    BSString rxcode(code);
    // char* pstr;
    size_type pos;
    BSString crlf("\r\n"),
             lf("\n"),
             splf(" \n"),
             lfspspspsp("\n    "),
             lfsp("\n "),
             commalf(",\n"),
             empty(""),
             semicolon(";");

    // change all CRLF to LF codes
    ULONG ulOfs = 0;
    while (rxcode._find_replace(crlf, lf, &ulOfs)
                != BSString::npos)
            ;
    // trim all lines (remove spaces)
    ulOfs = 0;
    while (rxcode._find_replace(lfspspspsp, lf, &ulOfs)
                != BSString::npos)
            ;
    ulOfs = 0;
    while (rxcode._find_replace(lfsp, lf, &ulOfs)
                != BSString::npos)
            ;
    ulOfs = 0;
    while (rxcode._find_replace(splf, lf, &ulOfs)
                != BSString::npos)
            ;
    // remove commas at line ends and merge lines
    ulOfs = 0;
    while (rxcode._find_replace(commalf, empty, &ulOfs)
                != BSString::npos)
            ;
    // replace line ends by semocolons
    ulOfs = 0;
    while (rxcode._find_replace(lf, semicolon, &ulOfs)
                != BSString::npos)
            ;
    // this should run now! Store it.
    G_RexxCodeStore[BSString(name)] = rxcode;
}

/*
 *@@ ExecuteRexxCode:
 *      pass the name of a REXX code segment that was stored.
 *
 *      The return code is 0 (WIREXX_NOERROR) if no error
 *      occurred or one of the other WIREXX_* values to
 *      indicate a problem or error.
 *
 *      If WIREXX_NOERROR is returned, the function return
 *      value is stored in a new buffer in **ppszRet, which
 *      the caller must free() when it's no longer needed.
 *
 *@@added V0.9.2 (2000-03-10) [cbo]
 */

int FERexx::ExecuteRexxCode(const char *pcszFuncName,   // in: function to execute
                            const char *pcszArgs,       // in: function parameters
                            char **ppszRet)             // out: function return value
{
    BSStringMap::iterator icode;
    RXSTRING        rxproc[2],
                    rxargv[1],
                    rxret;
    char            retbuf[256];
    SHORT           rexxrc;
    LONG            rc;

    // make sure we return a valid ret pointer
    *ppszRet = NULL;

    // no code - no call -> probably not what the caller intended to do!
    if ((NULL == pcszFuncName) || (0 == strlen(pcszFuncName)))
        return WIREXX_INVALIDNAME;

    icode = G_RexxCodeStore.find(BSString(pcszFuncName));
    if (G_RexxCodeStore.end() == icode)
        return WIREXX_NOCODE;

    // command: 0=source code, 1 would be parsed code -> not available!
    MAKERXSTRING(rxproc[0], (*icode).second.c_str(), (*icode).second.size());
    MAKERXSTRING(rxproc[1], NULL, 0);

    // argument string, if present
    if (    (NULL == pcszArgs)
         || (0 == strlen(pcszArgs))
       )
    {
        MAKERXSTRING(rxargv[0], NULL, 0);
    }
    else
    {
        MAKERXSTRING(rxargv[0], pcszArgs, strlen(pcszArgs));
    }

    // prepare a return buffer
    memset(retbuf, 0, sizeof(retbuf));
    MAKERXSTRING(rxret, retbuf, sizeof(retbuf));

    // do the call
    rc = RexxStart(1, rxargv, pcszFuncName, rxproc, NULL,
                   RXCOMMAND, NULL, &rexxrc, &rxret);
    if (0 != rc)
        return WIREXX_REXXFAILED;

    // allocate a new buffer for the result
    *ppszRet = (char *)malloc(RXSTRLEN(rxret) + 1);

    // copy the string to the result, even allowing for 0 characters!
    memcpy(*ppszRet, RXSTRPTR(rxret), RXSTRLEN(rxret));
    (*ppszRet)[RXSTRLEN(rxret)] = 0;

    // free extra allocated memory
    if (RXSTRPTR(rxret) != retbuf)
        DosFreeMem(RXSTRPTR(rxret));

    return WIREXX_NOERROR;
}

/*
 *@@ StoreExecuteRexxCode:
 *      calls FERexx::StoreRexxCode and
 *      returns the return value of
 *      FERexx::ExecuteRexxCode. Just a
 *      shortcut to the two.
 *
 *@@added V0.9.2 (2000-03-10) [cbo]
 */

int FERexx::StoreExecuteRexxCode(const char *pcszName,
                                 const char *pcszCode,
                                 const char *pcszArgs,
                                 char **ppszRet)
{
    StoreRexxCode(pcszName, pcszCode);
    return ExecuteRexxCode(pcszName, pcszArgs, ppszRet);
}

/*
 *@@ ParseRexxTags:
 *      pass the contents of a <HEAD>...</HEAD> tag to this function and
 *      it will find and interpret all contained <REXX NAME=xyz>code</REXX>
 *      tags. That means: the "code" is stored under the name "xyz" in the
 *      REXX code store within this REXX support class.
 *      The code is not veryfied at all in this function.
 *
 *      This calls FERexx::StoreRexxCode for each code block found.
 *
 *      Returns the no. of REXX tags found in the header.
 *
 *@@added V0.9.2 (2000-03-10) [cbo]
 *@@changed V0.9.12 (2001-05-31) [umoeller]: fixed return value
 */

ULONG FERexx::ParseRexxTags(const char *hdrstring)
{
    ULONG   ulrc = 0,           // 0 was missing V0.9.12 (2001-05-31) [umoeller]
            rc,
            ofs,
            start,
            end;
    PSZ block, attribs, name;

    // loop through the input string
    for (ofs = 0; 1; ofs = end)
    {
        // see if there is a next REXX tag, terminate if not
        rc = strhGetBlock(hdrstring, &ofs, "REXX", &block, &attribs, &start, &end);
        if (0 != rc)
            break;

        // if no name attribute found: ignore the rest: We cannot store it!
        if (    (NULL == attribs)
             || (NULL == (name = strhGetTextAttr(attribs, "NAME", &ofs)))
           )
        {
            free(block);
            free(attribs);
            continue;
        }

        // now we have all we need to store the code
        strupr(name);
        StoreRexxCode(name, block);
        ulrc++;

        // free all strings
        free(block);
        free(attribs);
        free(name);
    }

    return (ulrc);
}

/*
 *@@ ExecuteRexxMacros:
 *      pass the pointer to a string containing =("xyz arg1 arg2...")
 *      tags to this and it will call the REXX function with the name
 *      "xyz" and the argument string "arg1 arg2...". After that the
 *      tag is removed and replaced with the return value of the function,
 *      which may also be empty.
 *
 *      Since the string length can change with this call, the string
 *      storage may be reallocated using realloc(), i.e. it is assumed
 *      that it was originally allocated with alloc() and must be
 *      freed with free().
 *
 *      The call returns TRUE if the REXX call succeeded, otherwise
 *      FALSE. In the latter case the return string is set to empty
 *      so possibly an erroneous call to REXX can still return a
 *      useable string - but better catch these cases!
 *
 *      After an error occurred, the function still tries to handle
 *      the remaining string, so "an error occurred" could mean that
 *      actually several occurred.
 *
 *      If pulErrorOffset is not NULL (which is allowed), it returns the
 *      position of the first error within the passed string.
 *      If errorName is not NULL, it returns the name of the first
 *      offending REXX macro call.
 *
 *@@added V0.9.2 (2000-03-10) [cbo]
 */

BOOL FERexx::ExecuteRexxMacros(char **ppszStr,
                               PULONG pulErrorOffset,
                               char **ppszErrorCall)
{
    PSZ     ps, pe, macro, args, ret;
    ULONG   os, oe, oldtlen, newtlen, oldlen, newlen, rc;
    BOOL    ok = TRUE;

#define STAG "=(\""
#define ETAG "\")"
#define NO_ERROROFS 0xFFFFFFFF

    // initialize error return
    if (NULL != pulErrorOffset)
        *pulErrorOffset = NO_ERROROFS;
    if (NULL != ppszErrorCall)
        *ppszErrorCall = NULL;

    // do while we find start and end of a REXX macro
    for (ps = strstr(*ppszStr, STAG), pe = (NULL == ps) ? NULL : strstr(ps, ETAG);
         (NULL != ps) && (NULL != pe);
         ps = strstr(pe, STAG), pe = (NULL == ps) ? NULL : strstr(ps, ETAG))
    {
        // get the total code
        ps += strlen(STAG);
        macro = (char *)malloc(pe - ps + 1);
        memcpy(macro, ps, pe - ps);
        macro[pe - ps] = 0;

        // separate uppercase "macro" from "args"
        args = strchr(macro, ' ');
        if (NULL != args)
            *args++ = 0;
        strupr(macro);

        // call the macro
        if (WIREXX_NOERROR != (rc = ExecuteRexxCode(macro, args, &ret)))
        {
            // the call didn't succeed (e.g. because of errors in the REXX code)
            // create an empty return string: This will result in the erroneous
            // call just being deleted from the string
            ret = (char *)malloc(1);
            *ret = 0;
            ok = FALSE;
            if ((NULL != pulErrorOffset) && (NO_ERROROFS == *pulErrorOffset))
                *pulErrorOffset = ps - *ppszStr;
            if ((NULL != ppszErrorCall) && (NULL == *ppszErrorCall))
            {
                *ppszErrorCall = (char *)malloc(pe - ps + 1);
                strncpy(*ppszErrorCall, ps, pe - ps);
                (*ppszErrorCall)[pe - ps] = 0;
            }
        }

        // calculate all the needed lengths
        oldlen = pe - ps + strlen(STAG) + strlen(ETAG);
        newlen = strlen(ret);
        oldtlen = strlen(*ppszStr);
        newtlen = oldtlen + newlen - oldlen + 1;

        // now we need offsets instead of pointers
        os = ps - *ppszStr - strlen(STAG);
        oe = pe - *ppszStr + strlen(ETAG);

        // reallocate the string if it is growing
        if (newtlen > oldtlen)
            *ppszStr = (char *)realloc(*ppszStr, newtlen);

        // move the tail of the string
        memmove(*ppszStr + oe - oldlen + newlen, *ppszStr + oe, strlen(*ppszStr + oe) + 1);

        // insert the returned string
        memcpy(*ppszStr + os, ret, newlen);

        // reallocate the string if it is shrinking
        if (newtlen < oldtlen)
            *ppszStr = (char *)realloc(*ppszStr, newtlen);

        // free the returned string from REXX
        free(ret);

        // set pointers for further searching
        pe = ps = *ppszStr + os;
    }

    // TRUE or FALSE: success indicator
    return ok;
}

/*
 *@@ ExecuteRexxMacros:
 *      second overloaded version that uses BSStrings.
 *
 *@@added V0.9.2 (2000-03-10) [cbo]
 */

BOOL FERexx::ExecuteRexxMacros(BSString &str,
                               PULONG pulErrorOffset,
                               BSString &strErrorCall)
{
    char    *pstr,
            *perr = NULL;
    BOOL ok;

    // copy string to a common C string
    pstr = strhdup(str.c_str());

    // call the first overload version
    ok = ExecuteRexxMacros(&pstr, pulErrorOffset, &perr);

    // change back to BSStrings
    str = pstr;
    strErrorCall = perr;

    // free the temporary memory
    free(pstr);
    if (perr != NULL)
        free(perr);

    return ok;
}


