
/*
 *@@sourcefile bs_script.cpp:
 *      BSScriptBase and subclasses implementation.
 *
 *      These are the WarpIN script classes, responsible for
 *      parsing an installation script.
 *
 *      This has been extracted from FEArchive and FEPackageBase
 *      with V0.9.9 and moved to this file to allow for a
 *      completely different implementation of script parsing.
 *      See fe_archive.cpp for an overview how scripts are
 *      now implemented.
 *
 *      The concept is that BSScriptBase has a static method
 *      (BSScriptBase::CreateScript) which gets called by FEArchive
 *      when a script needs to be parsed. FEArchive no longer cares
 *      about how the script is interpreted, but only wants
 *      the member fields in BSScriptBase to point to the data
 *      in the script.
 *
 *      The actual parsing is then done by the virtual
 *      BSScriptBase::Parse method, which must be overridden
 *      by subclasses (script implementations).
 *
 *      This file only contains the abstract base classes:
 *
 *      -- BSScriptBase for the script;
 *
 *      -- BSDeclRoot, BSPckDeclBase, BSGroupDeclBase for
 *         packages in the script.
 *
 *      Presently, there is only one implementation for
 *      parsing an old-style WarpIN script. See bs_script_old.cpp.
 *
 *      To implement a different script language, a check has
 *      to be added to BSScriptBase::CreateScript, which must
 *      then create an instance of the new script subclass and
 *      invoke that class's Parse method.
 *
 *@@header "base\bs_script.h"
 *@@added V0.9.9 (2001-02-19) [umoeller]
 */

/*
 *      This file Copyright (C) 1999-2001 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define OS2EMX_PLAIN_CHAR
    // this is needed for "os2emx.h"; if this is defined,
    // emx will define PSZ as _signed_ char, otherwise
    // as unsigned char

#define INCL_REXXSAA
#define INCL_DOSERRORS
#include <os2.h>

#ifdef __IBMCPP__
    #include <rexxsaa.h>    // EMX has this in os2.h already
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "setup.h"

#include <list>
// #include <map>

#include "bldlevel.h"           // needed for version checks

#include "helpers\configsys.h"
#include "helpers\dosh.h"
#include "helpers\stringh.h"
#include "helpers\tmsgfile.h"
#include "helpers\xstring.h"

#include "base\bs_string.h"
#include "base\bs_errors.h"

#include "base\bs_logger.h"
#include "base\bs_config.h"
#include "base\bs_config_impl.h"

#include "base\bs_script.h"
#include "base\bs_script_old.h"

#include "base\bs_rexx.h"

/* ******************************************************************
 *
 *  Static variables
 *
 ********************************************************************/

BSVarPromptsMap BSScriptBase::VarPromptsMap;

// warpin.tmf message file; this is referenced by wpiGetMessage
// in bs_script.cpp
CHAR            G_szMessageFile[2*CCHMAXPATH] = "";

/* ******************************************************************
 *
 *  Helper functions
 *
 ********************************************************************/

/*
 *@@ wpiGetMessage:
 *      like DosGetMessage, but automatically finds the
 *      (NLS) WarpIN message file.
 *
 *      This has been moved here because otherwise we
 *      can't use NLS messages with the parsers.
 *
 *@@changed (99-10-22) [umoeller]: now using TMF functions.
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved here from fe_base.cpp
 */

APIRET wpiGetMessage(BSString &strBuf,  // out: composed message string
                     ULONG ulMsgNumber, // in: TMF message number to retrieve
                     PCHAR *pTable,     // array[ulTable] of PSZs as is DosGetMessage; defaults to NULL
                     ULONG ulTable)     // array item count (defaults to 0)
{
    CHAR    szBuf[5000];
    ULONG   ul;
    APIRET  arc;
    CHAR    szMsgID[30];
    sprintf(szMsgID, "WPI%04d", ulMsgNumber);
    arc = tmfGetMessageExt(pTable, ulTable,
                           szBuf, sizeof(szBuf),
                           szMsgID,
                           G_szMessageFile,
                           &ul);
    if (arc != NO_ERROR)
        sprintf(szBuf, "tmfGetMessageExt returned error %u for msg %u", arc, ulMsgNumber);

    strBuf = szBuf;

    return (arc);
}

/* ******************************************************************
 *
 *  BSScriptExcpt implementation
 *
 ********************************************************************/

/*
 *@@ BSScriptExcpt:
 *      first constructor with a plain string message.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

BSScriptExcpt::BSScriptExcpt(const BSString &str,
                             unsigned long ulLine,      // in: line, defaults to 0
                             unsigned long ulColumn)    // in: column, defaults to 0
{
    _strDescription = str;
    _ulLine = ulLine;
    _ulColumn = ulColumn;
}

/*
 *@@ BSScriptExcpt:
 *      second constructor for retrieving a TMF message.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

BSScriptExcpt::BSScriptExcpt(int iMessage,     // in: TMF msg id
                             char** apszTable,
                             int iCount)
{
    BSString    strMsg2;
    wpiGetMessage(strMsg2,
                  iMessage,
                  apszTable, iCount);
    _strDescription = strMsg2;
}

/* ******************************************************************
 *
 *  BSVarPrompt implementation
 *
 ********************************************************************/

/*
 *@@ Validate:
 *      returns TRUE only if the specified value is
 *      valid.
 *
 *@@added V0.9.2 (2000-03-11) [umoeller]
 *@@changed V0.9.3 (2000-04-28) [umoeller]: didn't check for pcszValue == NULL; fixed
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 */

bool BSVarPrompt::Validate(const char *pcszValue)
     const
{
    bool brc = FALSE;

    if (pcszValue)
        switch (_ulType)
        {
            case VPT_NUM:
            break;

            case VPT_ALPHANUM:
                brc = TRUE;
            break;

            case VPT_PATH:
                brc = (doshIsValidFileName(pcszValue, TRUE) == NO_ERROR);
            break;

            case VPT_FAIL:
            break;
        }

    return (brc);
}

int operator==(const BSVarPrompt &v1, const BSVarPrompt &v2)
{
    return (v1._strVarName == v2._strVarName);
}

int operator<(const BSVarPrompt &v1, const BSVarPrompt &v2)
{
    return (v1._strVarName < v2._strVarName);
}

/* ******************************************************************
 *
 *  BSDeclRoot implementation
 *
 ********************************************************************/

/*
 *@@ BSDeclRoot:
 *      constructor.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

BSDeclRoot::BSDeclRoot(BSGroupDeclBase *pGroup,
                       BOOL fIsGroup)
{
    _pGroup = pGroup;
    _fIsGroup = fIsGroup;
}

/* ******************************************************************
 *
 *  BSPckDeclBase implementation
 *
 ********************************************************************/

/*
 *@@ BSPckDeclBase:
 *      constructor for the base class.
 *
 *      This constructor is protected to make sure that
 *      instances of BSPckDeclBase can only be created
 *      as instances of one of the subclasses.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

BSPckDeclBase::BSPckDeclBase(BSGroupDeclBase *pGroup)     // the group this belongs to or NULL
               : BSDeclRoot(pGroup,
                            FALSE)              // not group
{
    _ulIndexInArchive = 0;

    _External = INTERNAL;

    _Selected = UNKNOWN;

    _ulPathType = 0;

    _ulConfigData = 0;
}

/* ******************************************************************
 *
 *  BSScriptBase implementation
 *
 ********************************************************************/

/*
 *@@ CreateScript:
 *      this static method is the magic entry point into
 *      script parsing. This creates a script instance
 *      and parses the script.
 *
 *      Once FEArchive has retrieved the script from the
 *      archive, it calls this function with the script
 *      as the only parameter.
 *
 *      This function then does a few quick checks on
 *      the script to determine its format.
 *
 *      --  If the script starts with "<?xml ", it is
 *          assumed to be a new-style XML script. In that
 *          case, an instance of BSXmlScript is created
 *          and returned.
 *
 *      --  Otherwise, it is assumed to be an old-style
 *          script. In that case, an instance of BSOldScript
 *          is created and returned.
 *
 *      Before returning, this function calls the virtual
 *      "Parse" method on the new script instance, which
 *      does the actual work of filling all the fields.
 *      As a result, the caller doesn't even need to know
 *      any more what format the script is in. It receives
 *      an instance of BSScriptBase (actually, of one of
 *      the subclasses) with all the members nicely set...
 *      how the parsing works doesn't matter.
 *
 *      If pfnConfirmRexxAllowed != NULL, that function gets
 *      called if any REXX code is encountered in the script.
 *      That function must then return TRUE if parsing should
 *      continue or FALSE if not (paranoid user selected "no"
 *      then).
 *
 *@@added V0.9.9 (2001-02-19) [umoeller]
 */

BSScriptBase* BSScriptBase::CreateScript(const BSString &strScript, // in: entire install script
                                         PFNCONFIRMREXXALLOWED pfnConfirmRexxAllowed)
{
    BSScriptBase *pScript = NULL;

    // check the script string for the magic <?xml string; if we
    // find it, we assume it's a new-style string, otherwise
    // we invoke the old parser
    ULONG ulFirstBracket = strScript.find('<');
    if (ulFirstBracket != BSString::npos)
    {
        if (strcmp(strScript.c_str() + ulFirstBracket,
                   "<?xml ") == 0)
        {
            // new script:
        }
        else
        {
            // old script:
            pScript = new BSOldScript;
        }

        pScript->Parse(strScript,
                       pfnConfirmRexxAllowed);
    }
    else
        // we don't even have an angle bracket in the script:
        // that's neither old nor new style...
        throw (BSScriptExcpt(201));

    return (pScript);
}


