
/*
 *@@sourcefile bs_script_old.cpp:
 *      implementation of script parsing for old-style scripts.
 *      See bs_script.cpp for an introduction about script
 *      parsing.
 *
 *      This file has the following implementations:
 *
 *      -- BSOldScript as an implementation for BSScriptBase
 *         (for parsing old-style scripts);
 *
 *      -- BSOldPckDecl as an implementation for BSPckDeclBase
 *         (for packages in an old-style script).
 *
 *      This has exactly the code which used to be in FEArchive...
 *      except that a lot of code had to be adjusted to this
 *      abstraction.
 *
 *@@header "base\bs_script_old.h"
 *@@added V0.9.9 (2001-02-19) [umoeller]
 */

/*
 *      This file Copyright (C) 1999-2001 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define OS2EMX_PLAIN_CHAR
    // this is needed for "os2emx.h"; if this is defined,
    // emx will define PSZ as _signed_ char, otherwise
    // as unsigned char

#define INCL_REXXSAA
#define INCL_DOSERRORS
#include <os2.h>

#ifdef __IBMCPP__
    #include <rexxsaa.h>    // EMX has this in os2.h already
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "setup.h"

#include <list>
// #include <map>

#include "bldlevel.h"           // needed for version checks

#include "helpers\configsys.h"
#include "helpers\dosh.h"
#include "helpers\stringh.h"
#include "helpers\tmsgfile.h"
#include "helpers\xstring.h"

#include "base\bs_string.h"
#include "base\bs_errors.h"

#include "base\bs_logger.h"
#include "base\bs_config.h"
#include "base\bs_config_impl.h"

#include "base\bs_script.h"
#include "base\bs_script_old.h"

#include "base\bs_rexx.h"

/* ******************************************************************
 *
 *  BSOldPckDecl implementation
 *
 ********************************************************************/

/*
 *@@ BSPckDeclBase:
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

BSOldPckDecl::BSOldPckDecl(BSOldScript *pScript,
                           BSGroupDeclBase *pGroup_,  // in: "parent" group; can be NULL
                           const char *pszInfo,     // in: stuff between PCK and /PCK (must be copied)
                           const char *pszPckAttribs, // in: attributes of PCK tag
                           ULONG ulPckOfs)          // in: offset of PCK tag in script (for errors)
              : BSPckDeclBase(pGroup_)
{
    _pScript = pScript;

    // "INDEX="
    if (!strhGetNumAttribValue(pszPckAttribs, "INDEX", (PLONG)&_ulIndexInArchive))
        pScript->ParseError(ulPckOfs,
                            "A <PCK> tag is lacking the required INDEX attribute.");

    _strDescription = pszInfo;

    ParsePckAttribsMain(pszPckAttribs,
                        ulPckOfs);
}

/*
 *@@ ParsePckAttribsMain:
 *      this private method gets called by the contructor
 *      (BSOldPckDecl::BSOldPckDecl)
 *      to parse the package tag attributes. This calls
 *      the other private ParsePckAttribs* methods in
 *      turn.
 *
 *@@added V0.9.1 (2000-02-03) [umoeller]
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 *@@changed V0.9.9 (2001-02-28) [umoeller]: fixed crash with EXTERNAL REQUIRED
 */

VOID BSOldPckDecl::ParsePckAttribsMain(const char *pszPckAttribs,
                                       ULONG ulAttrOfs)
{
    PSZ pszTitle = strhGetTextAttr(pszPckAttribs, "TITLE", NULL);
    if (!pszTitle)
        _pScript->ParseError(ulAttrOfs,
                             "TITLE attribute missing for package \"%d\".",
                             _ulIndexInArchive);
    else
    {
        _strTitle = pszTitle;
        free(pszTitle);
    }

    // "TARGET="
    PSZ pszTargetPath = strhGetTextAttr(pszPckAttribs, "TARGET", NULL);
    if (!pszTargetPath)
        _pScript->ParseError(ulAttrOfs,
                             "TARGET attribute missing for package \"%s\".",
                             _strTitle.c_str());
    else
    {
        _strTargetPath = pszTargetPath;
        free(pszTargetPath);
    }

    // "PACKAGEID="
    PSZ pszID = strhGetTextAttr(pszPckAttribs, "PACKAGEID", NULL);
    if (!pszID)
        _pScript->ParseError(ulAttrOfs,
                             "Package ID missing for package \"%s\".",
                             _strTitle.c_str());
    else
    {
        _strID = pszID;
        free(pszID);
    }

    // check ID; it must contain four or five backslashes
    int cBackslashes = strhCount(pszID, '\\');
    if (    (cBackslashes != 4)
         && (cBackslashes != 5)
       )
        _pScript->ParseError(ulAttrOfs,
                             "Package ID for package \"%s\" does not contain four or five backslashes. ",
                             _strTitle.c_str());

    // "REQUIRES="; this is optional
    ParsePckAttribsRequiresStrings(pszPckAttribs,
                                   ulAttrOfs);

    // CONFIGSYS attributes (errors are handled by that method)
    ParsePckAttribsConfigSys(pszPckAttribs,
                             ulAttrOfs);

    // "REGISTERCLASS=" attributes (errors are handled by that method)
    ParsePckAttribsWPSClasses(pszPckAttribs,
                              ulAttrOfs);

    // "CREATEOBJECT=" attributes (errors are handled by that method)
    ParsePckAttribsWPSObjects(pszPckAttribs,
                              ulAttrOfs);

    // "CLEARPROFILE=", "WRITEPROFILE=" attributes (errors are handled by that method)
    ParsePckAttribsProfiles(pszPckAttribs,
                            ulAttrOfs);

    // "EXECUTE=" attributes (errors are handled by that method)
    ParsePckAttribsExecutes(pszPckAttribs,
                            ulAttrOfs);

    // "KILLPROCESS=" attributes (errors are handled by that method)
    ParsePckAttribsKillProcesses(pszPckAttribs,
                                 ulAttrOfs);

    if (strhFindAttribValue(pszPckAttribs, "NODESELECT") != 0)
        _Selected = NODESELECT;
    else if (strhFindAttribValue(pszPckAttribs, "SELECT") != 0)
        _Selected = SELECTED;
    else
        _Selected = DESELECTED;

    _ulPathType = 0;
    if (strhFindAttribValue(pszPckAttribs, "BASE") != 0)
        _ulPathType |= PATH_BASE;
    if (strhFindAttribValue(pszPckAttribs, "FIXED") == 0)
        _ulPathType |= PATH_VARIABLE;

    // handle external packages
    // (totally rewritten with V0.9.9 (2001-02-28) [umoeller])

    PSZ         pszExternal = NULL;
    BOOL        fRequired = FALSE;
    if (NULL != (pszExternal = strhGetTextAttr(pszPckAttribs,
                                               "EXTERNAL",
                                               NULL)))
    {
        // EXTERNAL attribute specified:
        // pszExternal now has the EXTERNAL data, which is:
        // "[REQUIRED|]archive"

        PSZ     pArchiveName = pszExternal,
                pSep = 0;
        if (NULL != (pSep = strchr(pszExternal, '|')))
        {
            // separator found:
            if (strstr(pszExternal, "REQUIRED"))
            {
                // REQUIRED found:
                pArchiveName = pSep + 1;
                while (    (*pArchiveName)
                        && (*pArchiveName == ' ')
                      )
                    pArchiveName++;

                fRequired = TRUE;
            }
            else
                pArchiveName = NULL;    // error
        }

        if (    pArchiveName     // V0.9.9 (2001-02-28) [umoeller]
             && *pArchiveName
           )
        {
            // valid archive name, apparently:
            _strExternalArchive = pszExternal;

            if (fRequired)
                _External = REQEXTERNAL;
            else
                _External = EXTERNAL;
        }
        else
            _pScript->ParseError(ulAttrOfs,
                                 "Syntax error in EXTERNAL attribute. ", NULL);

        free(pszExternal);

    } // end if strhGetTextAttr(pszPckAttribs," EXTERNAL",
    // else: we still have the INTERNAL default from the constructor
}

/*
 *@@ ParsePckAttribsRequiresStrings:
 *      this private method gets called by
 *      BSOldPckDecl::ParsePckAttribsMain to parse
 *      the REQUIRES strings.
 *
 *      This terminates WarpIN upon errors (by calling
 *      FEArchive::ParseError), giving the user a
 *      meaningful error message.
 *
 *@@added V0.9.1 (2000-01-07) [umoeller]
 *@@changed V0.9.1 (2000-02-07) [umoeller]: fixed memory leak
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 */

LONG BSOldPckDecl::ParsePckAttribsRequiresStrings(const char *pszPckAttribs,  // in: attributes block of PCK tag
                                                  ULONG ulAttrOfs)    // in: offset of pszPckAttribs in script (for errors)
{
    LONG        lrc = 0;
    ULONG       ulOfs = 0;
    const char  *pszSearchIn = pszPckAttribs;

    while (lrc != -1)   // no error?
    {
        PSZ     pszRequiresThis = NULL;

        pszRequiresThis = strhGetTextAttr(pszSearchIn,
                                          "REQUIRES", &ulOfs);
                // this can either be a decimal index
                // or a five- or six-part package ID
        if (!pszRequiresThis)
            // no more attribs found:
                break;  // while

        // else: pszAttr now has the whole attribute

        if (    (atoi(pszRequiresThis) == 0)
             && (strhCount(pszRequiresThis, '\\') != 4)
             && (strhCount(pszRequiresThis, '\\') != 5)
           )
            _pScript->ParseError(ulAttrOfs, "\"REQUIRES\" ID for package %d has invalid syntax. ",
                                 _ulIndexInArchive);
        // else:
        // store in BSPckDeclBase logger;
        // this will be re-parsed after all archives
        // have been parsed to resolve indices
        // (FEArchive::ResolveRequirements)
        _logRequiresStrings.Append(pszRequiresThis);
        lrc++;

        if (pszRequiresThis)
            free(pszRequiresThis);
        // advance search pointer for next search
        pszSearchIn += ulOfs;
    }

    return (-1);
}

/*
 *@@ ParsePckAttribsConfigSys:
 *      this private method gets called by
 *      BSOldPckDecl::ParsePckAttribsMain to parse
 *      the CONFIGSYS attributes of a given PCK tag
 *      and create a list of BSConfigSys instances (bs_config.h)
 *      in pConfigSysList.
 *
 *      Returns the number of BSConfigSys instances created
 *      (0 if none found).
 *
 *      This terminates WarpIN upon errors (by calling
 *      FEArchive::ParseError), giving the user a
 *      meaningful error message.
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: this used to be a global function in warpin.cpp
 *@@changed V0.9.1 (2000-02-07) [umoeller]: fixed memory leak
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 */

LONG BSOldPckDecl::ParsePckAttribsConfigSys(const char *pszPckAttribs,  // in: attributes block of PCK tag
                                            ULONG ulAttrOfs)    // in: offset of pszPckAttribs in script (for errors)
{
    LONG        lrc = 0;
    ULONG       ulOfs = 0;
    const char  *pszSearchIn = pszPckAttribs;

    while (lrc != -1)   // no error?
    {
        PSZ     pszAttrThis = NULL;

        pszAttrThis = strhGetTextAttr(pszSearchIn,
                                      "CONFIGSYS", &ulOfs);
        if (!pszAttrThis)
            // no more attribs found:
                break;  // while

        // else: pszAttr now has the whole attribute

        // we now create an instance of BSCfgSysManip (cfgsys.h),
        // whose constructor is capable of parsing this attribute

        try
        {
            BSCfgSysManip* pBSCfgSysManip = new BSCfgSysManip(pszAttrThis);
                            // this may throw a BSConfigExcpt, caught below
            _listConfigObjects.push_back(pBSCfgSysManip);
            // _listCfgSysManips.push_back(pBSCfgSysManip);
            lrc++;

            _ulConfigData |= PCK_CONFIGSYS;

            // append this attribute string to the log string
            _logCfgSysAttrs.Append(pszAttrThis);
        }
        catch (BSConfigExcpt& X)
        {
            // thrown by BSCfgSysManip constructor:
            CHAR        szError[1000] = "unknown";

            switch (X._iErrorCode)
            {
                case CFGEXCPT_DOUBLEREPLACEMODE:
                    sprintf(szError,
                            "Replace mode specified twice (\"%s\")",
                            X._strDescription.c_str());
                    break;
                case CFGEXCPT_INVALIDVERTICAL:
                    sprintf(szError,
                            "Invalid modifier combination; vertical modifier not allowed (\"%s\")",
                            X._strDescription.c_str());
                    break;
                case CFGEXCPT_INVALIDSEARCH:
                    sprintf(szError,
                            "Invalid search string specified (\"%s\")",
                            X._strDescription.c_str());
                    break;
                case CFGEXCPT_NOTIMPLEMENTED:
                    sprintf(szError,
                            "Keyword not implemented yet (\"%s\")",
                            X._strDescription.c_str());
                default:
                    strcpy(szError, "unknown error");
            }

            _pScript->ParseError(ulAttrOfs, "Error parsing CONFIGSYS attribute of PCK tag:\n%s",
                                 szError);
                // this exits
        }

        if (pszAttrThis)
            free(pszAttrThis);
        // advance search pointer for next search
        pszSearchIn += ulOfs;
    }

    return (lrc);
}

/*
 *@@ ParsePckAttribsWPSClasses:
 *      this private method gets called by
 *      BSOldPckDecl::ParsePckAttribsMain to parse
 *      the REGISTERCLASS and REPLACECLASS attributes
 *      of a given PCK tag and create two lists of
 *      BSRegisterClass and BSReplaceClass instances (bs_config.h)
 *      in listRegisterClassAttrs/listReplaceClass.
 *
 *      Returns the total number of instances created (0 if none found).
 *
 *      This terminates WarpIN upon errors (by calling
 *      FEArchive::ParseError), giving the user a
 *      meaningful error message.
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: this used to be a global function in warpin.cpp
 *@@changed V0.9.1 (2000-02-07) [umoeller]: fixed memory leak
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 */

LONG BSOldPckDecl::ParsePckAttribsWPSClasses(const char *pszPckAttribs,  // in: attributes block of PCK tag
                                             ULONG ulAttrOfs)    // in: offset of pszPckAttribs in script (for errors)
{
    LONG        lrc = 0;
    ULONG       ulOfs = 0;

    // step 1: get the REGISTERCLASS attributes
    const char  *pszSearchIn = pszPckAttribs;

    while (lrc != -1)   // no error?
    {
        PSZ     pszAttrThis = NULL;
        try
        {
            pszAttrThis = strhGetTextAttr(pszSearchIn,
                                          "REGISTERCLASS", &ulOfs);
            if (!pszAttrThis)
                // no more attribs found:
                break;

            // else: pszAttr now has the whole attribute, which is like this:
            // "classname|classmodule"
            // (without quotes)

            BSRegisterClass *pRegisterClass = new BSRegisterClass(pszAttrThis);
                            // this may throw a BSConfigExcpt, caught below
            _listConfigObjects.push_back(pRegisterClass);
            // _listRegisterClasses
            lrc++;

            _ulConfigData |= PCK_WPSCLASSES;

            // append this attribute string to the log string
            _logRegisterClassAttrs.Append(pszAttrThis);

            // advance search pointer for next search
            pszSearchIn += ulOfs;
        }
        catch (BSConfigExcpt& X)
        {
            // thrown by BSCreateWPSObject constructor
            CHAR        szError[1000] = "unknown";

            switch (X._iErrorCode)
            {
                case REGEXCPT_SYNTAX:
                    sprintf(szError,
                            "Invalid class specification; must be \"classname|classdll\" (is: \"%s\")",
                            X._strDescription.c_str());
                break;
                default:
                    strcpy(szError, "unknown error");
            }
            _pScript->ParseError(ulAttrOfs, "Error parsing REGISTERCLASS attribute of PCK tag:\n%s",
                                 szError);
                // this exits
        }

        if (pszAttrThis)
            free(pszAttrThis);
    }

    // step 2: get the REPLACECLASS attributes

    // reset the search pointer
    pszSearchIn = pszPckAttribs;

    while (lrc != -1)   // no error?
    {
        PSZ     pszAttrThis = NULL;

        try
        {
            pszAttrThis = strhGetTextAttr(pszSearchIn,
                                          "REPLACECLASS", &ulOfs);
            if (!pszAttrThis)
                // no more attribs found:
                break;

            // else: pszAttr now has the whole attribute, which is like this:
            // "oldclassname|newclassname"
            // (without quotes)

            BSReplaceClass *pReplaceClass = new BSReplaceClass(pszAttrThis);
                            // this may throw a BSConfigExcpt, caught below
            // _listReplaceClasses
            _listConfigObjects.push_back(pReplaceClass);
            lrc++;

            _ulConfigData |= PCK_WPSCLASSES;

            // append this attribute string to the log string
            _logReplaceClassAttrs.Append(pszAttrThis);

            // advance search pointer for next search
            pszSearchIn += ulOfs;
        }
        catch (BSConfigExcpt& X)
        {
            // thrown by BSCreateWPSObject constructor
            CHAR        szError[1000] = "unknown";

            switch (X._iErrorCode)
            {
                case REGEXCPT_SYNTAX:
                    sprintf(szError,
                            "Invalid class specification; must be \"classname|classdll\" (is: \"%s\")",
                            X._strDescription.c_str());
                break;
                default:
                    strcpy(szError, "unknown error");
            }
            _pScript->ParseError(ulAttrOfs, "Error parsing REPLACECLASS attribute of PCK tag:\n%s",
                                 szError);
                // this exits
        }

        if (pszAttrThis)
            free(pszAttrThis);
    }

    return (lrc);
}

/*
 *@@ ParsePckAttribsWPSObjects:
 *      this private method gets called by
 *      BSOldPckDecl::ParsePckAttribsMain to parse
 *      the CREATEOBJECT attributes of a given PCK tag
 *      and create a list of BSCreateWPSObject instances (bs_config.h)
 *      in listWPSObjectAttrs.
 *
 *      Returns the number of BSCreateWPSObject instances created (0 if none found).
 *
 *      This terminates WarpIN upon errors (by calling
 *      FEArchive::ParseError), giving the user a
 *      meaningful error message.
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: this used to be a global function in warpin.cpp
 *@@changed V0.9.1 (2000-02-07) [umoeller]: fixed memory leak
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 */

LONG BSOldPckDecl::ParsePckAttribsWPSObjects(const char *pszPckAttribs,  // in: attributes block of PCK tag
                                             ULONG ulAttrOfs)    // in: offset of pszPckAttribs in script (for errors)
{
    LONG        lrc = 0;
    ULONG       ulOfs = 0;
    const char  *pszSearchIn = pszPckAttribs;

    while (lrc != -1)   // no error?
    {
        PSZ     pszAttrThis = NULL;

        try
        {
            pszAttrThis = strhGetTextAttr(pszSearchIn,
                                          "CREATEOBJECT", &ulOfs);
            if (!pszAttrThis)
                // no more attribs found:
                break;

            // else: pszAttr now has the whole attribute, which is like this:
            // "[REPLACE] classname|title|location[|config]"
            // (without quotes)

            BSCreateWPSObject *pCreateWPSObject = new BSCreateWPSObject(pszAttrThis);
                            // this may throw a BSConfigExcpt, caught below
            // _listWPSObjects
            _listConfigObjects.push_back(pCreateWPSObject);
            lrc++;

            _ulConfigData |= PCK_WPSOBJECTS;

            // append this attribute string to the log string
            _logWPSObjectAttrs.Append(pszAttrThis);
        }
        catch (BSConfigExcpt& X)
        {
            // thrown by BSCreateWPSObject constructor (ParseCreateObject)
            CHAR        szError[1000] = "unknown";

            switch (X._iErrorCode)
            {
                case WPOEXCPT_NOLOCATION:
                    sprintf(szError,
                            "No location specified (\"%s\")",
                            X._strDescription.c_str());
                    break;
                case WPOEXCPT_NOTITLE:
                    sprintf(szError,
                            "No title specified (\"%s\")",
                            X._strDescription.c_str());
                    break;
                case WPOEXCPT_NOCLASS:
                    sprintf(szError,
                            "No class specified (\"%s\")",
                            X._strDescription.c_str());
                    break;
                default:
                    strcpy(szError, "unknown error");
            }
            _pScript->ParseError(ulAttrOfs, "Error parsing CREATEOBJECT attribute of PCK tag:\n%s",
                                 szError);
                // this exits
        }

        if (pszAttrThis)
            free(pszAttrThis);
        // advance search pointer for next search
        pszSearchIn += ulOfs;
    }

    return (lrc);
}

/*
 *@@ ParsePckAttribsProfiles:
 *      this private method gets called by
 *      BSOldPckDecl::ParsePckAttribsMain
 *      to parse the CLEARPROFILE and WRITEPROFILE strings.
 *
 *      Returns the no. of matching attributes found.
 *
 *      This terminates WarpIN upon errors (by calling
 *      FEArchive::ParseError), giving the user a
 *      meaningful error message.
 *
 *@@added V0.9.1 (2000-02-07) [umoeller]
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 */

LONG BSOldPckDecl::ParsePckAttribsProfiles(const char *pszPckAttribs,
                                           ULONG ulAttrOfs)
{
    LONG        lrc = 0;
    const char  *pszSearchIn = pszPckAttribs;

    while (lrc != -1)   // no error?
    {
        ULONG   ulOfs = 0;
        PSZ     pszAttrThis = NULL;
        try
        {
            pszAttrThis = strhGetTextAttr(pszSearchIn,
                                          "CLEARPROFILE", &ulOfs);
            if (!pszAttrThis)
                // no more attribs found:
                break;

            // else: pszAttr now has the whole attribute
            lrc++;
            // we just store this in the logger, it's only
            // for de-installation
            _logClearPrfAttrs.Append(pszAttrThis);

            _ulConfigData |= PCK_CLEARPROFILES;
        }
        catch (BSConfigExcpt& X)
        {
            // thrown by BSCreateWPSObject constructor (ParseCreateObject)
            CHAR        szError[1000] = "unknown";

            switch (X._iErrorCode)
            {
                case PRFEXCPT_SYNTAX:
                    sprintf(szError,
                            "Invalid syntax (%s)",
                            X._strDescription.c_str());
                    break;
                default:
                    strcpy(szError, "unknown error");
            }
            _pScript->ParseError(ulAttrOfs, "Error parsing CLEARPROFILE attribute of PCK tag:\n%s",
                                 szError);
                // this exits
        }

        if (pszAttrThis)
            free(pszAttrThis);
        // advance search pointer for next search
        pszSearchIn += ulOfs;
    }

    // reset search buffer
    pszSearchIn = pszPckAttribs;

    while (lrc != -1)   // no error?
    {
        ULONG   ulOfs = 0;
        PSZ     pszAttrThis = NULL;
        try
        {
            pszAttrThis = strhGetTextAttr(pszSearchIn,
                                          "WRITEPROFILE", &ulOfs);
            if (!pszAttrThis)
                // no more attribs found:
                break;

            // else: pszAttr now has the whole attribute
            BSWriteProfile *pWriteProfile = new BSWriteProfile(pszAttrThis);
                            // this may throw a BSConfigExcpt, caught below
            // _listWriteProfiles
            _listConfigObjects.push_back(pWriteProfile);
            lrc++;

            // store this in the logger for de-installation
            _logWritePrfAttrs.Append(pszAttrThis);

            _ulConfigData |= PCK_WRITEPROFILES;
        }
        catch (BSConfigExcpt& X)
        {
            // thrown by BSCreateWPSObject constructor (ParseCreateObject)
            CHAR        szError[1000] = "unknown";

            switch (X._iErrorCode)
            {
                case KILLEXCPT_SYNTAX:
                    sprintf(szError,
                            "Invalid syntax (%s)",
                            X._strDescription.c_str());
                    break;
                default:
                    strcpy(szError, "unknown error");
            }
            _pScript->ParseError(ulAttrOfs, "Error parsing KILLPROCESS attribute of PCK tag:\n%s",
                                 szError);
                // this exits
        }

        if (pszAttrThis)
            free(pszAttrThis);
        // advance search pointer for next search
        pszSearchIn += ulOfs;
    }

    return (lrc);
}

/*
 *@@ ParsePckAttribsExecutes:
 *      this private method gets called by
 *      BSOldPckDecl::ParsePckAttribsMain to parse
 *      the EXECUTE and DEEXECUTE strings.
 *
 *      Returns the no. of EXECUTE/DEEXECUTE attributes found.
 *
 *      This terminates WarpIN upon errors (by calling
 *      FEArchive::ParseError), giving the user a
 *      meaningful error message.
 *
 *@@added V0.9.1 (2000-02-03) [umoeller]
 *@@changed V0.9.1 (2000-02-07) [umoeller]: fixed memory leak
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 *@@changed V0.9.9 (2001-03-27) [umoeller]: added DEEXECUTE support
 */

LONG BSOldPckDecl::ParsePckAttribsExecutes(const char *pszPckAttribs,
                                           ULONG ulAttrOfs)
{
    LONG        lrc = 0;
    ULONG       ulOfs = 0;
    const char  *pszSearchIn = pszPckAttribs;

    while (lrc != -1)   // no error?
    {
        PSZ     pszAttrThis = NULL;
        try
        {
            pszAttrThis = strhGetTextAttr(pszSearchIn,
                                          "EXECUTE", &ulOfs);
            if (!pszAttrThis)
                // no more attribs found:
                break;

            // else: pszAttr now has the whole attribute, which is like this:
            // "[context | ]execfile [params]"
            BSExecute *pExecute = new BSExecute(pszAttrThis);
                            // this may throw a BSConfigExcpt, caught below
            _listConfigObjects.push_back(pExecute);
            lrc++;

            // append this attribute string to the log string
            _logExecuteAttrs.Append(pszAttrThis);

            _ulConfigData |= PCK_EXECUTES;

            // copy the execute's config stuff to the package's one
            if (pExecute->_ulExecType == CFGT_CFGSYS)
                _ulConfigData |= PCK_CONFIGSYS;
            if (pExecute->_ulExecType == CFGT_REGISTERCLASS)
                _ulConfigData |= PCK_WPSCLASSES;
            if (pExecute->_ulExecType == CFGT_CREATEOBJECT)
                _ulConfigData |= PCK_WPSOBJECTS;
        }
        catch (BSConfigExcpt& X)
        {
            // thrown by BSExecute constructor
            CHAR        szError[1000] = "unknown";

            switch (X._iErrorCode)
            {
                case EXEEXCPT_SYNTAX:
                    sprintf(szError,
                            "Invalid syntax (%s)",
                            X._strDescription.c_str());
                    break;
                default:
                    strcpy(szError, "unknown error");
            }
            _pScript->ParseError(ulAttrOfs, "Error parsing EXECUTE attribute of PCK tag:\n%s",
                                 szError);
                // this exits
        }

        if (pszAttrThis)
            free(pszAttrThis);
        // advance search pointer for next search
        pszSearchIn += ulOfs;
    }

    // restart for DEEXECUTE
    ulOfs = 0;
    pszSearchIn = pszPckAttribs;

    while (lrc != -1)   // no error?
    {
        PSZ     pszAttrThis = NULL;
        try
        {
            pszAttrThis = strhGetTextAttr(pszSearchIn,
                                          "DEEXECUTE", &ulOfs);
            if (!pszAttrThis)
                // no more attribs found:
                break;

            // else: pszAttr now has the whole attribute, which is like this:
            // "execfile [params]"
            BSDeExecute *pExecute = new BSDeExecute(pszAttrThis);
                            // this may throw a BSConfigExcpt, caught below
            _listConfigObjects.push_back(pExecute);
            lrc++;

            // append this attribute string to the log string
            _logDeExecuteAttrs.Append(pszAttrThis);

            _ulConfigData |= PCK_DEEXECUTES;
        }
        catch (BSConfigExcpt& X)
        {
            // thrown by BSCreateWPSObject constructor (ParseCreateObject)
            CHAR        szError[1000] = "unknown";

            switch (X._iErrorCode)
            {
                case EXEEXCPT_SYNTAX:
                    sprintf(szError,
                            "Invalid syntax (%s)",
                            X._strDescription.c_str());
                    break;
                default:
                    strcpy(szError, "unknown error");
            }
            _pScript->ParseError(ulAttrOfs, "Error parsing DEEXECUTE attribute of PCK tag:\n%s",
                                 szError);
                // this exits
        }

        if (pszAttrThis)
            free(pszAttrThis);
        // advance search pointer for next search
        pszSearchIn += ulOfs;
    }

    return (lrc);
}

/*
 *@@ ParsePckAttribsKillProcesses:
 *      this private method gets called by
 *      BSOldPckDecl::ParsePckAttribsMain to parse
 *      the KILLPROCESS strings.
 *
 *      Returns the no. of KILLPROCESS attributes found.
 *
 *      This terminates WarpIN upon errors (by calling
 *      FEArchive::ParseError), giving the user a
 *      meaningful error message.
 *
 *@@added V0.9.1 (2000-02-12) [umoeller]
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 */

LONG BSOldPckDecl::ParsePckAttribsKillProcesses(const char *pszPckAttribs,
                                                ULONG ulAttrOfs)
{
    LONG        lrc = 0;
    ULONG       ulOfs = 0;
    const char  *pszSearchIn = pszPckAttribs;

    while (lrc != -1)   // no error?
    {
        PSZ     pszAttrThis = NULL;

        try
        {
            pszAttrThis = strhGetTextAttr(pszSearchIn,
                                          "KILLPROCESS", &ulOfs);
            if (!pszAttrThis)
                // no more attribs found:
                break;

            // else: pszAttr now has the whole attribute, which is like this:
            // "[REPLACE] classname|title|location[|config]"
            // (without quotes)

            BSKillProcess *pKillProcess = new BSKillProcess(pszAttrThis);
                            // this may throw a BSConfigExcpt, caught below
            _listConfigObjects.push_back(pKillProcess);
            lrc++;

            _ulConfigData |= PCK_KILLPROCESS;

            // append this attribute string to the log string
            _logKillProcessAttrs.Append(pszAttrThis);
        }
        catch (BSConfigExcpt& X)
        {
            // thrown by BSCreateWPSObject constructor (ParseCreateObject)
            CHAR        szError[1000] = "unknown";

            switch (X._iErrorCode)
            {
                case WPOEXCPT_NOLOCATION:
                    sprintf(szError,
                            "No location specified (\"%s\")",
                            X._strDescription.c_str());
                    break;
                case WPOEXCPT_NOTITLE:
                    sprintf(szError,
                            "No title specified (\"%s\")",
                            X._strDescription.c_str());
                    break;
                case WPOEXCPT_NOCLASS:
                    sprintf(szError,
                            "No class specified (\"%s\")",
                            X._strDescription.c_str());
                    break;
                default:
                    strcpy(szError, "unknown error");
            }
            _pScript->ParseError(ulAttrOfs, "Error parsing CREATEOBJECT attribute of PCK tag:\n%s",
                                 szError);
                // this exits
        }

        if (pszAttrThis)
            free(pszAttrThis);
        // advance search pointer for next search
        pszSearchIn += ulOfs;
    }

    return (lrc);
}

/* ******************************************************************
 *
 *  BSOldScript implementation
 *
 ********************************************************************/

/*
 *@@ Parse:
 *      virtual BSScriptBase method which is overridden
 *      here to parse old-style scripts.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

VOID BSOldScript::Parse(const BSString &strScript,
                        PFNCONFIRMREXXALLOWED pfnConfirmRexxAllowed)
{
    ParseWARPINTag(strScript);

    ParseHeader(strScript,
                pfnConfirmRexxAllowed);

    ParsePages(strScript);
}

/*
 *@@ ParseError:
 *      throws BSScriptExcpt when an error occured parsing
 *      the installation script. This will terminate WarpIN.
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: this was wpiParseError (warpin.cpp)
 *@@changed V0.9.0 (99-10-31) [umoeller]: now throws a FEScriptExcpt
 *@@changed V0.9.1 (2000-01-07) [umoeller]: now throws an FEFatalErrorExcpt
 *@@changed V0.9.9 (2001-02-28) [umoeller]: added vargs
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 */

VOID BSOldScript::ParseError(ULONG ulOfs,
                             const char *pcszFormat,
                             ...)
{
    BSString strText("Error");

    /* if (ulOfs)
    {
        // count lines up to offset
        ULONG   ulLine = 1;
        PSZ     p = (PSZ)_strProfile.c_str(),
                pEnd = p + ulOfs;
        while (p < pEnd)
        {
            if (*p == '\n')
                ulLine++;
            p++;
        }
        strText._printf("WarpIn encountered an error in the installation"
                        " profile of \"%s\" (line %u):\n",
                        _szArcFilename,
                        ulLine);
    } */

    BSString str2;
    str2._printf(pcszFormat); // , pszExtra);
    strText += str2;

    throw BSScriptExcpt(strText);
}

/*
 *@@ GetBlock:
 *      wrapper function for strhGetBlock, which
 *      works on BSString's and can also display an
 *      error message.
 *
 *      strhGetBlock returns 1 if pszTag was not
 *      found and 2 if the opening, but no closing
 *      tag was found.
 *
 *      This is only a private helper method.
 *
 *      This now has full REXX support if the flREXX
 *      flags are set to anything != 0. If so,
 *      FERexx::ExecuteRexxMacros gets called
 *      with the contents of *pstrBlock and/or
 *      *pstrAttribs.
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: this was wpiGetBlock (warpin.cpp)
 *@@changed V0.9.0 (99-11-07) [umoeller]: now using BSString's
 *@@changed V0.9.1 (2000-01-06) [umoeller]: updated prototype because line numbers were always wrong
 *@@changed V0.9.2 (2000-03-11) [umoeller]: added REXX support
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 */

BOOL BSOldScript::GetBlock(const char *pcszSearchIn,
                                    // in: buffer to search
                           PULONG pulSearchOfs,
                                      // in/out: where to begin search (or 0 for start)
                           PSZ pszTag,
                                      // in: tag to search for (without brackets)
                           ULONG flREXX,
                                      // in: REXX flags; any or none of the following:
                                      // -- GBLREXX_ATTRIBS: execute REXX code in attribs
                                      // -- GBLREXX_BLOCK: execute REXX code in block
                           BSString *pstrBlock,
                                      // out: stuff between opening and closing tag;
                                      // pointer can be NULL if not desired
                           BSString *pstrAttribs,
                                      // out: attributes of opening tag;
                                      // pointer can be NULL if not desired
                           PULONG pulOfsBeginTag,
                           PULONG pulOfsBeginBlock)
{
    PSZ     pszBlock = 0;
    PSZ     pszAttribs = 0;

    ULONG ulrc  = strhGetBlock(pcszSearchIn,
                                pulSearchOfs,
                               pszTag,
                               &pszBlock,
                               &pszAttribs,
                               pulOfsBeginTag,
                               pulOfsBeginBlock);

    if (ulrc == 2)
    {
        ParseError(*pulSearchOfs,
                   "Missing closing tag for \"<%s>\" tag.",
                   pszTag);
            // this terminates WarpIN
    }

    if (pstrBlock)
        *pstrBlock = pszBlock;
    if (pstrAttribs)
        *pstrAttribs = pszAttribs;

    if (pszBlock)
        free(pszBlock);
    if (pszAttribs)
        free(pszAttribs);

    if (flREXX & GBLREXX_ATTRIBS)
    {
        BSString strRexxMacroName;
        ULONG ulRexxOfs;

        if(!G_pFERexx->ExecuteRexxMacros(*pstrAttribs,
                                         &ulRexxOfs,
                                         strRexxMacroName))
        {
            CHAR    szMsg[1000];
            sprintf(szMsg,
                    "\nThe REXX code in the <%s> attributes failed.\n"
                    "Macro call: =(\"%s\")",
                    pszTag,
                    strRexxMacroName.c_str());
            ParseError(*pulSearchOfs + ulRexxOfs,
                       szMsg,
                       0);
        }
    }

    if (flREXX & GBLREXX_BLOCK)
    {
        BSString strRexxMacroName;
        ULONG ulRexxOfs;

        if(!G_pFERexx->ExecuteRexxMacros(*pstrBlock,
                                         &ulRexxOfs,
                                         strRexxMacroName))
        {
            CHAR    szMsg[1000];
            sprintf(szMsg,
                    "The REXX code within the <%s>...</%s> block failed.\n"
                    "Macro call: =(\"%s\")",
                    pszTag,
                    pszTag,
                    strRexxMacroName.c_str());
            ParseError(*pulSearchOfs + ulRexxOfs,
                       szMsg,
                       0);
        }
    }

    return (ulrc == 0);     // TRUE if tag was found
}

/*
 *@@ ParseWARPINTag:
 *      this gets called by FEArchive::FEArchive to parse
 *      the WARPIN tag attributes. This terminates WarpIN
 *      if the requirements are not met.
 *
 *      Throws:
 *      -- FEFatalErrorExcpt.
 *      -- BSCancelExcpt.
 *
 *@@added V0.9.0 (99-11-06) [umoeller]
 *@@changed V0.9.2 (2000-03-11) [umoeller]: added CODEPAGE attribute
 *@@changed V0.9.4 (2000-07-26) [umoeller]: added three-part WARPIN version checks
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 */

VOID BSOldScript::ParseWARPINTag(const BSString &strScript)
{
    while (TRUE)
    {
        // PSZ         pszProfile2 = (PSZ)_strProfile.c_str(); // pointer gets modified
        ULONG       ulOfs = 0;
        BSString    strWarpINBlock,
                    strWarpINAttrs;
        ULONG       ulWarpINTagOfs = 0,
                    ulWarINBlockOfs = 0;

        if (!GetBlock(strScript.c_str(), &ulOfs,
                      "WARPIN",
                      0, // no REXX
                      &strWarpINBlock, &strWarpINAttrs,
                      &ulWarpINTagOfs, &ulWarINBlockOfs))
        {
            throw (BSScriptExcpt(170));
        }

        // codepage
        if (!strhGetNumAttribValue(strWarpINAttrs.c_str(),
                                   "CODEPAGE", (PLONG)&_ulCodepage))
            _ulCodepage = 850;

    /*
    ULONG   ulrc = guiSetCodepage(_ulCodepage,
                                  _strProfile);
    if (ulrc == GUICP_TRANSLATED)
        continue;       // reparse stuff
    else if (ulrc == GUICP_FAILED)
    {
        CHAR szMsg[20];
        sprintf(szMsg, "%u", _ulCodepage);
        PSZ psz = szMsg;
        wpiMessageBox(107,   // warning
                      186,  // cp %1 not available
                      MB_ICONEXCLAMATION | MB_OK,
                      &psz, 1);
    } ### */

        // VERSION attribute
        PSZ pszReqVersion = strhGetTextAttr(strWarpINAttrs.c_str(), "VERSION", NULL);
        if (pszReqVersion)
        {
            ULONG   ulReqMajor,
                    ulReqMinor,
                    ulReqRevision = 0;
            ULONG ulScanned = sscanf(pszReqVersion,
                                     "%u.%u.%u",
                                     &ulReqMajor,
                                     &ulReqMinor,
                                     &ulReqRevision);
            ULONG   ulRunMajor, ulRunMinor, ulRunRevision;
            sscanf(BLDLEVEL_VERSION,
                   "%u.%u.%u",
                   &ulRunMajor,
                   &ulRunMinor,
                   &ulRunRevision);

            if (    // major higher
                    (ulReqMajor > ulRunMajor)
                 || (   // major equal:
                        (ulReqMajor == ulRunMajor)
                        // then minor must be higher
                     && (   (ulReqMinor > ulRunMinor)
                            // minor equal as well:
                         || (   (ulReqMinor == ulRunMinor)
                                // then revision must be higher
                             && (ulReqRevision > ulRunRevision)
                            )
                        )
                    )
               )
            {
                throw BSScriptExcpt(171,
                                    &pszReqVersion,
                                    1);
            }

            free(pszReqVersion);
        }

        break;
    } // end while TRUE;
}

/*
 *@@ ParseHeader:
 *      this private method gets called by the constructor
 *      to parse the HEAD.../HEAD block. This calls the
 *      other Parse* methods in turn.
 *
 *      Returns the no. of tags parsed.
 *
 *      Throws:
 *      -- BSCancelExcpt.
 *
 *@@added V0.9.0 (99-11-07) [umoeller]
 *@@changed V0.9.2 (2000-03-10) [cbo]: added bExternalCall
 *@@changed V0.9.2 (2000-03-10) [cbo]: added REXX support
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 */

ULONG BSOldScript::ParseHeader(const BSString &strScript,
                               PFNCONFIRMREXXALLOWED pfnConfirmRexxAllowed)
{
    ULONG   ulrc = 0;

    // first, extract the <HEAD>...</HEAD> part
    // from the profile

    ULONG       ulOfs = 0;
    BSString    strHeadBlock,
                strHeadAttrs,
                strRexxMacroName; // (cbo 6.3.00)
    ULONG       ulHeadTagOfs = 0,
                ulHeadBlockOfs = 0,
                ulBlockOfs = 0; // (cbo 6.3.00)
    if (!GetBlock(strScript.c_str(), &ulOfs,
                  "HEAD",
                  0, // no REXX
                  &strHeadBlock, &strHeadAttrs,
                  &ulHeadTagOfs, &ulHeadBlockOfs))
    {
        throw BSScriptExcpt(104);  // "no <HEAD>"
    }

    // parse the block for <REXX>...</REXX> blocks and store the code.
    // If a "name" attribute occurs twice, the second occurrence will
    // overwrite the stored code from the first time.
    // V0.9.2 (2000-03-10) [cbo]
    ULONG cRexxTags = G_pFERexx->ParseRexxTags(strHeadBlock.c_str());

    if ((cRexxTags) && (pfnConfirmRexxAllowed))
    {
        // there are REXX tags, and caller wants confirmation:
        if (!pfnConfirmRexxAllowed(strScript.c_str()))
            // confirmation func returned FALSE;
            throw(BSCancelExcpt());
    }

    ulrc += cRexxTags;
    // check whether a <MSG>...</MSG> block
    // exists in the HEAD block; if so, display
    // message and stop
    ulOfs = 0;
    BSString    strMessage;
    if (    (GetBlock(strHeadBlock.c_str(),
                      &ulOfs,
                      "MSG",
                      GBLREXX_BLOCK,
                      &strMessage, 0, 0, 0))
       )
    {
        throw BSScriptExcpt(strMessage);
    }

    // parse all VARPROMPT blocks
    ulrc += ParseVarPrompts(strHeadBlock,
                            ulHeadBlockOfs);

    // check if we have a <TITLE>...</TITLE> block;
    // this is optional and stored in the instance data
    ulOfs = 0;
    GetBlock(strHeadBlock.c_str(),
             &ulOfs,
             "TITLE",
             GBLREXX_BLOCK,
             &_strTitle, 0, 0, &ulBlockOfs);

    // set up the list of FEPackages
    ulrc += ParsePackages(strHeadBlock,
                          NULL,             // for first call (this might recurse)
                          ulHeadBlockOfs);

    return (ulrc);
}

/*
 *@@ ParseVarPrompts:
 *      this gets called by FEArchive::ParseHeader to parse
 *      all VARPROMPT tags in the HEAD block.
 *
 *@@added V0.9.2 (2000-03-11) [umoeller]
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 */

ULONG BSOldScript::ParseVarPrompts(BSString strBlock, // in: HEAD block to process (first call)
                                   ULONG ulTotalBlockOfs) // in: offset of strBlock in profile (for error messages)
{
    ULONG   ulrc = 0;
    PSZ     pszBlock = (PSZ)strBlock.c_str();  // pointer gets modified as we search on
    ULONG   ulSearchOfs = 0;

    do
    {
        BSString    strVarBlock,
                    strVarAttribs;
        ULONG       ulVarTagOfs = 0,
                    ulVarBlockOfs = 0;
        // extract new block
        if (GetBlock(pszBlock, // pszSearch,
                     &ulSearchOfs,      // advances search offset to after VARPROMPT
                     "VARPROMPT",
                     GBLREXX_BLOCK,
                     &strVarBlock,
                     &strVarAttribs,
                     &ulVarTagOfs, &ulVarBlockOfs))
        {
            // found: get name
            PSZ pszVarNameAttr;
            if (!(pszVarNameAttr = strhGetTextAttr(strVarAttribs.c_str(), "NAME", NULL)))
            {
                ParseError(ulTotalBlockOfs + ulVarTagOfs,
                           "The VARPROMPT tag contains no NAME attribute. ", NULL);
            }

            PSZ     pszVarTypeAttr;
            ULONG   ulType = 0;
            ULONG   ulMin = 0,
                    ulMax = 0;
            if (!(pszVarTypeAttr = strhGetTextAttr(strVarAttribs.c_str(), "TYPE", NULL)))
            {
                ParseError(ulTotalBlockOfs + ulVarTagOfs,
                           "The VARPROMPT tag contains no TYPE attribute. ", NULL);
            }
            if (stricmp(pszVarTypeAttr, "ALPHA") == 0)
                ulType = VPT_ALPHANUM;
            else if (stricmp(pszVarTypeAttr, "PATH") == 0)
                ulType = VPT_PATH;
            else if (stricmp(pszVarTypeAttr, "FAIL") == 0)
                ulType = VPT_FAIL;
            else if (strnicmp(pszVarTypeAttr, "NUM", 3) == 0)
            {
                ulType = VPT_NUM;
                PSZ psz2 = strhExtract(pszVarTypeAttr, '(', ')', NULL);
                if (psz2)
                {
                    sscanf(psz2, "%u,%u", &ulMin, &ulMax);
                }
                if (ulMax == 0)
                    ParseError(ulTotalBlockOfs + ulVarTagOfs,
                               "The VARPROMPT tag has an invalid TYPE syntax. ", NULL);
            }
            else ParseError(ulTotalBlockOfs + ulVarTagOfs,
                            "The VARPROMPT tag has an invalid TYPE attribute. ", NULL);

            BSVarPrompt VarPrompt(pszVarNameAttr);
            VarPrompt._strDescription = strVarBlock;
            VarPrompt._ulType = ulType;
            VarPrompt._ulMin = ulMin;
            VarPrompt._ulMax = ulMax;
            // append to static string map
            VarPromptsMap[BSString(pszVarNameAttr)] = VarPrompt;

            free(pszVarNameAttr);
            free(pszVarTypeAttr);

            ulrc++;
        }
        else
            // no more VARPROMPTs:
            break;

    } while (TRUE);
        // loop forever; there's a break in the loop body

    return (ulrc);
}

/*
 *@@ ParsePackages:
 *      this gets called by FEArchive::ParseHeader to set up
 *      the list of FEPackages. To do this, we analyze the
 *      PCK tags in the HEAD block of the installation profile.
 *
 *      This routine possibly calls itself recursively
 *      if GROUP tags are found.
 *
 *      It is initially called with pszBlock containing
 *      the HEAD block and will then call itself with a
 *      new block if GROUP tags are found therein.
 *
 *      This is a private method which gets called by the
 *      constructor and calls the other Parse* methods in turn.
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: this used to be a global function in warpin.cpp
 *@@changed V0.9.1 (2000-01-06) [umoeller]: fixed wrong line numbers in parse error msgs
 *@@changed V0.9.2 (2000-03-11) [umoeller]: added EXTERNAL support
 *@@changed V0.9.3 (2000-03-24) [umoeller]: added EXTERNAL REQUIRED support
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 */

ULONG BSOldScript::ParsePackages(BSString strBlock,
                                        // in: HEAD block to process (first call)
                                        // or GROUP block (recursive calls)
                                 BSGroupDeclBase *pGroup,
                                        // in: group package to append
                                        // new packages to; initially NULL
                                 ULONG ulTotalBlockOfs)
                                        // in: offset of strBlock in profile (for error messages)
{
    ULONG   ulrc = 0;
    PSZ     pszBlock = (PSZ)strBlock.c_str(),  // pointer gets modified as we search on
            pszPackage,
            pszGroup;
    ULONG   ulSearchOfs = 0;
            // ulRexxOfs = 0; // (cbo 7.3.00)
    BSString strRexxMacroName; // (cbo 7.3.00)

    // while we have packages to go
    while ((pszPackage = strhistr(pszBlock + ulSearchOfs, "<PCK")))
    {
        // check if we have a <GROUP> in this block
        pszGroup = strhistr(pszBlock + ulSearchOfs, "<GROUP");
        if (    // <GROUP> found?
                (pszGroup)
                // <GROUP> before next <PCK>?
             && (pszGroup < pszPackage)
           )
        {
            // if so: extract the text between <GROUP> tags
            // and recurse!
            BSString    strGroupBlock,
                        strGroupAttribs;
            ULONG       ulGroupTagOfs = 0,
                        ulGroupBlockOfs = 0;

            // extract new block
            GetBlock(pszBlock, // pszSearch,
                     &ulSearchOfs,      // advances search offset to after GROUP
                     "GROUP",
                     0, // no REXX
                     &strGroupBlock,
                     &strGroupAttribs,
                     &ulGroupTagOfs, &ulGroupBlockOfs);

            // get group attributes
            PSZ pszTitle = strhGetTextAttr(strGroupAttribs.c_str(), "TITLE", NULL);
            if (!pszTitle)
                ParseError(ulTotalBlockOfs + ulGroupBlockOfs,
                           "No \"TITLE\" attribute given for \"GROUP\" tag.",
                           NULL);

            BOOL fInitiallyExpanded
                = (strhFindAttribValue(strGroupAttribs.c_str(), "EXPANDED") != 0);

            BSGroupDeclBase *pGroupNew = new BSGroupDeclBase(pGroup,
                                                             pszTitle,
                                                             fInitiallyExpanded);
            free(pszTitle);

            // add this to the packages list
            if (pGroup)
                // we're in a subgroup:
                pGroup->_DeclarationsList.push_back(pGroupNew);
            else
                // we're on root level: store in script then
                _DeclarationsList.push_back(pGroupNew);

            // recurse with the group buffer
            ParsePackages(strGroupBlock,
                          pGroupNew,
                          ulTotalBlockOfs + ulGroupBlockOfs);
        }
        else
        {
            // no group, but package next:
            // extract text between <PCK> and </PCK> tags
            BSString    strPckBlock = 0,
                        strPckAttribs = 0;
            ULONG       ulPckTagOfsInBlock = 0,
                        ulPckBlockOfsInBlock = 0;
            if (!GetBlock(pszBlock, // &pszSearch,
                          &ulSearchOfs,
                          "PCK",
                          GBLREXX_BLOCK | GBLREXX_ATTRIBS,
                          &strPckBlock,
                          &strPckAttribs,
                          &ulPckTagOfsInBlock, &ulPckBlockOfsInBlock))
            {
                ParseError(ulTotalBlockOfs + ulPckTagOfsInBlock,
                           "You have specified no information within the PCK block. ", NULL);
            }

            // so now we have:
            //      -- pszInfo: the "..." stuff between the <PCK>...</PCK> tags
            //      -- pszPckAttribs: the "..." stuff after the <PCK ...> tag

            // create package from that; the constructor
            // will parse that string data by itself
            BSOldPckDecl *pPackageNew = new BSOldPckDecl(this, // pArchive,      // archive
                                                         pGroup,
                                                         strPckBlock.c_str(),
                                                         strPckAttribs.c_str(),
                                                         ulTotalBlockOfs + ulPckTagOfsInBlock);

            // add this to the packages list
            if (pGroup)
                // we're in a subgroup:
                pGroup->_DeclarationsList.push_back(pPackageNew);
            else
                // we're on root level: store in script then
                _DeclarationsList.push_back(pPackageNew);

            ulrc++;
        }
    }

    return (ulrc);
}

/*
 *@@ ParsePages:
 *      this gets called by EnterArchiveMode()
 *      to set up a list in WpiGlobals containing all
 *      the BSPageInfo instances. To do this, we analyze the
 *      <PAGE> blocks in the <BODY>...</BODY> block
 *      of the installation profile (in our instance data).
 *
 *      Throws:
 *      -- BSCancelExcpt.
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: this used to be a global function in warpin.cpp
 *@@changed V0.9.1 (2000-01-06) [umoeller]: fixed wrong line numbers in parse error msgs
 *@@changed V0.9.9 (2001-02-28) [umoeller]: moved this here from fe_archive.cpp
 *@@changed V0.9.9 (2001-04-04) [umoeller]: added EXTRACTFROMPCK attribute
 */

ULONG BSOldScript::ParsePages(const BSString &strScript)
{
    // extract the <BODY>...</BODY> part from the script
    // PSZ         pszProfile2 = (PSZ)_strProfile.c_str();
    ULONG       ulSearchOfsInProfile = 0;
    BSString    strBodyBlock,
                strBodyAttrs,
                strRexxMacroName; // (cbo 7.3.00)
    ULONG       ulBodyTagOfs = 0,
                ulBodyBlockOfs = 0,
                ulBlockOfs = 0; // (cbo 7.3.00)
                // ulRexxOfs = 0; // (cbo 7.3.00)
    if (!GetBlock(strScript.c_str(), // pszProfile2,
                  &ulSearchOfsInProfile,
                  "BODY",
                  0, // no REXX
                  &strBodyBlock, &strBodyAttrs,
                  &ulBodyTagOfs, &ulBodyBlockOfs))
    {
        throw BSScriptExcpt(105); // "no <BODY>"
    }

    // start with page one
    LONG    lCurrentPage = 1;

    do {
        const char  *pcszBodyBlock = strBodyBlock.c_str();
                        // start search at the beginning; pointer gets modified
        ULONG       ulBodySearchOfs = 0;
        BSString    strPageBlock, // this will receive the <PAGE>...</PAGE> block
                    strPageAttrs;
        ULONG       ulPageTagOfs = 0,       // offset in entire profile (for errors)
                    ulPageBlockOfs = 0;     // offset in entire profile (for errors)

        BOOL        fExit = FALSE;

        // find the page we're looking for (lCurrentPage)
        do
        {
            // get the next <PAGE> block in <BODY> block
            if (GetBlock(pcszBodyBlock,
                         &ulBodySearchOfs,  // advanced after </PAGE> if found
                         "PAGE",
                         GBLREXX_BLOCK,
                         &strPageBlock,
                         &strPageAttrs,
                         &ulPageTagOfs, &ulPageBlockOfs))
            {
                // PAGE block found:
                LONG lPageFound = 0;
                // another <PAGE> block found:
                // query page index
                if (!(strhGetNumAttribValue(strPageAttrs.c_str(),
                                            "INDEX", &lPageFound)))
                {
                    ParseError(ulBodyBlockOfs + ulPageTagOfs,
                               "The <PAGE> tag contains no INDEX attribute. ", NULL);
                }

                if (lPageFound == lCurrentPage)
                {
                    // it's the page we're looking for:
                    // get out of here
                    break;
                }
            }
            else
                // no other page block found:
                fExit = TRUE;
        } while (!fExit);

        if (fExit)
            // page not found: probably the
            // last page; exit the outer do loop
            break;

        // else page found: create a new WPIPAGEINFO structure
        BSPageInfo*  pPageInfoNew = new BSPageInfo();
                    // this now has a constructor

        pPageInfoNew->_lPageIndex = lCurrentPage;

        // determine page type
        PSZ pszTypeAttr;
        if (!(pszTypeAttr = strhGetTextAttr(strPageAttrs.c_str(), "TYPE", NULL)))
        {
            ParseError(ulBodyBlockOfs + ulPageTagOfs,
                       "The PAGE tag contains no TYPE attribute. ", NULL);
        }
        else if (stricmp(pszTypeAttr, "TEXT") == 0)
        {
            pPageInfoNew->_PageMode = BSPageInfo::MODE_TEXT;
        }
        else if (stricmp(pszTypeAttr, "README") == 0)
        {
            pPageInfoNew->_PageMode = BSPageInfo::MODE_README;
        }
        else if (stricmp(pszTypeAttr, "CONTAINER") == 0)
        {
            pPageInfoNew->_PageMode = BSPageInfo::MODE_CONTAINER;
        }
        else if (stricmp(pszTypeAttr, "CONFIGURE") == 0)
        {
            pPageInfoNew->_PageMode = BSPageInfo::MODE_CONFIGURE;
        }
        else
        {
            ParseError(ulBodyBlockOfs + ulPageTagOfs,
                       "The PAGE tag contains an invalid TYPE attribute (\"%s\"). ",
                       pszTypeAttr);
        }
        free(pszTypeAttr);

        // page intro text
        ULONG       ulPageSearchOfs = 0;
        BSString    strTextBody;
        if (!GetBlock(strPageBlock.c_str(),
                      &ulPageSearchOfs,
                      "TEXT",
                      GBLREXX_BLOCK,
                      &strTextBody, 0, 0, &ulBlockOfs))
        {
            ParseError(ulBodyBlockOfs + ulPageTagOfs,
                       "Every page must contain exactly one \"<TEXT ...>\" tag. "
                       "Page %d does not.",
                       (PSZ)lCurrentPage);
        }
        pPageInfoNew->_strInfoText = strTextBody;

        // get the buttons
        LONG        lTarget = -1;
        BSString    strButtonTitle,
                    strButtonAttrs;
        ULONG       ulButtonTagOfs = 0;

        ulPageSearchOfs = 0;
        if (GetBlock(strPageBlock.c_str(),
                     &ulPageSearchOfs,
                     "NEXTBUTTON",
                     GBLREXX_ATTRIBS | GBLREXX_BLOCK,
                     &strButtonTitle,
                     &strButtonAttrs,
                     &ulButtonTagOfs, &ulBlockOfs))
        {
            // query button target page
            if (!(strhGetNumAttribValue(strButtonAttrs.c_str(), "TARGET", &lTarget)))
                lTarget = -1;

            pPageInfoNew->_lNextButton = lTarget;
            if (strButtonTitle())
            {
                pPageInfoNew->_strNextButtonTitle = strButtonTitle;
            }
            else
                ParseError(ulBodyBlockOfs + ulPageBlockOfs + ulButtonTagOfs,
                           "The NEXTBUTTON tag has no title. ", NULL);
        }
        else
        {
            // no NEXTBUTTON specified:
            // use a default V0.9.2 (2000-03-11) [umoeller]

            // next page index: current page index + 1 (fits for most pages)
            pPageInfoNew->_lNextButton = pPageInfoNew->_lPageIndex + 1;

            pPageInfoNew->_strNextButtonTitle = "";    // the GUI will replace this
        }

        /*
         * MODE_README:
         *
         */

        if (pPageInfoNew->_PageMode == BSPageInfo::MODE_README)
        {
            BSString    strReadmeBody;
            BSString    strReadmeAttrs;
            ULONG       ulReadmeTagOfsInPageBlock = 0,
                        ulReadmeBlockOfsInPageBlock = 0;
            ulPageSearchOfs = 0;
            if (!GetBlock(strPageBlock.c_str(),
                           &ulPageSearchOfs,
                          "README",
                          GBLREXX_BLOCK,
                          &strReadmeBody, &strReadmeAttrs,
                          &ulReadmeTagOfsInPageBlock, &ulReadmeBlockOfsInPageBlock))
            {
                ParseError(ulBodyBlockOfs + ulPageTagOfs,
                           "You have specified that page %d should be a README page, but "
                           "the <README>...</README> block is missing.",
                           (PSZ)lCurrentPage);
            }

            PSZ pszFormatAttr;
            if ((pszFormatAttr = strhGetTextAttr(strReadmeAttrs.c_str(), "FORMAT", NULL)))
            {
                // FORMAT specified:
                if (stricmp(pszFormatAttr, "PLAIN") == 0)
                    pPageInfoNew->_ReadmeFormat = BSPageInfo::README_PLAIN;    // default anyway
                else if (stricmp(pszFormatAttr, "FLOW") == 0)
                    pPageInfoNew->_ReadmeFormat = BSPageInfo::README_FLOW;
                else if (stricmp(pszFormatAttr, "HTML") == 0)
                    pPageInfoNew->_ReadmeFormat = BSPageInfo::README_HTML;
                else
                    ParseError(ulBodyBlockOfs + ulPageBlockOfs + ulReadmeTagOfsInPageBlock,
                               "The README tag contains an invalid FORMAT attribute (\"%s\"). ",
                               pszFormatAttr);
            }

            LONG ulPckIndex = 0;
            if ((strhGetNumAttribValue(strReadmeAttrs.c_str(),
                                       "EXTRACTFROMPCK",
                                       &ulPckIndex)))
            {
                pPageInfoNew->_ulExtractFromPck = ulPckIndex;
            }

            pPageInfoNew->_strReadmeSrc = strReadmeBody;
        }

        // alright, done with this page info;
        // store the page info in the list
        _PageInfoList.push_back(pPageInfoNew);
        // go for next
        lCurrentPage++;

        /* free(pszPageBlock);
        free(pszPageAttrs); */

    } while (TRUE);
        // loop forever; there's a break in the loop

    return (lCurrentPage);
}


