
/*
 *@@sourcefile fe_archive.cpp:
 *      this implements the FEArchive class, which is new with
 *      V0.9.0 (99-10-31) [umoeller].
 *
 *      FEArchive represents a WarpIN archive file (.WPI) to the
 *      front end in "install" mode. FEArchive has an instance of
 *      WIArchive (the back-end archive class) as a member.
 *      Note that archives are not used at all in "database" mode.
 *
 *      The entire archive/package/script/packages handling has
 *      been reworked greatly for V0.9.9 to allow for replacing
 *      the script parsing. As a result, another layer had to
 *      be introduced between the FEArchive class and the package
 *      classes for script handling.
 *
 *      This now gives us four layers in "install" mode:
 *
 *      -- FEArchive represents an archive (WPI) file. To open
 *         an archive which was given on the command line,
 *         WarpIN's EnterArchiveMode() (which gets called from
 *         main()) creates an instance of FEArchive with the
 *         archive file name.
 *
 *         FEArchive then runs WIArchive::open on the member
 *         back-end archive class to open the archive file.
 *
 *         EnterArchiveMode then calls FEArchive::ParseScript.
 *
 *      -- FEArchive::ParseScript in turn creates a script
 *         instance. This has been completely encapsulated
 *         in BSScriptBase and the various subclasses with V0.9.9
 *         to allow for replacing the script language entirely
 *         (e.g. for XML support).
 *
 *         The magic of this is in the static BSScriptBase::CreateScript
 *         method, which creates an instance of a BSScriptBase subclass
 *         depending on the script format. This is magic because
 *         that method will find out what subclass needs to be used
 *         and call the proper parsing methods automatically, while
 *         the caller only needs to know the fields in the BSScriptBase
 *         instance which is returned. It therefore no longer needs
 *         to bother about the parsing itself.
 *
 *         BSScriptBase contains package declarations and page
 *         declarations, represented by the BSDeclRoot and BSPageInfo
 *         classes.
 *
 *      -- After BSScriptBase::CreateScript has parsed the script,
 *         FEArchive::ParseScript then constructs instances of
 *         the various package classes from the package declarations
 *         in the script.
 *
 *         Basically, FEPackageBase is the base of the package
 *         classes, the most important of which is FEArcPackageBase.
 *         So in short, FEArchive (the archive) contains
 *         FEArcPackageBase's (the packages), which in turn point
 *         to the BSDeclRoot package declarations.
 *
 *         See fe_package.cpp for a description of the package
 *         classes, which represent either archive packages or
 *         groups in the script (which are not really packages).
 *         Besides, there are package classes for the database also,
 *         which we are not worrying about in "install" mode.
 *
 *      -- The fourth layer is "install jobs" finally, which form
 *         another wrapper around the package classes to hold data
 *         which can be changed by the user. While the data in
 *         the package classes is assumed to be non-variable (because
 *         it comes from the script), the install jobs hold things
 *         like target paths and package selections, which can change.
 *
 *         The FEInstallJob instances are created after FEArchive has
 *         initialized itself by EnterArchiveMode() again.
 *         See fe_job.cpp for details.
 *
 *      To summarize:
 *
 *      1)  FEArchive creates BSScriptBase.
 *
 *      2)  BSScriptBase parses the script and creates BSDeclRoot
 *          and BSPageInfo instances.
 *
 *      3)  FEArchive then constructs FEPackageBase instances
 *          from the BSDeclRoot instances.
 *
 *      4)  FEInstallJob's are then created from the FEPackageBase
 *          before the install pages are displayed to the user.
 *
 *@@header "frontend\fe_archive.h"
 *@@added V0.9.0 (99-10-31) [umoeller]
 */

/*
 *      This file Copyright (C) 1999-2001 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

// Changes by Cornelis Bockemhl (cbo) for REXX implementation:
// -> search for "(cbo" to find all changes
// Most changes follow immediately after a GetBlock call and perform
// REXX calls in order to replace =("macro args") macros
// Be careful: In some cases the GetBlock() call had to get offset
// attribute(s) added. In some other cases the if(!GetBlock()) logic
// is reversed (in order to get the "positive" code block first).

#define OS2EMX_PLAIN_CHAR
    // this is needed for "os2emx.h"; if this is defined,
    // emx will define PSZ as _signed_ char, otherwise
    // as unsigned char

#define INCL_DOSERRORS
#define INCL_WINSHELLDATA
#define INCL_WINWORKPLACE
#define INCL_REXXSAA
#include <os2.h>
#ifdef __IBMCPP__
    #include <rexxsaa.h>    // EMX has this in os2.h already
#endif

#include <stdio.h>
#include <string.h>
#ifdef __IBMCPP__
    #include <direct.h>
#endif
#include <io.h>

#include "setup.h"
#include "bldlevel.h"           // needed for version checks

// include's from helpers
#include "helpers\dosh.h"
#include "helpers\stringh.h"
#include "helpers\xstring.h"

// back-end includes
#include "wiarchive\wiarchive.h"

// base includes (99-11-07) [umoeller]
#include "base\bs_string.h"
#include "base\bs_errors.h"
#include "base\bs_logger.h"
#include "base\bs_config.h"
#include "base\bs_script.h"

// front-end includes
#include "frontend\fe_base.h"
#include "frontend\fe_package.h"
#include "frontend\fe_package_arc.h"
#include "frontend\calbacks.h"
#include "frontend\fe_archive.h"

#pragma hdrstop

/* ******************************************************************
 *
 *  Static variables
 *
 ********************************************************************/

BSString FEArchive::_strArchivePath;
BSString FEArchive::_strArchivePathEnv;

/* ******************************************************************
 *
 *  FEArchive implementation
 *
 ********************************************************************/

/*
 *@@ FEArchive:
 *      FEArchive constructor. This must be created by
 *      passing the WIArchive file name.
 *
 *      NOTE: The archive file name must be fully qualified.
 *
 *      This constructor calls the back end by initializing
 *      the member instance of WIArchive from the given file name.
 *
 *      As opposed to versions before V0.9.9, it no longer parses
 *      the script because this should be deferred (and not done at
 *      all for dependent archives).
 *
 *      To parse the script, call FEArchive::ParseScript afterwards.
 *      Only that call will set the _pScript member.
 *
 *      See fe_archive.cpp for an introduction.
 *
 *      This throws:
 *
 *      -- FEFatalErrorExcpt (if archive could not be opened).
 *
 *@@added V0.9.0 (99-10-31) [umoeller]
 *@@changed V0.9.2 (2000-03-10) [umoeller]: added external archive support
 *@@changed V0.9.2 (2000-03-11) [umoeller]: added more error checking
 *@@changed V0.9.3 (2000-04-28) [umoeller]: finally accepting tabs in scripts
 *@@changewd V0.9.4 (2000-07-26) [umoeller]: added strNext to args list
 */

FEArchive::FEArchive(const char *pcszArcFilename,        // in: archive filename
                     FEArchive *pReferencedFrom) // in: archive this archive was
                                                // referenced from; NULL means this
                                                // is no external archive (default)
{
    _pReferencedFrom = pReferencedFrom;

    // if this is the base archive (not referenced from anywhere):
    // get the global archive path
    // (cbo 2000-05-23)
    BSString strArcFilename(pcszArcFilename);
    if (NULL == pReferencedFrom)
        ExtractArchivePath(strArcFilename);

    // prepend archive path in any case
    PrependArchivePath(strArcFilename);

    _pPackHeadersList = 0;

    _pArcHeader = NULL;

    _pScript = 0;

    // open the archive in param 1 (WIArchive back-end)
    int irc;
    if (irc = _Arc.open(strArcFilename.c_str()))
    {
        PSZ     apsz[2] = { (char*)strArcFilename.c_str(), 0 };
        ULONG   cpsz = 1;
        ULONG   ulMsg = 0;

        switch (irc)
        {
            case WIERR_FILENOTFOUND:
                if (pReferencedFrom)
                {
                    // external archive:
                    ulMsg = 184;    // unable to open archive %1 from %2.
                    apsz[1] = (PSZ)pReferencedFrom->_strArcFilename.c_str();
                    cpsz = 2;
                }
                else
                    ulMsg = 182;    // unable to open archive %1.
            break;

            case WIERR_OUTDATED_ARCHIVE:
                ulMsg = 168;    // archive %1 is outdated
            break;

            case WIERR_BZDECOMPRESS:
                ulMsg = 181;
            break;

            case WIERR_INVALID_HEADER:
                ulMsg = 183;    // archive %1 has invalid header
            break;

            case WIERR_NEWER_ARCHIVE:
                ulMsg = 208;    // archive %1 is for a newer WarpIN
            break;

            default:
                ulMsg = 106;    // not warpin archive
        }
        // error opening archive:
        throw (FEFatalErrorExcpt(ulMsg, apsz, cpsz));
    }
    else
    {
        // WPI file opened OK: call back-end to get more info

        _strArcFilename = strArcFilename;

        _pArcHeader = _Arc.getArcHeader();

        _pPackHeadersList = _Arc.getPackList();
    }
}

/*
 *@@ CreatePackages:
 *      little helper function called from FEArchive::ParseScript
 *      to create packages from script package declarations.
 *
 *      This recurses for groups.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

VOID FEArchive::CreatePackages(list<BSDeclRoot*> &DeclarationsList,
                               FEArcPackageGroup *pParentGroup, // in: parent group, NULL unless we're recursing
                               PFNWICALLBACK pWICallback)   // in: callback func for back-end
{
    static level = 0;
    _Pmpf(("CreatePackages: entering, level = %u", level++));

    // go thru all declarations given to us
    list<BSDeclRoot*>::iterator  DeclStart = DeclarationsList.begin(),
                                 DeclEnd   = DeclarationsList.end();
    for (;
         DeclStart != DeclEnd;
         DeclStart++)
    {
        BSDeclRoot *pDeclThis = *DeclStart;

        if (pDeclThis->_fIsGroup)
        {
            // declaration is for group: create a group then
            // and recurse for subpackages/-groups
            BSGroupDeclBase *pGroupDeclThis = (BSGroupDeclBase*)pDeclThis;

            FEArcPackageGroup *pGroupNew = new FEArcPackageGroup(this,
                                                                 pParentGroup, // initially NULL
                                                                 pGroupDeclThis);
            _PackagesList.push_back(pGroupNew);

            // recurse!!
            CreatePackages(pGroupDeclThis->_DeclarationsList,
                           pGroupNew,
                           pWICallback);
        }
        else
        {
            // regular package:
            BSPckDeclBase *pPckDeclThis = (BSPckDeclBase*)pDeclThis;

            FEArchive *pArchive = this;

            // check if it's external
            if (pPckDeclThis->_External != BSPckDeclBase::INTERNAL)
            {
                // yes:
                if (pPckDeclThis->_External != BSPckDeclBase::REQEXTERNAL)
                {
                    // external, but not required:

                    // (cbo 2000-05-23) prepend path name for checking existence
                    BSString strArchiveName(pPckDeclThis->_strExternalArchive);
                    PrependArchivePath(strArchiveName);

                    // not required: check for existence first,
                    // because the constructor throws exceptions...
                    if (access(strArchiveName.c_str(), 0) != 0)
                        // does not exist:
                        // go for next package
                        continue;
                }

                // (remark 2000-05-23) use the pure archive name here, not
                // the one with the path prepended: The path is prepended
                // in the constructor while this pure archive name is
                // stored in _szArcFilename
                pArchive = new FEArchive(pPckDeclThis->_strExternalArchive.c_str(),
                                         this); // FEArchive *pReferencedFrom
            }

            // create a package then
            FEArcPackagePck *pPckNew = new FEArcPackagePck(pArchive,
                                                           pParentGroup, // initially NULL
                                                           pPckDeclThis,
                                                           pWICallback);

            _PackagesList.push_back(pPckNew);
        }
    }

    _Pmpf(("CreatePackages: leaving level %u", --level));
}

/*
 *@@ ParseScript:
 *      this parses the script and sets _pScript to a
 *      BSScriptBase instance (one of the subclasses)
 *      by calling BSScriptBase::CreateScript.
 *
 *      If that doesn't throw, we create packages from
 *      the package declarations in the script by calling
 *      FEArchive::CreatePackages. That's what we need
 *      the callback param for.
 *
 *      See fe_archive.cpp for an introduction.
 *
 *      Note: Only call this for the "root" (internal)
 *      archive, never for external archives.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 *@@changed V0.9.12 (2001-05-31) [umoeller]: added REXX confirmation
 */

VOID FEArchive::ParseScript(PFNWICALLBACK pWICallback)   // in: callback func for back-end
{
    // get the installation profile from the archive
    PCSZ pcszProfileTemp = _Arc.getScript();
    if (pcszProfileTemp == NULL)
    {
        throw FEFatalErrorExcpt(103);
    }

    // replace all tabs with single spaces;
    // XML doesn't care, but the old parser does
    // V0.9.3 (2000-04-28) [umoeller]
    CHAR* p = (PSZ)pcszProfileTemp;
    while (*p)
    {
        if (*p == '\t')
            *p = ' ';
        p++;
    }

    _strProfile = pcszProfileTemp;

    // create BSScriptBase subclass instance depending
    // on script format and parse the entire thing
    _pScript = BSScriptBase::CreateScript(_strProfile,
                                          guiConfirmRexxAllowed);   // V0.9.12 (2001-05-31) [umoeller]

    if (!_pScript)
        throw FEFatalErrorExcpt(202);    // internal script error

    // now go thru all the script's package declarations and
    // create packages from that...
    CreatePackages(_pScript->_DeclarationsList,     // initial list
                   NULL,                            // group (initially NULL)
                   pWICallback);
            // this recurses!

    // resolve package requirements
    ResolveRequirements();
}

/*
 *@@ ExtractArchivePath:
 *      extracts any path information from the passed file name
 *      and stores it in _strArchivePath.
 *
 *      The path is also assigned to the environment variable
 *      WI_ARCHIVE_PATH so REXX script may access it.
 *      This environment assignemnt is put in _strArchivePathEnv
 *      for reasons I cannot see.
 *
 *      The passed in archive name is stripped for its path.
 *
 *@@added V0.9.3 (2000-05-23) [cbo]
 *@@changed V0.9.13 (2001-06-09) [bird]: rewritten, this failed with various path/filename combinations
 */

VOID FEArchive::ExtractArchivePath(BSString& strArchiveName)
{
#if 0
    CHAR          adrive[_MAX_DRIVE],
                  adir[_MAX_DIR],
                  afname[_MAX_FNAME],
                  aext[_MAX_EXT],
                  currpath[CCHMAXPATH],
                  cdrive[_MAX_DRIVE],
                  cdir[_MAX_DIR],
                  cfname[_MAX_FNAME],
                  cext[_MAX_EXT];

    _splitpath( (char*)strArchiveName.c_str(), adrive, adir, afname, aext );
    _getcwd( currpath, sizeof( currpath ));
    _splitpath( currpath, cdrive, cdir, cfname, cext );
    strcat( cdir, cfname );           // _getcwd returns path without terminating "\"
    strcat( cdir, cext );             // so the last directory becomes the file name!

    // drive
    if( 0 == strlen( adrive ))        // no explicit drive?
      _strArchivePath = cdrive;
    else
      _strArchivePath = adrive;

    // directory
    if(( 0 == strlen( adir )) ||      // no explicit directory?
       ( '\\' != adir[0] ))           // relative path?
      _strArchivePath += cdir;
    _strArchivePath += adir;
    if( '\\' != _strArchivePath[_strArchivePath.size()-1] )
      _strArchivePath += '\\';

    // put the archive path into the environment variable WI_ARCHIVE_PATH,
    // for use of any REXX scripts
    _strArchivePathEnv = BSString("WI_ARCHIVE_PATH=") + _strArchivePath;
    putenv( _strArchivePathEnv.c_str() );

    // archive name
    strArchiveName = afname;
    strArchiveName += aext;
#else

// rewrite by bird.
    CHAR    szArchiveName[CCHMAXPATH];
    if (!DosQueryPathInfo(strArchiveName.c_str(), FIL_QUERYFULLNAME, szArchiveName, sizeof(szArchiveName)))
    {
        char * psz = strrchr(szArchiveName, '\\');
        if (!psz || psz < strrchr(szArchiveName, '/'))
            psz = strrchr(szArchiveName, '/');
        if (!psz)
        {   //urg! this should _never_ happen!!!
            _strArchivePath = _strArchivePathEnv = strArchiveName = "";
            return;
        }

        // make filename and archive path.
        strArchiveName = ++psz;
        *psz = '\0';
        _strArchivePath = szArchiveName;

        // put the archive path into the environment variable WI_ARCHIVE_PATH,
        // for use of any REXX scripts
        _strArchivePathEnv = BSString("WI_ARCHIVE_PATH=") + _strArchivePath;
        putenv( _strArchivePathEnv.c_str() );
    }
    else // bad filename most likely!
        _strArchivePath = _strArchivePathEnv = "";
#endif

}

/*
 *@@ PrependArchivePath:
 *      prepends the archive path in _strArchivePath to the
 *      passed archive file name, thus making it a fully
 *      qualified file name.
 *
 *@@added V0.9.3 (2000-05-23) [cbo]
 */

VOID FEArchive::PrependArchivePath( BSString& strArchiveName )
{
    strArchiveName = _strArchivePath + strArchiveName;
}

/*
 *@@ ResolveRequirements:
 *      this gets called by FEArchive::ParseScript after
 *      FEArchive::ParsePackages (so that all packages
 *      have been parsed) to resolve package requirements.
 *
 *      In the script, package requirements can be specified
 *      either by package index or by full five- or six-part
 *      package ID. For all packages, we will now initialize
 *      the FEPackageBase::_logRequiresIDs member, which hasn't
 *      been initialized yet.
 *
 *      Since package requirements can also be specified by
 *      package index in the script, we can only resolve this
 *      after all packages have been parsed.
 *
 *      If a full package ID has been specified, this will
 *      simply be copied.
 *
 *      If a package index has been specified, that package's
 *      ID will be copied.
 *
 *      Throws:
 *      -- FEFatalErrorExcpt if an invalid index has been
 *         specified.
 *
 *@@added V0.9.1 (2000-01-07) [umoeller]
 */

VOID FEArchive::ResolveRequirements()
{
    // go thru all packages of this archive
    list<FEArcPackageBase*>::iterator PckStart = _PackagesList.begin(),
                                      PckEnd = _PackagesList.end();
    for (; PckStart != PckEnd; PckStart++)
    {
        FEArcPackageBase *pPckThis = *PckStart;

        FEArcPackagePck *pPckPckThis;

        // only for "real" packages; groups have
        // no dependencies:
        if ( (pPckPckThis = pPckThis->IsPackage()) )
        {
            // decode logger strings
            PSZ     pszLogStart = pPckPckThis->_pDecl->_logRequiresStrings._pabLogString,
                    pszLogThis = pszLogStart;
            int     iCount = 0;

            while (pszLogThis < pszLogStart + pPckPckThis->_pDecl->_logRequiresStrings._cbLogString)
            {
                ULONG   cbLogThis = strlen(pszLogThis);
                ULONG ulIndex = atoi(pszLogThis);

                if (ulIndex)
                {
                    // decimal: copy from package with that index
                    FEArcPackagePck *pPckSource = FindArcPackage(ulIndex);
                    if (pPckSource)
                    {
                        pPckPckThis->_logRequiresIDs.Append(pPckSource->_pDecl->_strID);
                    }
                    else
                    {
                        BSString strError;
                        strError._printf("Invalid package index specified with REQUIRES attribute: %s",
                                         pszLogThis);
                        throw(FEFatalErrorExcpt(strError));
                    }
                }
                else
                {
                    // not decimal: simply copy ID
                    pPckPckThis->_logRequiresIDs.Append(pszLogThis);
                }

                pszLogThis += cbLogThis + 1;    // go beyond null byte
                iCount++;
            }

        }
    }
}

/*
 *@@ FindArcPackage:
 *      returns the package with the specified index
 *      from the archive or NULL if not found.
 *
 *      This returns one of the items from the
 *      _PackagesList member.
 *
 *@@added V0.9.1 (2000-01-07) [umoeller]
 */

FEArcPackagePck* FEArchive::FindArcPackage(ULONG ulIndex)
{
    FEArcPackagePck *pPckReturn = NULL;

    list<FEArcPackageBase*>::iterator PckStart = _PackagesList.begin(),
                                      PckEnd = _PackagesList.end();
    for (; PckStart != PckEnd; PckStart++)
    {
        FEArcPackageBase *pPckThis = *PckStart;
        FEArcPackagePck *pPckPckThis;

        if ( (pPckPckThis = pPckThis->IsPackage()) )
        {
            if (pPckPckThis->_pDecl->_ulIndexInArchive == ulIndex)
            {
                pPckReturn = pPckPckThis;
                break;
            }
        }
    }

    return (pPckReturn);
}

/*
 *@@ IsInArchive:
 *      this checks whether the package with the
 *      specified five- or six-part package ID is
 *      in the archive.
 *
 *      If so, the corresponding item from the
 *      _PackagesList member is returned and
 *      *pulComp is set to the return value of
 *      FEPackageID::Compare. If the IDCOMP_THISOLDER
 *      flag is set, that means that the package in the
 *      archive has an insufficient version number
 *      compared to pszPckID.
 *
 *      If not, NULL is returned.
 *
 *      Note that pszID is understood as the
 *      _minimum_ package ID. If any higher
 *      version is in the archive, this is
 *      also returned.
 *
 *@@added V0.9.1 (2000-01-08) [umoeller]
 */

FEArcPackagePck* FEArchive::IsInArchive(const char *pszPckID,
                                        PULONG pulComp) // out: result of FEPackageID::Compare;
                                                        // ptr can be NULL if not desired
{
    FEArcPackagePck *pPckReturn = NULL;

    FEPackageID     PckIDSearch(pszPckID, __FILE__, __LINE__,
                                 FALSE);    // don't allow short format
            // this may throw a FEConstructorExcpt

    list<FEArcPackageBase*>::iterator PckStart = _PackagesList.begin(),
                                      PckEnd = _PackagesList.end();
    for (; PckStart != PckEnd; PckStart++)
    {
        FEArcPackageBase *pPckThis = *PckStart;
        FEArcPackagePck *pPckPckThis;

        if ( (pPckPckThis = pPckThis->IsPackage()) )
        {
            // get package ID for the one in the archive
            FEPackageID     PckIDThis(pPckPckThis->_pDecl->_strID.c_str(),
                                      __FILE__, __LINE__,
                                      FALSE);    // don't allow short format
                    // this may throw a FEConstructorExcpt
            // compare with the one to search for
            ULONG ulComp = PckIDThis.Compare(PckIDSearch);
            if (ulComp & IDCOMP_SAMEPCK)
            {
                // store result
                if (pulComp)
                    *pulComp = ulComp;
                // return this archive
                pPckReturn = pPckPckThis;
                break;
            }
        }
    }

    return (pPckReturn);
}


