
/*
 *@@sourcefile fe_database.cpp:
 *      this implements the FEDatabase class (declared in fe_database.h).
 *
 *      Creating a database instance (using FEDataBase::FEDataBase)
 *      will open the database file and initialize FEDBPackage instances.
 *
 *      After creation, you can, for example, invoke FEDatabase::FindPackageID
 *      or FEDatabase::FindDBPackage to query the database.
 *
 *      Also, the database has methods for verification and de-installation.
 *      See FEDatabase::VerifyPackages and FEDatabase::DeinstallPackages.
 *
 *      Currently, the database is stored in a possibly large
 *      INI file in the directory where WARPIN.EXE also resides.
 *
 *      With V0.9.0 (99-10-31) [umoeller], the database has been transformed
 *      into a C++ class so that more than one database may possibly
 *      be created (even though this is not used yet).
 *
 *      The functions in here are _not_ thread-safe, so only
 *      thread 1 may call any of the database members.
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: renamed from database.cpp
 *@@header "frontend\fe_database.h"
 */

/*
 *      This file Copyright (C) 1999-2000 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define OS2EMX_PLAIN_CHAR
    // this is needed for "os2emx.h"; if this is defined,
    // emx will define PSZ as _signed_ char, otherwise
    // as unsigned char

#define INCL_DOS
#define INCL_DOSERRORS
#define INCL_WIN
#define INCL_WINWORKPLACE
#include <os2.h>
#include <stdlib.h>
#include <stdio.h>
#include <io.h>

#ifdef __IBMCPP__
    #include <direct.h>
#endif

#include <string.h>
#include <ctype.h>
#include <setjmp.h>             // needed for except.h
#include <assert.h>             // needed for except.h

#include "setup.h"

// include's from helpers
#include "helpers\dosh.h"
#include "helpers\except.h"
#include "helpers\prfh.h"
#include "helpers\stringh.h"
#include "helpers\threads.h"
#include "helpers\winh.h"
#include "helpers\xstring.h"

// back-end includes
#include "wiarchive\wiarchive.h"

// base includes (99-11-07) [umoeller]
#include "base\bs_string.h"
#include "base\bs_errors.h"
#include "base\bs_logger.h"
#include "base\bs_config.h"

// front-end includes
#include "frontend\fe_base.h"
#include "frontend\fe_package.h"
#include "frontend\fe_package_db.h"
#include "frontend\calbacks.h"
#include "frontend\fe_database.h"
#include "frontend\fe_db_thread.h"

#pragma hdrstop

/* ******************************************************************
 *
 *  Global variables
 *
 ********************************************************************/

THREADINFO         G_tiDatabaseThread = {0};

/* ******************************************************************
 *
 *  FEDatabaseExcpt implementation
 *
 ********************************************************************/

/*
 *@@ FEDatabaseExcpt:
 *
 *      ulErrNo specifies the error message in warpin.tmf and can be:
 *      -- 157: PrfOpenProfile failed.
 *      -- 164: file list is NULL. This is recoverable.
 *      -- 176: FEPackageID constructor failed. This is recoverable.
 *      -- 179: FEDBPackage::Load failed with a BSConfigExcpt.
 *
 *@@added V0.9.1 (2000-01-08) [umoeller]
 */

FEDatabaseExcpt::FEDatabaseExcpt(const char *pszPckID,
                                 ULONG ulErrNo,
                                 BOOL fRecover)
                : _strFailingPckID(pszPckID)
{
    BSString strMsg;
    wpiGetMessage(strMsg, ulErrNo);

    _strDescription = "WarpIN detected an error in the global database.\n";
    if (pszPckID)
    {
        _strDescription += "Affected package: "
                           + BSString(pszPckID);
    }
    _strDescription += "\nError description: "
                       + strMsg;
    _fRecover = fRecover;
};

/*
 *@@ FEDatabaseExcpt:
 *
 *@@added V0.9.1 (2000-01-08) [umoeller]
 */

FEDatabaseExcpt::FEDatabaseExcpt(const char *pszPckID,
                            const char *pszErrMsg,
                            BOOL fRecover)
                : _strFailingPckID(pszPckID)
{
    _strDescription = "WarpIN detected an error in the global database.\n";
    if (pszPckID)
    {
        _strDescription += "Affected package: "
                           + BSString(pszPckID);
    }
    _strDescription += "\nError description: "
                       + BSString(pszErrMsg);
    _fRecover = fRecover;
};

/* ******************************************************************
 *
 *  FEDatabase implementation
 *
 ********************************************************************/

/*
 *@@ FEDatabase:
 *      database constructor. This opens the database and
 *      constructs the public member list of FEDBPackage
 *      instances by creating as many instances as entries
 *      are found in the database. So after constructing
 *      the database, you get all database packages constructed
 *      as well.
 *
 *      Throws:
 *      -- FEDatabaseExcpt.
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: this was datInitialize
 *@@changed V0.9.1 (2000-01-08) [umoeller]: bugfixing mode added
 */

FEDatabase::FEDatabase(HAB habThread1,      // in: application anchor block
                       const char *pszWarpINPath, // in: path of WARPIN.EXE (without trailing "\")
                       const char *pszDeleteID)  // in: whether to delete a package ID for fixing;
                                                 // set this to 0 unless you know what you're doing;
                                                 // defaults to 0
{
    hiniDatabase = NULLHANDLE;

    CHAR szOldDatabaseFile[2*CCHMAXPATH],
         szNewDatabaseFile[2*CCHMAXPATH];

    sprintf(szOldDatabaseFile,
            "%s\\database.ini",
            pszWarpINPath);
    sprintf(szNewDatabaseFile,
            "%s\\datbas_%c.ini",
            pszWarpINPath,
            doshQueryBootDrive());

    // check if old-style database still exists
    if (access(szOldDatabaseFile, 0) == 0)
    {
        // yes: rename
        rename(szOldDatabaseFile, szNewDatabaseFile);
    }

    hiniDatabase = PrfOpenProfile(habThread1,
                                  szNewDatabaseFile);

    if (hiniDatabase == NULLHANDLE)
        throw (FEDatabaseExcpt(NULL, 157,
                               FALSE));     // not recoverabe

    if (pszDeleteID == NULL)
    {
        // initialize packages list
        ULONG               ulPckCount = 0;
        PSZ                 pszPckID = EnumPackages(NULL);

        while (pszPckID)
        {
            // any packages are installed:
            // create a new package info from database
            try
            {
                FEDBPackage*    pPackageNew = new FEDBPackage(this, pszPckID);
                                    // this can throw a FEDatabaseExcpt
                DBPackagesList.push_back(pPackageNew);
            }
            catch(FEConstructorExcpt &X)
            {
                throw FEDatabaseExcpt(pszPckID,
                                      X._strDescription.c_str(),
                                      TRUE);       // not recoverable
                        // this gets caught in main() only
            }

            PSZ pszPckIDNext = EnumPackages(pszPckID);

            free(pszPckID);
            pszPckID = pszPckIDNext;
            ulPckCount++;
        }
    }
    else
    {
        // bugfixing mode:
        PrfWriteProfileString(hiniDatabase,
                              pszDeleteID,
                              NULL,             // delete whole application
                              NULL);
    }
}

/*
 *@@ ~FEDatabase:
 *      database destructor.
 */

FEDatabase::~FEDatabase()
{
    PrfCloseProfile(hiniDatabase);
}

/*
 *@@ EnumPackages:
 *      this enumerates the packages which are stored in
 *      the database.
 *
 *      This works in two modes:
 *
 *      --  If pszPackageSearch is NULL, the first package
 *          in the database is returned.
 *
 *      --  If pszPackageSearch is != NULL, the package
 *          which comes after pszPackageSearch in the
 *          database is returned.
 *
 *      In any case, this function returns a PSZ to a
 *      new buffer, which must be free()'d after use.
 *      If no (more) packages are found, NULL is returned.
 *
 *      The return value and pszPackageIDSearch are both
 *      in the "package ID" format (five- or six-part
 *      package ID, separated by four backslashes).
 *
 *      This does _not_ work on the member database lists,
 *      but only on the database INI file directly.
 *
 *      However, you can pass the return PSZ to
 *      FEDatabase::FindDBPackage to get a database member
 *      package.
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: this was datEnumPackages
 *@@changed V0.9.1 (2000-01-08) [umoeller]: this used to be private; now it's public
 */

PSZ FEDatabase::EnumPackages(PSZ pszPackageIDSearch)
{
    PSZ pszReturn = NULL;

    APIRET arc;
    PSZ pszPackageList = NULL;

    if (!(arc = prfhQueryKeysForApp(hiniDatabase,
                                    NULL, // query applications list
                                    &pszPackageList)))
    {
        if (pszPackageIDSearch == NULL)
            // first package queried:
            pszReturn = strdup(pszPackageList); // first app is null-terminated
        else
        {
            // search list for pszPackageSearch
            PSZ pPackage = pszPackageList;

            while (*pPackage != 0)
            {
                // pPackage has the current key now
                if (strcmp(pszPackageIDSearch, pPackage) == 0)
                {
                    // pPackage found: return next
                    pPackage += strlen(pPackage)+1;
                    if (*pPackage)
                        // not last:
                        pszReturn = strdup(pPackage);

                    // in any case, stop search
                    break;
                }

                // else: next package (application)
                pPackage += strlen(pPackage)+1;
            }
        }
        free(pszPackageList);
    }

    // return found package or NULL
    return (pszReturn);
}

/*
 *@@ FindPackageID:
 *      this returns the package ID for the package
 *      which matches the given attributes. This
 *      can be used to find out whether a given
 *      package has already been installed and if
 *      so, which version is installed.
 *
 *      Since WarpIN does not allow several packages
 *      with the same author/application/name, but
 *      different version numbers, the returned
 *      information is guaranteed to be unique.
 *
 *      This returns NULL if no matching package was found.
 *
 *      The return value must be delete'd after use.
 *
 *      This does _not_ work on the member database lists,
 *      but only on the database INI file directly.
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: this was datFindPackage
 *@@changed V0.9.0 (99-11-03) [umoeller]: now returning FEPackageID*
 */

FEPackageID* FEDatabase::FindPackageID(const char *pcszAuthor,
                                       const char *pcszApplication,
                                       const char *pcszPackageName)
             const
{
    FEPackageID* pReturn = NULL;

    CHAR szCompare[2000];
    sprintf(szCompare, "%s\\%s\\%s\\", pcszAuthor,
                                       pcszApplication,
                                       pcszPackageName);

    ULONG cbCompare = strlen(szCompare);

    APIRET arc;
    PSZ pszPackageList = NULL;

    if (!(arc = prfhQueryKeysForApp(hiniDatabase,
                                    NULL, // query applications list
                                    &pszPackageList)))
    {
        PSZ pPackageIDThis = pszPackageList;

        while (*pPackageIDThis != 0)
        {
            // pPackageID has the current application (== package ID) now
            if (strlen(pPackageIDThis) > cbCompare)
                // compare the first part
                if (memcmp(pPackageIDThis, szCompare, cbCompare) == 0)
                {
                    // pPackageID found: return
                    pReturn = new FEPackageID(pPackageIDThis, __FILE__, __LINE__,
                                              FALSE);    // don't allow short format
                    // stop search
                    break;
                }

            // else: next package (application)
            pPackageIDThis += strlen(pPackageIDThis)+1;
        }

        free(pszPackageList);
    }

    // return found package or NULL
    return (pReturn);
}

/*
 *@@ FindDBPackage:
 *      kinda similar to FEDatabase::FindPackageID,
 *      but this searches the member list of FEDBPackage
 *      instances and returns the list item which matches
 *      the given attributes.
 *
 *      The returned instance comes from the database
 *      packages list, so you better not delete() the
 *      returned package.
 *
 *      If none is found, NULL is returned.
 *
 *@@added V0.9.0 (99-11-04) [umoeller]
 */

const FEDBPackage* FEDatabase::FindDBPackage(const char *pcszAuthor,
                                             const char *pcszApplication,
                                             const char *pcszPackageName)
{
    FEDBPackage *pReturn = NULL;

    list<FEDBPackage*>::iterator PckBegin = DBPackagesList.begin(),
                                 PckEnd = DBPackagesList.end();

    for (; PckBegin != PckEnd; PckBegin++)
    {
        FEDBPackage *pPackageThis = *PckBegin;
        FEPackageID IDThis(pPackageThis, __FILE__, __LINE__,
                           FALSE);    // don't allow short format

        if (    (IDThis._strAuthor == pcszAuthor)
             && (IDThis._strApplication == pcszApplication)
             && (IDThis._strPackageName == pcszPackageName)
           )
        {
            // matches: return that one
            pReturn = pPackageThis;
            break;
        }
    }

    return (pReturn);
}

/*
 *@@ FindDBPackage:
 *      second version of FindDBPackage, which takes a
 *      package ID as input.
 *
 *      Note that this does NOT check the version numbers
 *      in the package ID, but will return the first database
 *      package which matches author, application, and
 *      package name.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

const FEDBPackage* FEDatabase::FindDBPackage(const FEPackageID &PckID)
{
    return (FindDBPackage(PckID._strAuthor.c_str(),
                          PckID._strApplication.c_str(),
                          PckID._strPackageName.c_str()));
}

/*
 *@@ RemovePackage:
 *      this removes a single package from the database.
 *      Note that this removes _only_ the database
 *      information. It does _not_ de-install the
 *      package (which is not possible after calling
 *      this function, because all database info
 *      for that package will be lost).
 *
 *      This calls guiDatabaseCallback with DBC_PACKAGEREMOVED.
 *
 *      Note that pPckInfo is removed from WPIGLOBALS.DBPckList,
 *      but is not deleted itself. It's still valid after
 *      this call.
 *
 *      Returns FALSE upon errors.
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: this was datRemovePackage
 */

BOOL FEDatabase::RemovePackage(const FEDBPackage *pPackage)
{
    BOOL brc = FALSE;

    try
    {
        FEPackageID PckID(pPackage, __FILE__, __LINE__,
                          FALSE);    // don't allow short format

        BSString strIDSix;
        PckID.CreateStringSix(strIDSix);

        if (PrfWriteProfileString(hiniDatabase,
                                  strIDSix.c_str(),
                                  NULL,             // delete whole application
                                  NULL))
        {
            // OK, package deleted from database:

            // remove it from the list in memory too
            DBPackagesList.remove((FEDBPackage*)pPackage);
            // and call the GUI callback
            guiDatabaseCallback(DBC_PACKAGEREMOVED,
                                NULL,
                                0,
                                (ULONG)pPackage);

            brc = TRUE;
        }
    }
    catch (...) {}

    return (brc);
}

/*
 *@@ IsPackageRequired:
 *      returns the number of packages from the database which
 *      require the specified package to be installed, or returns
 *      0 if the specified package is not required by any other
 *      package.
 *
 *      The packages from the database which require the specified
 *      packages are appended to the specified list so they can
 *      be evaluated by the caller.
 *
 *      All packages on *pIgnoreList are completely ignored during
 *      these checks. That is, if a package requires PckIDThis,
 *      but that package is also on *pIgnoreList, that requirement
 *      is not added to the list.
 *
 *      pIgnoreList can be NULL if you don't want any packages
 *      to be ignored. However, this can be useful if
 *      you have several packages to be deinstalled
 *      and all those packages should be ignored in the checks,
 *      because if the requiring packages are de-installed as
 *      well, there's no problem.
 *
 *@@added V0.9.4 (2000-07-01) [umoeller]
 */

ULONG FEDatabase::IsPackageRequired(FEPackageID &PckIDThis,     // in: package to check
                                    list<FEDBPackage*> &RequiredByList, // out: packages which require this
                                    list<FEDBPackage*> *pIgnoreList) // in: packages to ignore in checks
{
    ULONG ulrc = 0;

    // check if a package in the database
    // requires the given package:

    // enumerate all packages in the database
    PSZ     pszDBPckIDThis = EnumPackages(NULL);
    while (pszDBPckIDThis)
    {
        // get database package from that ID
        FEPackageID DBPckIDThis(pszDBPckIDThis, __FILE__, __LINE__,
                                FALSE);    // don't allow short format
        const FEDBPackage *pDBPckThis
            = FindDBPackage(DBPckIDThis);
        if (!pDBPckThis)
            throw FEFatalErrorExcpt("database package not found");

        // decode "requires" logger strings for that DBPackage
        PSZ     pszLogStart = pDBPckThis->_logRequiresIDs._pabLogString,
                pszDBRequiresIDThis = pszLogStart;
        // int     iCount = 0;

        while (pszDBRequiresIDThis < pszLogStart + pDBPckThis->_logRequiresIDs._cbLogString)
        {
            ULONG   cbLogThis = strlen(pszDBRequiresIDThis);

            // pszDBRequiresIDThis now has a requirement for the DB package;
            // check if that ID is the same as that of the job which is
            // to be removed
            FEPackageID DBRequiresIDThis(pszDBRequiresIDThis, __FILE__, __LINE__,
                                         FALSE);    // don't allow short format
            if (DBRequiresIDThis.Compare(PckIDThis))
            {
                // matches:
                // check if this package is on the "ignore" list
                BOOL fIgnore = FALSE;
                if (pIgnoreList)
                {
                    list<FEDBPackage*>::iterator IgnoreStart = pIgnoreList->begin(),
                                                 IgnoreEnd = pIgnoreList->end();
                    for (; IgnoreStart != IgnoreEnd; IgnoreStart++)
                    {
                        FEDBPackage *pIgnorePck = *IgnoreStart;
                        FEPackageID IgnoreID(pIgnorePck, __FILE__, __LINE__,
                                             FALSE);    // don't allow short format
                        if (IgnoreID.Compare(DBPckIDThis))
                        {
                            fIgnore = TRUE;
                            break;
                        }
                    }
                }

                if (!fIgnore)
                {
                    // requiring package is not on ignore list:
                    // store the DB package in the list which was
                    // given to us
                    RequiredByList.push_back((FEDBPackage*)pDBPckThis);
                    ulrc++;
                }
            }

            pszDBRequiresIDThis += cbLogThis + 1;    // go beyond null byte
        }

        // next package in the database
        PSZ pszPckIDNext = EnumPackages(pszDBPckIDThis);
        free(pszDBPckIDThis);
        pszDBPckIDThis = pszPckIDNext;
    }

    return (ulrc);
}

/*
 *@@ VerifyPackages:
 *      this starts the Database thread to verify the
 *      given list of packages.
 *
 *      The list is copied and then passed to the database
 *      thread, so you may pass a temporary list to this
 *      function (i.e. it does not have to be valid all the
 *      time while the database thread is running). However,
 *      since the list points to FEPackage instances, those
 *      instances _must_ remain in memory all the time.
 *
 *      The pPackagesList is copied into a new list for the Database
 *      thread, so this does not have to be permanent.
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: this was datVerifyPackages
 *@@changed V0.9.0 (99-11-02) [umoeller]: reworked database thread
 *@@changed V0.9.9 (2001-03-30) [umoeller]: added log file
 */

BOOL FEDatabase::VerifyPackages(list<FEDBPackage*>* pPackagesList,
                                ULONG ulUser, // in: user parameter passed to
                                              // guiDatabaseCallback (could be HWND)
                                BSFileLogger *pLogFile) // in: log file name or NULL
{
    BOOL brc = FALSE;

    if (pPackagesList)
    {
        // now we have the package info which we need to verify;
        // create a thread info for the database thread
        FEDatabaseThread
                *pdbti = new FEDatabaseThread(this,  // database
                                              pPackagesList, // list
                                              DBT_VERIFY,  // task
                                              0, // deinstall flags, don't care
                                              ulUser,
                                              pLogFile);
        brc = TRUE;
    }

    return (brc);
}

/*
 *@@ DeinstallPackages:
 *      this starts the Database thread to de-install the
 *      given list of packages.
 *
 *      The list is copied and then passed to the database
 *      thread, so you may pass a temporary list to this
 *      function (i.e. it does not have to be valid all the
 *      time while the database thread is running). However,
 *      since the list points to FEPackage instances, those
 *      instances _must_ remain in memory all the time.
 *
 *      If ulFlags is 0, only the files will be removed.
 *      You can however OR any of the following:
 *
 *      -- DBDEINST_CONFIGSYS:  remove CONFIG.SYS entries
 *
 *      -- DBDEINST_WPSCLASSES: de-install WPS classes
 *
 *      -- DBDEINST_WPSOBJECTS: destroy WPS objects
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: this was datDeinstallPackages
 *@@changed V0.9.0 (99-11-02) [umoeller]: reworked database thread
 *@@changed V0.9.9 (2001-03-30) [umoeller]: added log file
 */

BOOL FEDatabase::DeinstallPackages(list<FEDBPackage*>* pPackagesList,
                                   ULONG ulFlags,        // DBDEINST_* flags
                                   ULONG ulUser,         // in: user parameter passed to
                                                         // guiDatabaseCallback (could be HWND)
                                   BSFileLogger *pLogFile) // in: log file name or NULL
{
    BOOL brc = FALSE;

    if (pPackagesList)
    {
        // now we have the package info which we need to verify;
        // create a thread info for the database thread
        FEDatabaseThread
                *pdbti = new FEDatabaseThread(this,     // database
                                              pPackagesList, // list
                                              DBT_DEINSTALL,  // task
                                              ulFlags,
                                              ulUser,
                                              pLogFile);

        brc = TRUE;
    }

    return (brc);
}


