
/*
 *@@sourcefile fe_db_thread.cpp:
 *      this implements the database thread used by the
 *      FEDatabase class. This thing is running only when
 *      packages are being verified or de-installed.
 *
 *      No need to call anything in here manually. The
 *      thread is properly controlled by the FEDatabase
 *      class.
 *
 *      This has been separated from fe_database.cpp.
 *
 *@@added V0.9.1 (2000-02-06) [umoeller]
 *@@header "frontend\fe_db_thread.h"
 */

/*
 *      This file Copyright (C) 1999-2000 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define OS2EMX_PLAIN_CHAR
    // this is needed for "os2emx.h"; if this is defined,
    // emx will define PSZ as _signed_ char, otherwise
    // as unsigned char

#define INCL_DOS
#define INCL_DOSERRORS
#define INCL_WIN
#define INCL_WINWORKPLACE
#include <os2.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef __IBMCPP__
    #include <direct.h>
#endif

#include <string.h>
#include <ctype.h>
#include <setjmp.h>             // needed for except.h
#include <assert.h>             // needed for except.h

#include "setup.h"

// include's from helpers
#include "helpers\configsys.h"
#include "helpers\dosh.h"
#include "helpers\except.h"
#include "helpers\prfh.h"
#include "helpers\stringh.h"
#include "helpers\threads.h"
#include "helpers\winh.h"
#include "helpers\xstring.h"

// back-end includes
#include "wiarchive\wiarchive.h"

// base includes (99-11-07) [umoeller]
#include "base\bs_string.h"
#include "base\bs_errors.h"
#include "base\bs_logger.h"
#include "base\bs_config.h"
#include "base\bs_config_impl.h"
#include "base\bs_script.h"

// front-end includes
#include "frontend\warpin.h"
#include "frontend\fe_base.h"
#include "frontend\fe_package.h"
#include "frontend\fe_package_db.h"
#include "frontend\calbacks.h"
#include "frontend\fe_database.h"
#include "frontend\fe_db_thread.h"

#pragma hdrstop

/* ******************************************************************
 *
 *  Global variables
 *
 ********************************************************************/

extern WPIGLOBALS      G_WpiGlobals;
            // in warpin.cpp

extern THREADINFO      G_tiDatabaseThread;
            // in fe_db_thread.cpp

/* ******************************************************************
 *
 *  FEDatabaseThreadInfo implementation
 *
 ********************************************************************/

/*
 *@@ FEDatabaseThread:
 *      constructor with parameters, all of which
 *      _must_ be specified. This copies the
 *      specified list into a new list in the
 *      instance data. However, the FEPackage
 *      instances themselves are _not_ copied.
 *
 *      This constructor initializes _all_ data
 *      for the database thread and starts the
 *      database thread (datDatabaseThread)
 *      automatically, which keeps running after
 *      this constructor returns.
 *
 *      If you pass in a BSFileLogger* for a log
 *      file, you must close that log file (by
 *      invoking delete() on the logger) yourself
 *      when the database thread has finished its
 *      work. The database thread will only write
 *      into the logger.
 *
 *@@changed V0.9.0 (99-11-02) [umoeller]: reworked completely
 *@@changed V0.9.9 (2001-03-30) [umoeller]: added log file
 *@@changed V0.9.12 (2001-05-22) [umoeller]: now creating flag with PM message queue
 */

FEDatabaseThread::FEDatabaseThread(FEDatabase* pDatabase_,
                                        // database instance to work on
                                   list<FEDBPackage*>  *pPackagesList_,
                                        // packages list to copy
                                   ULONG ulTask_,
                                        // DBT_* flag (what to do)
                                   ULONG flDeinstall_,
                                        // DBDEINST_* flags (warpin.h)
                                   ULONG ulUser_,
                                        // user parameter passed to callback
                                   BSFileLogger *pLogFile)
                                        // in: log file or NULL
{
    // _hab = NULLHANDLE;          // changed by datDatabaseThread
            // removed V0.9.12 (2001-05-22) [umoeller]

    _pDatabase = pDatabase_;
    _ulTask   = ulTask_;
    _flDeinstall = flDeinstall_;
    _ulUser   = ulUser_;

    _pLogFile = pLogFile;

    // copy the list
    list<FEDBPackage*>::iterator PckStart = pPackagesList_->begin(),
                                 PckEnd = pPackagesList_->end();
    for (; PckStart != PckEnd; PckStart++)
        _PackagesList.push_back(*PckStart);

    // start database thread
    thrCreate(&G_tiDatabaseThread ,      // global PTHREADINFO at top
              datDatabaseThread,        // thread func
              NULL,
              "Database",
              THRF_PMMSGQUEUE | THRF_WAIT,    // changed V0.9.12 (2001-05-22) [umoeller]
              (ULONG)this);             // data passed to thread
    // our instance will be deleted by the database thread
}

/* ******************************************************************
 *
 *  Database thread
 *
 ********************************************************************/

/*
 *@@ UndoConfig:
 *      called by datDatabaseThread for each package
 *      to be de-installed. This removes config.sys
 *      entries, wps objects and classes, etc.
 *
 *@@added V0.9.4 (2000-07-27) [umoeller]
 *@@changed V0.9.9 (2001-03-30) [umoeller]: added de-executes
 */

VOID UndoConfig(FEDBPackage *pPackage,      // in: current package being de-installed
                FEDatabaseThread *pdbti,    // in: db thread info
                BSConfigSys* pConfigSys,    // in: config.sys to manipulate
                PBOOL pfConfigSysChanged)   // out: config.sys changed?
{
    /*
     *  1) delete WPS objects:
     *
     */

    // use a reverse_iterator for deleting the
    // objects because there might be a folder
    // first which contains the other objects
    // and we'd get error messages then
    list<BSConfigBase*>::reverse_iterator
            RConfigStart = pPackage->_listUndoConfig.rbegin(),
            RConfigEnd = pPackage->_listUndoConfig.rend();
    for (; RConfigStart != RConfigEnd; RConfigStart++)
    {
        // now typecast this to the proper BSConfigBase
        // subclass we need
        BSDeleteWPSObject *pDelThis = (**RConfigStart).IsDeleteWPSObject();
        if (!pDelThis)
            // not proper subclass:
            continue;

        if (pDelThis->IsSelected()) // V0.9.4 (2000-07-01) [umoeller]
            try
            {
                pDelThis->Delete(pdbti->_pLogFile);
                guiDatabaseCallback(DBC_OBJECTDELETED,
                                    &pdbti->_Status,
                                    (ULONG)(pDelThis->_pszObjectID),
                                    pdbti->_ulUser);
            }
            catch(BSConfigExcpt& X)
            {
                // error deleting WPS object:
                // call GUI callback with error
                guiDatabaseCallback(DBC_OBJECTDELETEERROR,
                                    &pdbti->_Status,
                                    (ULONG)(pDelThis->_pszObjectID),
                                    pdbti->_ulUser);
                pPackage->_ulStatus = 1;         // == errors found
            }
    }

    /*
     *  2)   unreplace classes
     *
     */

    list<BSConfigBase*>::iterator
        ConfigStart = pPackage->_listUndoConfig.begin(),
        ConfigEnd = pPackage->_listUndoConfig.end();
    for (; ConfigStart != ConfigEnd; ConfigStart++)
    {
        // now typecast this to the proper BSConfigBase
        // subclass we need
        BSUnreplaceClass *pReplThis = (**ConfigStart).IsUnreplaceClass();
        if (!pReplThis)
            // not proper subclass:
            continue;

        if (pReplThis->IsSelected()) // V0.9.4 (2000-07-01) [umoeller]
            try
            {
                pReplThis->Unreplace(pdbti->_pLogFile);
                guiDatabaseCallback(DBC_CLASSUNREPLACED,
                                    &pdbti->_Status,
                                    (ULONG)(pReplThis->_pszOldClassName),
                                    pdbti->_ulUser);
                G_WpiGlobals.fWPSClassesChanged = TRUE;
            }
            catch(BSConfigExcpt& X)
            {
                // error deregistering class:
                // call GUI callback with error
                guiDatabaseCallback(DBC_UNREPLACECLASSERROR,
                                    &pdbti->_Status,
                                    (ULONG)(pReplThis->_pszOldClassName),
                                    pdbti->_ulUser);
                pPackage->_ulStatus = 1;         // == errors found
            }
    }

    /*
     *  3)   then deregister classes
     *
     */

    ConfigStart = pPackage->_listUndoConfig.begin(),
    ConfigEnd = pPackage->_listUndoConfig.end();
    for (; ConfigStart != ConfigEnd; ConfigStart++)
    {
        // now typecast this to the proper BSConfigBase
        // subclass we need
        BSDeregisterClass *pDeregThis = (**ConfigStart).IsDeregisterClass();
        if (!pDeregThis)
            // not proper subclass:
            continue;

        if (pDeregThis->IsSelected()) // V0.9.4 (2000-07-01) [umoeller]
            try
            {
                pDeregThis->Deregister(pdbti->_pLogFile);
                guiDatabaseCallback(DBC_CLASSDEREGISTERED,
                                    &pdbti->_Status,
                                    (ULONG)(pDeregThis->_pszClassName),
                                    pdbti->_ulUser);
                G_WpiGlobals.fWPSClassesChanged = TRUE;
            }
            catch(BSConfigExcpt& X)
            {
                // error deregistering class:
                // call GUI callback with error
                guiDatabaseCallback(DBC_DEREGISTERCLASSERROR,
                                    &pdbti->_Status,
                                    (ULONG)(X._strDescription.c_str()),
                                    pdbti->_ulUser);
                pPackage->_ulStatus = 1;         // == errors found
            }
    }

    /*
     *  4) undo CONFIG.SYS changes:
     *
     */

    ConfigStart = pPackage->_listUndoConfig.begin(),
    ConfigEnd = pPackage->_listUndoConfig.end();
    for (; ConfigStart != ConfigEnd; ConfigStart++)
    {
        // now typecast this to the proper BSConfigBase
        // subclass we need
        BSCfgSysManip *pManipThis = (**ConfigStart).IsCfgSysManip();
        if (!pManipThis)
            // not proper subclass:
            continue;

        if (pManipThis->IsSelected()) // V0.9.4 (2000-07-01) [umoeller]
            try
            {
                // the items on this list have been
                // created by the FEPackage constructor
                // just according to what needs to be undone
                pConfigSys->Manipulate(*pManipThis,
                                       NULL,   // no logger here
                                       pdbti->_pLogFile);
                *pfConfigSysChanged = TRUE;
                guiDatabaseCallback(DBC_CFGSYSUNDONE,
                                    &pdbti->_Status,
                                    (ULONG)pManipThis,
                                    pdbti->_ulUser);
                G_WpiGlobals.fConfigSysChanged = TRUE;
            }
            catch (BSConfigExcpt& X)
            {
                // error deregistering class:
                // call GUI callback with error
                guiDatabaseCallback(DBC_UNDOCFGSYSERROR,
                                    &pdbti->_Status,
                                    (ULONG)pManipThis,
                                    pdbti->_ulUser);
                pPackage->_ulStatus = 1;         // == errors found
            }
    }

    /*
     *  5) clear profile entries
     *
     */

    ConfigStart = pPackage->_listUndoConfig.begin(),
    ConfigEnd = pPackage->_listUndoConfig.end();
    for (; ConfigStart != ConfigEnd; ConfigStart++)
    {
        // now typecast this to the proper BSConfigBase
        // subclass we need
        BSClearProfile *pManipThis = (**ConfigStart).IsClearProfile();
        if (!pManipThis)
            // not proper subclass:
            continue;

        if (pManipThis->IsSelected()) // V0.9.4 (2000-07-01) [umoeller]
            try
            {
                pManipThis->Clear(G_tiDatabaseThread.hab,
                                  pdbti->_pLogFile);
                guiDatabaseCallback(DBC_PROFILECLEARED,
                                    &pdbti->_Status,
                                    (ULONG)pManipThis,
                                    pdbti->_ulUser);
            }
            catch (BSConfigExcpt& X)
            {
                // error deregistering class:
                // call GUI callback with error
                guiDatabaseCallback(DBC_CLEARPROFILEERROR,
                                    &pdbti->_Status,
                                    (ULONG)pManipThis,
                                    pdbti->_ulUser);
                pPackage->_ulStatus = 1;         // == errors found
            }
    }

    /*
     *  6) call de-executes V0.9.9 (2001-03-30) [umoeller]
     *
     */

    ConfigStart = pPackage->_listUndoConfig.begin(),
    ConfigEnd = pPackage->_listUndoConfig.end();
    for (; ConfigStart != ConfigEnd; ConfigStart++)
    {
        // now typecast this to the proper BSConfigBase
        // subclass we need
        BSDeExecute *pManipThis = (**ConfigStart).IsDeExecute();
        if (!pManipThis)
            // not proper subclass:
            continue;

        if (pManipThis->IsSelected()) // V0.9.4 (2000-07-01) [umoeller]
            try
            {
                guiDatabaseCallback(DBC_DEEXECUTING,
                                    &pdbti->_Status,
                                    (ULONG)pManipThis,
                                    pdbti->_ulUser);
                pManipThis->Execute(pdbti->_pLogFile);
                guiDatabaseCallback(DBC_DEEXECUTE_DONE,
                                    &pdbti->_Status,
                                    (ULONG)pManipThis,
                                    pdbti->_ulUser);
            }
            catch (BSConfigExcpt& X)
            {
                // error de-executing:
                // call GUI callback with error
                guiDatabaseCallback(DBC_DEEXECUTE_ERROR,
                                    &pdbti->_Status,
                                    (ULONG)pManipThis,
                                    pdbti->_ulUser);
                pPackage->_ulStatus = 1;         // == errors found
            }
    }
}

/*
 *@@ VerifyFile:
 *      this gets called from datDatabaseThread to
 *      verify a single file.
 *
 *      The "Status" field in the thread info has
 *      the current package and file.
 *
 *      Returns FALSE if an error was found or TRUE
 *      if the file is OK.
 *
 *@@changed V0.9.0 (99-11-02) [umoeller]: reworked database thread
 *@@changed V0.9.9 (2001-04-06) [umoeller]: didn't work with root dir target path
 */

BOOL VerifyFile(FEDatabaseThread *pdbti)    // in: database thread info
{
    BOOL brc = FALSE;

    FEDBPackage *pPackage = pdbti->_Status.pPackage;
    WIFileHeader *pwifh = pdbti->_Status.pwifh;

    CHAR szFullFile[2*CCHMAXPATH];

    // fixed root directory V0.9.9 (2001-04-06) [umoeller]
    strcpy(szFullFile, pPackage->QueryTargetPath());
    if (szFullFile[strlen(szFullFile) - 1] != '\\')
        strcat(szFullFile, "\\");
    strcat(szFullFile, pwifh->name);

    FILESTATUS3 fs3;
    APIRET arc = DosQueryPathInfo(szFullFile, FIL_STANDARD, &fs3, sizeof(fs3));
    if (arc)
    {
        // error: complain at callback
        guiDatabaseCallback(DBC_FILEERROR,
                            &pdbti->_Status,
                            arc,            // error code
                            pdbti->_ulUser);
    }
    else
    {
        // file found: compare last write dates
        // of database and real file
        FDATE fdateHeaderLastWrite;
        FTIME ftimeHeaderLastWrite;
        int iComp = wpiCompareFileDates(pwifh,
                                        &fs3,
                                        &fdateHeaderLastWrite,
                                        &ftimeHeaderLastWrite);
        if (iComp < 0)
            guiDatabaseCallback(DBC_EXISTINGFILENEWER,
                                &pdbti->_Status,
                                (ULONG)&fs3,
                                pdbti->_ulUser);
        else if (iComp > 0)
            guiDatabaseCallback(DBC_EXISTINGFILEOLDER,
                                &pdbti->_Status,
                                (ULONG)&fs3,
                                pdbti->_ulUser);
        else
            // else 0: file dates are the same
            brc = TRUE;
    }
    return (brc);
}

/*
 *@@ DeinstallFile:
 *      this gets called from datDatabaseThread to
 *      deinstall (= delete) a single file.
 *
 *      The "Status" field in the thread info has
 *      the current package and file.
 *
 *      Returns FALSE if an error was found or TRUE
 *      if the file is OK.
 *
 *@@changed V0.9.0 (99-11-02) [umoeller]: reworked database thread
 *@@changed V0.9.9 (2001-04-06) [umoeller]: didn't work with root dir target path
 */

BOOL DeinstallFile(FEDatabaseThread *pdbti)     // in: database thread info
{
    BOOL brc = FALSE;

    FEDBPackage *pPackage = pdbti->_Status.pPackage;
    WIFileHeader *pwifh = pdbti->_Status.pwifh;

    CHAR szFullFile[2*CCHMAXPATH];

    // fixed root directory V0.9.9 (2001-04-06) [umoeller]
    strcpy(szFullFile, pPackage->QueryTargetPath());
    if (szFullFile[strlen(szFullFile) - 1] != '\\')
        strcat(szFullFile, "\\");
    strcat(szFullFile, pwifh->name);

    // check file date; do not delete if newer V0.9.9 (2001-04-06) [umoeller]
    BOOL fProceed = TRUE;

    FILESTATUS3 fs3;
    APIRET arc = DosQueryPathInfo(szFullFile, FIL_STANDARD, &fs3, sizeof(fs3));
    if (arc)
    {
        // error: complain at callback
        guiDatabaseCallback(DBC_FILEERROR,
                            &pdbti->_Status,
                            arc,            // error code
                            pdbti->_ulUser);
        if (pdbti->_pLogFile)
            pdbti->_pLogFile->Append("Error %d checking file \"%s\".\n",
                                     arc,
                                     szFullFile);
        fProceed = FALSE;
    }
    else
    {
        // file found: compare last write dates
        // of database and real file
        FDATE fdateHeaderLastWrite;
        FTIME ftimeHeaderLastWrite;
        int iComp = wpiCompareFileDates(pwifh,
                                        &fs3,
                                        &fdateHeaderLastWrite,
                                        &ftimeHeaderLastWrite);
        if (iComp < 0)
        {
            if (pdbti->_pLogFile)
            {
                CHAR        szExistingDate[40] = "",
                            szExistingTime[40] = "",
                            szDatabaseDate[40] = "",
                            szDatabaseTime[40] = "";
                FDATE       fdateLastWrite;
                FTIME       ftimeLastWrite;

                strhFileDate(szExistingDate,
                             &(fs3.fdateLastWrite),
                             2, '-');
                strhFileTime(szExistingTime,
                             &(fs3.ftimeLastWrite),
                             1, ':');
                // convert file header's time information
                wpiCTime2FTime(&pwifh->lastwrite,
                               &fdateLastWrite,
                               &ftimeLastWrite);
                strhFileDate(szDatabaseDate,
                             &fdateLastWrite,
                             2, '-');
                strhFileTime(szDatabaseTime,
                             &ftimeLastWrite,
                             1, ':');

                pdbti->_pLogFile->Append("File \"%s\" is newer than in database:\n",
                                         szFullFile);
                pdbti->_pLogFile->Append("  Existing file: date %s, time %s\n",
                                         szExistingDate, szExistingTime);
                pdbti->_pLogFile->Append("  Database: date %s, time %s\n",
                                         szDatabaseDate, szDatabaseTime);
            }

            GUIFILEERROR gfe;
            wpiPrepareGUIFILEERROR(&gfe,
                                   pPackage->QueryTargetPath(),
                                   pwifh);
            fProceed = guiDatabaseCallback(DBC_DELETE_EXISTINGFILENEWER,
                                           &pdbti->_Status,
                                           (ULONG)&gfe,
                                           pdbti->_ulUser);
            if (pdbti->_pLogFile)
            {
                pdbti->_pLogFile->Append("  User selected: %s\n",
                                         (fProceed) ? "delete anyway" : "skip");
            }
        }
    }
    // end check file date V0.9.9 (2001-04-06) [umoeller]

    if (fProceed)
    {
        // reset the file's attributes, because
        // DosDelete fails if this has READONLY
        BOOL fRetry = FALSE;
        do
        {
            fRetry = FALSE;
            APIRET arc = doshSetPathAttr(szFullFile, 0);
            if (arc != NO_ERROR)
            {
                if (arc == ERROR_SHARING_VIOLATION)
                {
                    // file is locked:
                    if (guiDatabaseCallback(DBC_FILELOCKED,
                                            &pdbti->_Status,
                                            arc,                // error code
                                            pdbti->_ulUser)
                            == TRUE)
                        // unlock:
                    {
                        #ifndef __EMX__
                            // we use the undocumented DosReplaceModule API.
                            DosReplaceModule(szFullFile,        // old module
                                             NULL,              // new module
                                             NULL);             // backup module
                        #endif

                        fRetry = TRUE;
                    }
                }

                if (!fRetry)
                {
                    // other error or no unlock:
                    if (pdbti->_pLogFile)
                        pdbti->_pLogFile->Append("Error %d checking file \"%s\".\n",
                                                 arc,
                                                 szFullFile);
                    // complain at callback
                    guiDatabaseCallback(DBC_FILEERROR,
                                        &pdbti->_Status,
                                        arc,                // error code
                                        pdbti->_ulUser);
                }
            }
            else
            {
                // file exists: delete it
                arc = DosDelete(szFullFile);
                if (arc != NO_ERROR)
                {
                    // report error
                    if (pdbti->_pLogFile)
                        pdbti->_pLogFile->Append("Error deleting file \"%s\".\n",
                                                 szFullFile);

                    guiDatabaseCallback(DBC_DELETEERROR,
                                        &pdbti->_Status,
                                        arc,            // error code
                                        pdbti->_ulUser);
                }
                else
                {
                    if (pdbti->_pLogFile)
                        pdbti->_pLogFile->Append("Deleted file \"%s\".\n",
                                                 szFullFile);
                    brc = TRUE;
                }
            }
        } while (fRetry);
    }
    else
        pdbti->_pLogFile->Append("Skipped file \"%s\".\n",
                                 szFullFile);

    return (brc);
}

/*
 *@@ datDatabaseThread:
 *      this thread is started by FEDatabase::VerifyPackages
 *      and FEDatabase::DeinstallPackages to compare packages
 *      with the actual files on disk or to remove them.
 *
 *      ((PFEDatabaseThreadInfo)ptiMyself)->ulData is a
 *      pointer to the FEDatabaseThread instance
 *      which specifies what needs to be done.
 *      It is the responsibility of this thread to free
 *      this structure.
 *
 *      This calls guiDatabaseCallback all the time.
 *
 *@@changed V0.9.0 (99-11-02) [umoeller]: reworked database thread
 *@@changed V0.9.1 (2000-01-05) [umoeller]: reworked configuration undoing
 *@@changed V0.9.1 (2000-01-05) [umoeller]: CONFIG.SYS was always rewritten to disk; fixed
 *@@changed V0.9.4 (2000-07-01) [umoeller]: added individual undo-config selection
 *@@changed V0.9.4 (2000-07-01) [umoeller]: added deletion of empty subdirs
 *@@changed V0.9.6 (2000-10-27) [umoeller]: CONFIG.SYS backup filename wasn't reported right, fixed.
 */

void _Optlink datDatabaseThread(PTHREADINFO pti)
{
    FEDatabaseThread *pdbti
        = (FEDatabaseThread*)(pti->ulData);

    // get anchor block; needed for profile functions
    // pdbti->_hab = WinInitialize(0);
            // removed V0.9.12 (2001-05-22) [umoeller]

    // install the LOUD exception handler (except.h)
    TRY_LOUD(excpt1)
    {
        if (pdbti->_pLogFile)
            pdbti->_pLogFile->Append("WarpIN Database thread started.\n");

        // to this list, we'll add the target directories of
        // the packages which we have removed
        list<FEDBPackage*> TargetDirList;

        // if we're in de-install mode and we're to undo
        // changes to CONFIG.SYS, load CONFIG.SYS now
        // (by creating a BSConfigSys instance)
        BSConfigSys*        pConfigSys = 0;
        BOOL                fConfigSysChanged = FALSE;
        if (    (pdbti->_flDeinstall & DBT_UNDOCONFIG)
             && (pdbti->_ulTask == DBT_DEINSTALL)
           )
            pConfigSys = new(BSConfigSys);

        // PASS 1: count packages and files
        list<FEDBPackage*>::iterator PckBegin = pdbti->_PackagesList.begin(),
                                     PckEnd = pdbti->_PackagesList.end();
        for (; (PckBegin != PckEnd); PckBegin++)
        {
            pdbti->_Status.cPackages++;
            pdbti->_Status.cFiles += (**PckBegin)._PackHeader.files;
        }

        // PASS 2: go thru all packages given to us
        PckBegin = pdbti->_PackagesList.begin();
        PckEnd = pdbti->_PackagesList.end();

        for (; (PckBegin != PckEnd); PckBegin++)
        {
            FEDBPackage *pPackage = *PckBegin;

            if (pdbti->_pLogFile)
            {
                pdbti->_pLogFile->Append("Deinstalling package \"%s\".\n",
                                         pPackage->QueryPackageID());
                pdbti->_pLogFile->IncIndent(+4);
            }

            pPackage->_ulStatus = 2;
                    // package status = "no errors found"
                    // (was either 0 or 2 before)

            // set package in status; this is used
            // for the callback and all subsequent actions
            pdbti->_Status.pPackage = pPackage;
            pdbti->_Status.pwifh = 0;        // for now, changed later

            // report package to callback
            guiDatabaseCallback(DBC_NEXTPACKAGE,
                                &pdbti->_Status,
                                0,
                                pdbti->_ulUser);

            if (!doshQueryDirExist(pPackage->QueryTargetPath()))
            {
                // target directory not found:
                // call GUI callback with error
                guiDatabaseCallback(DBC_TARGETDIRNOTFOUND,
                                    &pdbti->_Status,
                                    0,
                                    pdbti->_ulUser);
            }

            // in de-install mode, undo CONFIG.SYS/WPS classes/WPS objects
            if (    (pPackage->_ulStatus == 2)        // no errors found
                 && (pdbti->_ulTask == DBT_DEINSTALL) // de-install mode
                 && (pdbti->_flDeinstall & DBT_UNDOCONFIG) // undo configuration
               )
            {
                UndoConfig(pPackage,
                           pdbti,
                           pConfigSys,
                           &fConfigSysChanged);
            } // end configuration (CONFIG.SYS, WPS classes, WPS objects)

            // only in verify mode or if files are to be
            // removed, go thru files list
            if (    (pdbti->_ulTask == DBT_VERIFY)
                 || (pdbti->_flDeinstall & DBT_DELETEFILES)
               )
            {
                 // get list of file headers from package
                 ULONG ulHeaderCount = 0;            // count of headers in this package
                 list<WIFileHeader*>* pFileHeadersList = pPackage->GetFileHeaders(&ulHeaderCount);

                 // now verify or delete all files
                 if (pFileHeadersList)
                 {
                     list<WIFileHeader*>::iterator HeaderStart = pFileHeadersList->begin(),
                                                   HeaderEnd = pFileHeadersList->end();

                     for (; HeaderStart != HeaderEnd; HeaderStart++)
                     {
                         WIFileHeader *pwifh = (*HeaderStart);
                         pdbti->_Status.pwifh = pwifh;

                         // report file
                         guiDatabaseCallback(DBC_NEXTFILE,
                                             &pdbti->_Status,
                                             NULL,               // ulExtra
                                             pdbti->_ulUser);

                         switch (pdbti->_ulTask)
                         {
                             case DBT_VERIFY:
                                 if (!VerifyFile(pdbti))
                                     pPackage->_ulStatus = 1;         // == errors found
                             break;

                             case DBT_DEINSTALL:
                                 if (!DeinstallFile(pdbti))
                                     pPackage->_ulStatus = 1;         // == errors found
                             break;
                         }

                         pdbti->_Status.ulCurrentFile++;
                     }

                     delete pFileHeadersList;
                 }
                 // now done with the files of this package

                 // in de-install mode, prepare removal
                 // of directories too: this must be done
                 // at the very end, because several packages
                 // might use the same directory, so we collect
                 // these for all packages first
                 if (    (pPackage->_ulStatus == 2)        // no errors found
                      && (pdbti->_ulTask == DBT_DEINSTALL)
                      && (pdbti->_flDeinstall & DBT_DELETEFILES) // files are to be removed
                    )
                 {
                     // check if the target dir of this package is
                     // already in the list
                     list<FEDBPackage*>::iterator TargetDirFirst = TargetDirList.begin(),
                                                  TargetDirLast = TargetDirList.end();
                     BOOL fFound = FALSE;
                     for (; TargetDirFirst != TargetDirLast; TargetDirFirst++)
                     {
                         FEDBPackage    *pTargetThis = *TargetDirFirst;
                         if (strcmp(pTargetThis->QueryTargetPath(),
                                    pPackage->QueryTargetPath())
                               == 0)
                         {
                             fFound = TRUE;
                             break;
                         }
                     }
                     if (!fFound)
                         // not in list: add
                         TargetDirList.push_back(pPackage);

                     // remove this package from the database;
                     // this will call guiDatabaseCallback again
                     pdbti->_pDatabase->RemovePackage(pPackage);

                     if (pdbti->_pLogFile)
                         pdbti->_pLogFile->Append("Removed package \"%s\" from the database.\n",
                                                  pPackage->QueryPackageID());
                 }

                 // increase package counter
                 pdbti->_Status.ulCurrentPackage++;
            }

            if (pdbti->_pLogFile)
            {
                pdbti->_pLogFile->IncIndent(-4);
                pdbti->_pLogFile->Append("Done with package \"%s\".\n",
                                         pPackage->QueryPackageID());
            }

        } // end for (; (PckBegin != PckEnd); PckBegin++)

        // done with packages!

        // remove all directories on the target dir list
        // which have been collected from all packages above
        // (this list is empty for verify mode)
        list<FEDBPackage*>::iterator TargetDirFirst = TargetDirList.begin(),
                                     TargetDirLast = TargetDirList.end();
        for (; TargetDirFirst != TargetDirLast; TargetDirFirst++)
        {
            FEDBPackage *pPackage = *TargetDirFirst;
            pdbti->_Status.pPackage = pPackage;
            pdbti->_Status.pwifh = 0;

            pdbti->_Status.cDirsInDir = 0;
            pdbti->_Status.cFilesInDir = 0;
            const char *pcszDir = pPackage->QueryTargetPath();
            APIRET arc = doshDeleteDir(pcszDir,
                                       DOSHDELDIR_RECURSE,
                                       &pdbti->_Status.cDirsInDir,
                                       &pdbti->_Status.cFilesInDir);
            if (arc != NO_ERROR)
            {
                if (pdbti->_pLogFile)
                    pdbti->_pLogFile->Append("Error %d removing directory \"%s\".\n",
                                             arc,
                                             pcszDir);

                guiDatabaseCallback(DBC_DELETEDIRERROR,
                                    &pdbti->_Status,
                                    arc,
                                    pdbti->_ulUser);
                (*TargetDirFirst)->_ulStatus = 1;         // == errors found
            }
            else
                if (pdbti->_pLogFile)
                    pdbti->_pLogFile->Append("Removed directory \"%s\".\n",
                                             pcszDir);
        }

        // write BSConfigSys back to disk
        if (fConfigSysChanged)
        {
            if (pConfigSys->Flush(G_WpiGlobals.szConfigSysBackup,
                                  pdbti->_pLogFile) == 0)
                G_WpiGlobals.fConfigSysChanged = TRUE;
            delete(pConfigSys);
        }

        // report NULL package (== we're done)
        guiDatabaseCallback(DBC_NEXTPACKAGE,
                            NULL,
                            0,
                            pdbti->_ulUser);

        if (pdbti->_pLogFile)
            pdbti->_pLogFile->Append("WarpIN Database thread is exiting.\n");

        delete pdbti;       // this has a destructor, which
                            // also closes the log file

        G_WpiGlobals.fDeinstallComplete = TRUE;
                // provoke exit message
    }
    CATCH (excpt1)
    {
    } END_CATCH();

    // WinTerminate(pdbti->_hab);
            // removed V0.9.12 (2001-05-22) [umoeller]
}




