
/*
 *@@sourcefile fe_job.cpp:
 *      this implements the job classes, based on FEJobBase,
 *      which represent the actual job of installing/
 *      deinstalling for "install" mode (see warpin.cpp for
 *      an introduction to these modes).
 *
 *      While installing, each job represents a package to
 *      be installed or deinstalled. Jobs contain additional
 *      data for packages which can be deinstalled, while
 *      the packages themselves only represent the (read-only)
 *      data from the script.
 *
 *      For example, the package has the target path as specified
 *      in the script, but the job has the target path which was
 *      actually entered by the user.
 *
 *      Jobs are _not_ used in database mode, because the database
 *      operates on (database) packages directly (FEDBPackage).
 *
 *      Each instance of FEInstallJob has a member pointer
 *      to the FEArcPackagePck which is to be installed
 *      or deinstalled.
 *
 *      A list of instances of these classes is built in
 *      WPIGLOBALS before warpin.cpp calls the GUI to
 *      display the dialogs. The GUI is responsible
 *      for marking the packages for install, using
 *      FEInstallJob::Select.
 *
 *      After the GUI returns control to the front end,
 *      this list will then get processed. For each package
 *      which has been successfully installed, we call
 *      FEInstallJob::Store() to have it stored in the
 *      database.
 *
 *      The job classes have been added for several reasons.
 *      For one, with the previous design (working on
 *      packages directly), it was nearly impossible to
 *      implement a working add/remove mode. Secondly, this
 *      design allows WarpIN to work on several archives
 *      at the same time, because the job instances can
 *      now point to packages from several archives.
 *
 *@@added V0.9.0 (99-11-01) [umoeller]
 *@@header "frontend\fe_job.h"
 */

/*
 *      This file Copyright (C) 1999-2000 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define OS2EMX_PLAIN_CHAR
    // this is needed for "os2emx.h"; if this is defined,
    // emx will define PSZ as _signed_ char, otherwise
    // as unsigned char

#define INCL_WINSHELLDATA
#define INCL_WINWORKPLACE
#include <os2.h>

#include <stdio.h>
#include <string.h>

#include "setup.h"

#include "helpers\xstring.h"

// back-end includes
#include "wiarchive\wiarchive.h"

// base includes (99-11-07) [umoeller]
#include "base\bs_string.h"
#include "base\bs_errors.h"
#include "base\bs_logger.h"
#include "base\bs_config.h"
#include "base\bs_script.h"

// front-end includes
#include "frontend\fe_base.h"
#include "frontend\fe_job.h"
#include "frontend\fe_database.h"
#include "frontend\fe_package.h"
#include "frontend\fe_package_arc.h"

#include "frontend\calbacks.h"

#pragma hdrstop

/* ******************************************************************
 *
 *  FEJobBase implementation
 *
 ********************************************************************/

/*
 *@@ FEJobBase:
 *      protected default constructor.
 */

FEJobBase::FEJobBase(FEGroupJob *pGroupJob_)
{
    _ulJobClass = JOBCLASS_ERROR;
    _ulJobID = JOB_UNDEFINED;
    _pGroupJob = pGroupJob_;
    if (_pGroupJob)
        // we're part of a group:
        // store ourselves in the group's member list
        _pGroupJob->_GroupMembersList.push_back(this);
    _pvGuiData = 0;
}

/*
 *@@ IsInstallJob:
 *      virtual method overridden by FEInstallJob only.
 *      The default returns NULL.
 *
 *      This can be used to quickly check if a given
 *      FEJobBase instance is actually of the FEInstallJob
 *      subclass (and not FEGroupJob) and to typecast the
 *      pointer at the same time if it is.
 */

FEInstallJob* FEJobBase::IsInstallJob()
{
    return (NULL);
}

/*
 *@@ IsGroupJob:
 *      virtual method overridden by FEGroupJob only.
 *      The default returns NULL.
 *
 *      This can be used to quickly check if a given
 *      FEJobBase instance is actually of the FEGroupJob
 *      subclass (and not FEInstallJob) and to typecast
 *      the pointer at the same time if it is.
 */

FEGroupJob* FEJobBase::IsGroupJob()
{
    return (NULL);
}

/*
 *@@ QueryJobID:
 *      returns the current job status, which is one
 *      of:
 *
 *      -- JOB_INSTALL_IGNORE: well, do nothing
 *      -- JOB_INSTALL: install package
 *      -- JOB_DEINSTALL: remove package
 *
 *      If this job is an FEGroupJob, in addition:
 *
 *      -- JOB_GROUP_MIXED: sub-packages have
 *         varying flags.
 */

ULONG FEJobBase::QueryJobID()
{
    return (_ulJobID);
}

/* ******************************************************************
 *
 *  FEInstallJob implementation
 *
 ********************************************************************/

/*
 *@@ FEInstallJob:
 *      constructor to create an install job from a
 *      given package. This is used while the front
 *      end sets up the packages from the archive.
 *
 *      All the constructor arguments are required.
 *
 *      Note: To give the ulInstall* fields a meaningful
 *      value, you must explicitly call FEInstallJob::CheckInstall.
 *      This cannot be done by the constructor because while
 *      all the jobs are being created, we know little about
 *      the database yet.
 */

FEInstallJob::FEInstallJob(FEArcPackagePck *pPackage,  // in: new member package
                           FEGroupJob *pGroupJob)      // in: the group we belong to or NULL
              : FEJobBase(pGroupJob)
{
    _ulJobClass = JOBCLASS_INSTALLJOB;
    _ulJobID = JOB_INSTALL_IGNORE;

    _pArcPackage = pPackage;
    _pPackageID = 0;
    _pPackageIDInstalled = 0;

    _ulInstallStatus = INSTALLED_UNKNOWN;
}

/*
 *@@ FEInstallJob:
 *      destructor.
 */

FEInstallJob::~FEInstallJob()
{
    if (_pPackageID)
        delete _pPackageID;
    if (_pPackageIDInstalled)
        delete _pPackageIDInstalled;
}

/*
 *@@ IsInstallJob:
 *      since we're of the FEInstallJob class,
 *      we return "this" instead of the base class's NULL.
 */

FEInstallJob* FEInstallJob::IsInstallJob()
{
    return (this);
}

/*
 *@@ CheckInstall:
 *      this queries the given database for whether
 *      the package of this job is already installed.
 *
 *      After this call, the ulInstall* fields have
 *      a meaningful value.
 *
 *      This returns TRUE if the package is already installed.
 */

BOOL FEInstallJob::CheckInstall(FEDatabase *pDatabase)
{
    BOOL brc = FALSE;

    _ulInstallStatus = INSTALLED_NO;

    // create package ID
    _pPackageID = new FEPackageID(_pArcPackage, __FILE__, __LINE__,
                                  FALSE);    // don't allow short format

    _pPackageIDInstalled = pDatabase->FindPackageID(_pPackageID->_strAuthor.c_str(),
                                                    _pPackageID->_strApplication.c_str(),
                                                    _pPackageID->_strPackageName.c_str());
    if (_pPackageIDInstalled)
    {
        // package found:

        // compare versions
        CHAR    szVersionThis[40],
                szVersionInstalled[40];
        sprintf(szVersionThis, "%010d.%010d.%010d",
                _pPackageID->_ulVersionMajor,
                _pPackageID->_ulVersionMinor,
                _pPackageID->_ulVersionRevision);
        sprintf(szVersionInstalled, "%010d.%010d.%010d",
                _pPackageIDInstalled->_ulVersionMajor,
                _pPackageIDInstalled->_ulVersionMinor,
                _pPackageIDInstalled->_ulVersionRevision);
        int icomp = strcmp(szVersionInstalled, szVersionThis);
        if (icomp > 0)
            this->_ulInstallStatus = INSTALLED_IS_NEWER;
        else if (icomp < 0)
            this->_ulInstallStatus = INSTALLED_IS_OLDER;
        else    // 0:
            this->_ulInstallStatus = INSTALLED_IS_SAME;

        brc = TRUE;
    }

    return (brc);
}

/*
 *@@ QueryTargetPath:
 *      returns a const pointer to the target path
 *      member.
 */

const char* FEInstallJob::QueryTargetPath()
            const
{
    return (_strTargetPath.c_str());
}

/*
 *@@ QueryPckPackageID:
 *      returns the package ID of the
 *      install job's archive package.
 *
 *      This has only been added so that this
 *      can be queried without having to include
 *      all the package headers.
 *
 *@@added V0.9.9 (2001-03-30) [umoeller]
 */

const char* FEInstallJob::QueryPckPackageID()
            const
{
    if (_pArcPackage)
        return (_pArcPackage->QueryPackageID());

    return (0);
}

/*
 *@@ QueryPckSizeRounded:
 *      returns the _ulSizeRounded member of the
 *      install job's archive package.
 *
 *      This has only been added so that this
 *      can be queried without having to include
 *      all the package headers.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

ULONG FEInstallJob::QueryPckSizeRounded()
      const
{
    if (_pArcPackage)
        return (_pArcPackage->_ulSizeRounded);

    return (0);
}

/*
 *@@ QueryPckFilesCount:
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

ULONG FEInstallJob::QueryPckFilesCount()
            const
{
    if (_pArcPackage)
        return (_pArcPackage->_PackHeader.files);

    return (0);
}

/*
 *@@ QueryPckOrigSize:
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

ULONG FEInstallJob::QueryPckOrigSize()
            const
{
    if (_pArcPackage)
        return (_pArcPackage->_PackHeader.origsize);

    return (0);
}


/*
 *@@ QueryPckTitle:
 *      invokes FEPackageBase::QueryTitle on the
 *      install job's archive package.
 *
 *      This has only been added so that this
 *      can be queried without having to include
 *      all the package headers.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

const char* FEInstallJob::QueryPckTitle()
            const
{
    if (_pArcPackage)
        return (_pArcPackage->QueryTitle());

    return (0);
}

/*
 *@@ SetTargetPath:
 *      sets the new target path for the package
 *      and updates other packages' target paths
 *      if BASE attributes have been set for them.
 *
 *      Returns TRUE if the drive letter has changed.
 *      The caller should update the drives dialog then.
 *
 *@@changed V0.9.7 (2001-01-07) [umoeller]: fixed BASE handling not to update paths of installed packages
 */

BOOL FEInstallJob::SetTargetPath(const char *pszNewTargetPath,
                                        // in: new target path
                                 list<FEJobBase*> &JobsList)
                                        // in: list of all install jobs; this is processed
                                        // if "this" has the BASE attribute
{
    BOOL    brc = FALSE;

    // is our member package is a base package?
    if (_pArcPackage->_pDecl->_ulPathType & PATH_BASE)
    {
        // yes: update all other packages
        list<FEJobBase*>::iterator JobStart = JobsList.begin(),
                                   JobEnd = JobsList.end();
        for (; JobStart != JobEnd; JobStart++)
        {
            FEJobBase *pTemp = *JobStart;
            FEInstallJob *pJobThat = pTemp->IsInstallJob();
            if (pJobThat)
                if (    (pJobThat != this)
                     && (pJobThat->_pPackageIDInstalled == NULL)
                            // only if package is not installed yet...
                            // because otherwise it already has the target
                            // path of the installed package, which we must
                            // not change!!
                   )
                {
                    BSString strSearch(_strTargetPath);
                    BSString strTargetPathThat(pJobThat->QueryTargetPath());
                    size_type ulOfs = 0;
                    strTargetPathThat._find_replace(strSearch,
                                                       // search string: our old target path
                                                    pszNewTargetPath,
                                                       // replace with new path
                                                    &ulOfs);
                    pJobThat->_strTargetPath = strTargetPathThat;
                }
        }
    }

    if (_strTargetPath[0] != *pszNewTargetPath)
        // drive letter changed:
        // update drive info, because the drive might
        // have changed
        brc = TRUE;

    _strTargetPath = pszNewTargetPath;

    return (brc);
}

/*
 *@@ DefaultSelect:
 *      this selects the job for installation according
 *      to the default selection which was specified in
 *      the script. A package is however not selected
 *      if a newer version is already installed.
 *
 *      This calls FEInstallJob::Select in turn.
 *
 *      Returns FALSE if CheckInstall hasn't been
 *      called yet.
 *
 *@@added V0.9.0 (99-11-06) [umoeller]
 *@@changed V0.9.12 (2001-05-31) [umoeller]: fixed NODESELECT
 */

BOOL FEInstallJob::DefaultSelect()
{
    if (    (    (_pArcPackage->_pDecl->_Selected == BSPckDeclBase::SELECTED)
              || (_pArcPackage->_pDecl->_Selected == BSPckDeclBase::NODESELECT)
                    // NODESELECT was missing V0.9.12 (2001-05-31) [umoeller]
            )
         && (_ulInstallStatus != INSTALLED_IS_NEWER)
       )
        Select(JOB_INSTALL);
    else
        Select(JOB_INSTALL_IGNORE);

    return (TRUE);
}

/*
 *@@ RepairSelect:
 *      this compares the job's package version with
 *      the installed version, if any, and selects
 *      the package for installation if it's not
 *      yet installed, older, or same version --
 *      but not if the installed version is newer.
 *
 *      This calls FEInstallJob::Select in turn.
 *
 *      Returns FALSE if CheckInstall hasn't been
 *      called yet.
 *
 *@@added V0.9.0 (99-11-06) [umoeller]
 */

BOOL FEInstallJob::RepairSelect()
{
    switch (_ulInstallStatus)
    {
        case INSTALLED_UNKNOWN:
            return (FALSE);

        case INSTALLED_NO:
            // not installed:
        case INSTALLED_IS_NEWER:
            // newer package installed:
            // deselect package per default, we don't want
            // to overwrite with outdated version
            Select(JOB_INSTALL_IGNORE);
        break;

        case INSTALLED_IS_OLDER:
        case INSTALLED_IS_SAME:
            Select(JOB_INSTALL);
        break;
    }

    return (TRUE);
}

/*
 *@@ UpdateSelect:
 *      this compares the job's package version with
 *      the installed version, if any, and selects
 *      the package only if the one in the archive
 *      is newer than the installed version.
 *
 *      This calls FEInstallJob::Select in turn.
 *
 *      Returns FALSE if CheckInstall hasn't been
 *      called yet.
 *
 *@@added V0.9.0 (99-11-06) [umoeller]
 */

BOOL FEInstallJob::UpdateSelect()
{
    switch (_ulInstallStatus)
    {
        case INSTALLED_UNKNOWN:
            return (FALSE);

        case INSTALLED_IS_OLDER:
            Select(JOB_INSTALL);
        break;

        default:
            Select(JOB_INSTALL_IGNORE);
    }

    return (TRUE);
}

/*
 *@@ Select:
 *      this virtual FEJobBase method is overridden to
 *      set the job status for this job.
 *
 *      ulJobNew can be one of the following:
 *
 *      -- JOB_INSTALL:   select the member package for installation.
 *      -- JOB_DEINSTALL:    select the member package for removal (de-install).
 *                  If this is specified and the package isn't installed,
 *                  JOB_INSTALL_IGNORE is chosen instead.
 *      -- JOB_INSTALL_IGNORE: do nothing. If the package is not yet installed,
 *                  it will not be installed. If the package is
 *                  already installed, it will not be deinstalled.
 *
 *      This calls guiJobSelectionChanged if anything has changed
 *      and automatically updates the group to which this job
 *      belongs to, if any.
 *
 *@@changed V0.9.1 (2000-01-05) [umoeller]: it was impossible to deinstall "no-deselect" packages; fixed
 *@@changed V0.9.1 (2000-01-08) [umoeller]: renamed from "SetJobID"
 */

BOOL FEInstallJob::Select(ULONG ulJobNew)
{
    BOOL brc = TRUE;

    if (_ulJobID != ulJobNew)
    {
        // prohibit ignore if package must be installed
        if (    (ulJobNew == JOB_INSTALL_IGNORE)
             && (_pArcPackage->_pDecl->_Selected == BSPckDeclBase::NODESELECT)
           )
            ulJobNew = JOB_INSTALL;
        // prohibit deinstall?
        else if (ulJobNew == JOB_DEINSTALL)
        {
            // prohibit deinstall if package is not installed
            if (_ulInstallStatus == INSTALLED_NO)
                ulJobNew = JOB_INSTALL_IGNORE;
        }

        _ulJobID = ulJobNew;

        guiOnJobSelectionChanged(this);

        if (_pGroupJob)
        {
            _pGroupJob->Update();
        }
    }

    return (brc);
}

/*
 *@@ ToggleSelection:
 *      similar to FEInstallJob::Select, but this toggles
 *      the selection.
 *
 *      If the member package is already installed, this
 *      circles thru the three job statuses. Otherwise,
 *      "deinstall" is not used, of course.
 *
 *      This is a virtual method defined abstractly by
 *      FEJobBase. This way, packages can toggle differently
 *      from groups.
 *
 *      This returns the new selection status.
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: PckInfoList param removed
 */

ULONG FEInstallJob::ToggleSelection()
{
    switch (_ulJobID)
    {
        case JOB_INSTALL_IGNORE:
            Select(JOB_INSTALL);
        break;

        case JOB_INSTALL:
            if (    (_ulInstallStatus != INSTALLED_UNKNOWN)
                 && (_ulInstallStatus != INSTALLED_NO)
               )
                // package already installed: deinstall then
                Select(JOB_DEINSTALL);
            else
                Select(JOB_INSTALL_IGNORE);
        break;

        case JOB_DEINSTALL:
            Select(JOB_INSTALL_IGNORE);
        break;
    }

    return (_ulJobID);
}

/*
 *@@ Is2BeInstalled:
 *      returns TRUE if the job represents a
 *      package (not a group) and has been
 *      selected for installation.
 */

BOOL FEInstallJob::Is2BeInstalled()
{
    BOOL brc = FALSE;

    if (_ulJobID == JOB_INSTALL)
        if (_pArcPackage->IsPackage())
            brc = TRUE;

    return (brc);
}

/*
 *@@ Is2BeRemoved:
 *      returns TRUE if the job represents a
 *      package (not a group) and has been
 *      selected for de-installation.
 */

BOOL FEInstallJob::Is2BeRemoved()
{
    BOOL brc = FALSE;

    if (_ulJobID == JOB_DEINSTALL)
        if (_pArcPackage->IsPackage())
            brc = TRUE;

    return (brc);
}

/*
 *@@ Unpack:
 *      simple wrapper around FEArcPackagePck::Unpack,
 *      which in turn calls WIArchive::unpack. This
 *      unpacks all files in the member package and
 *      does not return until all files have been unpacked
 *      or a fatal error occured.
 *
 *      That method is called on the archive to which
 *      the member package belongs, with the package
 *      index of the member package.
 *
 *      This produces tons of calls to the callback that
 *      was specified with the FEArcPackagePck constructor.
 *
 *      Gets called on the WarpIN install thread for each
 *      package that is to be installed.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

VOID FEInstallJob::Unpack(VOID)
{
    _pArcPackage->Unpack();
}

/*
 *@@ AddFileHeader:
 *      simple wrapper around FEArcPackagePck::AddFileHeader.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

VOID FEInstallJob::AddFileHeader(WIFileHeader* pwifh)
{
    _pArcPackage->AddFileHeader(pwifh);
}

/*
 *@@ Store:
 *      this stores the member package in the given database.
 *
 *      This gets called from warpin.cpp after installation
 *      was complete and successful to store the package
 *      information in the global database.
 *
 *      We store the whole package data in the database then,
 *      using the package ID stored in _pszID.
 *      This better be valid.
 *
 *      If this package is already stored in the database,
 *      even if it's an older version, the old data is
 *      overwritten without further notice.
 *
 *      The format is so that the package info can later
 *      be recreated as an instance of FEDBPackage simply
 *      by passing the package ID to the FEDBPackage::FEDBPackage
 *      constructor, which in turn calls FEDBPackage::Load.
 *
 *      The data stored here must match exactly the data which
 *      is restored in FEDBPackage::Load.
 *
 *      This returns FALSE upon errors.
 *
 *@@changed V0.9.0 (99-10-26) [umoeller]: six-part ID support added
 *@@changed V0.9.0 (99-11-07) [umoeller]: added install date/time
 *@@changed V0.9.1 (2000-01-07) [umoeller]: changed "requires" storage
 *@@changed V0.9.1 (2000-02-03) [umoeller]: added "execute" loggers support
 *@@changed V0.9.2 (2000-03-11) [umoeller]: added description and codepage
 */

BOOL FEInstallJob::Store(FEDatabase* pDatabase)
{
    BOOL    brc = FALSE;

    // normalize package ID by creating an FEPackageID
    // and re-parsing that
    FEPackageID pckid(_pArcPackage, __FILE__, __LINE__,
                      FALSE);    // don't allow short format
    BSString    strIDSix;
    pckid.CreateStringSix(strIDSix);
    const char *pcszIDSix = strIDSix.c_str();

    // check if another version of this package has
    // been stored already
    const FEDBPackage *pPckInstalled = pDatabase->FindDBPackage(pckid);
    if (pPckInstalled)
        // yes: remove that first
        pDatabase->RemovePackage(pPckInstalled);

    // write files list which was set by the
    // Install thread
    brc = PrfWriteProfileData(pDatabase->hiniDatabase,
                              pcszIDSix,
                              WPIKEY_FILESLIST,
                              _pArcPackage->_pszFilesList,
                              _pArcPackage->_cbFilesList);

    PrfWriteProfileString(pDatabase->hiniDatabase,
                          pcszIDSix,
                          WPIKEY_TARGETPATH,
                          _strTargetPath.c_str());

    _pArcPackage->_logRequiresIDs.Store(pDatabase->hiniDatabase,
                                        pcszIDSix,
                                        WPIKEY_REQUIRES);

    _pArcPackage->_pDecl->_logCfgSysAttrs.Store(pDatabase->hiniDatabase,
                                                pcszIDSix,
                                                WPIKEY_CONFIGATTRS);
    _pArcPackage->_logCfgSysDone.Store(pDatabase->hiniDatabase,
                                       pcszIDSix,
                                       WPIKEY_CONFIGDONE);

    _pArcPackage->_pDecl->_logRegisterClassAttrs.Store(pDatabase->hiniDatabase,
                                                       pcszIDSix,
                                                       WPIKEY_REGISTERATTRS);
    _pArcPackage->_logRegisterClassesDone.Store(pDatabase->hiniDatabase,
                                                pcszIDSix,
                                                WPIKEY_REGISTERDONE);

    _pArcPackage->_pDecl->_logReplaceClassAttrs.Store(pDatabase->hiniDatabase,
                                                      pcszIDSix,
                                                      WPIKEY_REPLACEATTRS);
    _pArcPackage->_logReplaceClassDone.Store(pDatabase->hiniDatabase,
                                             pcszIDSix,
                                             WPIKEY_REPLACEDONE);

    _pArcPackage->_pDecl->_logWPSObjectAttrs.Store(pDatabase->hiniDatabase,
                                                   pcszIDSix,
                                                   WPIKEY_WPSOBJECTATTRS);
    _pArcPackage->_logWPSObjectsDone.Store(pDatabase->hiniDatabase,
                                           pcszIDSix,
                                           WPIKEY_WPSOBJECTDONE);

    _pArcPackage->_pDecl->_logClearPrfAttrs.Store(pDatabase->hiniDatabase,
                                                  pcszIDSix,
                                                  WPIKEY_CLEARPRFATTRS);
    _pArcPackage->_pDecl->_logWritePrfAttrs.Store(pDatabase->hiniDatabase,
                                                  pcszIDSix,
                                                  WPIKEY_WRITEPRFATTRS);
    _pArcPackage->_logWritePrfsDone.Store(pDatabase->hiniDatabase,
                                          pcszIDSix,
                                          WPIKEY_WRITEPRFDONE);

    _pArcPackage->_pDecl->_logExecuteAttrs.Store(pDatabase->hiniDatabase,
                                                 pcszIDSix,
                                                 WPIKEY_EXECUTEATTRS);
    _pArcPackage->_logExecutesDone.Store(pDatabase->hiniDatabase,
                                         pcszIDSix,
                                         WPIKEY_EXECUTEDONE);

    _pArcPackage->_pDecl->_logDeExecuteAttrs.Store(pDatabase->hiniDatabase,
                                                   pcszIDSix,
                                                   WPIKEY_DEEXECUTEATTRS);
    _pArcPackage->_logDeExecutesResolved.Store(pDatabase->hiniDatabase,
                                               pcszIDSix,
                                               WPIKEY_DEEXECUTERESOLVED);

    _pArcPackage->_pDecl->_logKillProcessAttrs.Store(pDatabase->hiniDatabase,
                                                     pcszIDSix,
                                                     WPIKEY_KILLPROCESSATTRS);

    PrfWriteProfileData(pDatabase->hiniDatabase,
                        pcszIDSix,
                        WPIKEY_WIPACKHEADER,
                        (PVOID)(&_pArcPackage->_PackHeader),
                        sizeof(_pArcPackage->_PackHeader));

    // store current date/time as install date/time
    DATETIME dt;
    DosGetDateTime(&dt);
    PrfWriteProfileData(pDatabase->hiniDatabase,
                        pcszIDSix,
                        WPIKEY_INSTALLDATETIME,
                        &dt,
                        sizeof(DATETIME));

    // store description V0.9.2 (2000-03-11) [umoeller]
    CHAR szCodepage[20];
    sprintf(szCodepage, "%u", _pArcPackage->QueryCodepage());
    PrfWriteProfileString(pDatabase->hiniDatabase,
                          pcszIDSix,
                          WPIKEY_CODEPAGE,
                          szCodepage);

    return (brc);
}

/* ******************************************************************
 *
 *  FEGroupJob implementation
 *
 ********************************************************************/

FEGroupJob::FEGroupJob(FEArcPackageGroup *pGroup_, // in: group package (required)
                       FEGroupJob *pGroupJob_)     // in: our own group (or NULL)
            : FEJobBase(pGroupJob_)
{
    _ulJobClass = JOBCLASS_INSTALLJOB;

    _pGroupPackage = pGroup_;
}

/*
 *@@ IsGroupJob:
 *      since we're of the FEGroupJob class,
 *      we return "this" instead of the base class's NULL.
 */

FEGroupJob* FEGroupJob::IsGroupJob()
{
    return (this);
}

/*
 *@@ ToggleSelection:
 *      this toggles the selection for the group and all
 *      its members.
 *
 *      This is a virtual method defined abstractly by
 *      FEJobBase. This way, packages can toggle differently
 *      from groups.
 *
 *      This returns the new selection status.
 *
 *@@changed V0.9.0 (99-10-31) [umoeller]: PckInfoList param removed
 */

ULONG FEGroupJob::ToggleSelection()
{
    ULONG   ulNewJobID;

    switch (_ulJobID)
    {
        case JOB_INSTALL:
            ulNewJobID = JOB_DEINSTALL;
        break;

        case JOB_DEINSTALL:
            ulNewJobID = JOB_INSTALL_IGNORE;
        break;

        case JOB_INSTALL_IGNORE:
            ulNewJobID = JOB_INSTALL;
        break;

        case JOB_GROUP_MIXED:
            ulNewJobID = JOB_INSTALL;
        break;
    }

    Select(ulNewJobID);

    return (ulNewJobID);
}

/*
 *@@ Select:
 *      this virtual FEJobBase method is overridden to
 *      set the job status for this group job and all of
 *      its member jobs.
 *
 *      ulJobNew can be one of the following:
 *
 *      -- JOB_INSTALL:   select the member package for installation.
 *      -- JOB_DEINSTALL:    select the member package for removal (de-install).
 *      -- JOB_INSTALL_IGNORE: do nothing. If the package is not yet installed,
 *                        it will not be installed. If the package is
 *                        already installed, it will not be deinstalled.
 *
 *      This calls guiJobSelectionChanged if anything has changed
 *      and automatically updates the group to which this group job
 *      belongs to, if any.
 *
 *@@changed V0.9.1 (2000-01-08) [umoeller]: renamed from "SetJobID"
 */

ULONG FEGroupJob::Select(ULONG ulJobNew)
{
    ULONG ulrc = 0;
    list<FEJobBase*>::iterator JobStart = _GroupMembersList.begin(),
                               JobEnd = _GroupMembersList.end();
    for (; JobStart != JobEnd; JobStart++)
    {
        (**JobStart).Select(ulJobNew);
            // this will call Update in turn
        ulrc++;
    }

    return (ulrc);
}

/*
 *@@ Update:
 *      this gets called by FEInstallJob::Select
 *      if the selection has changed for a package
 *      which belongs to a group, so that the group
 *      can update itself according to the members.
 */

VOID FEGroupJob::Update()
{
    const ULONG FOUND_MIXED     = 0x0001,
                FOUND_INSTALL   = 0x0002,
                FOUND_DEINSTALL = 0x0004,
                FOUND_IGNORE    = 0x0008;

    ULONG   ulFound = 0;

    // check our members for selections;
    // we will OR the flags above in ulFound
    // so if all our members have the same
    // selections, we'll get a "straight" value
    // in ulValue, otherwise more than one bit
    // will be set
    list<FEJobBase*>::iterator JobStart = _GroupMembersList.begin(),
                               JobEnd = _GroupMembersList.end();
    for (; JobStart != JobEnd; JobStart++)
    {
        FEJobBase *pJobThis = *JobStart;
        switch (pJobThis->QueryJobID())
        {
            case JOB_INSTALL:
                ulFound |= FOUND_INSTALL;
            break;

            case JOB_DEINSTALL:
                ulFound |= FOUND_DEINSTALL;
            break;

            case JOB_INSTALL_IGNORE:
                ulFound |= FOUND_IGNORE;
            break;

            case JOB_GROUP_MIXED:
                ulFound |= FOUND_MIXED;
            break;
        }
    }

    // now check if we have a "straight" one-bit value
    // and set our own value accordingly
    if (ulFound == FOUND_INSTALL)
        _ulJobID = JOB_INSTALL;
    else if (ulFound == FOUND_DEINSTALL)
        _ulJobID = JOB_DEINSTALL;
    else if (ulFound == FOUND_IGNORE)
        _ulJobID = JOB_INSTALL_IGNORE;
    else
        // any other combination:
        _ulJobID = JOB_GROUP_MIXED;

    guiOnJobSelectionChanged(this);

    if (_pGroupJob)
    {
        // we belong to a group ourselves:
        // update parent group too (this will
        // climb up the group tree until the
        // root has been reached)
        _pGroupJob->Update();
    }
}


