
/*
 *@@sourcefile fe_package.cpp:
 *      this implements the following classes (all declared in fe_package.h):
 *
 *      --  FEPackageID helper class (for parsing package ID strings).
 *
 *      --  The actual package classes. The hierarchy is as follows:
 *
 +          FEPackageBase (abstract base class)
 +                |
 +                +-- FEArcPackageBase (abstract base class for
 +                |     |               packages in archive)
 +                |     |
 +                |     +-- FEArcPackagePck (package in archive)
 +                |     |
 +                |     +-- FEArcPackageGroup (group def'd in script;
 +                |                            no real package)
 +                |
 +                +-- FEDBPackage (package retrieved from database)
 +
 *      "Install" mode uses FEArcPackageBase and subclasses.
 *      "Database" mode uses FEDBPackage. If packages are de-installed
 *      in "Install" mode, FEDBPackage instances are used as well.
 *
 *      In "install" mode, the packages contain data as specified
 *      in the install script; in addition, for each package,
 *      instances of FEJobBase are created to store additional
 *      configuration data (according to user selections).
 *
 *      In "database" mode, instances of FEDBPackage are created
 *      from the database directly, without jobs.
 *
 *      Before V0.9.0, 99-10-31, we only had a single "PackageInfo" class.
 *      Now this is much more complicated, as C++ code should be. ;-)
 *
 *@@header "frontend\fe_package.h"
 */

/*
 *      This file Copyright (C) 1999-2000 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define OS2EMX_PLAIN_CHAR
    // this is needed for "os2emx.h"; if this is defined,
    // emx will define PSZ as _signed_ char, otherwise
    // as unsigned char

#include <os2.h>

#include <stdio.h>
#include <string.h>

#include "setup.h"

// include's from helpers
#include "helpers\stringh.h"
#include "helpers\xstring.h"

// back-end includes
#include "wiarchive\wiarchive.h"

// base includes (99-11-07) [umoeller]
#include "base\bs_string.h"
#include "base\bs_errors.h"
#include "base\bs_logger.h"
#include "base\bs_config.h"
#include "base\bs_script.h"

// front-end includes
#include "frontend\fe_base.h"
#include "frontend\fe_package.h"
#include "frontend\fe_package_arc.h"
#include "frontend\fe_package_db.h"

#include "frontend\calbacks.h"

#pragma hdrstop

/* ******************************************************************
 *
 *  PackageID class
 *
 ********************************************************************/

/*
 *@@ Init:
 *
 *      Returns 0 on success, an error message number otherwise.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

ULONG FEPackageID::Init(const char *pszPackageID,
                        BOOL fAllowShort)
{
    ULONG   ulMsg = 0;

    if (pszPackageID)
    {
        PSZ     p,
                p2;
                // pszAuthor_ = 0,
                // pszApplication_ = 0,
                // pszPackageName_ = 0;
        ULONG   ulVersionMajor_ = 0,
                ulVersionMinor_ = 0,
                ulVersionRevision_ = 0;

        BOOL    fAllOK = FALSE;

        p = strchr(pszPackageID, '\\');
        if (p)
        {
            // valid author:
            _strAuthor._extract_from(pszPackageID, p);

            p2 = strchr(p+1, '\\');
            if (p2)
            {
                // valid application:
                _strApplication._extract_from(p+1, p2);

                BOOL fMinorValid = FALSE;

                p = p2;
                p2 = strchr(p+1, '\\');
                if (!p2)
                {
                    // '\' after application not found:
                    // this is only valid in "allow short" mode
                    if (fAllowShort)
                        _strPackageName = p + 1;
                }
                else
                {
                    // valid package name:
                    _strPackageName._extract_from(p+1, p2);

                    p = p2;
                    p2 = strchr(p+1, '\\');
                    if (p2)
                    {
                        // valid major version:
                        PSZ _pszMajorVersion = strhSubstr(p+1, p2);
                        sscanf(_pszMajorVersion, "%u", &ulVersionMajor_);
                        free(_pszMajorVersion);

                        fAllOK = TRUE;

                        p = p2;
                        p2 = strchr(p+1, '\\');
                        if (p2)
                        {
                            // revision specified: scan minor revision
                            PSZ _pszMinor = strhSubstr(p+1, p2);
                            sscanf(_pszMinor, "%u", &ulVersionMinor_);
                            free(_pszMinor);
                            fMinorValid = TRUE;

                            // scan revision
                            if (strlen(p2+1))
                                sscanf(p2+1, "%u", &ulVersionRevision_);
                        }
                        else
                            // no revision specified:
                            if (strlen(p+1))
                            {
                                sscanf(p+1, "%u", &ulVersionMinor_);
                                fMinorValid = TRUE;
                            }
                    }
                    else
                        ulMsg = 159;        // invalid major version
                }

                if (_strPackageName())
                {
                    if ((fMinorValid) || (fAllowShort))
                    {
                        // all OK: store in instance data
                        _ulVersionMajor = ulVersionMajor_;
                        _ulVersionMinor = ulVersionMinor_;
                        _ulVersionRevision = ulVersionRevision_;

                        return (0);     // success
                    }
                    else
                        ulMsg = 158;        // invalid minor version

                } // end valid package name
                else
                    ulMsg = 160;            // invalid package name

            }
            else
                ulMsg = 161;  // Invalid application name

        }
        else
            ulMsg = 162; // Invalid author name
    }
    else
        ulMsg = 163;  // package ID string is NULL

    return (ulMsg);
}

/*
 *@@ PackageID:
 *      this constructor splits the five- or six-part package ID
 *      into the six member buffers.
 *
 *      The package ID must have the following format:
 +          author\application\packagename\majorversion\minorversion\revision
 *
 *      If "revision" is not specified (that is, the ID string
 *      only contains four backslashes), the ulVersionRevision
 *      member is set to 0.
 *
 *      This constructor throws a FEConstructorExcpt upon
 *      parsing errors, which gets caught in main() only. You
 *      should therefore use this constructor in a try-catch
 *      block yourself.
 *
 *      If (fAllowShort == FALSE), this throws an exception even
 *      if the version numbers are not specified.
 *      If (fAllowShort == TRUE), this throws an exception only
 *      if the author\application\packagename part is broken and
 *      thus accepts a package ID without version numbers.
 *
 *      Throws:
 *      -- FEConstructorExcpt.
 *
 *@@changed V0.9.0 (99-10-26) [umoeller]: made PackageID a separate class and this the default constructor
 *@@changed V0.9.1 (2000-01-03) [umoeller]: version revision was scanned wrong; fixed
 *@@changed V0.9.1 (2000-01-03) [umoeller]: now having safe defaults for all members
 *@@changed V0.9.3 (2000-06-05) [umoeller]: added fAllowShort
 */

FEPackageID::FEPackageID(const char *pszPackageID,
                         const char *pszParentSource,   // in: caller's file (__FILE__)
                         int iParentLine,               // in: caller's line (__LINE__)
                         BOOL fAllowShort)              // in: allow short specification?
{
    ULONG ulMsg = Init(pszPackageID, fAllowShort);
    if (ulMsg)
        throw (FEConstructorExcpt("PackageID", pszParentSource, iParentLine, ulMsg));
}

/*
 *@@ FEPackageID:
 *      second constructor, which calls the first constructor
 *      in turn. This retrieves pszPackageID from the specified
 *      "pPackage".
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 *@@changed V0.9.12 (2001-05-17) [umoeller]: now accepting FEArcPackagePck only
 */

FEPackageID::FEPackageID(const FEArcPackagePck *pPackage,
                         const char *pszParentSource,
                         int iParentLine,
                         BOOL fAllowShort)
{
    ULONG ulMsg = Init(pPackage->_pDecl->_strID.c_str(),
                       fAllowShort);
    if (ulMsg)
        throw (FEConstructorExcpt("PackageID", pszParentSource, iParentLine, ulMsg));
}

/*
 *@@ FEPackageID:
 *      third constructor, which takes a DB package instead.
 *
 *@@added V0.9.12 (2001-05-17) [umoeller]
 */

FEPackageID::FEPackageID(const FEDBPackage *pPackage,
                         const char *pszParentSource,
                         int iParentLine,
                         BOOL fAllowShort)
{
    ULONG ulMsg = Init(pPackage->_pDecl->_strID.c_str(),
                       fAllowShort);
    if (ulMsg)
        throw (FEConstructorExcpt("PackageID", pszParentSource, iParentLine, ulMsg));
}

/*
 *@@ ~FEPackageID:
 *      destructor.
 */

FEPackageID::~FEPackageID()
{
}

/*
 *@@ Compare:
 *      this compares the member versions with
 *      those of the specified package ID.
 *
 *      This returns a ULONG bit field as follows:
 *
 *      If IDCOMP_SAMEPCK (0x10000)is set, this means
 *      that the author, application, and package-name
 *      fields are equal.
 *
 *      In that case, one of the following bits is
 *      also set:
 *
 *      -- IDCOMP_SAMEVERSION (0x00001): both version numbers
 *                  are exactly the same.
 *      -- IDCOMP_THISNEWER (0x00002): "this" has a higher
 *                  version number than "id".
 *      -- IDCOMP_THISOLDER (0x00004): "this" has a lower
 *                  version number than "id".
 *
 *      If IDCOMP_SAMEPCK is not set, no other bits will
 *      be set (i.e. 0 will always be returned).
 *
 *@@added V0.9.1 (2000-01-08) [umoeller]
 */

ULONG FEPackageID::Compare(FEPackageID &id)
                   const
{
    ULONG ulrc = 0;
    if (_strAuthor == id._strAuthor)
        if (_strApplication == id._strApplication)
            if (_strPackageName == id._strPackageName)
            {
                // author, application, and package name match:
                ulrc = IDCOMP_SAMEPCK;

                // now compare major version numbers
                if (_ulVersionMajor < id._ulVersionMajor)
                    ulrc |= IDCOMP_THISOLDER;
                else if (_ulVersionMajor > id._ulVersionMajor)
                    ulrc |= IDCOMP_THISNEWER;
                else
                {
                    // major versions match: compare minor versions
                    if (_ulVersionMinor < id._ulVersionMinor)
                        ulrc |= IDCOMP_THISOLDER;
                    else if (_ulVersionMinor > id._ulVersionMinor)
                        ulrc |= IDCOMP_THISNEWER;
                    else
                    {
                        // minor versions match as well: compare revisions
                        if (_ulVersionRevision < id._ulVersionRevision)
                            ulrc |= IDCOMP_THISOLDER;
                        else if (_ulVersionRevision > id._ulVersionRevision)
                            ulrc |= IDCOMP_THISNEWER;
                        else
                            // revisions match too: that's the same
                            ulrc |= IDCOMP_SAMEVERSION;
                    }
                }
            }


    return (ulrc);
}

/*
 *@@ CreateStringSix:
 *      creates a six-part package ID in the specified
 *      string, which is cleared first.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

VOID FEPackageID::CreateStringSix(BSString &str)
                  const
{
    str = _strAuthor;
    str += '\\';
    str += _strApplication;
    str += '\\';
    str += _strPackageName;

    CHAR    szVersion[100];
    sprintf(szVersion, "\\%u\\%u\\%u",
            _ulVersionMajor,
            _ulVersionMinor,
            _ulVersionRevision);
    str += szVersion;
}

/* ******************************************************************
 *
 *  FEPackageBase implementation
 *
 ********************************************************************/

/*
 *@@ FEPackageBase:
 *      protected default constructor.
 *
 *      This is only to be used by descendants.
 *
 *@@changed V0.9.12 (2001-05-17) [umoeller]: removed declaration parameter
 */

FEPackageBase::FEPackageBase(ULONG ulType)
{
    _ulType = ulType;

    memset(&_PackHeader, 0, sizeof(_PackHeader));

    _pszFilesList = 0;
    _cbFilesList = 0;
}

/*
 *@@ ~FEPackageBase:
 *      destructor.
 */

FEPackageBase::~FEPackageBase()
{
    if (_pszFilesList)
        free(_pszFilesList);
}

/*
 *@@ QueryTargetPath:
 *      returns the target path of the package as a const C string.
 *
 *      This is either the original target path as specified
 *      in the script (if this is an archive package) or the
 *      target path as stored in the database (for database
 *      packages).
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 *@@changed V0.9.12 (2001-05-17) [umoeller]: adjusted for _pDecl move to subclasses
 */

const char* FEPackageBase::QueryTargetPath() const
{
    switch (_ulType)
    {
        case PCK_PCK:
            return (((FEArcPackagePck*)this)->_pDecl->_strTargetPath.c_str());

        case PCK_DBPCK:
            return (((FEDBPackage*)this)->_pDecl->_strTargetPath.c_str());
    }

    return (NULL);
}

/*
 *@@ QueryPackageID:
 *      returns the package's ID as a const C string.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

const char* FEPackageBase::QueryPackageID() const
{
    switch (_ulType)
    {
        case PCK_PCK:
            return (((FEArcPackagePck*)this)->_pDecl->_strID.c_str());

        case PCK_DBPCK:
            return (((FEDBPackage*)this)->_pDecl->_strID.c_str());
    }

    return (NULL);
}

/*
 *@@ QueryTitle:
 *      returns the package's title as a const C string.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

const char* FEPackageBase::QueryTitle() const
{
    switch (_ulType)
    {
        case PCK_PCK:
            return (((FEArcPackagePck*)this)->_pDecl->_strTitle.c_str());

        case PCK_DBPCK:
            return (((FEDBPackage*)this)->_pDecl->_strTitle.c_str());
    }

    return (NULL);
}
