
/*
 *@@sourcefile fe_package_arc.cpp:
 *      this implements the "archive package" classes
 *      (FEArcPackageBase and the two subclasses
 *      FEArcPackagePck and FEArcPackageGroup).
 *
 *      These are created from the data in an install script.
 *      As a result, the constructors in here parse install
 *      script data.
 *
 *      See fe_package.cpp for an overview.
 *
 *      This has been separated from fe_package.cpp to reduce
 *      source code dependencies and speed up compile times.
 *
 *@@added V0.9.1 (2000-02-08) [umoeller]
 *@@header "frontend\fe_package_arc.h"
 */

/*
 *      This file Copyright (C) 1999-2000 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define OS2EMX_PLAIN_CHAR
    // this is needed for "os2emx.h"; if this is defined,
    // emx will define PSZ as _signed_ char, otherwise
    // as unsigned char

#include <os2.h>

#include <stdio.h>
#include <string.h>

#include "setup.h"

// include's from helpers
#include "helpers\configsys.h"
#include "helpers\stringh.h"
#include "helpers\xstring.h"

// back-end includes
#include "wiarchive\wiarchive.h"

// base includes (99-11-07) [umoeller]
#include "base\bs_string.h"
#include "base\bs_errors.h"

#include "base\bs_logger.h"
#include "base\bs_config.h"
#include "base\bs_config_impl.h"
#include "base\bs_script.h"

// front-end includes
#include "frontend\fe_base.h"
#include "frontend\fe_archive.h"
#include "frontend\fe_package.h"
#include "frontend\fe_package_arc.h"

#include "frontend\calbacks.h"

#pragma hdrstop

/* ******************************************************************
 *
 *  FEArcPackageBase implementation
 *
 ********************************************************************/

/*
 *@@ FEArcPackageBase:
 *      protected default constructor.
 *
 *      This is only to be used by descendants.
 *
 *      Throws:
 *      -- FEConstructorExcpt.
 */

FEArcPackageBase::FEArcPackageBase(FEArchive *pArchive,
                                   FEArcPackageGroup *pGroup,
                                   ULONG ulType)
                  : FEPackageBase(ulType)
{
    _pArchive = pArchive;
    _pGroup = pGroup;
    _ulSizeRounded = 0;

    if (!_pArchive)
        throw FEConstructorExcpt("FEArcPackageBase", __FILE__, __LINE__, 169);

    if (_pGroup)
        // group specified: store ourselves in the group's member list
        _pGroup->_GroupMembersList.push_back(this);
}

/*
 *@@ QueryCodepage:
 *      returns the codepage which has been used in
 *      the script of the archive where this package
 *      was created from.
 *
 *@@added V0.9.2 (2000-03-11) [umoeller]
 *@@changed V0.9.9 (2001-02-28) [umoeller]: now always returning "mother" archive
 */

ULONG FEArcPackageBase::QueryCodepage()
      const
{
    // begin V0.9.9 (2001-02-28) [umoeller]
    if (!_pArchive)
        throw(FEFatalErrorExcpt(BSString(__FUNCTION__) + ": _pArchive is NULL."));

    FEArchive *pArchive = _pArchive;
    if (_pArchive->_pReferencedFrom)
        pArchive = _pArchive->_pReferencedFrom;

    if (!pArchive->_pScript)
        throw(FEFatalErrorExcpt(BSString(__FUNCTION__) + ": _pScript is NULL."));
    // end V0.9.9 (2001-02-28) [umoeller]

    return (pArchive->_pScript->_ulCodepage);
}

/* ******************************************************************
 *
 *  FEArcPackagePck implementation
 *
 ********************************************************************/

/*
 *@@ FEArcPackagePck:
 *      constructor for creating an FEArcPackagePck
 *      from an archive.
 *
 *      This is used by FEArchive::ParsePackages
 *      to create packages filled with the data from
 *      the install script.
 *
 *      This is the only constructor for this class,
 *      so there's no way to create an archive package
 *      other than thru install scripts.
 *
 *@@changed V0.9.9 (2001-02-28) [umoeller]: added callback param
 */

FEArcPackagePck::FEArcPackagePck(FEArchive *pArchive,    // in: "parent" archive; must be valid!
                                 FEArcPackageGroup *pGroup,  // in: "parent" group; can be NULL
                                 BSPckDeclBase *pDecl,
                                 PFNWICALLBACK pWICallback)   // in: callback func for back-end
                 : FEArcPackageBase(pArchive,
                                    pGroup,
                                    // mark this as "package":
                                    PCK_PCK)
{
    _pDecl = pDecl;     // V0.9.12 (2001-05-17) [umoeller]

    // look up the package in the archive
    WIPackHeader *pPackHeader = NULL;
    list<WIPackHeader*>::iterator start = _pArchive->_pPackHeadersList->begin(),
                                  end = _pArchive->_pPackHeadersList->end();
    for (; start != end; start++)
    {
        WIPackHeader *pThis = *start;

        if (pThis->number == pDecl->_ulIndexInArchive)
        {
            pPackHeader = pThis;
            break;
        }
    }

    if (!pPackHeader)
    {
        CHAR    sz[40];
        sprintf(sz, "%d", _pDecl->_ulIndexInArchive);
        PSZ     apsz = sz;
        throw(FEFatalErrorExcpt(203,            // invalid package index
                                &apsz,
                                1));
    }

    // copy package header from WIArchive
    memcpy(&_PackHeader, pPackHeader, sizeof(*pPackHeader));

    // convert package size to KByte
    _ulSizeRounded = (pPackHeader->origsize + 512) / 1024;

    _pArchive->_pWICallback2Use = pWICallback;
        // V0.9.9 (2001-02-28) [umoeller]
}

/*
 *@@ Unpack:
 *      simple wrapper around WIArchive::unpack. This
 *      unpacks all files in the package and does
 *      not return until all files have been unpacked
 *      or a fatal error occured.
 *
 *      Gets called by FEInstallJob::Unpack, which in
 *      turn gets called on the WarpIN install thread.
 *
 *      This produces tons of calls to the callback that
 *      was specified with the FEArcPackagePck constructor.
 *
 *@@added V0.9.9 (2001-02-28) [umoeller]
 */

VOID FEArcPackagePck::Unpack(VOID)
{
    _pArchive->_Arc.setCallbackFunc(_pArchive->_pWICallback2Use,
                                    NULL);
    _pArchive->_Arc.unpack(_PackHeader.number);
}

/*
 *@@ AddFileHeader:
 *      this appends the given WIFileHeader to
 *      the pszFilesList member. This is used while
 *      files are being unpacked to create a files
 *      list which can later be stored in the database.
 *
 *      This function updates pszFilesList and
 *      cbFilesList in pPckInfo. With each call,
 *      the buffer that pszFilesList points to
 *      is reallocated.
 *
 *      The format of each entry in pszFilesList is as follows:
 *
 +          CHAR[x]   BYTE time_t  time_t  ULONG    BYTE
 +          filename   0   writedt creatdt filesize  0
 +                                         (uncompressed)
 *      All file entries are simply stored in sequential
 *      order.
 *
 *      <B>Preconditions:</B>
 *
 *      If this is the first call, pszFilesList and
 *      cbFilesList in pPckInfo must be 0.
 *      No check is made for whether the given file
 *      is already in the list.
 *
 *@@changed V0.9.0 (99-10-26) [umoeller]: this used to be datAddFileHeader; made this a FEPackage method
 */

VOID FEArcPackagePck::AddFileHeader(WIFileHeader* pwifh)
{
    ULONG cbFilesListOld = _cbFilesList;
    ULONG cbFilename = strlen(pwifh->name);

    _cbFilesList += (cbFilename
                    + (2 * sizeof(time_t))
                    + sizeof(ULONG)
                    + 2);
                        // two null bytes

    PSZ pszTemp = _pszFilesList;
    _pszFilesList = (PSZ)malloc(_cbFilesList);

    PSZ pszTarget = _pszFilesList + cbFilesListOld;

    if (pszTemp)
    {
        // not first call:
        // copy old file list
        memcpy(_pszFilesList, pszTemp, cbFilesListOld);
        // append new file name
        free(pszTemp);
    }

    strcpy(pszTarget, pwifh->name);     // this appends null terminator
    pszTarget += (cbFilename + 1);
    memcpy(pszTarget, &(pwifh->lastwrite), sizeof(time_t));
    pszTarget += sizeof(time_t);        // with VAC++, this is a long
    memcpy(pszTarget, &(pwifh->creation), sizeof(time_t));
    pszTarget += sizeof(time_t);        // with VAC++, this is a long
    memcpy(pszTarget, &(pwifh->origsize), sizeof(ULONG));
    pszTarget += sizeof(ULONG);
    *pszTarget = 0;
}

/* ******************************************************************
 *
 *  FEArcPackageGroup implementation
 *
 ********************************************************************/

/*
 *@@ FEArcPackageGroup:
 *      constructor.
 */

FEArcPackageGroup::FEArcPackageGroup(FEArchive *pArchive,       // in: archive we're being created from
                                     FEArcPackageGroup *pGroup, // in: group this group belongs to or NULL if none
                                     BSGroupDeclBase *pDecl)
                   : FEArcPackageBase(pArchive,
                                      pGroup,
                                      // mark as group:
                                      PCK_GROUP)
{
    _pGroupDecl = pDecl;
}


