
/*
 *@@sourcefile fe_package_db.cpp:
 *      this implements the "database package" class, which
 *      is a subclass for FEPackageBase.
 *      See fe_package.cpp for details.
 *
 *      This has been separated from fe_package.cpp to reduce
 *      source code dependencies and speed up compile times.
 *
 *@@added V0.9.1 (2000-02-08) [umoeller]
 *@@header "frontend\fe_package_db.h"
 */

/*
 *      This file Copyright (C) 1999-2000 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define OS2EMX_PLAIN_CHAR
    // this is needed for "os2emx.h"; if this is defined,
    // emx will define PSZ as _signed_ char, otherwise
    // as unsigned char

#define INCL_WINSHELLDATA
#define INCL_WINWORKPLACE
#include <os2.h>

#include <stdio.h>
#include <string.h>

#include "setup.h"

// include's from helpers
#include "helpers\configsys.h"
#include "helpers\prfh.h"
#include "helpers\xstring.h"

// back-end includes
#include "wiarchive\wiarchive.h"

// base includes (99-11-07) [umoeller]
#include "base\bs_string.h"
#include "base\bs_errors.h"

#include "base\bs_logger.h"
#include "base\bs_config.h"
#include "base\bs_config_impl.h"
#include "base\bs_script.h"

// front-end includes
#include "frontend\fe_base.h"
#include "frontend\fe_database.h"
#include "frontend\fe_package.h"
#include "frontend\fe_package_db.h"

#include "frontend\calbacks.h"

#pragma hdrstop

/* ******************************************************************
 *
 *  Undo-config helpers
 *
 ********************************************************************/

/*
 *@@ AddCfgSysManipToUndoList:
 *      helper func to create BSCfgSysManip
 *      instances from the logger instance that was
 *      previously used with BSConfigSys::Manipulate.
 *
 *      The new BSCfgSysManip objects are appended to
 *      the specified list and are exact opposites
 *      to the BSCfgSysManip objects that were stored
 *      in the logger. That is, if the logger registered
 *      that something was added to CONFIG.SYS, we
 *      create an object which removes that text again,
 *      and vice versa.
 *
 *      Use this method to undo CONFIG.SYS changes
 *      and pass the objects on the list to BSConfigSys::Manipulate,
 *      since there is no corresponding "undo" class
 *      for BSConfigSys.
 *
 *      See BSCfgSysManip::BSCfgSysManip for an example usage.
 *
 *      This returns the number of items created.
 *
 *      Throws:
 *      -- BSConfigExcpt.
 */

int AddCfgSysManipToUndoList(list<BSConfigBase*> &List,
                             BSCfgSysDoneLogger &logger)
{
    // the logger in this case has a list of special PSZs,
    // where the first three characters of each entry
    // signify to us what was changed in CONFIG.SYS

    PSZ     pszLogStart = logger._pabLogString,
            pszLogThis = pszLogStart;
    int     iCount = 0;

    while (pszLogThis < pszLogStart + logger._cbLogString)
    {
        ULONG   cbLogThis = strlen(pszLogThis);
        BSString strNewModifiers;

        // now check what we have in the log; the
        // first three characters (followed by a
        // space) signify what happened:
        //  --  "DLL":      deleted line
        //  --  "DLP":      deleted part of line
        //  --  "NWL":      added an all new line
        //  --  "NWP":      added a new part to an existing line

        if (memicmp(pszLogThis, "DLL ", 4) == 0)
        {
            // line was deleted: re-insert that line (UNIQUE mode)
            // strNewModifiers = "UNIQUE";
        }
        else if (memicmp(pszLogThis, "DLP ", 4) == 0)
        {
            // part of line was deleted: re-insert that line (ADDRIGHT mode)
            strNewModifiers = "ADDRIGHT";
        }
        else if (memicmp(pszLogThis, "NWL ", 4) == 0)
        {
            // line was added: remove that whole line
            strNewModifiers = "REMOVELINE";
                    // #### no, no, no!!! this removes an entire line...
        }
        else if (memicmp(pszLogThis, "NWP ", 4) == 0)
            // part of line was added: remove that part
            strNewModifiers = "REMOVEPART";
        else
            // none of the above: error
            throw(BSConfigExcpt(CFGEXCPT_PARSELOG, strdup(pszLogThis)));

        // something found: compose attribute string for manipulator
        PSZ pszNewAttrs = (PSZ)malloc(strlen(pszLogThis)
                                      + strNewModifiers.size()
                                      + 30);
        sprintf(pszNewAttrs, "%s | %s",
                pszLogThis + 4,    // stuff after "DLL " or whatever
                strNewModifiers.c_str());

        // add the undo manipulator to the _front_ of the
        // list so that items are undone in reverse order
        // (because if a line was replaced originally,
        // we first have a "delete line" and then an
        // "add line" entry in the log)
        List.push_front(new BSCfgSysManip(pszNewAttrs));
        free(pszNewAttrs);

        pszLogThis += cbLogThis + 1;    // go beyond null byte
        iCount++;
    }

    return (iCount);
}

/*
 *@@ AddDeregisterClassToUndoList:
 *      helper func to create BSDeregisterClass
 *      instances from the logger instance that was
 *      previously used with BSRegisterClass::Register.
 *
 *      The new BSDeregisterClass objects are appended to
 *      the specified list.
 *
 *      See BSRegisterClass::Register for an example usage.
 *
 *      This returns the number of items created.
 */

int AddDeregisterClassToUndoList(list<BSConfigBase*> &List,
                                 BSRegisterDoneLogger &logger)
{
    // the logger in this case has a simple list of PSZs with all
    // the class names that were registered, each of which is terminated
    // with a null character

    PSZ     pszLogStart = logger._pabLogString,
            pszLogThis = pszLogStart;
    int     iCount = 0;

    while (pszLogThis < pszLogStart + logger._cbLogString)
    {
        ULONG cbLogThis = strlen(pszLogThis);
        List.push_back(new BSDeregisterClass(pszLogThis));
        pszLogThis += cbLogThis + 1;    // go beyond null byte
        iCount++;
    }

    return (iCount);
}

/*
 *@@ AddUnreplaceClassToUndoList:
 *      helper func to create BSUnreplaceClass
 *      instances from the logger instance that was
 *      previously used with BSRegisterClass::Register.
 *
 *      The new BSUnreplaceClass objects are appended to
 *      the specified list.
 *
 *      See BSReplaceClass::Replace for an example usage.
 *
 *      This returns the number of items created.
 */

int AddUnreplaceClassToUndoList(list<BSConfigBase*> &List,
                                    BSReplaceDoneLogger &logger)
{
    // the logger in this case has a simple list of PSZs with all
    // the class names that were registered, each of which is terminated
    // with a null character

    PSZ     pszLogStart = logger._pabLogString,
            pszLogThis = pszLogStart;
    int     iCount = 0;

    while (pszLogThis < pszLogStart + logger._cbLogString)
    {
        ULONG cbLogThis = strlen(pszLogThis);
        List.push_back(new BSUnreplaceClass(pszLogThis));
                                    // this calls the inherited BSReplaceClass
                                    // constructor, which will parse the thing
        pszLogThis += cbLogThis + 1;    // go beyond null byte
        iCount++;
    }

    return (iCount);
}

/*
 *@@ AddDeleteWPSObjectToUndoList:
 *      helper func to create BSDeleteWPSObject
 *      instances from the logger instance that was
 *      previously used with BSRegisterClass::Register.
 *
 *      The new BSDeleteWPSObject objects are appended to
 *      the specified list.
 *
 *      This returns the number of items created.
 */

int AddDeleteWPSObjectToUndoList(list<BSConfigBase*> &List,
                                     BSWPSObjectsDoneLogger &logger)
{
    // the logger in this case has a simple list of PSZs with all
    // the object IDs that were created, each of which is terminated
    // with a null character

    PSZ     pszLogStart = logger._pabLogString,
            pszLogThis = pszLogStart;
    int     iCount = 0;

    while (pszLogThis < pszLogStart + logger._cbLogString)
    {
        ULONG cbLogThis = strlen(pszLogThis);
        List.push_back(new BSDeleteWPSObject(pszLogThis));
        pszLogThis += cbLogThis + 1;    // go beyond null byte
        iCount++;
    }

    return (iCount);
}

/*
 *@@ AddClearProfileToUndoList:
 *      helper func to create BSClearProfile
 *      instances from the specified logger instance.
 *      Note that this class method will be used both
 *      for "CLEARPROFILE" tags specified in the script
 *      as well as "WRITEPROFILE" tags which were actually
 *      executed (written to a profile).
 *
 *      The new BSClearProfile objects are appended to
 *      the specified list.
 *
 *      This returns the number of items created.
 */

int AddClearProfileToUndoList(list<BSConfigBase*> &List,
                                  BSClearPrfAttrsLogger &logger)
{
    // the logger in this case has a simple list of PSZs with all
    // the profile keys that were created, each of which is terminated
    // with a null character

    PSZ     pszLogStart = logger._pabLogString,
            pszLogThis = pszLogStart;
    int     iCount = 0;

    while (pszLogThis < pszLogStart + logger._cbLogString)
    {
        ULONG cbLogThis = strlen(pszLogThis);
        List.push_back(new BSClearProfile(pszLogThis));
        pszLogThis += cbLogThis + 1;    // go beyond null byte
        iCount++;
    }

    return (iCount);
}

/*
 *@@ AddDeExecuteToUndoList:
 *      helper func to create BSDeExecute instances from
 *      the specified logger instance.
 *
 *      The new BSKillProcess objects are appended to
 *      the specified list.
 *
 *      This returns the number of items created.
 *
 *      This isn't really an "Undo" method like the other
 *      static AddToUndoList methods (for the other classes),
 *      but has been named like this for conformity. This
 *      is during package de-install to create the same
 *      BSDeExecute list which was initially created from
 *      the script so that processes can be killed during
 *      de-install as well.
 *
 *@@added V0.9.9 (2001-03-27) [umoeller]
 */

int AddDeExecuteToUndoList(list <BSConfigBase*> &List,
                           BSDeExecuteResolvedLogger &logger)
{
    PSZ     pszLogStart = logger._pabLogString,
            pszLogThis = pszLogStart;
    int     iCount = 0;

    while (pszLogThis < pszLogStart + logger._cbLogString)
    {
        ULONG cbLogThis = strlen(pszLogThis);
        List.push_back(new BSDeExecute(pszLogThis));
        pszLogThis += cbLogThis + 1;    // go beyond null byte
        iCount++;
    }
}

/*
 *@@ AddKillProcessToUndoList:
 *      helper func to create BSKillProcess
 *      instances from the specified logger instance.
 *
 *      The new BSKillProcess objects are appended to
 *      the specified list.
 *
 *      This returns the number of items created.
 *
 *      This isn't really an "Undo" method like the other
 *      static AddToUndoList methods (for the other classes),
 *      but has been named like this for conformity. This
 *      is during package de-install to create the same
 *      BSKillProcess list which was initially created from
 *      the script so that processes can be killed during
 *      de-install as well.
 *
 *@@added V0.9.4 (2000-07-01) [umoeller]
 */

int AddKillProcessToUndoList(list<BSConfigBase*> &List,
                             BSKillProcessAttrsLogger &logger)
{
    // the logger in this case has a simple list of PSZs with all
    // the executable names that were created, each of which is
    // terminated with a null character

    PSZ     pszLogStart = logger._pabLogString,
            pszLogThis = pszLogStart;
    int     iCount = 0;

    while (pszLogThis < pszLogStart + logger._cbLogString)
    {
        ULONG cbLogThis = strlen(pszLogThis);
        List.push_back(new BSKillProcess(pszLogThis));
        pszLogThis += cbLogThis + 1;    // go beyond null byte
        iCount++;
    }

    return (iCount);
}


/* ******************************************************************
 *
 *  FEDBPackage implementation
 *
 ********************************************************************/

/*
 *@@ FEPackage:
 *      second constructor for creating an FEPackage
 *      from the database, given a full package ID.
 *
 *      Use FEDatabase::FindPackageID to search the
 *      database if you don't know the exact package
 *      version which is installed.
 *
 *      Upon errors, a FEConstructorExcpt is thrown.
 *
 *      This initializes all fields in the returned
 *      FEDBPackage structure, including the WIPackHeader
 *      substructure.
 *
 *      Throws:
 *      -- FEDatabaseExcpt.
 *
 *@@changed V0.9.0 (99-10-26) [umoeller]: six-part ID support added
 *@@changed V0.9.0 (99-11-03) [umoeller]: now using FEPackageID
 *@@changed V0.9.1 (2000-01-08) [umoeller]: switched back to string ID for exception handling
 */

FEDBPackage::FEDBPackage(FEDatabase* pDatabase,
                         const char *pszPckID)
             : FEPackageBase(PCK_DBPCK)
{
    _ulStatus = 0;
    _pvGuiData = 0;

    // create pseudo-package declaration V0.9.9 (2001-02-28) [umoeller]
    _pDecl = new BSDBPckDecl();     // V0.9.12 (2001-05-17) [umoeller]

    FEPackageID PackageID(pszPckID, __FILE__, __LINE__,
                          FALSE);    // don't allow short format
                        // this can throw a FEConstructorExcpt
    if (!Load(pDatabase, PackageID))
        throw (FEDatabaseExcpt(pszPckID,
                               164,    // file list is null
                               TRUE));  // recover
}

/*
 *@@ FEDBPackage:
 *      second constructor to load from a FEPackageID
 *      instance.
 *
 *      Throws:
 *      -- FEDatabaseExcpt.
 *
 *@@added V0.9.1 (2000-01-08) [umoeller]
 */

FEDBPackage::FEDBPackage(FEDatabase* pDatabase,
                         FEPackageID &PckID)
             : FEPackageBase(PCK_DBPCK)
{
    _ulStatus = 0;
    _pvGuiData = 0;

    // create pseudo-package declaration V0.9.9 (2001-02-28) [umoeller]
    _pDecl = new BSDBPckDecl();     // V0.9.12 (2001-05-17) [umoeller]

    if (!Load(pDatabase, PckID))
        throw (FEDatabaseExcpt(NULL,
                               164,     // file list is null
                               TRUE));  // recover
}

/*
 *@@ Load:
 *      private helper method called from the constructors
 *      to load data from the database.
 *
 *      The data which is retrieved here must exactly match
 *      the data which has been stored in FEInstallJob::Store.
 *
 *      Returns FALSE upon errors.
 *
 *      Throws:
 *      -- FEDatabaseExcpt.
 *
 *@@added V0.9.1 (2000-01-08) [umoeller]
 *@@changed V0.9.1 (2000-01-07) [umoeller]: changed "requires" storage
 *@@changed V0.9.1 (2000-02-03) [umoeller]: added "execute" loggers support
 *@@changed V0.9.2 (2000-03-11) [umoeller]: added description and codepage
 *@@changed V0.9.4 (2000-07-01) [umoeller]: restore of "kill processes" was missing; fixed.
 */

BOOL FEDBPackage::Load(FEDatabase* pDatabase,       // in: database
                       FEPackageID &PckID)          // in: package ID to load
{
    BOOL brc = TRUE;

    BSString strIDSix;
    PckID.CreateStringSix(strIDSix);
    const char *pcszIDSix = strIDSix.c_str();
    /* CHAR    szPckIDSix[1000];
    sprintf(szPckIDSix, "%s\\%s\\%s\\%d\\%d\\%d",
            PckID._pszAuthor,
            PckID._pszApplication,
            PckID._pszPackageName,
            PckID._ulVersionMajor,
            PckID._ulVersionMinor,
            PckID._ulVersionRevision); */

    // check if file list exists in database
    _pszFilesList = prfhQueryProfileData(pDatabase->hiniDatabase,
                                         pcszIDSix,
                                         WPIKEY_FILESLIST,
                                         &_cbFilesList);
    if (_pszFilesList)
    {
        // create pseudo-package declaration so we can access
        // the data

        // all the static AddToUndoList methods can throw
        // exceptions, so protect this
        try
        {
            // store package ID
            _pDecl->_strID = pcszIDSix;

            // target path
            PSZ pszTargetPath = prfhQueryProfileData(pDatabase->hiniDatabase,
                                                     pcszIDSix,
                                                     WPIKEY_TARGETPATH,
                                                     NULL);
            if (!pszTargetPath)
                return (FALSE);
            else
            {
                _pDecl->_strTargetPath = pszTargetPath;
                free(pszTargetPath);
            }

            _logRequiresIDs.Load(pDatabase->hiniDatabase,
                                 pcszIDSix,
                                 WPIKEY_REQUIRES);

            _pDecl->_logCfgSysAttrs.Load(pDatabase->hiniDatabase,
                                         pcszIDSix,
                                         WPIKEY_CONFIGATTRS);
            _logCfgSysDone.Load(pDatabase->hiniDatabase,
                                pcszIDSix,
                                WPIKEY_CONFIGDONE);
            // create undo list
            AddCfgSysManipToUndoList(_listUndoConfig,  // list to append to
                                         _logCfgSysDone);  // logger restored above

            _pDecl->_logRegisterClassAttrs.Load(pDatabase->hiniDatabase,
                                                pcszIDSix,
                                                WPIKEY_REGISTERATTRS);
            _logRegisterClassesDone.Load(pDatabase->hiniDatabase,
                                         pcszIDSix,
                                         WPIKEY_REGISTERDONE);
            // create undo list
            AddDeregisterClassToUndoList(_listUndoConfig,  // list to append to
                                             _logRegisterClassesDone);  // logger restored above

            // class replacements
            _pDecl->_logReplaceClassAttrs.Load(pDatabase->hiniDatabase,
                                               pcszIDSix,
                                               WPIKEY_REPLACEATTRS);
            _logReplaceClassDone.Load(pDatabase->hiniDatabase,
                                      pcszIDSix,
                                      WPIKEY_REPLACEDONE);
            // create undo list
            AddUnreplaceClassToUndoList(_listUndoConfig,  // list to append to
                                            _logReplaceClassDone);  // logger restored above

            // WPS objects
            _pDecl->_logWPSObjectAttrs.Load(pDatabase->hiniDatabase,
                                            pcszIDSix,
                                            WPIKEY_WPSOBJECTATTRS);
            _logWPSObjectsDone.Load(pDatabase->hiniDatabase,
                                    pcszIDSix,
                                    WPIKEY_WPSOBJECTDONE);
            AddDeleteWPSObjectToUndoList(_listUndoConfig,
                                         _logWPSObjectsDone);

            // clear profiles
            _pDecl->_logClearPrfAttrs.Load(pDatabase->hiniDatabase,
                                           pcszIDSix,
                                           WPIKEY_CLEARPRFATTRS);
            // create BSClearProfile instances for the _attributes_!
            AddClearProfileToUndoList(_listUndoConfig,
                                      _pDecl->_logClearPrfAttrs);

            _pDecl->_logWritePrfAttrs.Load(pDatabase->hiniDatabase,
                                           pcszIDSix,
                                           WPIKEY_WRITEPRFATTRS);
            _logWritePrfsDone.Load(pDatabase->hiniDatabase,
                                   pcszIDSix,
                                   WPIKEY_WRITEPRFDONE);
            // create BSClearProfile instances for the done BSWriteProfile's
            AddClearProfileToUndoList(_listUndoConfig,
                                      _logWritePrfsDone);

            // executes
            _pDecl->_logExecuteAttrs.Load(pDatabase->hiniDatabase,
                                          pcszIDSix,
                                          WPIKEY_EXECUTEATTRS);
            _logExecutesDone.Load(pDatabase->hiniDatabase,
                                  pcszIDSix,
                                  WPIKEY_EXECUTEDONE);

            // de-executes V0.9.9 (2001-03-27) [umoeller]
            _pDecl->_logDeExecuteAttrs.Load(pDatabase->hiniDatabase,
                                          pcszIDSix,
                                          WPIKEY_DEEXECUTEATTRS);
            _logDeExecutesResolved.Load(pDatabase->hiniDatabase,
                                        pcszIDSix,
                                        WPIKEY_DEEXECUTERESOLVED);
            // create BSDeExecute instances for those attribs
            AddDeExecuteToUndoList(_listUndoConfig,
                                   _logDeExecutesResolved);

            // kill processes
            _pDecl->_logKillProcessAttrs.Load(pDatabase->hiniDatabase,
                                              pcszIDSix,
                                              WPIKEY_KILLPROCESSATTRS);
                                        // V0.9.4 (2000-07-01) [umoeller]
            AddKillProcessToUndoList(_listUndoConfig,
                                     _pDecl->_logKillProcessAttrs);

            // package header
            ULONG cbPackHeader = sizeof(WIPackHeader);
            PrfQueryProfileData(pDatabase->hiniDatabase,
                                pcszIDSix,
                                WPIKEY_WIPACKHEADER,
                                &_PackHeader,
                                &cbPackHeader);

            // install date/time
            ULONG cbDateTime = sizeof(DATETIME);
            PrfQueryProfileData(pDatabase->hiniDatabase,
                                pcszIDSix,
                                WPIKEY_INSTALLDATETIME,
                                &_InstallDateTime,
                                &cbDateTime);

            // description V0.9.2 (2000-03-11) [umoeller]
            PSZ pszInfo = prfhQueryProfileData(pDatabase->hiniDatabase,
                                               pcszIDSix,
                                               WPIKEY_DESCRIPTION,
                                               NULL);
            _pDecl->_strDescription = pszInfo;
            free(pszInfo);

            // codepage V0.9.2 (2000-03-11) [umoeller]
            _ulCodepage = PrfQueryProfileInt(pDatabase->hiniDatabase,
                                             pcszIDSix,
                                             WPIKEY_CODEPAGE,
                                             850);      // default


        } // end try
        catch(BSConfigExcpt &X)
        {
            throw(FEDatabaseExcpt(pcszIDSix,
                                  179, // "load" failed
                                  TRUE));
        }
    }
    else
        brc = FALSE;

    return (brc);
}

/*
 *@@ GetFileHeaders:
 *      quite reverse to FEArcPackagePck::AddFileHeader, this
 *      creates a new list of WIFileHeaders which have all
 *      the files in this package which have already
 *      been installed.
 *
 *      If this package is not yet installed, NULL is
 *      returned.
 *
 *      If this package was created from an archive,
 *      this returns
 *
 *      The number of list items is returned in *pulHeaderCount.
 *
 *      Returns NULL upon errors.
 *
 *      Each file header on the list is created using new,
 *      just as the returned list itself.
 *
 *@@changed V0.9.0 (99-10-26) [umoeller]: this used to be datAddFileHeader; made this a FEPackage method
 */

list<WIFileHeader*>* FEDBPackage::GetFileHeaders(PULONG pulHeaderCount)   // out: no. of list items
                                                                // in returned list
{
    list<WIFileHeader*>*    pList = 0;
    ULONG                   ulHeaderCount = 0;

    if (_cbFilesList)
    {
        pList = new list<WIFileHeader*>;

        PSZ p = _pszFilesList;

        while (p < (_pszFilesList + _cbFilesList) - 1)
        {
            WIFileHeader* pHeader = new WIFileHeader;

            // format of each entry:
            //    CHAR[x]   BYTE time_t  time_t  ULONG    BYTE
            //    filename   0   writedt creatdt filesize  0
            //                                   (uncompressed)

            // p now points to the filename
            strcpy(pHeader->name, p);
            p += strlen(p) + 1;     // go beyond null terminator

            // p now points to last write date
            pHeader->lastwrite = *((time_t*)p);
            p += sizeof(time_t);

            // p now points to last write date
            pHeader->creation = *((time_t*)p);
            p += sizeof(time_t);

            // p now points to the uncompressed file size
            pHeader->origsize = *((long*)p);

            p += sizeof(long);
            if (*p != 0)
            {
                delete pList;
                pList = 0;
                break;
            }

            p++;    // go beyond last null terminator

            pList->push_back(pHeader);
            ulHeaderCount++;
        }

        if (pulHeaderCount)
            *pulHeaderCount = ulHeaderCount;
    }

    return (pList);
}


