
/*
 *@@sourcefile fe_rexx.cpp:
 *      WarpIN REXX support (V0.9.2 (2000-03-10) [cbo]).
 *
 *      This REXX support module allows for the following features:
 *
 *      -- Store a REXX code string together with a name.
 *
 *      -- Execute a stored REXX code string by name, with argument string
 *         and return value.
 *
 *      -- Provide WarpIN functionality to the REXX comands and functions.
 *
 *      WarpIN REXX support is encapsulated in the FERexx class, of
 *      which one single instance is created by main() at WarpIN startup.
 *
 *@@header "frontend\fe_rexx.h"
 *@@added V0.9.2 (2000-03-10) [cbo]
 */

/*
 *      Copyright (C) 2000 Cornelis Bockemhl.
 *      Copyright (C) 2000 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define OS2EMX_PLAIN_CHAR
    // this is needed for "os2emx.h"; if this is defined,
    // emx will define PSZ as _signed_ char, otherwise
    // as unsigned char

// REXX support
#define INCL_REXXSAA
#include <os2.h>
#ifdef __IBMCPP__
    #include <rexxsaa.h>    // EMX has this in os2.h already
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// include all these headers because we need calbacks.h
#include "setup.h"

// include's from helpers
#include "helpers\stringh.h"
#include "helpers\xstring.h"

// back-end includes
#include "wiarchive\wiarchive.h"

// base includes
#include "base\bs_string.h"
#include "base\bs_errors.h"
#include "base\bs_rexx.h"

// front-end includes
#include "frontend\calbacks.h"
#include "frontend\fe_rexx.h"

#pragma hdrstop

/* ******************************************************************
 *
 *  Global variables
 *
 *******************************************************************/

/*
 *@@ G_aFERexxFunctions:
 *      global array of all WarpIN functions which
 *      are made available to REXX code. These are
 *      registered with the REXX interpreter in the
 *      FERexx::FERexx constructor and deregistered
 *      in the destructor.
 *
 *      To add a new WarpIN REXX function, just add
 *      the two fields to this array. The size of the
 *      array is automatically taken into account in
 *      the constructor/destructor, so there's nothing
 *      to adjust.
 *
 *@@added V0.9.2 (2000-03-10) [umoeller]
 */

WIREXXFUNCTION  G_aRexxFunctions[]
        = {
            "WirexxPutEnv", ::WirexxPutEnv,
            "WirexxGetEnv", ::WirexxGetEnv,
            "WirexxShowMessage", ::WirexxShowMessage
          };

// size of that array
ULONG           G_cRexxFunctions = sizeof(G_aRexxFunctions) / sizeof(G_aRexxFunctions[0]);

BSStringMap G_EnvStrings;

/* ******************************************************************
 *
 *  Helper funcs
 *
 ********************************************************************/

/*
 *@@ SetRexxReturn:
 *      helper function for REXX functions to correctly return a REXX
 *      return string (for implementation details see docs extract
 *      at the end of this file).
 *
 *@@added V0.9.2 (2000-03-10) [cbo]
 */

void SetRexxReturn(char *retstr,
                   PRXSTRING rxstr)
{
    ULONG rc;

    if (255 < strlen(retstr))
    {
        // enlarge the REXX string storage
        rc = DosAllocMem((PPVOID) rxstr->strptr,
                         strlen(retstr),
                         PAG_READ | PAG_WRITE | PAG_COMMIT);

        // if an error occurs we return a string indicating that
        if (0 != rc)
        {
            strcpy(rxstr->strptr, "ERROR allocating memory for REXX return");
            rxstr->strlength = strlen(rxstr->strptr);
            return;
        }
    }

    // copy the string to it's place
    strcpy(rxstr->strptr, retstr);
    rxstr->strlength = strlen(retstr);
}

/* ******************************************************************
 *
 *  Exported REXX functions
 *
 *******************************************************************/

//////////////////////////////////////////////////////////////////
// Implement functions that can be called by REXX scripts
//
// Each function has to be defined twice:
// - once as a static member function of FERexx. This allows e.g.
//   for storing static data there (like the EnvStrings map).
// - then as a C style stand-alone function as a wrapper: This
//   is needed because a member function cannot be exported with
//   APIENTRY linkage.
// The C function is registered in the FERexx constructor using
// the G_aFERexxFunctions array.

/*
 * V0.9.9 (2001-02-28) [umoeller]:
 *      the above is no longer true. For each REXX function,
 *      we now have a single C APIENTRY function, an array
 *      of which is passed to the FERexx constructor to
 *      register the functions with the REXX interpreter
 *      when the FERexx instance gets created on WarpIN
 *      startup.
 *
 *      See src\base\bs_rexx.cpp for this separation now.
 */

/*
 *@@ WirexxPutEnv:
 *      sets an environment value in the WarpIN environment.
 *
 *      REXX syntax: rc=WirexxPutEnv("name","value");
 *
 *      If the function is called without a "value" parameter, the variable
 *      "name" is removed from the environment.
 *
 *      The implementation is in FERexx::WirexxPutEnv.
 *
 *@@added V0.9.2 (2000-03-10) [cbo]
 */

ULONG WirexxPutEnv(PCSZ name,
                   ULONG argc,
                   RXSTRING argv[],
                   PCSZ queuename,
                   PRXSTRING retstr)
{
    int rc;
    BSString strKey, strExpression;

    BSStringMap::const_iterator cit;

    // no name -> nothing to do
    // Remark: ALL parameters of a REXX script have to be checked for
    // validity: Unlike for C/C++ programs you cannot assume that if
    // one argument exists also all the arguments "before" this do
    // exist!
    if ((1 > argc) || !RXVALIDSTRING(argv[0]))
        return 1;

    // create environment set string
    strupr(RXSTRPTR(argv[0]));
    strKey = RXSTRPTR(argv[0]);
    strExpression = RXSTRPTR(argv[0]);
    if ((2 <= argc) && RXVALIDSTRING(argv[1]))
    {
        strExpression += "=";
        strExpression += RXSTRPTR(argv[1]);
    }
    // if argv[1] doesn't exist the variable is removed

    // the environment string is put into this static map structure
    // in order to keep the memory valid after leaving this function
    G_EnvStrings[strKey] = strExpression;

    // Set the _stored_ string pointer which is persistent!
    // Remark: The putenv() function seems to store only a _pointer_ to the
    // passed string, i.e. the string storage must not be local, but
    // persistent!
    // Be careful: This retrieval with a const_iterator seems to really get
    // the pointer to the actual string, but this is not necessarily true
    // in any STL implementation, I suppose...(!?)
    cit = G_EnvStrings.find(strKey);
    rc = putenv((*cit).second.c_str());

    // Now comes the return value for the REXX caller
    // Remark: You should always return _something_, and if it is an
    // empty string (explicitly, i.e. with SetRexxReturn("",retstr),
    // otherwise your function will just return the garbage that was
    // accidentally in retstr on calling!
    if (0 != rc)
        SetRexxReturn("1", retstr);
    else
        SetRexxReturn("0", retstr);

    // Here you should always return 0, otherwise the REXX interpreter
    // just stops because it means that some kind of "internal error"
    // occurred!
    // (Return error codes to the REXX caller using retstr: See above)
    return 0;
}

/*
 *@@ WirexxGetEnv:
 *      gets an environment value in the WarpIN environment.
 *
 *      remark: this is necessary for retrieving values set by WarpIN,
 *      because these are not in the inherited environment of the REXX
 *      function calls!
 *
 *      Example: WI_ARCHIVE_PATH contains the path of the WarpIN archives
 *
 *      REXX syntax: value=WirexxPutEnv("name");
 *
 *      The implementation is in FERexx::WirexxGetEnv.
 *
 *@@added V0.9.3 (2000-05-26) [cbo]
 */

ULONG WirexxGetEnv(PCSZ name,
                   ULONG argc,
                   RXSTRING argv[],
                   PCSZ queuename,
                   PRXSTRING retstr)
{
    BSString strKey, strExpression;

    // no name -> nothing to do
    // Remark: ALL parameters of a REXX script have to be checked for
    // validity: Unlike for C/C++ programs you cannot assume that if
    // one argument exists also all the arguments "before" this do
    // exist!
    if ((1 > argc) || !RXVALIDSTRING(argv[0]))
        return 1;

    // get environment value string
    PSZ pszVal = getenv(RXSTRPTR(argv[0]));

    // return a value, if not NULL
    if (NULL != pszVal)
        SetRexxReturn(pszVal, retstr);
    else
        SetRexxReturn("", retstr);

    // Here you should always return 0, otherwise the REXX interpreter
    // just stops because it means that some kind of "internal error"
    // occurred!
    // (Return error codes to the REXX caller using retstr: See above)
    return 0;
}

/*
 *@@ WirexxShowMessage:
 *      shows a message using the WarpIN callback function
 *      guiShowMessage.
 *
 *      REXX syntax: rc=WirexxShowMessage("title","message",flags);
 *
 *      The flags are those that are passed to guiShowMessage directly,
 *      i.e. you may "or" the following values together. The default
 *      value is 0.
 *
 *      -- 0x0000 MB_OK               Message box contains an OK push button.
 *      -- 0x0001 MB_OKCANCEL         Message box contains both OK and CANCEL push
 *                                    buttons.
 *      -- 0x0006 MB_CANCEL           Message box contains a CANCEL push button.
 *      -- 0x0007 MB_ENTER            Message box contains an ENTER push button.
 *      -- 0x0008 MB_ENTERCANCEL      Message box contains both ENTER and CANCEL push
 *                                    buttons.
 *      -- 0x0002 MB_RETRYCANCEL      Message box contains both RETRY and CANCEL
 *                                    push buttons.
 *      -- 0x0003 MB_ABORTRETRYIGNORE Message box contains ABORT, RETRY, and
 *                                    IGNORE push buttons.
 *      -- 0x0004 MB_YESNO            Message box contains both YES and NO push
 *                                    buttons.
 *      -- 0x0005 MB_YESNOCANCEL      Message box contains YES, NO, and CANCEL push
 *                                    buttons.
 *
 *      -- 0x0030 MB_ERROR            Message box contains a small red circle with a red
 *                                    line across it.
 *      -- 0x0020 MB_ICONASTERISK     Message box contains an information (i) icon.
 *      -- 0x0040 MB_ICONEXCLAMATION  Message box contains an exclamation point (!)
 *                                    icon.
 *      -- 0x0030 MB_ICONHAND         Message box contains a small red circle with a red
 *                                    line across it.
 *      -- 0x0000 MB_ICONQUESTION     Message box contains a question mark (?) icon.
 *      -- 0x0010 MB_INFORMATION      Message box contains an information (i) icon.
 *      -- 0x0020 MB_NOICON           Message box is not to contain an icon.
 *      -- 0x0010 MB_QUERY            Message box contains a question mark (?) icon.
 *      -- 0x0020 MB_WARNING          Message box contains an exclamation point (!) icon.
 *
 *      -- 0x0000 MB_DEFBUTTON1       The first button is the default selection. This is the
 *                                    default case, if none of MB_DEFBUTTON1,
 *                                    MB_DEFBUTTON2, and MB_DEFBUTTON3 is
 *                                    specified.
 *      -- 0x0100 MB_DEFBUTTON2       The second button is the default selection.
 *      -- 0x0200 MB_DEFBUTTON3       The third button is the default selection.
 *
 *      -- 0x0000 MB_APPLMODAL        Message box is application modal. This is the
 *                                    default case. Its owner is disabled; therefore, do
 *                                    not specify the owner as the parent if this option
 *                                    is used.
 *      -- 0x1000 MB_SYSTEMMODAL      Message box is system modal.
 *      -- 0x4000 MB_MOVEABLE         Message box is moveable.
 *
 *      With rc you get feedback about what with which button the user terminated
 *      the message display:
 *      -- 9      MBID_ENTER  ENTER push button was selected
 *      -- 1      MBID_OK     OK push button was selected
 *      -- 2      MBID_CANCEL CANCEL push button was selected
 *      -- 3      MBID_ABORT  ABORT push button was selected
 *      -- 4      MBID_RETRY  RETRY push button was selected
 *      -- 5      MBID_IGNORE IGNORE push button was selected
 *      -- 6      MBID_YES    YES push button was selected
 *      -- 7      MBID_NO     NO push button was selected
 *      -- 0xffff MBID_ERROR  Function not successful; an error occurred.
 *
 *      The implementation is in FERexx::WirexxShowMessage.
 *
 *@@added V0.9.2 (2000-03-10) [cbo]
 */

ULONG WirexxShowMessage(PCSZ name,
                        ULONG argc,
                        RXSTRING argv[],
                        PCSZ queuename,
                        PRXSTRING retstr)
{
    BSString title, message;
    char str[32];
    ULONG flags, rc;

    // title
    if ((1 > argc) || !RXVALIDSTRING(argv[0]))
        title = BSString("");
    else
        title = BSString(RXSTRPTR(argv[0]));

    // message
    if ((2 > argc) || !RXVALIDSTRING(argv[1]))
        message = BSString("");
    else
        message = BSString(RXSTRPTR(argv[1]));

    // flags
    if ((3 > argc) || !RXVALIDSTRING(argv[2]))
        flags = 0;
    else
        flags = atoi(RXSTRPTR(argv[2]));

    // display the message
    rc = guiShowMessage(title, message, flags);

    // put the result code into retstr
    _itoa(rc, str, sizeof(str) - 1);
    SetRexxReturn(str, retstr);

    // Here you should always return 0, otherwise the REXX interpreter
    // just stops because it means that some kind of "internal error"
    // occurred!
    // (Return error codes to the REXX caller using retstr: See above)
    return 0;
}

//////////////////////////////////////////////////////////////////
//   From the docs about writing REXX functions:
//
//   The following is a sample external function definition:
//
//   ULONG SysLoadFuncs(
//      PSZ       Name,                   // name of the function
//      LONG      Argc,                   // number of arguments
//      RXSTRING  Argv[],                 // list of argument strings
//      PSZ       Queuename,              // current queue name
//      PRXSTRING Retstr)                 // returned result string
//
//   Where
//
//    Name        Address of ASCIIZ external function name.
//
//    Argc        The size of the argument list. Argv will contain Argc RXSTRINGs.
//
//    Argv        An array of null-terminated RXSTRINGs for the function arguments.
//
//    Queue       The name of the currently defined REXX external data queue.
//
//    Retstr      Address of an RXSTRING for the returned value. Retstr is a
//                character string function or subroutine return value. When a
//                REXX program calls an external function with the REXX CALL
//                instruction, Retstr is assigned to the REXX special variable
//                RESULT. When the REXX program calls an external function as a
//                function, Retstr is used directly within the REXX expression.
//
//                The REXX interpreter provides a default 256-byte RXSTRING in
//                Retstr. A longer RXSTRING can allocated with "DosAllocMem" if
//                the returned string is longer name 256 bytes. The REXX interpreter
//                releases Retstr with "DosFreeMem" when the external function
//                completes.
//
//    Returns     An integer return code from the function. When the external
//                function returns 0, the function completed successfully. Retstr
//                contains the function return value. When external function return
//                code is not 0, the REXX interpreter raises REXX error 40
//                ("Invalid call to routine").  The Retstr value is ignored.
//
//                If the external function does not have a return value, the
//                function must set Retstr to an an empty RXSTRING (NULL strptr).
//                When a function does not return a value, the interpreter raises
//                error 44, "Function did not return data". When an external
//                function invoked with the REXX CALL instruction does not return
//                a value, the REXX interpreter drops (unassigns) the special
//                variable RESULT.

