
/*
 *@@sourcefile warpin.cpp:
 *      this is the main source file for WARPIN.EXE.
 *
 *      The WarpIn installer source code is organized into three
 *      major layers:
 *
 *      1)  The WIArchive class ("back end"). This one is standard
 *          C++ code which should be completely independent of OS/2
 *          and handles compression/decompression.
 *          This is in the ../wiarchive and ../libbz2 directories.
 *
 *      2)  The "naked front end", which is OS/2-, but not PM-specific.
 *          This handles everything the scripts can do plus the database.
 *
 *          This consists of a fairly large number of C++ classes to
 *          handle scripts, archives, packages, and install jobs.
 *          For all this, see fe_archive.cpp for an introduction.
 *
 *          The idea behind these classes is to separate the actual program
 *          logic from what is displayed to the user so that WarpIN can be
 *          ported to a new user interface (e.g. a text-mode version) without
 *          having to rewrite everything.
 *
 *          WarpIN operates in different "modes" depending on whether
 *          an archive was passed to WARPIN.EXE on the command line:
 *
 *          --  "install" mode, if an archive was passed on the
 *              command line; this will open the archive, load the
 *              install script, parse it, and display the pages as
 *              specified in the install script.
 *
 *          --  "database" mode, if no archive was given. This opens
 *              the entire database in a different window, from where
 *              packages can be de-installed as well.
 *
 *          See main() for more details and which functions in this
 *          file are called in which order.
 *
 *      3)  The GUI of the front end, which does all the PM-specific
 *          stuff for all of these modes. This is in the gui_*.cpp
 *          source files.
 *
 *          The GUI is not implemented thru classes, but only thru
 *          a number of callbacks which are required by the "naked
 *          front end". All these GUI callbacks are listed in
 *          include\frontend\calbacks.h.
 *
 *          Note that WarpIn is multithreaded. When installation
 *          actually starts, i.e. on the "Install" page, a second
 *          thread is started, which calls the WIArchive methods
 *          upon the archive to unpack the files. The GUI part must
 *          have a number of callbacks for while installation is in
 *          progress, for showing the database, etc. If you wish to
 *          rewrite the gui_*.cpp part, you must be aware of which
 *          function gets called by which thread, or everything will
 *          crash pretty soon.
 *
 *      <B>Exception handling</B>
 *
 *      The entire main() routine is enclosed in a try{} block,
 *      so that any C++ exception thrown anywhere will finally end
 *      up in main()'s catch{} blocks. This applies especially
 *      to BSCancelExcpt, which can be thrown anywhere to terminate
 *      installation. All exceptions are defined in include\base\bs_errors.h.
 *
 *@@header "frontend\warpin.h"
 *@@header "base\bs_errors.h"
 *@@header "frontend\calbacks.h"
 */

/*      This file Copyright (C) 1998-2000 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define OS2EMX_PLAIN_CHAR
    // this is needed for "os2emx.h"; if this is defined,
    // emx will define PSZ as _signed_ char, otherwise
    // as unsigned char

#define INCL_DOS
#define INCL_DOSERRORS
#define INCL_WINSHELLDATA
#define INCL_WINWORKPLACE
#define INCL_REXXSAA
#include <os2.h>
#ifdef __IBMCPP__
    #include <rexxsaa.h>    // EMX has this in os2.h already
#endif

#include <stdlib.h>
#ifdef __IBMCPP__
    #include <direct.h>
#endif
#include <string.h>
#include <setjmp.h>             // needed for except.h
#include <assert.h>             // needed for except.h

#include <ctype.h>

#ifdef __IBMCPP__
    #include <terminat.h>
#endif

#include "setup.h"              // was missing (99-12-23) [umoeller]
#include "bldlevel.h"           // added (99-10-24) [umoeller]

// include's from helpers
#include "helpers\dosh.h"
#include "helpers\stringh.h"
#include "helpers\threads.h"
#include "helpers\winh.h"
#include "helpers\xstring.h"

// back-end includes
#include "wiarchive\wiarchive.h"

// base includes (99-11-07) [umoeller]
#include "base\bs_string.h"
#include "base\bs_errors.h"
#include "base\bs_logger.h"
#include "base\bs_config.h"
#include "base\bs_script.h"
#include "base\bs_rexx.h"

// front-end includes
#include "frontend\fe_base.h"
#include "frontend\fe_archive.h"
#include "frontend\fe_database.h"
#include "frontend\fe_job.h"
#include "frontend\fe_package.h"
#include "frontend\fe_package_arc.h"
#include "frontend\fe_package_db.h"

#include "frontend\warpin.h"
#include "frontend\calbacks.h"
#include "frontend\fe_rexx.h"

#include "helpers\except.h"     // must be at the bottom because wiarchive includes stdio

#pragma hdrstop

/* ******************************************************************
 *
 *  More declarations
 *
 ********************************************************************/

void EnterArchiveMode(PSZ pszArchiveFile);

/* ******************************************************************
 *
 *  Global variables
 *
 ********************************************************************/

// this is the large structure of variables which
// are shared with gui.cpp, declared in warpin.h.
WPIGLOBALS      G_WpiGlobals;

// WarpIN database (created in main())
FEDatabase*     G_pDatabase;

/* ******************************************************************
 *
 *  PART 1
 *  ------
 *  The following code is all running on thread 1 (the main PM
 *  thread). This includes main(), which will open the archive,
 *  analyze the install script, build a list of dialog pages etc.
 *
 ********************************************************************/

/*
 *@@ ExceptOpenLogFile:
 *      hook for exceptions, registered with excRegisterHooks
 *      (src\helpers\except.c).
 *
 *@@changed V0.9.2 (2000-03-10) [umoeller]: switched date format to ISO
 */

FILE* APIENTRY ExceptOpenLogFile(VOID)
{
    CHAR        szFileName[CCHMAXPATH];
    FILE        *file;
    DATETIME    DT;

    // ULONG ulCopied;

    sprintf(szFileName, "%c:\\%s", doshQueryBootDrive(), "wpitrap.log");
    file = fopen(szFileName, "a");

    if (file)
    {
        DosGetDateTime(&DT);
        fprintf(file, "\nWarpIN trap message -- Date: %04d-%02d-%02d, Time: %02d:%02d:%02d\n",
            DT.year, DT.month, DT.day,
            DT.hours, DT.minutes, DT.seconds);
        fprintf(file, "-------------------------------------------------------\n"
                      "\nAn internal error occurred in WarpIN.\n"
                      "Please contact the author so that this error may be removed\n"
                      "in future WarpIN versions. The author's e-mail address may\n"
                      "be obtained from the WarpIN User Guide. Please supply this\n"
                      "file (?:\\WPITRAP.LOG) with your e-mail and describe as\n"
                      "exactly as possible the conditions under which the error\n"
                      "occured.\n"
                      "\nRunning WarpIN version: " BLDLEVEL_VERSION " built " __DATE__ "\n");

    }
    return (file);
}

/*
 *@@ WarpINTerminate:
 *      function used with C++ set_terminate().
 *
 *@@added V0.9.1 (2000-01-08) [umoeller]
 */

void WarpINTerminate()
{
    guiShowMessage("WarpIN: Internal Error",
                   "Unexpected C++ exception caught.",
                   MB_ICONEXCLAMATION | MB_CANCEL);
}

/*
 *@@ GetGeneralEnvironment:
 *      this gets called from EnterArchiveMode() or main()
 *      before the archive is opened (if so) to set up
 *      the general environment. This gets called in "Archive"
 *      and "Database" modes.
 *
 *      Still, this gets called _after_ guiInitialize so
 *      environment settings override GUI settings.
 *
 *      Throws:
 *      -- FEEnvironmentExcpt if invalid values are found.
 *
 *@@added V0.9.4 (2000-07-28) [umoeller]
 *@@changed V0.9.9 (2001-03-30) [umoeller]: added WARPIN_LOGFILEPATH
 */

VOID GetGeneralEnvironment()
{
    const char  *pcszCommand = 0,
                *pcszValue = 0;

    // SET WARPIN_DEFAULTAPPSPATH=[path]
    pcszCommand = "WARPIN_DEFAULTAPPSPATH";
    if (    (pcszValue = getenv(pcszCommand)) == NULL)
    {
        // not set:
        sprintf(G_WpiGlobals.szDefaultAppsPath, "WARPIN_DEFAULTAPPSPATH=%c:\\APPS", doshQueryBootDrive());
        putenv(G_WpiGlobals.szDefaultAppsPath);
    }

    // SET WARPIN_DEFAULTTOOLSPATH=[path]
    pcszCommand = "WARPIN_DEFAULTTOOLSPATH";
    if (    (pcszValue = getenv(pcszCommand)) == NULL)
    {
        // not set:
        sprintf(G_WpiGlobals.szDefaultToolsPath, "WARPIN_DEFAULTTOOLSPATH=%c:\\TOOLS", doshQueryBootDrive());
        putenv(G_WpiGlobals.szDefaultToolsPath);
    }

    // SET WARPIN_IFSAMEDATE={PROMPT|SKIP|OVERWRITE}
    pcszCommand = "WARPIN_IFSAMEDATE";
    if ((pcszValue = getenv(pcszCommand)))
        if (stricmp(pcszValue, "PROMPT") == 0)
            G_WpiGlobals.ulIfSameDate = FE_PROMPT;
        else if (stricmp(pcszValue, "SKIP") == 0)
            G_WpiGlobals.ulIfSameDate = FE_SKIP;
        else if (stricmp(pcszValue, "OVERWRITE") == 0)
            G_WpiGlobals.ulIfSameDate = FE_OVERWRITE;
        else
            // invalid value:
            throw(FEEnvironmentExcept(pcszCommand, pcszValue));

    // SET WARPIN_IFEXISTINGOLDER={PROMPT|SKIP|OVERWRITE}
    pcszCommand = "WARPIN_IFEXISTINGOLDER";
    if ((pcszValue = getenv(pcszCommand)))
        if (stricmp(pcszValue, "PROMPT") == 0)
            G_WpiGlobals.ulIfExistingOlder = FE_PROMPT;
        else if (stricmp(pcszValue, "SKIP") == 0)
            G_WpiGlobals.ulIfExistingOlder = FE_SKIP;
        else if (stricmp(pcszValue, "OVERWRITE") == 0)
            G_WpiGlobals.ulIfExistingOlder = FE_OVERWRITE;
        else
            // invalid value:
            throw(FEEnvironmentExcept(pcszCommand, pcszValue));

    // SET WARPIN_IFEXISTINGNEWER={PROMPT|SKIP|OVERWRITE}
    pcszCommand = "WARPIN_IFEXISTINGNEWER";
    if ((pcszValue = getenv(pcszCommand)))
        if (stricmp(pcszValue, "PROMPT") == 0)
            G_WpiGlobals.ulIfExistingNewer = FE_PROMPT;
        else if (stricmp(pcszValue, "SKIP") == 0)
            G_WpiGlobals.ulIfExistingNewer = FE_SKIP;
        else if (stricmp(pcszValue, "OVERWRITE") == 0)
            G_WpiGlobals.ulIfExistingNewer = FE_OVERWRITE;
        else
            // invalid value:
            throw(FEEnvironmentExcept(pcszCommand, pcszValue));

    // SET WARPIN_LOGFILEPATH=<logfilepath> V0.9.9 (2001-03-30) [umoeller]
    pcszCommand = "WARPIN_LOGFILEPATH";
    if ((pcszValue = getenv(pcszCommand)))
        if (0 == doshIsValidFileName(pcszValue,
                                     TRUE))         // must be fully qualified
            strcpy(G_WpiGlobals.szLogFilename, pcszValue);
        else
            // invalid value:
            throw(FEEnvironmentExcept(pcszCommand, pcszValue));
}

/*
 *@@ GetArchiveEnvironment:
 *      this gets called from EnterArchiveMode() after everything
 *      has been set up and right before we'd normally
 *      go into "Archive" mode. This gets called in "Archive"
 *      mode only.
 *
 *      This can override various archive settings
 *      according to the currently defined environment
 *      variables.
 *
 *      Throws:
 *      -- FEEnvironmentExcpt if invalid values are found.
 *
 *@@changed V0.9.6 (2000-11-23) [umoeller]: added DISPLAYPAGES, fixed SELECTPACKAGE
 */

VOID GetArchiveEnvironment()
{
    const char  *pcszCommand = 0,
                *pcszValue = 0;

    pcszCommand = "WARPIN_DISPLAYPAGES";
    pcszValue = getenv(pcszCommand);
    if (pcszValue)
    {
        if (stricmp(pcszValue, "YES") == 0)
            G_WpiGlobals.fDisplayPages = TRUE;
        else if (stricmp(pcszValue, "NO") == 0)
            G_WpiGlobals.fDisplayPages = FALSE;
            // the constructor has defaulted this to YES anyway
        else
            // invalid value:
            throw(FEEnvironmentExcept(pcszCommand, pcszValue));
    }

    // modify packages
    list<FEJobBase*>::iterator JobStart = G_WpiGlobals.AllJobsList.begin(),
                               JobEnd = G_WpiGlobals.AllJobsList.end();
    for (; JobStart != JobEnd; JobStart++)
    {
        FEJobBase *pTemp = *JobStart;
        FEInstallJob *pJobThis = pTemp->IsInstallJob();
        if (pJobThis)
        {
            const FEArcPackageBase *pPackageThis = pJobThis->_pArcPackage;

            CHAR szVarname[100];
            // SET WARPIN_SELECTPACKAGEx={YES|NO}
            sprintf(szVarname, "WARPIN_SELECTPACKAGE%d", pPackageThis->_PackHeader.number);
            pcszValue = getenv(szVarname);
            if (pcszValue)
                if (stricmp(pcszValue, "IGNORE") == 0)
                    pJobThis->Select(JOB_INSTALL_IGNORE);
                else if (stricmp(pcszValue, "INSTALL") == 0)
                    pJobThis->Select(JOB_INSTALL);
                else if (stricmp(pcszValue, "DEINSTALL") == 0)
                    pJobThis->Select(JOB_DEINSTALL);
                else
                    // invalid value:
                    throw(FEEnvironmentExcept(szVarname, pcszValue));

            // SET WARPIN_TARGETPATHx=path
            sprintf(szVarname, "WARPIN_TARGETPATH%d", pPackageThis->_PackHeader.number);
            pcszValue = getenv(szVarname);
            if (pcszValue)
                if (doshIsValidFileName(pcszValue, TRUE) == NO_ERROR)
                    pJobThis->SetTargetPath(pcszValue,
                                            G_WpiGlobals.AllJobsList);
                else
                    // invalid value:
                    throw(FEEnvironmentExcept(szVarname, pcszValue));
        }
    }

    // SET WARPIN_UPDATECONFIGSYS={YES|NO}
    pcszCommand = "WARPIN_UPDATECONFIGSYS";
    if ((pcszValue = getenv(pcszCommand)))
        if (stricmp(pcszValue, "YES") == 0)
            G_WpiGlobals.ulAllowConfig |= PCK_CONFIGSYS;
        else if (stricmp(pcszValue, "NO") == 0)
            G_WpiGlobals.ulAllowConfig &= ~PCK_CONFIGSYS;
        else
            // invalid value:
            throw(FEEnvironmentExcept(pcszCommand, pcszValue));

    // SET WARPIN_INSTALLWPSCLASSES={YES|NO}
    pcszCommand = "WARPIN_INSTALLWPSCLASSES";
    if ((pcszValue = getenv(pcszCommand)))
        if (stricmp(pcszValue, "YES") == 0)
            G_WpiGlobals.ulAllowConfig |= PCK_WPSCLASSES;
        else if (stricmp(pcszValue, "NO") == 0)
            G_WpiGlobals.ulAllowConfig &= ~PCK_WPSCLASSES;
        else
            // invalid value:
            throw(FEEnvironmentExcept(pcszCommand, pcszValue));

    // SET WARPIN_CREATEWPSOBJECTS={YES|NO}
    pcszCommand = "WARPIN_CREATEWPSOBJECTS";
    if ((pcszValue = getenv(pcszCommand)))
        if (stricmp(pcszValue, "YES") == 0)
            G_WpiGlobals.ulAllowConfig |= PCK_WPSOBJECTS;
        else if (stricmp(pcszValue, "NO") == 0)
            G_WpiGlobals.ulAllowConfig &= ~PCK_WPSOBJECTS;
        else
            // invalid value:
            throw(FEEnvironmentExcept(pcszCommand, pcszValue));

}

/* ******************************************************************
 *
 *  Program entry point
 *
 ********************************************************************/

/*
 *@@ main:
 *      Program entry point.
 *      This first initializes the GUI front-end branches
 *      into several subroutines, depending on the command
 *      line parameters:
 *
 *      --  EnterArchiveMode(), if a WarpIN archive was
 *          passed on the command line; that function
 *          checks whether archives have already been
 *          installed and will then enter "full install"
 *          or "add/remove" mode;
 *
 *      --  guiBeginDatabaseMode(), if no archive was
 *          passed and we should switch into "database"
 *          mode.
 *
 *@@changed V0.9.2 (2000-03-11) [umoeller]: added REXX support
 *@@changed V0.9.4 (2000-07-01) [umoeller]: added auto-restart WPS option
 *@@changed V0.9.6 (2000-10-27) [umoeller]: disabled white-error popups
 *@@changed V0.9.9 (2001-03-30) [umoeller]: disabled stupid finish message in CID mode
 */

int main(int argc, char *argv[])
{
    // register exception hooks for helpers\except.c
    excRegisterHooks(ExceptOpenLogFile, NULL, NULL, TRUE);

    set_terminate(WarpINTerminate);

    DosError(FERR_DISABLEHARDERR | FERR_ENABLEEXCEPTION);
        // V0.9.6 (2000-10-27) [umoeller]

    // register our exit list
    /* DosExitList(EXLST_ADD
                    | 0x00002A00,         // Invocation order is 42 (0x2A)
                    (PFNEXITLIST)WarpInExitList); */

    // install the LOUD exception handler (except.h);
    // all the string parsing might crash...
    TRY_LOUD(excpt1)
    {
        // open generic try block for exceptions
        // which are not caught in the subroutines;
        // this especially applies to BSCancelExcpt
        try
        {
            /*
             * initialize
             *
             */

            // store the WARPIN.EXE executable in the globals
            // for later use
            strcpy(G_WpiGlobals.szWarpINFile, argv[0]);

            // extract the path from that (99-10-24) [umoeller]
            strhncpy0(G_WpiGlobals.szWarpINPath,
                      G_WpiGlobals.szWarpINFile,
                      strrchr(G_WpiGlobals.szWarpINFile, '\\') - G_WpiGlobals.szWarpINFile);

            // compose the name of the message (tmf) file (99-10-24) [umoeller];
            // this is referenced by bs_script.cpp
            sprintf(G_szMessageFile, "%s\\warpin.tmf", G_WpiGlobals.szWarpINPath);

            // initialize the PM environment (this is
            // needed for the profile functions)
            if (!(G_WpiGlobals.habThread1 = WinInitialize(0)))
                return FALSE;

            // open the WARPIN.INI file
            CHAR szINIFile[2*CCHMAXPATH];
            sprintf(szINIFile, "%s\\warpin.ini", G_WpiGlobals.szWarpINPath);
            G_WpiGlobals.hiniWarpIN = PrfOpenProfile(G_WpiGlobals.habThread1, szINIFile);

            // set log file name V0.9.9 (2001-03-30) [umoeller]
            sprintf(G_WpiGlobals.szLogFilename,
                    "%s\\install.log",
                    G_WpiGlobals.szWarpINPath);

            // initialize the interface; this checks for
            // whether WARPIN.EXE is already installed (99-10-22) [umoeller]
            // and may override the log file name V0.9.9 (2001-03-30) [umoeller]
            if (!guiInitialize())
                return (1);

            HMTX    hmtx;
            // check the "one-instance" semaphore, which
            // we create just for testing that WarpIN is
            // started only once
            if (DosCreateMutexSem(IDMUTEX_ONEINSTANCE,
                                  &hmtx,
                                  DC_SEM_SHARED,
                                  TRUE)
                 != NO_ERROR)
            {
                // error creating semaphore: this means
                // that another instance is running
                wpiMessageBox(102,
                              155,     // "another instance"
                              MB_ICONEXCLAMATION | MB_CANCEL);
                throw BSCancelExcpt();
            }

            // initialize the database
            G_pDatabase = new FEDatabase(G_WpiGlobals.habThread1,
                                         G_WpiGlobals.szWarpINPath);

            // initialize REXX support (global variable is in src\base\bs_rexx.cpp)
            G_pFERexx = new FERexx(G_aRexxFunctions,
                                   G_cRexxFunctions);

            // now check arguments
            if (argc > 1)
            {
                /*
                 *  "Archive" mode:
                 *      an archive was given on the command line.
                 */

                EnterArchiveMode(argv[1]);               // archive

            } // end if (argc > 1)
            else
            {
                /*
                 *  "Database" mode:
                 *      if no archive given on the command line.
                 */

                GetGeneralEnvironment();

                if (G_pDatabase->DBPackagesList.begin() == G_pDatabase->DBPackagesList.end())
                {
                    // no packages found (first run or
                    // database not found):
                    guiProductInfo(HWND_DESKTOP, TRUE);
                }
                else
                    guiBeginDatabaseMode();
            }

            // done: check what we've done and display an according message
            if (G_WpiGlobals.fDisplayPages)     // V0.9.9 (2001-03-30) [umoeller]
            {
                if (    (G_WpiGlobals.fInstallComplete)
                     || (G_WpiGlobals.fDeinstallComplete)
                   )
                {
                    BSString    strDone;

                    if (G_WpiGlobals.fInstallComplete)
                    {
                        wpiGetMessage(strDone,
                                      143      // installation has completed
                                      );
                        strDone += " ";
                        if (G_WpiGlobals.fWPSObjectsCreated)
                        {
                            // created objects:
                            BSString str2;
                            wpiGetMessage(str2,
                                          144);  // "objects created"
                            strDone += str2;
                        }
                    }

                    if (    (G_WpiGlobals.fConfigSysChanged)
                         && (!G_WpiGlobals.fWPSClassesChanged)
                       )
                    {
                        PSZ pTable = G_WpiGlobals.szConfigSysBackup;
                        BSString str2;
                        wpiGetMessage(str2,
                                      145, // "CONFIG.SYS changed"
                                      &pTable,      // replacement for %1
                                      1);
                        strDone += str2;
                    }
                    else if (    (!G_WpiGlobals.fConfigSysChanged)
                              && (G_WpiGlobals.fWPSClassesChanged)
                            )
                    {
                        BSString str2;
                        wpiGetMessage(str2,
                                      146);  // "WPS class list changed",
                        strDone += str2;
                    }
                    else if (    (G_WpiGlobals.fConfigSysChanged)
                              && (G_WpiGlobals.fWPSClassesChanged)
                            )
                    {
                        BSString str2;
                        wpiGetMessage(str2,
                                      147);  // "both CONFIG.SYS and WPS class list changed",
                        strDone += str2;
                    }

                    if (strDone())
                        // we got something to say:
                        guiShowMessage("WarpIN",
                                       strDone,
                                       MB_INFORMATION | MB_OK);

                    // offer restart WPS if only the WPS class list,
                    // but not CONFIG.SYS has changed.
                    if (    (!G_WpiGlobals.fConfigSysChanged)
                         && (G_WpiGlobals.fWPSClassesChanged)
                       )
                    {
                        if (wpiMessageBox(192,
                                          193,  // restart WPS?
                                          MB_YESNO)
                            == MBID_YES)
                        winhResetWPS(G_WpiGlobals.habThread1);
                    }
                }
            }
        }
        catch(BSCancelExcpt &X)
        {       // do nothing, this is for exiting only;
        }
        catch (FEDatabaseExcpt &X)
        {
            BSString str2 = X._strDescription;

            if (X._fRecover)
            {
                // recoverable: delete that key
                FEDatabase Database2(G_WpiGlobals.habThread1,
                                     G_WpiGlobals.szWarpINPath,
                                     X._strFailingPckID.c_str());     // bugfixing
                str2 += "\nThe offending database entry has been deleted. Please restart WarpIN.";
            }

            guiShowMessage("WarpIN: Database Error",
                           str2,
                           MB_ICONEXCLAMATION | MB_CANCEL);
        }
        catch (BSExcptBase &X)
        {
            // base exception class: all exception classes
            // are derived from this
            guiShowMessage("WarpIN: Error",
                           X._strDescription,
                           MB_ICONEXCLAMATION | MB_CANCEL);
        }
    }
    CATCH(excpt1)
    {
        // exception occured:
        // kill Install thread if it's running
        if (thrQueryID(&G_WpiGlobals.tiInstallThread))
            thrKill(&G_WpiGlobals.tiInstallThread);

        wpiMessageBox(165,      // internal error
                      107,         // "exception in main thread"
                      MB_ICONEXCLAMATION | MB_CANCEL);
    } END_CATCH();

    // delete all temp files
    if (G_WpiGlobals.strTempFiles())
    {
        const char *pStart = G_WpiGlobals.strTempFiles.c_str(),
                   *pEnd;

        while (    (*pStart)
                && (pEnd = strchr(pStart, '|'))
              )
        {
            string strThis(pStart, pEnd);
            if (strThis())
            {
                DosDelete(strThis.c_str());
                pStart = pEnd + 1;
            }
            else
                break;
        }
    }

    // clean up
    PrfCloseProfile(G_WpiGlobals.hiniWarpIN);
    guiCleanupBeforeExit();
    WinTerminate(G_WpiGlobals.habThread1);

    return (0);         // no error
}

/*
 *@@ AddArchive2JobList:
 *      this gets called from EnterArchiveMode to create
 *      FEInstallJob instances from the given archive.
 *      This also checks the database for whether the
 *      packages are already installed.
 *
 *      Note: this leaves all packages unselected at this
 *      point.
 *
 *@@added V0.9.0 (99-10-31) [umoeller]
 *@@changed V0.9.12 (2001-05-17) [umoeller]: fixed crash with group jobs
 */

VOID AddArchive2JobList(FEArchive *pArchive)
{
    BOOL fJobsAdded = FALSE;

    list<FEArcPackageBase*>::iterator PckStart = pArchive->_PackagesList.begin(),
                                      PckEnd = pArchive->_PackagesList.end();
    for (; PckStart != PckEnd; PckStart++)
    {
        FEArcPackageBase *pPackageThis = *PckStart;

        // first find if we're part of a group ourselves
        FEGroupJob *pGroupJob = NULL;
        if (pPackageThis->_pGroup)
            // yes: find group job for that (must exist already)
            pGroupJob = wpiFindGroupJob(pPackageThis->_pGroup);

        FEArcPackagePck *pPckPckThis = pPackageThis->IsPackage();

        if (!pPckPckThis)
        {
            // group: create group job
            FEGroupJob *pJobNew = new FEGroupJob((FEArcPackageGroup*)pPackageThis,
                                                 pGroupJob);
            // G_WpiGlobals.GroupJobsList.push_back(pJobNew);
            G_WpiGlobals.AllJobsList.push_back(pJobNew);
            fJobsAdded = TRUE;
        }
        else
        {
            // add config info to global config allow flags
            // (moved this here, _pDecl is NULL for groups!)
            // V0.9.12 (2001-05-17) [umoeller]
            G_WpiGlobals.ulAllowConfig |= pPckPckThis->_pDecl->_ulConfigData;

            // create job for package
            FEInstallJob *pJobNew = new FEInstallJob(pPckPckThis,
                                                     pGroupJob);
            if (pJobNew->CheckInstall(G_pDatabase))
            {
                G_WpiGlobals.fAnyAlreadyInstalled = TRUE;
            }
            // G_WpiGlobals.InstallJobsList.push_back(pJobNew);
            G_WpiGlobals.AllJobsList.push_back(pJobNew);
            fJobsAdded = TRUE;
        }

    } // end for (; PckStart != PckEnd; PckStart++)

    if (fJobsAdded)
    {
        // anything added: store archive in global archives list
        G_WpiGlobals.ArchivesList.push_back(pArchive);
        // is it the first archive?
        if (G_WpiGlobals.pFirstArchive == NULL)
            G_WpiGlobals.pFirstArchive = pArchive;
    }
}

/*
 *@@ DefaultSelectAllJobs:
 *      this gets called from EnterArchiveMode to
 *      go thru all install jobs in the package
 *      and check whether the required packages
 *      are already installed and default-select
 *      all jobs.
 *
 *@@added V0.9.1 (2000-01-08) [umoeller]
 *@@changed V0.9.3 (2000-04-20) [umoeller]: finally made macros work in target paths
 *@@changed V0.9.3 (2000-06-05) [umoeller]: added dependency checks
 *@@changed V0.9.7 (2001-01-07) [umoeller]: fixed problems with target paths and other checks...
 *@@changed V0.9.7 (2001-01-07) [umoeller]: version checks were missing for requirements
 */

VOID DefaultSelectAllJobs()
{
    BSString strMissingPackages;

    // STEP 1... check requirements

    list<FEJobBase*>::iterator JobStart,
                               JobEnd = G_WpiGlobals.AllJobsList.end();
    for (JobStart = G_WpiGlobals.AllJobsList.begin();
         JobStart != JobEnd;
         JobStart++)
    {
        FEJobBase *pTemp = *JobStart;
        FEInstallJob *pJobThis = pTemp->IsInstallJob();
        if (pJobThis)
        {
            // check dependencies: V0.9.3 (2000-06-05) [umoeller]
            // go thru the package ID's which have been stored in
            // the package's logger
            const FEArcPackagePck *pPckPckThis = pJobThis->_pArcPackage;

            PSZ     pszLogStart = pPckPckThis->_logRequiresIDs._pabLogString,
                    pszLogThis = pszLogStart;
            while (pszLogThis < pszLogStart + pPckPckThis->_logRequiresIDs._cbLogString)
            {
                ULONG   cbLogThis = strlen(pszLogThis);
                BOOL    fSufficient = FALSE;
                // pszLogThis now points to the package ID which is required
                // by the job
                FEPackageID pckidReq(pszLogThis, __FILE__, __LINE__,
                                     FALSE);   // do not allow short format

                const FEDBPackage *pDBPackage
                    = G_pDatabase->FindDBPackage(pckidReq); /* ._pszAuthor,
                                                 pckidReq._pszApplication,
                                                 pckidReq._pszPackageName); */
                if (pDBPackage)
                {
                    // check version
                    FEPackageID pckidDB(pDBPackage,
                                        __FILE__, __LINE__,
                                        FALSE);   // do not allow short format
                    // check version V0.9.7 (2001-01-07) [umoeller]
                    if (    (pckidDB.Compare(pckidReq)
                                   & IDCOMP_VERSIONMASK)
                            != IDCOMP_THISOLDER
                       )
                        fSufficient = TRUE;
                }

                // else: fSufficient still FALSE

                if (!fSufficient)
                {
                    // required package not found in database:
                    // is it on the install list then?

                    ULONG ulComp = 0;
                    if (!wpiFindPackageInArchives(pszLogThis, &ulComp))
                    {
                        // no: add to list of missing packages, which will
                        // be reported below
                        BSString strIDSix;
                        pckidReq.CreateStringSix(strIDSix);
                        /* CHAR szID[2000];
                        sprintf(szID, "%s\\%s\\%s\\%d\\%d\\%d",
                                pckidReq._pszAuthor,
                                pckidReq._pszApplication,
                                pckidReq._pszPackageName,
                                pckidReq._ulVersionMajor,
                                pckidReq._ulVersionMinor,
                                pckidReq._ulVersionRevision); */
                        strMissingPackages += strIDSix;
                        strMissingPackages += "\n";
                    }
                }

                pszLogThis += cbLogThis + 1;    // go beyond null byte
            } // while (pszLogThis < pszLogStart + pPckPckThis->_logRequiresIDs._cbLogString)
        } // if (pJobThis)
    } // for (; JobStart != JobEnd; JobStart++)

    if (strMissingPackages())
    {
        // dependencies failed:
        PSZ apsz = (PSZ)strMissingPackages.c_str();

        wpiMessageBox(102,  // error
                      190,  // missing packages
                      MB_CANCEL | MB_ICONEXCLAMATION,
                      &apsz,
                      1);
        // terminate!
        throw BSCancelExcpt();
    }

    // STEP 2... set job target paths according to package defaults

    for (JobStart = G_WpiGlobals.AllJobsList.begin();
         JobStart != JobEnd;
         JobStart++)
    {
        FEJobBase *pTemp = *JobStart;
        FEInstallJob *pJobThis = pTemp->IsInstallJob();
        if (pJobThis)
        {
            // set job's target path:
            // first use default target path from package
            pJobThis->SetTargetPath(pJobThis->_pArcPackage->QueryTargetPath(),
                                    G_WpiGlobals.AllJobsList);
        }
    }

    // STEP 3... if a package is already installed,
    // set the target path to that of the database package
    // or resolve macros...
    // we have to do a third loop because otherwise BASE
    // paths can't get updated
    for (JobStart = G_WpiGlobals.AllJobsList.begin();
         JobStart != JobEnd;
         JobStart++)
    {
        FEJobBase *pTemp = *JobStart;
        FEInstallJob *pJobThis = pTemp->IsInstallJob();
        if (pJobThis)
        {
            if (pJobThis->_pPackageIDInstalled)
            {
                // package already installed:
                // reset target path to that of the installed package;
                // we must set this twice because otherwise non-base
                // packages won't get updated
                const FEDBPackage *pPackageInstalled
                        = G_pDatabase->FindDBPackage(*pJobThis->_pPackageIDInstalled); /* ->_pszAuthor,
                                                     pJobThis->_pPackageIDInstalled->_pszApplication,
                                                     pJobThis->_pPackageIDInstalled->_pszPackageName); */
                pJobThis->SetTargetPath(pPackageInstalled->QueryTargetPath(),
                                        G_WpiGlobals.AllJobsList);
            }
            else
            {
                // not installed:
                // resolve macros for target path
                BSString strTemp = pJobThis->QueryTargetPath();
                wpiResolveMacros(strTemp);
                pJobThis->SetTargetPath(strTemp.c_str(),
                                        G_WpiGlobals.AllJobsList);
            }

            // use default selection for job;
            // this will update the parent group too
            pJobThis->DefaultSelect();
        } // end if (!strMissingPackages())
    }
}

/*
 *@@ fntParseScript:
 *
 *@@added V0.9.12 (2001-05-31) [umoeller]
 */

VOID _Optlink fntParseScript(PTHREADINFO ptiMyself)
{
    APIRET arc = 0;
    FEArchive *pArchive = (FEArchive*)ptiMyself->ulData;

    try
    {
        pArchive->ParseScript(WICallback);
    }
    catch(BSCancelExcpt &X)
    {       // do nothing, this is for exiting only;
        arc = 1;
    }
    catch (BSExcptBase &X)
    {
        // base exception class: all exception classes
        // are derived from this
        guiShowMessage("WarpIN: Error",
                       X._strDescription,
                       MB_ICONEXCLAMATION | MB_CANCEL);
        arc = 1;
    }

    // notify "done"
    WinPostMsg(ptiMyself->hwndNotify,
               WM_USER,
               (MPARAM)arc,
               0);
}

/*
 *@@ EnterArchiveMode:
 *      this gets called from main() when WarpIN was
 *      started with an archive as a parameter.
 *
 *      We create an instance of FEArchive in G_WpiGlobals
 *      from the file name which was given to us here.
 *      FEArchive::FEArchive will then parse
 *      the entire install script in the archive and
 *      create instances of FEInstallJob accordingly,
 *      each containing an instance of FEArcPackageBase.
 *
 *      We then call GetArchiveEnvironment() to override
 *      certain default settings, if environment variables
 *      have been set by the user.
 *
 *      If WARPIN_DISPLAYPAGES=YES (which is the default),
 *      we then call guiBeginInstallMode to have the GUI
 *      display the various pages in the script. That
 *      function only returns after the user has gone
 *      thru all pages.
 *
 *      If any packages have been selected for removal
 *      from the GUI (or thru environment variables),
 *      we call guiStartRemove.
 *
 *      If any packages must be installed, we then call
 *      guiStartInstall and invoke FEInstallJob::Store
 *      on all jobs which have been installed.
 *
 *      In any case, afterwards we call wpiConfigure to have
 *      CONFIG.SYS and the WPS changed, if necessary.
 *
 *      So this function only returns after everything has
 *      been done, or installation has been cancelled, or
 *      an error occured.
 *
 *@@changed V0.9.0 (99-11-01) [umoeller]: adjusted for FEInstallJob
 *@@changed V0.9.0 (99-11-02) [umoeller]: completely reworked add/remove mode
 *@@changed V0.9.3 (2000-04-28) [umoeller]: removed add/remove mode
 *@@changed V0.9.3 (2000-05-12) [umoeller]: added pszCfgSysBackup
 *@@changed V0.9.4 (2000-07-28) [umoeller]: fixed environment problems
 *@@changed V0.9.6 (2000-10-27) [umoeller]: made CONFIG.SYS backup part of WPIGLOBALS
 *@@changed V0.9.6 (2000-11-23) [umoeller]: added "no display pages" support
 *@@changed V0.9.9 (2001-03-30) [umoeller]: added log file
 *@@changed V0.9.12 (2001-05-31) [umoeller]: moved script parsing to fntParseScript thread
 */

void EnterArchiveMode(PSZ pszArchiveFile)
{
    /*
     *  "Archive" mode:
     *      archive given on command line
     */

    GetGeneralEnvironment();

    // create instance of FEArchive on the heap
    // and store it in G_WpiGlobals.
    // This constructor is complex: it reads the
    // archive and does all the install script parsing.
    FEArchive *pArchive = new FEArchive(pszArchiveFile);

    // parse script; this executes REXX code etc. and
    // might take a while, so use a second transient thread
    // for this V0.9.12 (2001-05-31) [umoeller]
    if (thrRunSync(G_WpiGlobals.habThread1,
                   fntParseScript,
                   "ParseScript",
                   (ULONG)pArchive))        // thread param
        // error occured:
        throw(BSCancelExcpt());

    // copy archive title to WPIGLOBALS
    G_WpiGlobals.strAppTitle = pArchive->_pScript->_strTitle;

    // we now have the list of FEPackages in
    // pArchive->PackagesList;
    // now create jobs for this and query the database
    // for whether any of these packages are already installed
    AddArchive2JobList(pArchive);
        // this updates WPIGLOBALS
        // and selects packages which are already installed and
        //  a)  outdated
        //  b)  same version

    DefaultSelectAllJobs();

    // store that list in WPIGLOBALS so the GUI
    // can access it
    G_WpiGlobals.pPageInfoList = &pArchive->_pScript->_PageInfoList;

    // now parse all the environment variables to
    // override certain settings
    GetArchiveEnvironment();       // install mode
            // this can set WPIGLOBALS.fDisplayPages to FALSE

    guiSetupGlobalStatus(TRUE);
        // moved this here V0.9.6 (2000-11-23) [umoeller]

    // enter "full install" mode
    BOOL fStartInstall = FALSE;

    if (G_WpiGlobals.fDisplayPages)
        // WARPIN_DISPLAYPAGES=YES (default): well, display pages then
        fStartInstall = guiBeginInstallMode();
    else
    {
        // pages are NOT to be displayed: we need to do the
        // checking here which is done by the GUI otherwise
        fStartInstall = wpiVerifySelections(FALSE); // do not prompt for missing dirs
                // this also checks dependencies
    }

    if (fStartInstall)
    {
        BSFileLogger *pLogFile = NULL;
        if (G_WpiGlobals.szLogFilename[0])
            // logging enabled:
            pLogFile = new BSFileLogger(G_WpiGlobals.szLogFilename);

        // 1) process KILLPROCESS attributes
        wpiKillProcesses(pLogFile);

        // 2) deinstall packages, if any

        // create a list of FEDBPackage instances
        // from all the FEArcPackages which have
        // been selected for de-installation
        list<FEDBPackage*> DeinstallPackagesList;

        // go thru jobs
        list<FEJobBase*>::iterator JobStart = G_WpiGlobals.AllJobsList.begin(),
                                   JobEnd = G_WpiGlobals.AllJobsList.end();
        ULONG   c2BeRemoved = 0;
        for (; JobStart != JobEnd; JobStart++)
        {
            FEJobBase *pTemp = *JobStart;
            FEInstallJob *pJob = pTemp->IsInstallJob();
            if (pJob)
                if (pJob->Is2BeRemoved())
                {
                    FEDBPackage *pDBPck = new FEDBPackage(G_pDatabase,
                                                          *pJob->_pPackageIDInstalled);
                    DeinstallPackagesList.push_back(pDBPck);
                    c2BeRemoved++;
                }
        }

        if (c2BeRemoved)
        {
            guiUpdateGlobalStatus(GSTAT_REMOVING);

            // anything selected for removal:
            guiStartRemove(&DeinstallPackagesList,
                           PCK_CONFIGSYS
                                | PCK_WPSCLASSES
                                | PCK_WPSOBJECTS,
                           TRUE,       // modal
                           pLogFile);
        }

        // clean up list
        list<FEDBPackage*>::iterator DBPckStart = DeinstallPackagesList.begin(),
                                     DBPckEnd = DeinstallPackagesList.end();
        for (; DBPckStart != DBPckEnd; DBPckStart++)
        {
            FEDBPackage *pDBPck = *DBPckStart;
            delete pDBPck;
        }

        // 2) unpack files

        guiUpdateGlobalStatus(GSTAT_UNPACKING);
        // TRUE means go ahead:
        guiStartInstall(pLogFile);

        if (G_WpiGlobals.fInstallComplete)
        {
            // success so far:
            // configure (CONFIG.SYS, WPS classes, objects)
            wpiConfigure(pLogFile);

            // new packages installed:
            // store all packages in the database
            list<FEJobBase*>::iterator JobStart = G_WpiGlobals.AllJobsList.begin(),
                                       JobEnd = G_WpiGlobals.AllJobsList.end();
            for (; JobStart != JobEnd; JobStart++)
            {
                FEJobBase *pTemp = *JobStart;
                FEInstallJob *pJobThis = pTemp->IsInstallJob();
                if (pJobThis)
                    if (pJobThis->Is2BeInstalled())
                        pJobThis->Store(G_pDatabase);
            }

            guiUpdateGlobalStatus(GSTAT_DONEWITHALL);
        }

        if (pLogFile)
            delete pLogFile;

    } // end if (fStartInstall)
}

/* ******************************************************************
 *
 *  PART 2
 *  ------
 *  The following code is started by part 1 when installation
 *  is actually about to begin, i.e. the packages shall be
 *  extracted from the archive. All this is running on a separate
 *  thread, the Install thread.
 *
 ********************************************************************/

/*
 *@@ wpiStartInstallThread:
 *      this starts the Install thread.
 *
 *      This returns immediately. From now on,
 *      the GUI gets lots of callbacks on the
 *      Install thread.
 *
 *@@changed V0.9.12 (2001-05-22) [umoeller]: now creating flag with PM message queue
 */

VOID wpiStartInstallThread(BSFileLogger *pLogFile)
{
    thrCreate(&G_WpiGlobals.tiInstallThread,
              wpiInstallThread,
              NULL,
              "Install",
              THRF_PMMSGQUEUE | THRF_WAIT,
              (ULONG)pLogFile);
}


