/*
 *@@sourcefile arcsubsys.cpp:
 *  archive subsystem
 *
 *
 */

/*      Copyright (C) 2000-2001 Teemu Ahola.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of this distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define INCL_DOSPROCESS
#define INCL_DOSERRORS
#define INCL_DOSQUEUES
#define INCL_DOSSEMAPHORES
#include <os2.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "setup.h"
#include <deque>
#include <vector>
#include <map>

#include "helpers\stringh.h"
#include "helpers\xstring.h"
#include "base\bs_string.h"

// Base include
#include "wipengine\wipebase.h"
// File database unit include
#include "wipengine\fdu.h"
// Thread base include
#include "wipengine\threadbase.h"
// Archive subsystem include
#include "wipengine\arcsubsys.h"
// Message logging unit include.
#include "wipengine\mlu.h"

// XML related includes. Refer to helpers\xml.c.
#include "expat\expat.h" // must come before xml.h
#include "helpers\linklist.h"
#include "helpers\tree.h"
#include "helpers\xml.h"

//#define PRINT_DEBUG_INFO


/*
 *@@assDefaultCallBack:
 *  Default callback function is used when the user has not provided
 *  an callback function. This function simply frees the data passed
 *  with the message to the logger thread. Only if the callback type
 *  is CB_TYPE_FILE_QUERYINGCOMPLETED, the memory is not freed since
 *  the user wants to get that memory.
 */
LONG assDefaultCallBack(SHORT p_sCBType, LONG, PVOID p_pData, PVOID) {

    #ifdef PRINT_DEBUG_INFO
     printf("assDefaultCallBack(SHORT, LONG, PVOID, PVOID):\n"
            " - parameters:\n"
            "    p_sCBType = %i\n"
            "    p_pData = 0x%X\n\n",
            p_sCBType, p_pData);
     fflush(stdout);
    #endif

    // Free the data if the message was not QUERY since user wants the
    // data.
    if ((p_pData) && (p_sCBType != CB_TYPE_FILE_QUERYINGCOMPLETED))
        delete p_pData;

    return RET_OK;
    }


/*
 *@@InitSemaphores:
 *  initializes map mutex semaphore and init event semaphore by
 *  using DosCreateMutexSem and DosCreateEventSem. <I>Map mutex semaphore</I>
 *  is used to coordinate _Messages multimap usage. <I>Init event semaphore</I>
 *  is used during initialization phase and it is closed soon after
 *  the initialization is completed.
 */
VOID ArcSubSystem::InitSemaphores(VOID) {
    APIRET ulrc = 0;
    // Create mutex semaphore for coordination of using of _Files.
    ulrc = DosCreateMutexSem(NULL, &_mtxSemFiles, 0L, FALSE);
    #ifdef PRINT_DEBUG_INFO
     printf("DosCreateMutexSem in InitSemaphores returned: %i\n", ulrc);
     fflush(stdout);
    #endif
    // Create mutex semaphore for coordination of using of _bssQueueName.
    ulrc = DosCreateMutexSem(NULL, &_mtxSemQueName, 0L, FALSE);
    #ifdef PRINT_DEBUG_INFO
     printf("DosCreateMutexSem in InitSemaphores returned: %i\n", ulrc);
     fflush(stdout);
    #endif
    // Create event semaphore for initialization syncronization.
    ulrc = DosCreateEventSem(NULL, &_hevSemInit, 0L, FALSE);
    #ifdef PRINT_DEBUG_INFO
     printf("DosCreateEventSem in InitSemaphores returned: %i\n", ulrc);
     fflush(stdout);
    #endif
    // Create event semaphore for initialization error indication.
    ulrc = DosCreateEventSem(NULL, &_hevSemError, 0L, FALSE);
    #ifdef PRINT_DEBUG_INFO
     printf("DosCreateEventSem in InitSemaphores returned: %i\n", ulrc);
     fflush(stdout);
    #endif
    }


/*
 *@@Start:
 *  this function calls 'execute' function which was inherit from
 *  threadbase class. After that this function waits until an event
 *  semaphore is posted from '_ThdFunction' in a newly created thread.
 *  Next the event semaphore is closed since there is no need for it
 *  anymore. If an error has been occurred during the thread initialization
 *  phase, error count semaphore is increased.
 */
USHORT ArcSubSystem::Start(VOID) {
    APIRET ulrc = 0;
    ULONG ulOwned, ulError = 0;
    PID pid;
    TID tid;
    if (_hQMessageLogger == NULLHANDLE)
        return RET_ERR_PARAMETERERROR;

    if (execute() == 0)
        return RET_ERR_THREADCREATIONFAILED;

    // Wait until the initialization is completed.
    ulrc = DosWaitEventSem(_hevSemInit, SEM_INDEFINITE_WAIT);
    #ifdef PRINT_DEBUG_INFO
     printf("DosWaitEventSem in Start returned: %i\n", ulrc);
     fflush(stdout);
    #endif

    // Query post counts of error semaphore which indicates
    // if an error has been occurred during the initialization.
    ulrc = DosQueryEventSem(_hevSemError, &ulError);
    #ifdef PRINT_DEBUG_INFO
     printf("Error semaphore post count: %i\n", ulError);
     printf("DosQueryEventSem in Start returned: %i\n", ulrc);
     fflush(stdout);
    #endif

    // If error occurred, close semaphores and return an error code.
    // Detailed description about the error can be found from the MLU.
    if (ulError > 0) {
        DosCloseEventSem(_hevSemError);
        DosCloseEventSem(_hevSemInit);
        return RET_ERR_INITIALIZATIONFAILED;
        }

    // Close the error semaphore since there is no need for it anymore.
    ulrc = DosCloseEventSem(_hevSemError);
    #ifdef PRINT_DEBUG_INFO
     printf("DosCloseEventSem in Start of returned: %i\n", ulrc);
     fflush(stdout);
    #endif

    // Close the initialization event semaphore since there is
    // no need for it anymore.
    ulrc = DosCloseEventSem(_hevSemInit);
    #ifdef PRINT_DEBUG_INFO
     printf("DosCloseEventSem in Start of returned: %i\n", ulrc);
     fflush(stdout);
    #endif

    _hevSemInit = NULLHANDLE;
    _hevSemError = NULLHANDLE;

    return RET_OK;
    }


/*
 *@@ InsertFiles:
 *  This method inserts files given in p_files vector to
 *  the file database unit. Multi-user protection has been
 *  achieved by using mutex semaphores.
 */
USHORT ArcSubSystem::InsertFiles(FileVector &p_files) {
    FileVector::iterator itrBegin;

    DosRequestMutexSem(_mtxSemFiles, SEM_INDEFINITE_WAIT);

    itrBegin = p_files.begin();
    for (;itrBegin < p_files.end(); itrBegin++)
        _Files.InsertFile((FileClass)*itrBegin);

    DosReleaseMutexSem(_mtxSemFiles);
    return RET_OK;
    }


/*
 *@@ InsertFiles:
 *  This method queries all files in the file database unit.
 *  Multi-user protection has been achieved by using mutex
 *  semaphores.
 */
USHORT ArcSubSystem::QueryFiles(FileVector &p_pFileVector) {

    DosRequestMutexSem(_mtxSemFiles, SEM_INDEFINITE_WAIT);

    _Files.QueryFiles(p_pFileVector);

    DosReleaseMutexSem(_mtxSemFiles);
    return RET_OK;
    }


/*
 *@@ DeleteFiles:
 *  This method deletes all files in the file database unit
 *  which are listed in the give FileVector parameter.
 *  Multi-user protection has been achieved by using mutex
 *  semaphores.
 */
USHORT ArcSubSystem::DeleteFiles(FileVector &p_files) {
    FileVector::iterator itrBegin;

    DosRequestMutexSem(_mtxSemFiles, SEM_INDEFINITE_WAIT);
    itrBegin = p_files.begin();
    for (; itrBegin < p_files.end(); itrBegin++)
        _Files.DeleteFile((FileClass)*itrBegin);

    DosReleaseMutexSem(_mtxSemFiles);
    return RET_OK;
    }


/*
 *@@_ThdFunction:
 *  this function is an implementation for a virtual function in
 *  ThreadBase class and therefore required. This function is called
 *  from inside ThreadStarted function which is called from inside
 *  _begingthread function which actually creates a new thread. In
 *  this function an unique message queue is created and after that
 *  Start function is notified by posting event semaphore. Next
 *  the function starts to wait messages in an indefinite loop.
 *  When a message arrives to the message queue needed operations
 *  are accomplishend depending the message type. After that the function
 *  calls a call-back function if provided and posts event semaphore.
 *  Also the memory allocated for the receivced data is deleted here so
 *  it is important not to send pointers to automatic variables or free
 *  allocated memory in the sender thread.
 *
 *  <B>Operations:</B>
 *      <I>Add files</I>: Adds files in a file vector to the file database
 *                        unit's database.
 *
 */
VOID ArcSubSystem::_ThdFunction() {
    BSString bssErrorMessage;
    TID tidSenderThread = 0;
    APIRET ulrc = ERROR_QUE_DUPLICATE;
    USHORT usIndexNbr = 0;
    CHAR szQueueName[256];
    REQUESTDATA request = {0};
    ULONG ulDataLength = 0;
    BYTE bPriority = 0;
    ULONG ulCommand;
    PVOID pReceivedData = NULL;
    MsgInfo *pMsgInfo = NULL;
    FileVector *pFileVector = NULL;


    // Create unique queue. DosCreateQueue returns ERROR_QUE_DUPLICATED, if there is
    // already a queue with same name.
    while (ulrc == ERROR_QUE_DUPLICATE) {
        // Format an unique queue name.
        sprintf(szQueueName, "\\QUEUES\\WIPENGINE\\ASUBSYS%i.QUE", usIndexNbr);
        // Create a new queue with FIFO priority-ordering and address-converting
        // flags.
        ulrc = DosCreateQueue(&_hQReceiver, QUE_FIFO | QUE_CONVERT_ADDRESS,
                              szQueueName);
        usIndexNbr++;
        }

    // If error occurred, return immediately. This also completes the thread.
    // Before returning, post error semaphore so starter function knows
    // that something went wrong. Sends error message to the MLU.
    // _aidSender will be NULLHANDLE.
    if (ulrc != NO_ERROR) {
        #ifdef PRINT_DEBUG_INFO
         printf("Initialization error occurred in arcsubsys.cpp (_ThdFunction).\n");
         fflush(stdout);
        #endif
        DosPostEventSem(_hevSemError);

        // Compose error message.
        bssErrorMessage = STATUS_MSG_FAIL;
        bssErrorMessage += " in ";
        bssErrorMessage += __FILE__;
        // Log the error to MLU. Sender in NULLHANDLE and type is critical.
        MsgLoggingUnit::LogMessage(NULLHANDLE, bssErrorMessage, TYPE_MSG_CRITICAL,
                                   RET_ERR_INITIALIZATIONFAILED, _hQMessageLogger);
        // Release initialization semaphore.
        DosPostEventSem(_hevSemInit);
        return;
        }

    #ifdef PRINT_DEBUG_INFO
     printf("Message queue created with name: %s\n", szQueueName);
     fflush(stdout);
    #endif

    // Store the queue name.
    DosRequestMutexSem(_mtxSemQueName, SEM_INDEFINITE_WAIT);
    _bssQueueName = szQueueName;
    DosReleaseMutexSem(_mtxSemQueName);

    // Release initialization semaphore.
    DosPostEventSem(_hevSemInit);

    while (TRUE) {
        // Wait until there is a message in the queue created earlier.
        ulrc = DosReadQueue(_hQReceiver, &request, &ulDataLength, (PPVOID)&pReceivedData,
                            0L, DCWW_WAIT, &bPriority, 0L);
        ulCommand = request.ulData;

        #ifdef PRINT_DEBUG_INFO
         printf("Message received.\n");
         fflush(stdout);
        #endif

        if (pReceivedData) {
            switch (ulCommand) {
                case QUE_MSG_ADD_FILES: {
                    pMsgInfo = (MsgInfo *)pReceivedData;
                    pFileVector = (FileVector *)pMsgInfo->_pData;
                    if (pFileVector) {
                        InsertFiles(*pFileVector);
                        if (_pfnWipeCallBack)
                            (*_pfnWipeCallBack)(CB_TYPE_FILE_ADDINGCOMPLETED, 0,
                                                pMsgInfo->_pData, _pUserData);
                        if (pMsgInfo->_hevSem)
                            DosPostEventSem(pMsgInfo->_hevSem);
                        #ifdef PRINT_DEBUG_INFO
                         printf("Add files.\n");
                         fflush(stdout);
                        #endif
                        }
                    delete pMsgInfo;
                    break;
                    }
                case QUE_MSG_QUERY_FILES: {
                    pMsgInfo = (MsgInfo *)pReceivedData;
                    pFileVector = (FileVector *)pMsgInfo->_pData;
                    if (pMsgInfo) {
                        QueryFiles(*pFileVector);
                        if (_pfnWipeCallBack)
                            (*_pfnWipeCallBack)(CB_TYPE_FILE_QUERYINGCOMPLETED, 0,
                                                pMsgInfo->_pData, _pUserData);
                        if (pMsgInfo->_hevSem)
                            DosPostEventSem(pMsgInfo->_hevSem);
                        #ifdef PRINT_DEBUG_INFO
                         printf("Query files.\n");
                         fflush(stdout);
                        #endif
                        }
                    delete pMsgInfo;
                    break;
                    }
                case QUE_MSG_DELETE_FILES: {
                    pMsgInfo = (MsgInfo *)pReceivedData;
                    pFileVector = (FileVector *)pMsgInfo->_pData;
                    if (pFileVector) {
                        DeleteFiles(*pFileVector);
                        if (_pfnWipeCallBack)
                            (*_pfnWipeCallBack)(CB_TYPE_FILE_QUERYINGCOMPLETED, 0,
                                                pMsgInfo->_pData, _pUserData);
                        if (pMsgInfo->_hevSem)
                            DosPostEventSem(pMsgInfo->_hevSem);
                        #ifdef PRINT_DEBUG_INFO
                         printf("Delete files.\n");
                         fflush(stdout);
                        #endif
                        }
                    delete pMsgInfo;
                    break;
                    }
                case QUE_MSG_QUIT: {
                    pMsgInfo = (MsgInfo *)pReceivedData;
                    #ifdef PRINT_DEBUG_INFO
                     printf("Quit\n\n");
                     fflush(stdout);
                    #endif
                    // Close the queue.
                    DosCloseQueue(_hQReceiver);
                    _hQReceiver = NULLHANDLE;
                    #ifdef PRINT_DEBUG_INFO
                     printf("Message queue closed.\n");
                     fflush(stdout);
                    #endif
                    // Make a call-back if a call-back function has been provided.
                    if (_pfnWipeCallBack)
                        (*_pfnWipeCallBack)(CB_TYPE_QUITCOMPLETED, 0,
                                            pMsgInfo->_pData, _pUserData);
                    // If there is an event semaphore, post it since a thread which
                    // wrote this message may want us to notify that quitting was
                    // completed.
                    if (pMsgInfo->_hevSem != NULLHANDLE)
                        DosPostEventSem(pMsgInfo->_hevSem);
                    // Free the memory received for the message.
                    delete pMsgInfo;
                    // Leave the function.
                    #ifdef PRINT_DEBUG_INFO
                     printf("Leaving _ThdFunction in arcsubssys.cpp.\n");
                     fflush(stdout);
                    #endif
                    return;
                    }
                } /* switch (ulCommand) { */
            } /* switch (ulCommand) { */
        } /* while (TRUE) { */

    }




/*
 *@@CloseArchive:
 *  This method closes an archive (an instance of the archive subsystem)
 *  which Archive Identifier is specified. Also the specified semaphore
 *  will be signalled by the archive after closing has been completed.
 *  This method is actually a dummy interface method for sending
 *  QUE_MSG_QUIT message to the specified archive subsystem.
 */
USHORT ArcSubSystem::CloseArchive(HEV p_hevSem) {

    MsgInfo *pMsgInfo = new MsgInfo;

    pMsgInfo->_aidSender = NULLHANDLE;
    pMsgInfo->_pData = NULL;
    pMsgInfo->_hevSem = p_hevSem;

    // Send message to the specified archive sub system which ID is
    // p_aidASS (HMQUEUE). This message is sent with lowest priority
    // so if some other thread sends other messages they will be
    // performed before this message.
    return DosWriteQueue(_hQReceiver, QUE_MSG_QUIT, sizeof(MsgInfo), (PVOID)pMsgInfo,
                         QUE_MSG_PRIORITY_LOWEST);


    }


/*
 *@@InsertFiles:
 *  Inserts files in the specified FileVector to the specified archive.
 *  The FileVector will be free so it should be dynamically allocated and
 *  should not be freed in the calling thread. Also the specified semaphore
 *  will be signalled by the archive after adding of the files has been
 *  completed. This method is actually a dummy interface method for sending
 *  QUE_MSG_ADD_FILES message to the specified archive subsystem.
 */
USHORT ArcSubSystem::InsertFiles(FileVector *p_pFileVector,
                                 HEV p_hevSem
                                ) {
    if (!p_pFileVector) {
        printf("*******ERRRROOOORRRRR QueryFiles******\n");
        return RET_ERR_PARAMETERERROR;
        }

    MsgInfo *pMsgInfo = new MsgInfo;

    pMsgInfo->_aidSender = NULLHANDLE;
    pMsgInfo->_pData = p_pFileVector;
    pMsgInfo->_hevSem = p_hevSem;

    // Send message to the specified archive sub system which ID is
    // p_aidASS (HMQUEUE).
    return DosWriteQueue(_hQReceiver, QUE_MSG_ADD_FILES, sizeof(MsgInfo),
                        (PVOID)pMsgInfo, QUE_MSG_PRIORITY_NORMAL);
    }


/*
 *@@DeleteFiles:
 *  Deletes files, specified in the FileVector, from the specified archive.
 *  The FileVector will be free so it should be dynamically allocated and
 *  should not be freed in the calling thread. Also the specified semaphore
 *  will be signalled by the archive after removing of the files has been
 *  completed. This method is actually a dummy interface method for sending
 *  QUE_MSG_DELETE_FILES message to the specified archive subsystem.
 */
USHORT ArcSubSystem::DeleteFiles(FileVector *p_pFileVector,
                                 HEV p_hevSem
                                ) {

    if (!p_pFileVector) {
        printf("*******ERRRROOOORRRRR QueryFiles******\n");
        return RET_ERR_PARAMETERERROR;
        }

    MsgInfo *pMsgInfo = new MsgInfo;

    pMsgInfo->_aidSender = NULLHANDLE;
    pMsgInfo->_pData = p_pFileVector;
    pMsgInfo->_hevSem = p_hevSem;

    // Send message to the specified archive sub system which ID is
    // p_aidASS (HMQUEUE).
    return DosWriteQueue(_hQReceiver, QUE_MSG_DELETE_FILES, sizeof(MsgInfo),
                        (PVOID)pMsgInfo, QUE_MSG_PRIORITY_NORMAL);
    }


/*
 *@@QueryFiles:
 *  Returns all files from the specified archive. A pointer to a FileVector
 *  must be provided. Also the specified semaphore will be signalled by
 *  the archive after removing of the files has been completed. This method
 *  is actually a dummy interface method for sending QUE_MSG_DELETE_FILES
 *  message to the specified archive subsystem.
 */
USHORT ArcSubSystem::QueryFiles(FileVector *p_pFileVector,
                                HEV p_hevSem
                                ) {

    if (!p_pFileVector) {
        printf("*******ERRRROOOORRRRR QueryFiles******\n");
        return RET_ERR_PARAMETERERROR;
        }

    MsgInfo *pMsgInfo = new MsgInfo;

    pMsgInfo->_aidSender = NULLHANDLE;
    pMsgInfo->_pData = p_pFileVector;
    pMsgInfo->_hevSem = p_hevSem;

    // Send message to the specified archive sub system which ID is
    // p_aidASS (HMQUEUE).
    return DosWriteQueue(_hQReceiver, QUE_MSG_QUERY_FILES, sizeof(MsgInfo),
                        (PVOID)pMsgInfo, QUE_MSG_PRIORITY_LOWEST);


    }

/*
 *@@~ArcSubSystem:
 *  Destructor for this class. Sends QUE_MSG_QUIT message to the thread
 *  and waits until the thread is dead.
 */
ArcSubSystem::~ArcSubSystem() {
    MsgInfo *pMsgInfo = new MsgInfo;

    #ifdef PRINT_DEBUG_INFO
     printf("~ArcSubSystem():\n"
            "   - _hQReceiver      = %i\n"
            "   - _Files           = %i\n"
            "   - _hQMessageLogger = %i\n"
            "   - _mtxSemFiles     = %i\n"
            "   - _mtxSemQueName   = %i\n"
            "   - _pfnWipeCallBack = 0x%X\n"
            "   - _pUserData       = 0x%X\n"
            "   - _bssQueueName    = %s",
            _hQReceiver, _Files, _hQMessageLogger,
            _mtxSemFiles, _mtxSemQueName, _pfnWipeCallBack,
            _pUserData, _bssQueueName.c_str());
     fflush(stdout);
    #endif

    pMsgInfo->_aidSender = NULLHANDLE;
    pMsgInfo->_pData = NULL;
    pMsgInfo->_hevSem = NULLHANDLE;
    // Send killer message to the thread.
    DosWriteQueue(_hQReceiver, QUE_MSG_QUIT, sizeof(MsgInfo), (PVOID)pMsgInfo,
                  QUE_MSG_PRIORITY_LOWEST);
    // Wait until the thread has died.
    waitThread();
    }



